/* Debugger.
   Copyright (C) 1992-1995 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define TABWIDTH 8

PUBLIC COUNTER stepMode=STEP_NO,resumeStepMode=STEP_NO;
#ifdef CODEPOINTS
/* if in launched as an emacs process (otherwise, leave them empty): */
PRIVATE char *enterBreakFmt="\033%d<";
PRIVATE char *exitBreakFmt="\033%d>";
#endif

PUBLIC MapPtr breakPoints;
PUBLIC INT breakPointCount;

PUBLIC ExprPtr stepFile;
PUBLIC COUNTER stepLine,stepPos;
EXTERN ExprPtr specialMarker;
EXTERN ExprPtr fntrue;

PRIVATE COUNTER maxWhere = 10;

PUBLIC ExprPtr applStkList;
PUBLIC ExprPtr applStkTable[NDBGSLOTS];
PUBLIC ExprPtr *applStkTop;

PUBLIC void flushApplStk()
{
  ExprPtr aux;
  VecPtr vp;

  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(NDBGSLOTS));
  vp->n = NDBGSLOTS;
  memcpy(vp->expr,applStkTable,sizeof(applStkTable));
  MAKEENV1_NOGC(vp,applStkList,aux);
  applStkList = aux;
  applStkTop = applStkTable;
}

PUBLIC void getbackApplStk()
{
  VecPtr vp;

  vp = EVARS(applStkList);
  memcpy(applStkTable,vp->expr,vp->n*sizeof(ExprPtr));
  applStkTop = applStkTable+vp->n;
  applStkList = ENEXT(applStkList);
}

PUBLIC ExprPtr getApplStk()
{
  COUNTER n;
  VecPtr vp;
  ExprPtr aux;

  if (applStkTable==applStkTop)
    return applStkList;
  n = applStkTop-applStkTable;
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  memcpy(vp->expr,applStkTable,n*sizeof(ExprPtr));
  MAKEENV1_NOGC(vp,applStkList,aux);
  applStkList = aux;
  applStkTop = applStkTable;
  return aux;
}

PUBLIC void setApplStk(wrap)
IN CONST ExprPtr wrap;
{
  applStkList = wrap;
  applStkTop = applStkTable;
}

PUBLIC void initApplStk()
{
  applStkList = NIL;
  applStkTop = applStkTable;
}

PRIVATE void startApplStk(p)
OUT struct applStkPtr *p;
{
  if (applStkTop==applStkTable)
    {
      p->ll = applStkList;
      p->i = (applStkList==NULL)?0:(EVARS(applStkList)->n-NFRAMEEXPRS);
    }
  else
    {
      p->ll = NOEXPR;
      p->i = applStkTop-applStkTable-NFRAMEEXPRS;
    }
}

#define APPLSTKPTR(p) ((p->ll==NOEXPR)?applStkTable:EVARS(p->ll)->expr)

PRIVATE Boolean endApplStk(p)
IN CONST struct applStkPtr *p;
{
  return p->ll==NIL;
}

PRIVATE ExprPtr applStkCode(p)
IN CONST struct applStkPtr *p;
{
  return APPLSTKPTR(p)[p->i];
}

PRIVATE ExprPtr applStkEnv(p)
IN CONST struct applStkPtr *p;
{
  return APPLSTKPTR(p)[p->i+1];
}

PRIVATE ExprPtr applStkCallingEnv(p)
IN CONST struct applStkPtr *p;
{
  return APPLSTKPTR(p)[p->i+2];
}

PRIVATE ExprPtr applStkZone(p)
IN CONST struct applStkPtr *p;
{
  return APPLSTKPTR(p)[p->i+3];
}

PRIVATE void nextApplStk(p)
INOUT struct applStkPtr *p;
{
  if (p->i<NFRAMEEXPRS)
    {
      if (p->ll==NOEXPR)
	p->ll = applStkList;
      else p->ll = ENEXT(p->ll);
      p->i = EVARS(p->ll)->n;
    }
  p->i -= NFRAMEEXPRS;
}

PRIVATE ExprPtr CONS(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return KEEP(z);
}

DECL_ACTION(apply_polyfun_closure)

PRIVATE void insertInfo(dz,listref)
IN CONST ExprPtr dz;
INOUT ExprPtr listref;
/* listref is a mutable list, whose cdr points to list;
   list is done with E_CONS (for beginning of zones)
   and with E_MAPLET (for ending of zones) */
{
  REGISTER ExprPtr lastl,ll;
  REGISTER ExprPtr info,info2;
  ExprPtr aux;

  info = DZINFO(dz);
  lastl = listref;
  for (; (ll=CDR(lastl))!=NIL; )
    switch (ll->kind)
      {
      case E_CONS:
	info2 = DZINFO(CAR(ll));
	if (ZONE(info)->line1<ZONE(info2)->line1 ||
	    (ZONE(info)->line1==ZONE(info2)->line1 &&
	     ZONE(info)->pos1<ZONE(info2)->pos1))
	  goto insertbegin;
	lastl = ll;
	break;
      case E_MAPLET:
	info2 = DZINFO(MAPLETX(ll));
	if (ZONE(info)->line1<ZONE(info2)->line2 ||
	    (ZONE(info)->line1==ZONE(info2)->line2 &&
	     ZONE(info)->pos1<ZONE(info2)->pos2))
	  goto insertbegin;
	lastl = ll; /* works only because MAPLETY==CDR */
	break;
      }
insertbegin:
  MAKECONS(dz,ll,aux);
  SETCDR(lastl,aux);
  lastl = aux;
  for (; (ll=CDR(lastl))!=NIL; )
    switch (ll->kind)
      {
      case E_CONS:
	info2 = DZINFO(CAR(ll));
	if (ZONE(info)->line2<ZONE(info2)->line1 ||
	    (ZONE(info)->line2==ZONE(info2)->line1 &&
	     ZONE(info)->pos2<=ZONE(info2)->pos1))
	  goto insertend;
	lastl = ll;
	break;
      case E_MAPLET:
	info2 = DZINFO(MAPLETX(ll));
	if (ZONE(info)->line2<ZONE(info2)->line2 ||
	    (ZONE(info)->line2==ZONE(info2)->line2 &&
	     ZONE(info)->pos2<=ZONE(info2)->pos2))
	  goto insertend;
	lastl = ll; /* works only because MAPLETY==CDR */
	break;
      }
insertend:
  MAKEMAPLET(dz,ll,aux);
  SETCDR(lastl,aux);
}

PRIVATE void sortInfos P((IN CONST ExprPtr code,INOUT ExprPtr listref));

#if STOP_AND_COPY_VECTORS
PRIVATE void sortInfosVec(vpp,listref)
IN Vector *HEAP *vpp;
INOUT ExprPtr listref;
{
  COUNTER i,n;

  for (i=0, n=(*vpp)->n; i<n; i++)
    sortInfos((*vpp)->expr[i],listref);
}
#else
PRIVATE void sortInfosVec(vp,listref)
IN CONST VecPtr vp;
INOUT ExprPtr listref;
{
  COUNTER i,n;
  ExprPtr *ep;

  for (i=0, n=vp->n, ep=vp->expr; i<n; i++, ep++)
    sortInfos(*ep,listref);
}
#endif

MAPSWEEPFUNC(do_sortinfos)
{
  ExprPtr listref;

  listref = MAPDATA(ExprPtr);
  sortInfos(x,listref);
  sortInfos(y,listref);
  return false;
}

PRIVATE void sortInfosBindings(bindlist,listref)
IN CONST ExprPtr bindlist;
INOUT ExprPtr listref;
{
  ExprPtr binding;

  while (bindlist!=NIL)
    {
      binding = CAR(bindlist);
      if (binding->kind==(CHAR)E_LET)
	{
	  sortInfosBindings(CDR(LETBIND(binding)),listref);
	  sortInfosBindings(LETBODY(binding),listref);
	}
      else
	{
	  sortInfos(CDR(binding),listref);
	}
      bindlist = CDR(bindlist);
    }
}

PRIVATE void sortInfosDoms(domlist,listref)
IN CONST ExprPtr domlist;
INOUT ExprPtr listref;
{
  while (domlist!=NIL)
    switch (domlist->kind)
      {
      case E_PAT_RECORD_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
	sortInfos(CDR(CAR(domlist)),listref);
	domlist = CDR(domlist);
	break;
      case E_TUPLE_1:
	sortInfosBindings(REFVAL(domlist),listref);
	domlist = NIL;
	break;
      }
}

DEF_PRIVATE_ACTION(sortinfos)
{
  struct sortinfos *data;

  data = (struct sortinfos *)env;
  sortInfos(data->code,data->listref);
  return NIL;
}

PRIVATE void sortInfos1(code,listref)
IN CONST ExprPtr code;
INOUT ExprPtr listref;
{
  struct sortinfos *data = (struct sortinfos *)alloctmp(sizeof(struct sortinfos));

  data->code = code;
  data->listref = listref;
  (void) CallCC(ACTION_NAME(sortinfos),(ExprPtr)data,true);
}

PRIVATE void sortInfos(code,listref)
IN CONST ExprPtr code;
INOUT ExprPtr listref;
{
  if (STKFULL(code))
    {
      sortInfos1(code,listref);
      return;
    }
again:
  /*if (code==NIL)
    return;*/
  switch (code->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV:
    case E_STACK: case E_EXTERNAL:
    case E_QUOTE: case E_VARREF: case E_EXCEPTION:
    case E_ARRAY_0: case E_DATACON:
    case E_VARINTRO: case E_PAT_ANY:
    case E_COMPILED_CLOSURE:
      break;
    case E_BYTECODE:
      /*!!!*/
      break;
    case E_DEBUGNAME:
      code = DNPAT(code);
      goto again;
    case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
      /* we go look inside interpreted closures */
      code = QENV(code);
      goto again;
    case E_REF:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_ARRAY_1:
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
    case E_PAT_CONS:
    case E_NOT:
      code = REFVAL(code);
      goto again;
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      code = CDR(code);
      goto again;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
    case E_MAPGET:
    case E_PAT_REVCONS: case E_PAT_REVTUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
    case E_ARRAY_2:
    case E_PAT_MAPADD:
    case E_PATCHOICE:
    case E_OVERWRITE: case E_UNDERWRITE:
    case E_PAT_AS:
      sortInfos(CAR(code),listref);
      code = CDR(code);
      goto again;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
    case E_ARRAY_GNRL:
#if STOP_AND_COPY_VECTORS
      sortInfosVec(&AVAL(code),listref);
#else
      sortInfosVec(AVAL(code),listref);
#endif
      break;
    case E_MAP:
    case E_MAPLET:
      (void) MapSweep(code,do_sortinfos,listref);
      break;
    case E_LET:
      sortInfosBindings(CDR(LETBIND(code)),listref);
      code = LETBODY(code);
      goto again;
    case E_POLYFUN:
      code = COUNTEDVAL(code); /* search in patterns */
      goto again;
    case E_MEMO_FN:
      code = MEMOFUN(code);
      goto again;
    case E_APPL: case E_GOTO:
      sortInfos(APPLFUN(code),listref);
      code = APPLARG(code);
      goto again;
    case E_GET:
      code = GETREF(code);
      goto again;
    case E_SET:
      sortInfos(SETREF(code),listref);
      code = SETVAL(code);
      goto again;
    case E_IF:
      sortInfos(TESTIF(code),listref);
      code = TESTFORK(code);
      goto again;
    case E_PROG:
    case E_AND:
    case E_OR:
    case E_EQUAL:
    case E_NEQUAL:
      sortInfos(PROGDO(code),listref);
      code = PROGTHEN(code);
      goto again;
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF: case E_MAP_OF:
    case E_MAP_UNDER_OF: case E_ALL_OF: case E_EXISTS_OF:
    case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP: case E_MAP_MAP:
    case E_MAP_UNDER_MAP: case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      sortInfos(CAR(OFFP(code)),listref);
      sortInfosDoms(CDR(OFDOM(code)),listref);
      code = CDR(OFFP(code));
      goto again;
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      sortInfos(CAR(WCOND(code)),listref);
      sortInfos(CDR(WCOND(code)),listref);
      code = WBODY(code);
      goto again;
    case E_HANDLE:
      sortInfos(HBODY(code),listref);
      code = HHANDLERS(code);
      goto again;
    case E_RAISE:
      code = RBODY(code);
      goto again;
    case E_DELAY:
      code = DCODE(code);
      goto again;
    case E_PACK:
    case E_PAT_PACK:
      code = PACKVAL(code);
      goto again;
    case E_DEBUGZONE: case E_DEBUGFUNC:
    case E_DZ_BRK_ENTRY: case E_DF_BRK_ENTRY:
    case E_DZ_BRK_EXIT: case E_DF_BRK_EXIT:
    case E_DZ_BRK_BOTH: case E_DF_BRK_BOTH:
      insertInfo(code,listref);
      code = DZCODE(code);
      goto again;
    case E_PROFILE:
      code = CAR(code);
      goto again;
    case E_PAT_CHECK:
      code = PATCHECKVAL(code);
      goto again;
    case E_TYPE_APPL: case E_NUMTYPE:
    case E_TYPEFX: case E_COUNTED:
    default:
      kinderror(message(415) /*"sortInfos"*/,code->kind);
      break;
    }
}

PRIVATE ExprPtr getMaxInfo(zlist,linep)
IN CONST ExprPtr zlist;
IN Boolean linep;
{
  ExprPtr next,args,info,maxinfo;
  TZone zone;

  if (zlist==NIL)
    {
      return NIL;
    }
  /* first cell in zlist must be a E_CONS */
  info = DZINFO(CAR(zlist));
  zone.line1 = ZONE(info)->line1;
  zone.pos1 = linep?0:ZONE(info)->pos1;
  args = ARGS(info);
  next = CDR(zlist);
  while (next!=NIL)
    {
      zlist = next;
      switch (zlist->kind)
	{
	case E_CONS:
	  next = CDR(zlist);
	  break;
	case E_MAPLET:
	  next = MAPLETY(zlist);
	  break;
	}
    }
  /* last cell in zlist must be a E_MAPLET */
  info = DZINFO(MAPLETX(zlist));
  zone.line2 = ZONE(info)->line2;
  zone.pos2 = ZONE(info)->pos2;
  if (linep && zone.pos2!=0)
    {
      zone.line2++;
      zone.pos2 = 0;
    }
  maxinfo = ZBuild(&zone,0,args);
  return KEEP(maxinfo);
}

#ifdef HAS_PROTOTYPES
PRIVATE void showLineNo(IN unsigned short line)
#else
PRIVATE void showLineNo(line)
IN unsigned short line;
#endif
{
  hfprintf(hstderr,"%5hu: ",line
#if OS!=MAC_OS
	   +1
#endif
	   );
}

#ifdef CODEPOINTS
PRIVATE void showDebugZones(listp,line,pos)
INOUT ExprPtr *listp;
IN COUNTER line,pos;
{
  char msg[16];
  ExprPtr info;
  COUNTER n;

  while (*listp!=NIL)
    {
      msg[0] = 0;
      switch ((*listp)->kind)
	{
	case E_CONS:
	  info = CAR(*listp);
	  if (ZONE(info)->line1==line && ZONE(info)->pos1==pos)
	    {
	      n = TYPE(info);
	      sprintf(msg,enterBreakFmt,(int)n);
	      *listp = CDR(*listp);
	      break;
	    }
	  return;
	case E_MAPLET:
	  info = CAR(*listp);
	  if (ZONE(info)->line2==line && ZONE(info)->pos2==pos)
	    {
	      n = TYPE(info);
	      sprintf(msg,exitBreakFmt,(int)n);
	      *listp = MAPLETY(*listp);
	      break;
	    }
	  return;
	}
      hfputs(msg,hstderr);
    }
}
#endif

PRIVATE ExprPtr getInfos(code)
IN CONST ExprPtr code;
{
  ExprPtr listref,list;
  ENTER;

  MAKECONS(NIL,NIL,listref);
  (void) KEEP(listref);
  if (code!=NIL)
    sortInfos(code,listref);
  list = CDR(listref);
  KEEPONLY(list);
  return list;
}

PRIVATE void showVal1(r)
IN CONST ExprPtr r;
{
  COUNTER save_maxlines;

  save_maxlines = crtab_maxlines;
  crtab_maxlines = IVAL(REFVAL(maxLinesRef));
  if (crtab_maxlines>0)
    {
      (void) dynPretty(r,do_printfile,(char *)hstderr,
		       0,IVAL(REFVAL(rightMarginRef)));
      endPretty(do_printfile,(char *)hstderr);
      hfputc('\n',hstderr);
    }
  crtab_maxlines = save_maxlines;
}

PRIVATE void showVal(val,type)
IN CONST ExprPtr val,type;
{
  ENTER;

  val = MakePair(E_PACK,val,type);
  (void) KEEP(val);
  showVal1(val);
  QUIT;
}

PRIVATE void showSource(info,val,status
#ifdef CODEPOINTS
			,zlist
#endif
			,fullp,rmargin)
IN CONST ExprPtr info;
IN CONST ExprPtr val;
IN COUNTER status;
#ifdef CODEPOINTS
IN CONST ExprPtr zlist;
#endif
IN Boolean fullp;
IN COUNTER rmargin;
{
  ExprPtr filename;
  ExprPtr type;
  /*char *blanks;*/
  INT len;
  hFILE *file;
  TZone *zone;
  unsigned short i,ii,line,goal,line1,line2/*,blen*/;
#ifdef CODEPOINTS
  unsigned short pos;
#endif
  int c;
  Boolean nodotsp,startp;

  if (status!=0)
  {
    hfprintf(hstderr,message(380) /*"stopped"*/,
	     message(380+status) /*"entry", "exit", or "break"*/);
    if (info!=NIL && status!=2)
      {
	filename = CAR(CDR(CDR(ARGS(info))));
	if (filename!=NIL)
	  {
	    hfprintf(hstderr,message(384) /*" at line %d in "*/,
		     (COUNTER)((status==2)?(ZONE(info)->line1
#if OS!=MAC_OS
					    +1
#endif
					    ):(ZONE(info)->line2
#if OS!=MAC_OS
					       +1
#endif
					       )));
	    hfputc('\"',hstderr);
	    (void) printSym(filename,do_printfile,(char *)hstderr);
	    hfputc('\"',hstderr);
	  }
      }
    hfputc('\n',hstderr);
  }
  if (status==2)
    {
      type = (info==NIL)?alpha:CAR(CDR(ARGS(info)));
      if (val!=specialMarker)
	showVal(val,type);
    }
  if (status!=0)
    hfputc('\n',hstderr);
  if (info!=NIL)
    {
      filename = CAR(CDR(CDR(ARGS(info))));
      zone = ZONE(info);
      if (filename!=NIL)
	{
	  len = symNameLen(filename);
	  {
	    char *name;

	    name = (char *)alloctmp(len+1);
	    memcpy(name,symName(filename),len);
	    name[len] = 0;
	    printZone((CHAR *)name,zone,hstderr,true);
	    hfputc('\n',hstderr);
	    file = hfopen(name,"r");
	    if (file==NULL)
	      {
		bmsg(message(413) /*"Source unavailable: '%s'\n"*/,name);
	      }
	  }
	  /*
	  blanks = message(442);
	  blen = strlen(blanks);
	  */
	  if (file!=NULL)
	    {
	      c = 0;
	      for (line=0, goal=zone->line1; line<goal; )
		{
		  c = hgetc(file);
		  if (c==-1)
		    break;
		  if (c=='\n')
		    line++;
		}
	      ii = 0;
	      if (c!=-1 && zone->pos1>0)
		{
		  if (zone->pos1>0)
		    hfputs("   ... ",hstderr);
		  for (i=0; i<zone->pos1; i++)
		    {
		      c = hgetc(file);
		      if (c==-1)
			break;
		      hfputc(c,hstderr);
		      ii++;
		    }
		  hfputc('\n',hstderr);
		}
	      if (c!=-1)
		{
		  showLineNo(line);
		  for (i=0; i<ii; i++)
		    {
		      hfputc(' ',hstderr);
		    }
		  if (fullp)
		  {
		    line1 = (unsigned short)-1;
		    line2 = 0;
		  }
		  else
		  {
		    line1 = zone->line1+3;
		    line2 = zone->line2-2;
		  }
		  nodotsp = true;
		  startp = false;
#ifdef CODEPOINTS
		  if (fullp)
		    pos = zone->pos1;
#endif
		  for (goal = zone->line2; line<goal; )
		    {
		      c = hgetc(file);
		      if (c==-1)
			break;
		      if (line<=line1 || line>=line2)
			{
			  if (startp)
			    {
			      showLineNo(line);
			      startp = false;
			    }
#ifdef CODEPOINTS
			  if (fullp)
			  {
			    showDebugZones(&zlist,line,pos);
			    pos++;
			  }
#endif
			  hfputc(c,hstderr);
			}
		      else if (nodotsp)
			{
			  nodotsp = false;
			  hfputs("\t...\n",hstderr);
			}
		      if (c=='\n')
		      {
			line++;
#ifdef CODEPOINTS
			if (fullp)
			  pos = 0;
#endif
			startp = true;
		      }
		    }
		  if (c!=-1)
		  {
		    if (zone->line1!=zone->line2)
		      i = 0;
		    for (; i<zone->pos2; i++)
		      {
			c = hgetc(file);
			if (c==-1)
			  break;
#ifdef CODEPOINTS
			if (fullp)
			{
			  showDebugZones(&zlist,line,pos);
			  pos++;
			}
#endif
			hfputc(c,hstderr);
		      }
		   }
		}
	      hfclose(file);
	    }
	  hfputc('\n',hstderr);
	}
    }
    hfflush(hstderr);
}

PRIVATE Boolean isDebugVp(vp)
IN CONST VecPtr vp;
{
  COUNTER i,n;
  ExprPtr e;

  if (vp->n & 1)
    return false;
  n = vp->n;
  for (i=1; i<n; i += 2)
    {
      e = vp->expr[i];
      if (e==NOEXPR)
	continue;
      if (e->kind==(CHAR)E_DEBUGNAME)
	/*if (e->kind==(CHAR)E_LET || e->kind==(CHAR)E_APPL || e->kind==(CHAR)E_GOTO)*/
	continue;
      return false;
    }
  return true;
}

PRIVATE COUNTER analyzeVp(env,level)
IN CONST ExprPtr env;
IN COUNTER level;
{
  ExprPtr info,dn;
  COUNTER nlocals,i;
  VecPtr vp;
  ENTER;

  vp = EVARS(env);
  nlocals = vp->n;
  if (isDebugVp(vp))
    {
      for (i=1; i<nlocals; i+=2)
	{
	  dn = EVARS(env)->expr[i];
	  if (dn!=NOEXPR)
	    {
	      info = COUNTEDVAL(DNINFO(dn));
	      addIdTypes(CAR(info),CDR(info),basisAssum);
	      if (info->kind==(CHAR)E_APPL) /* datatype constructor */
	        addConsTypes(CAR(info),CDR(info),basisAssum);
	      else remConsTypes(CAR(info),basisAssum);
	      if (info->kind!=(CHAR)E_LET) /* constructor */
	        setConstructor(CAR(info)); /* inform lexer */
	      /* there is no other case, since a constructor
		 name cannot be redefined in the same scope. */
	      addIdLocs(CAR(info),KONS(KINT(VARRANK(DNPAT(dn))),KINT(level)),basisTable);
	      QUIT;
	    }
	}
    }
  return nlocals;
}

PRIVATE COUNTER envHeight(env)
IN CONST ExprPtr env;
{
  ExprPtr ee;
  COUNTER n;

  for (ee=env, n=0; ee!=NIL; ee=ENEXT(ee), n++);
  return n;
}

PRIVATE COUNTER analyzeEnv(env)
IN CONST ExprPtr env;
{
  COUNTER level,n;
  ExprPtr *envs,*ep;
  ExprPtr ee;

#ifdef DEBUG
hfprintf(hstderr,"(f) env = 0x%x\n",env); hfflush(hstderr);
hfprintf(hstderr,"%x %x %x %x\n",((long *)env)[0],((long *)env)[1],((long *)env)[2],((long *)env)[3]); hfflush(hstderr);
#endif

  for (ee=env, n=0; ee!=NIL; ee=ENEXT(ee), n++);
#ifdef DEBUG
    {hfprintf(hstderr,"(loop) env = 0x%x\n",ee); hfflush(hstderr);
hfprintf(hstderr,"%x %x %x %x\n",((long *)ee)[0],((long *)ee)[1],((long *)ee)[2],((long *)ee)[3]); hfflush(hstderr);
}
hfputs("g\n",hstderr); hfflush(hstderr);
#endif

  envs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (ee=env, ep=envs; ee!=NIL; ee=ENEXT(ee))
    *ep++ = ee;
#ifdef DEBUG
hfputs("h\n",hstderr); hfflush(hstderr);
#endif
  for (level=0, ep=envs+n; --ep>=envs; level++)
    (void) analyzeVp(*ep,level);
#ifdef DEBUG
hfputs("i\n",hstderr); hfflush(hstderr);
#endif
  return n;
}

PRIVATE void createDebugEnv(env)
IN CONST ExprPtr env;
{
  COUNTER level;

#ifdef DEBUG
hfprintf(hstderr,"(d) env = 0x%x\n",env); hfflush(hstderr);
#endif
  level = analyzeEnv(env);
#ifdef DEBUG
hfputs("e\n",hstderr); hfflush(hstderr);
#endif
  C_LEVEL(basisTable) = C_ILEVEL(basisTable) = level-1;
  C_NLOCALS(basisTable) = EVARS(env)->n;
}

EXTERN int errno;

PRIVATE char *getline()
{
  COUNTER i,n;
  char *line,*s;
  int c;

  n = 128;
  line = (char *)MALLOC(n);
  for (i=0, s=line; c=hgetc(hstdin),c!='\n'; )
    {
      if (c==-1)
	{
#ifdef EINTR
	  if (errno==EINTR)
	    continue;
#endif
	  break;
	}
      ++i;
      if (i>=n)
	{
	  MARKFREE(line);
	  n *= 2;
	  line = MALLOC(n);
	  s = line+i-1;
	}
      *s++ = c;
    }
  if (c==-1 && i==0)
    { /* transform end-of-file seen at start of line into quit command */
      *s++ = 'q';
    }
  *s = 0;
  return line;
}

PRIVATE char *skipWord(s)
IN REGISTER CONST char *s;
{
  REGISTER char c;

  for (; c = *s, c!=0 && c!=' ' && c!='\t'; s++);
  return s;
}

PRIVATE char *skipBlank(s)
IN REGISTER CONST char *s;
{
  REGISTER char c;

  for (; c = *s, c==' ' || c=='\t'; s++);
  return s;
}

PRIVATE char *nextWord(s)
IN REGISTER CONST char *s;
{
  s = skipWord(s);
  s = skipBlank(s);
  return s;
}

PRIVATE void installEnvUp(up)
IN COUNTER up;
{
#ifdef BUGGY
  ExprPtr ll;
  VecPtr vp;
  COUNTER n;
#endif

  if (up==0)
    return;
  {
    struct applStkPtr p;

    startApplStk(&p);
    for (; --up!=0 && !endApplStk(&p); nextApplStk(&p));
    if (!endApplStk(&p))
      theEnvironment = applStkCallingEnv(&p);
  }
#ifdef BUGGY
  up *= NFRAMEEXPRS;
  n = (applStkTop-applStkTable);
  if (up<=n)
    theEnvironment = applStkTop[1-up];
  else
    for (ll=applStkList; up -= n, ll!=NIL; ll=ENEXT(ll))
      {
	vp = EVARS(applStkList);
	n = vp->n;
	if (up<=n)
	  {
	    theEnvironment = vp->expr[1-up];
	    break;
	  }
      }
#endif
}

PRIVATE ExprPtr readValueUp(s,type,up)
IN REGISTER CONST char *s;
IN ExprPtr type;
IN COUNTER up;
{
  ExprPtr val,saveenv;
  ENTER;

  saveenv = KEEP(theEnvironment);
  installEnvUp(up);
  val = readValue(s,type,createDebugEnv);
  theEnvironment = saveenv;
  if (val==NOEXPR)
    { QUIT; }
  else { KEEPONLY(val); }
  return val;
}

PRIVATE ExprPtr readValueByName(s,len,up,foundp)
IN REGISTER CONST char *s;
IN COUNTER len;
IN COUNTER up;
OUT Boolean *foundp;
{
  ExprPtr debugBasis,originalBasis;
  ExprPtr name,code;
  ENTER;

  originalBasis = basisNew(false);
  (void) KEEP(originalBasis);
  debugBasis = basisNew(true);
  (void) KEEP(debugBasis);
  basisInstall(debugBasis,true,false); /* install duplicated basis */
  installEnvUp(up); /* go up the call stack */
  createDebugEnv(theEnvironment); /* modify the basis to reflect current env */
  name = Kstring(len,(CHAR *)s);
  (void) KEEP(name);
  code = getValueByName(name,foundp,theEnvironment);
  basisInstall(originalBasis,true,false); /* reinstall original basis */
  KEEPONLY(code);
  return code;
}

PRIVATE ExprPtr getInfo(code,line,pos,after)
IN CONST ExprPtr code;
IN COUNTER line,pos;
IN Boolean after;
{
  ExprPtr list,info,res;
  ENTER;

  res = NIL;
  list = getInfos(code);
  for (; list!=NIL; list=CDR(list))
    {
      if (after)
	{
	  if (list->kind==(CHAR)E_CONS)
	    continue;
	}
      else
	{
	  if (list->kind==(CHAR)E_MAPLET)
	    continue;
	}
      info = DZINFO(CAR(list));
      if (ZONE(info)->line1>=line)
	{
	  if (ZONE(info)->line1>line || ZONE(info)->pos1>=pos)
	    {
	      res = CAR(list);
	      break;
	    }
	}
    }
  QUIT;
  return res;
}

PRIVATE ExprPtr inFn P((IN CONST ExprPtr polyfun,IN CONST ExprPtr code,
			IN CONST ExprPtr nestlist));

PRIVATE ExprPtr inFnPolyfun(polyfun,code,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr code;
IN CONST ExprPtr nestlist;
{
  ExprPtr e;
  ENTER;

  nestlist = CONS(code,nestlist);
  e = inFn(polyfun,COUNTEDVAL(code),nestlist);
  QUIT;
  return e;
}

#if STOP_AND_COPY_VECTORS
PRIVATE ExprPtr inFnVec(polyfun,vpp,nestlist)
IN CONST ExprPtr polyfun;
IN Vector *HEAP *vpp;
IN CONST ExprPtr nestlist;
{
  COUNTER i,n;
  ExprPtr e;

  for (i=0, n=(*vpp)->n; i<n; i++)
    {
      e = inFn(polyfun,(*vpp)->expr[i],nestlist);
      if (e!=falseValue)
	return e;
    }
  return falseValue;
}
#else
PRIVATE ExprPtr inFnVec(polyfun,vp,nestlist)
IN CONST ExprPtr polyfun;
IN CONST VecPtr vp;
IN CONST ExprPtr nestlist;
{
  COUNTER i,n;
  ExprPtr *ep,e;

  for (i=0, n=vp->n, ep=vp->expr; i<n; i++, ep++)
    {
      e = inFn(polyfun,*ep,nestlist);
      if (e!=falseValue)
	return e;
    }
  return falseValue;
}
#endif

MAPSWEEPFUNC(do_infn)
{
  struct infn *data;
  ExprPtr e;

  data = MAPDATA(struct infn *);
  e = inFn(data->polyfun,x,data->nestlist);
  if (e!=falseValue)
    {
      data->nestlist = e;
      return true;
    }
  e = inFn(data->polyfun,y,data->nestlist);
  if (e!=falseValue)
    {
      data->nestlist = e;
      return true;
    }
  return false;
}

PRIVATE ExprPtr inFnMap(polyfun,code,nestlist)
IN CONST ExprPtr polyfun,code;
IN CONST ExprPtr nestlist;
{
  struct infn *data = (struct infn *)alloctmp(sizeof(struct infn));

  data->polyfun = polyfun;
  data->nestlist = nestlist;
  if (MapSweep(code,do_infn,data))
    return data->nestlist;
  return falseValue;
}

PRIVATE ExprPtr inFnBindings(polyfun,bindlist,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr bindlist;
IN CONST ExprPtr nestlist;
{
  ExprPtr binding,e;

  while (bindlist!=NIL)
    {
      binding = CAR(bindlist);
      if (binding->kind==(CHAR)E_LET)
	{
	  ENTER;

	  nestlist = CONS(binding,nestlist);
	  e = inFnBindings(polyfun,CDR(LETBIND(binding)),nestlist);
	  if (e!=falseValue)
	    return e;
	  e = inFnBindings(polyfun,LETBODY(binding),nestlist);
	  if (e!=falseValue)
	    return e;
	  QUIT;
	}
      else
	{
	  e = inFn(polyfun,CDR(binding),nestlist);
	  if (e!=falseValue)
	    return e;
	}
      bindlist = CDR(bindlist);
    }
  return falseValue;
}

PRIVATE ExprPtr inFnLet(polyfun,code,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr code;
IN CONST ExprPtr nestlist;
{
  ENTER;

  nestlist = CONS(code,nestlist);
  nestlist = inFnBindings(polyfun,CDR(LETBIND(code)),nestlist);
  QUIT;
  return nestlist;
}

PRIVATE ExprPtr inFnDoms(polyfun,domlist,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr domlist;
IN CONST ExprPtr nestlist;
{
  ExprPtr e;

  while (domlist!=NIL)
    switch (domlist->kind)
      {
      case E_PAT_RECORD_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
	e = inFn(polyfun,CDR(CAR(domlist)),nestlist);
	if (e!=falseValue)
	  return e;
	domlist = CDR(domlist);
	break;
      case E_TUPLE_1:
	{
	  ENTER;

	  nestlist = CONS(domlist,nestlist);
	  e = inFnBindings(polyfun,REFVAL(domlist),nestlist);
	  QUIT;
	  if (e!=falseValue)
	    return e;
	  domlist = NIL;
	}
	break;
      }
  return falseValue;
}

DEF_PRIVATE_ACTION(infn)
{
  struct infn1 *data;

  data = (struct infn1 *)env;
  return inFn(data->polyfun,data->code,data->nestlist);
}

PRIVATE ExprPtr inFn1(polyfun,code,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr code;
IN CONST ExprPtr nestlist;
{
  struct infn1 *data = (struct infn1 *)alloctmp(sizeof(struct infn1));

  data->polyfun = polyfun;
  data->code = code;
  data->nestlist = nestlist;
  return CallCC(ACTION_NAME(infn),(ExprPtr)data,true);
}

PRIVATE ExprPtr inFn(polyfun,code,nestlist)
IN CONST ExprPtr polyfun;
IN CONST ExprPtr code;
IN CONST ExprPtr nestlist;
{
  ExprPtr e;

  if (STKFULL(code))
    return inFn1(polyfun,code,nestlist);
again:
  switch (code->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV:
    case E_STACK: case E_EXTERNAL:
    case E_QUOTE: case E_VARREF: case E_EXCEPTION:
    case E_ARRAY_0: case E_DATACON:
    case E_VARINTRO: case E_PAT_ANY:
    case E_COMPILED_CLOSURE:
      break;
    case E_BYTECODE:
      /*!!!*/
      break;
    case E_DEBUGNAME:
      code = DNPAT(code);
      goto again;
    case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
      /* we go look inside interpreted closures */
      code = CAR(QENV(code));
      if (code==polyfun)
	return nestlist;
      goto again;
    case E_REF:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_ARRAY_1:
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
    case E_PAT_CONS:
    case E_NOT:
      code = REFVAL(code);
      goto again;
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      code = CDR(code);
      goto again;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
    case E_MAPGET:
    case E_PAT_REVCONS: case E_PAT_REVTUPLE_2:
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
    case E_ARRAY_2:
    case E_PAT_MAPADD:
    case E_PATCHOICE:
    case E_OVERWRITE: case E_UNDERWRITE:
    case E_PAT_AS:
      e = inFn(polyfun,CAR(code),nestlist);
      if (e!=falseValue)
	return e;
      code = CDR(code);
      goto again;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
    case E_ARRAY_GNRL:
      e =
#if STOP_AND_COPY_VECTORS
	  inFnVec(polyfun,&AVAL(code),nestlist)
#else
	  inFnVec(polyfun,AVAL(code),nestlist)
#endif
	    ;
      if (e!=falseValue)
	return e;
      break;
    case E_MAP:
    case E_MAPLET:
      return inFnMap(polyfun,code,nestlist);
    case E_LET:
      e = inFnLet(polyfun,code,nestlist);
      if (e!=falseValue)
	return e;
      code = LETBODY(code);
      goto again;
    case E_POLYFUN:
      if (code==polyfun)
	return nestlist;
      return inFnPolyfun(polyfun,code,nestlist);
    case E_MEMO_FN:
      code = MEMOFUN(code);
      goto again;
    case E_APPL: case E_GOTO:
      e = inFn(polyfun,APPLFUN(code),nestlist);
      if (e!=falseValue)
	return e;
      code = APPLARG(code);
      goto again;
    case E_GET:
      code = GETREF(code);
      goto again;
    case E_SET:
      e = inFn(polyfun,SETREF(code),nestlist);
      if (e!=falseValue)
	return e;
      code = SETVAL(code);
      goto again;
    case E_IF:
      e = inFn(polyfun,TESTIF(code),nestlist);
      if (e!=falseValue)
	return e;
      code = TESTFORK(code);
      goto again;
    case E_PROG:
    case E_AND:
    case E_OR:
    case E_EQUAL:
    case E_NEQUAL:
      e = inFn(polyfun,PROGDO(code),nestlist);
      if (e!=falseValue)
	return e;
      code = PROGTHEN(code);
      goto again;
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF: case E_MAP_OF:
    case E_MAP_UNDER_OF: case E_ALL_OF: case E_EXISTS_OF:
    case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP: case E_MAP_MAP:
    case E_MAP_UNDER_MAP: case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      e = inFnPolyfun(polyfun,OFFP(code),nestlist);
      if (e!=falseValue)
	return e;
      return inFnDoms(polyfun,CDR(OFDOM(code)),nestlist);
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
    case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      e = inFn(polyfun,WCOND(code),nestlist);
      if (e!=falseValue)
	return e;
      code = WBODY(code);
      goto again;
    case E_HANDLE:
      e = inFn(polyfun,HBODY(code),nestlist);
      if (e!=falseValue)
	return e;
      code = HHANDLERS(code);
      goto again;
    case E_RAISE:
      code = RBODY(code);
      goto again;
    case E_DELAY:
      code = DCODE(code);
      goto again;
    case E_PACK:
    case E_PAT_PACK:
      code = PACKVAL(code);
      goto again;
    case E_DEBUGZONE: case E_DEBUGFUNC:
    case E_DZ_BRK_ENTRY: case E_DF_BRK_ENTRY:
    case E_DZ_BRK_EXIT: case E_DF_BRK_EXIT:
    case E_DZ_BRK_BOTH: case E_DF_BRK_BOTH:
      code = DZCODE(code);
      goto again;
    case E_PROFILE:
      code = CAR(code);
      goto again;
    case E_PAT_CHECK:
      code = PATCHECKVAL(code);
      goto again;
    case E_TYPE_APPL: case E_NUMTYPE:
    case E_TYPEFX: case E_COUNTED:
    default:
      kinderror(message(411) /*"inFn"*/,code->kind);
      break;
    }
  return falseValue;
}

MAPSWEEPFUNC(do_getfnname)
{
  struct getfnname *data;

  data = MAPDATA(struct getfnname *);
  if (y->kind==(CHAR)E_INT)
    {
      data->nest = inFn(data->polyfun,EVARS(data->topenv)->expr[IVAL(y)],NIL);
      if (data->nest!=falseValue)
	{
	  data->name = x;
	  return true;
	}
    }
  return false;
}

PRIVATE ExprPtr getFnName(polyfun,defaultname,nestp)
IN CONST ExprPtr polyfun;
IN CONST SymPtr defaultname;
OUT ExprPtr *nestp;
{
  ExprPtr env,val,next;
  VecPtr vp;
  COUNTER nlocals,i;

  if (polyfun==NIL)
    return NIL;
  for (env=theEnvironment; next=ENEXT(env), next!=NIL; env=next)
    {
      vp = EVARS(env);
      nlocals = vp->n;
      if (isDebugVp(vp))
	{
	  for (i=1; i<nlocals; i+=2)
	    {
	      ExprPtr dn;

	      dn = vp->expr[i];
	      if (dn!=NOEXPR)
		{
		  ExprPtr info;

		  info = COUNTEDVAL(DNINFO(dn));
		  val = vp->expr[VARRANK(DNPAT(dn))];
		  if (val!=NOEXPR)
		    {
		      *nestp = inFn(polyfun,val,NIL);
		      if (*nestp!=falseValue)
			return CAR(info);
		    }
		}
	    }
	}
    }
  /* if we could not find the function in local environments, then it
     is probably in the toplevel environment (where no debug information
     is ever put) */
  {
    struct getfnname *data = (struct getfnname *)alloctmp(sizeof(struct getfnname));

    data->polyfun = polyfun;
    data->topenv = env;
    if (MapSweep(C_IDS(basisTable),do_getfnname,data))
      {
	*nestp = data->nest;
	return data->name;
      }
  }
  return defaultname;
}

MAPSWEEPFUNC(do_showbreaks)
{
  struct showbreak **sbp;

  sbp = MAPDATA(struct showbreak **);
  (*sbp)->i = IVAL(x);
  (*sbp)->msg = CDR(y);
  ++*sbp;
  if (stopwanted)
    return true;
  return false;
}

PRIVATE int swpbrk(_sb1,_sb2)
IN CONST ANSI_CONST VOIDPTR _sb1;
IN CONST ANSI_CONST VOIDPTR _sb2;
{
  REGISTER struct showbreak *sb1 = (struct showbreak *)_sb1;
  REGISTER struct showbreak *sb2 = (struct showbreak *)_sb2;

  if (sb1->i<sb2->i)
    return -1;
  if (sb1->i>sb2->i)
    return 1;
  return 0;
}

PRIVATE void showBreaks()
{
  INT i,n;
  struct showbreak *sb,*sbp;

  n = MapCard(breakPoints);
  if (n==0)
    return;
  sb = (struct showbreak *)alloctmp(n*sizeof(struct showbreak));
  sbp = sb;
  (void) MapSweep(breakPoints,do_showbreaks,&sbp);
  qsort(sb,(long)n,(long)sizeof(struct showbreak),swpbrk);
  for (i=0, sbp=sb; i<n; i++, sbp++)
    {
      if (stopwanted)
	{
	  stopwanted = false;
	  break;
	}
      (void) printSym(sbp->msg,do_printfile,(char *)hstderr);
    }
}

PRINTFUNC(do_prtmax)
{
  struct prtstrmax *data;
  COUNTER len2,i;
  char *s;
  Boolean overflow;

  data = PRINTDATA(struct prtstrmax *);
  if (data->len<data->maxlen)
    {
      len2 = data->maxlen-data->len;
      if (overflow=(len>len2))
	len = len2;
      for (i=0, s=(char *)text; i<len; i++, s++)
	hfputc(*s,hstderr);
      data->len += len;
    }
}

PRIVATE Boolean showEnv(env,commap,lmargin)
IN CONST ExprPtr env;
IN Boolean commap;
IN COUNTER lmargin;
{
  VecPtr vp;
  COUNTER i,n;
  ExprPtr val,info,dn;
  struct prtstrmax *data = (struct prtstrmax *)alloctmp(sizeof(struct prtstrmax));

  data->maxlen = IVAL(REFVAL(rightMarginRef))-lmargin-6;
  vp = EVARS(env);
  if (isDebugVp(vp))
    {
      n = vp->n;
      for (i=1; i<n; i+=2)
	{
	  vp = EVARS(env);
	  dn = vp->expr[i];
	  if (dn==NOEXPR)
	    continue;
	  info = COUNTEDVAL(DNINFO(dn));
	  val = vp->expr[VARRANK(DNPAT(dn))];
	  if (val!=NOEXPR && info!=NIL)
	    {
	      if (commap)
		{
		  COUNTER j;

		  hfputs(",\n",hstderr);
		  for (j=0; j<lmargin; j++)
		    hfputc(' ',hstderr);
		}
	      commap = true;
	      printSym(LETBIND(info),do_printfile,(char *)hstderr);
	      hfputs(message(239) /*" = "*/,hstderr);
	      data->len = 0;
	      valPrint(val,LETBODY(info),do_prtmax,(char *)data,0);
	      if (data->len>=data->maxlen)
		hfputs(message(505) /*" ..."*/,hstderr);
	    }
	}
    }
  else
    hfputs(message(238) /*"?"*/,hstderr);
  return commap;
}

PRIVATE ExprPtr getZone(code)
IN CONST ExprPtr code;
{
  ExprPtr zlist;

  switch (code->kind)
    {
    case E_DEBUGZONE: case E_DEBUGFUNC:
    case E_DZ_BRK_ENTRY: case E_DF_BRK_ENTRY:
    case E_DZ_BRK_EXIT: case E_DF_BRK_EXIT:
    case E_DZ_BRK_BOTH: case E_DF_BRK_BOTH:
      return DZINFO(code);
    }
  zlist = getInfos(code);
  /* get code, and build maximal info from code */
  return getMaxInfo(zlist,false);
}

PRIVATE void showStack(up,mw,topinfo)
IN COUNTER up;
IN COUNTER mw;
IN CONST ExprPtr topinfo;
{
  struct applStkPtr p;
  ExprPtr code,info,fn,filename,nest,env;
  COUNTER i,j,n,lmargin;
  ExprPtr *envs;
  ENTER;

  up++;
  startApplStk(&p);
  i = 1;
  info = topinfo;
  if (up>mw/2) /* then scroll through the stack first */
    {
      n = up - mw/2;
      for (; i<n && !endApplStk(&p); nextApplStk(&p), i++)
	info = applStkZone(&p);
    }
  for (; !endApplStk(&p) && --mw>0; nextApplStk(&p), i++)
    {
      if (stopwanted)
	{
	  stopwanted = false;
	  return;
	}
      hfputc((i==up)?'>':' ',hstderr);
      hfprintf(hstderr,message(74) /*"%3d: "*/,i);
      code = applStkCode(&p);
      fn = getFnName(code,NIL,&nest);
      (void) KEEP(nest);
      if (fn==NIL)
	hfputs(message(235) /*"<fn>"*/,hstderr);
      else (void) printSym(fn,do_printfile,(char *)hstderr);
      /* info is the debugzone of the previous entry in ApplStk */
      if (info!=NIL)
	info = getZone(info);
      if (info!=NIL)
	{
	  filename = CAR(CDR(CDR(ARGS(info))));
	  if (filename!=NIL)
	    {
	      CHAR *buf;
	      COUNTER len;

	      hfputs(message(236) /*"    ;in "*/,hstderr);
	      len = SLEN(filename);
	      buf = (CHAR *)MALLOC(len+1);
	      memcpy(buf,STEXT(filename),len);
	      buf[len] = 0;
	      printZone(buf,ZONE(info),hstderr,false);
	      MARKFREE(buf);
	    }
	}
      info = applStkZone(&p);
      nest = CONS(code,nest);
      n = ListLength(nest);
      envs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
      for (env = applStkEnv(&p), j=0; nest!=NIL && env!=NIL; nest = CDR(nest))
	switch (CAR(nest)->kind)
	  {
	  case E_POLYFUN:
	    envs[j++] = env;
	    env = ENEXT(env);
	    break;
	  case E_LET:
	    if (IVAL(CAR(LETBIND(CAR(nest))))<0)
	      env = ENEXT(env);
	    if (env==NIL)
	      break;
	    /*fallthrough*/
	  default:
	    env = ENEXT(env);
	    break;
	  }
      for (; --j>=0; )
	{
	  hfputs(message(378) /*"\n        ("*/,hstderr);
	  lmargin = 9;
	  (void) showEnv(envs[j],false,lmargin);
	  hfputc(')',hstderr);
	}
      hfputc('\n',hstderr);
      hfflush(hstderr);
    }
  if (endApplStk(&p))
    {
      hfputc((i==up)?'>':' ',hstderr);
      hfprintf(hstderr,message(74) /*"%3d: "*/,i);
      hfputs(message(385) /*"toplevel.\n"*/,hstderr);
    }
  else
    hfputs(message(442) /*"   ...\n"*/,hstderr);
  QUIT;
}

struct Entry {
  char *str;
  char cmd;
};

PRIVATE struct Entry dbgCommands[] = {
  { "help", 'h' },
  { "step", 's' },
  { "next", 'n' },
  { "continue", 'c' },
  { "print", 'p' },
  { "break", 'b' },
#ifdef BREAKAFTER
  { "breakafter", 'B' },
  { "bafter", 'B' },
#endif
  { "erase", 'e' },
  { "showbreakpoints", 'S' },
  { "sbreakpoints", 'S' },
  { "where", 'w' },
  { "up", 'u' },
  { "down", 'd' },
  { "list", 'l' },
  { "quit", 'q' },
  { NULL, 0 }
};

PRIVATE char getCommand(s)
IN CONST char *s;
{
  struct Entry *entry;
  char *ss,*ts,c,cc;

  for (entry = dbgCommands; entry->str!=NULL; entry++)
    {
      ss = s;
      ts = entry->str;
      while (c = *ts++)
	{
	  cc = *ss++;
	  if (cc==' ' || cc=='\t' || cc==0)
	    return entry->cmd;
	  if (c!=cc)
	    break;
	}
      if (c==0 && (*ss==0 || *ss==' ' || *ss=='\t'))
	return entry->cmd;
    }
  return 0;
}

PRIVATE ExprPtr getCurrentCode()
{
  struct applStkPtr p;

  startApplStk(&p);
  return endApplStk(&p)?NIL:applStkCode(&p);
}

PRIVATE Boolean clrBrkAfter(code)
INOUT ExprPtr code;
{
  switch (code->kind)
    {
    case E_DF_BRK_EXIT: code->kind = E_DEBUGFUNC; return true;
    case E_DZ_BRK_EXIT: code->kind = E_DEBUGZONE; return true;
    case E_DF_BRK_BOTH: code->kind = E_DF_BRK_ENTRY; return true;
    case E_DZ_BRK_BOTH: code->kind = E_DZ_BRK_ENTRY; return true;
    }
  return false;
}

PRIVATE Boolean setBrkAfter(code)
INOUT ExprPtr code;
{
  switch (code->kind)
    {
    case E_DEBUGFUNC: code->kind = E_DF_BRK_EXIT; return true;
    case E_DEBUGZONE: code->kind = E_DZ_BRK_EXIT; return true;
    case E_DF_BRK_ENTRY: code->kind = E_DF_BRK_BOTH; return true;
    case E_DZ_BRK_ENTRY: code->kind = E_DZ_BRK_BOTH; return true;
    }
  return false;
}

PRIVATE Boolean clrBrk(code)
INOUT ExprPtr code;
{
  switch (code->kind)
    {
    case E_DF_BRK_ENTRY: code->kind = E_DEBUGFUNC; return true;
    case E_DZ_BRK_ENTRY: code->kind = E_DEBUGZONE; return true;
    case E_DF_BRK_BOTH: code->kind = E_DF_BRK_EXIT; return true;
    case E_DZ_BRK_BOTH: code->kind = E_DZ_BRK_EXIT; return true;
    }
  return false;
}

PRIVATE Boolean setBrk(code)
INOUT ExprPtr code;
{
  switch (code->kind)
    {
    case E_DEBUGFUNC: code->kind = E_DF_BRK_ENTRY; return true;
    case E_DEBUGZONE: code->kind = E_DZ_BRK_ENTRY; return true;
    case E_DF_BRK_EXIT: code->kind = E_DF_BRK_BOTH; return true;
    case E_DZ_BRK_EXIT: code->kind = E_DZ_BRK_BOTH; return true;
    }
  return false;
}

PUBLIC ExprPtr debugger(code,val,status)
IN CONST ExprPtr code;
IN CONST ExprPtr val;
IN COUNTER status;
{
  ExprPtr info;
  char *buf,*s,c;
  COUNTER up,maxup;
  Boolean cont;
  char *defaultfuncname = NULL;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  ENTER;

  if (debugMode) /* occurs when we ask to evaluate an expression from
		    inside the debugger. */
    return val;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = getApplStk();
  theStackFrames = (gcStackView *)&_x;

  maxup = envHeight(theEnvironment)-1;
  debugMode = true;
  up = 0;
  stepMode = STEP_NO;
  resumeStepMode = STEP_NO; /* so that, if we break while executing a 'n' command,
			       the 'n' command is forgotten, i.e. it won't break on
			       exit of something. */
  if (code==NIL)
    info = NIL;
  else
    {
      info = DZINFO(code);
      SETQ(stepFile,CAR(CDR(CDR(ARGS(info)))));
      stepLine = ZONE(info)->line1;
      stepPos = ZONE(info)->pos1;
    }
  showSource(info,val,status
#ifdef CODEPOINTS
	     ,NIL
#endif
	     ,false,IVAL(REFVAL(rightMarginRef))-9);
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputc('\n',coreTraceFile);
      hfflush(coreTraceFile);
    }
#endif
  for (cont = true; cont; )
    {
      QUIT;
      setApplStk(_x.p[0]);
      hfputs(message(90) /*"(debug) "*/,hstderr);
      hfflush(hstderr);
      buf = getline();
      stopwanted = false;
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	{
	  hfputs(buf,coreTraceFile);
	  hfputc('\n',coreTraceFile);
	  hfflush(coreTraceFile);
	}
#endif
      s = buf;
      s = skipBlank(s);
again:
      switch (c = getCommand(s))
	{
	case 'c': /* continue evaluation */
	  s = nextWord(s);
	  if (*s!=0)
	    {
	      ExprPtr r;

	      r = readValueUp(s,
			      (status!=0 && info!=NIL)?CAR(CDR(ARGS(info))):NOEXPR,
			      up);
	      if (r==specialMarker)
		break;
	      val = CAR(r);
	    }
	  cont = false;
	  break;
	case 'p': /* print value */
	  s = nextWord(s);
	  if (*s!=0)
	    {
	      ExprPtr r;

	      r = readValueUp(s,NOEXPR,up);
	      if (r==specialMarker)
		break;
	      showVal1(r);
	    }
	  break;
	case 'S': /* show breakpoints */
	  showBreaks();
	  break;
	case 's': /* single-step (enter procedures, too) */
	  stepMode = STEP_STEP;
	  cont = false;
	  break;
	case 'n': /* single-step (don't enter procedures) */
	  stepMode = STEP_NEXT;
	  cont = false;
	  break;
	case 'q': /* quit evaluation, and return to toplevel */
	  debugMode = false;
	  Toplevel(ione); /* with an error code, so that
			     declarations are not installed */
	  break;
	case 'u': /* up the call stack */
	  s = nextWord(s);
	  if (*s!=0)
	    up += atoi(s);
	  else up++;
stkmove:
	  if (up<=0)
	    up = 0;
	  if (up>=maxup)
	    up = maxup;
	  hfprintf(hstderr,message(796) /*"current depth in call stack: %d.\n"*/,up+1);
	  break;
	case 'd': /* down the call stack */
	  s = nextWord(s);
	  if (*s!=0)
	    up -= atoi(s);
	  else up--;
	  goto stkmove;
	case 'b': /* set a breakpoint */
#ifdef BREAKAFTER
	case 'B':
#endif
	  {
	    Boolean after;
	    COUNTER line,pos;
	    ExprPtr code2,defaultname;

	    after = (c=='B'); /* break after */
	    line = pos = after?MAX_SIGNED_INT:0;
	    code2 = getCurrentCode();
	    defaultname = NIL;
	    if (defaultfuncname!=NULL)
	      {
		Boolean found;

		code2 = readValueByName(defaultfuncname,strlen(defaultfuncname),up,&found);
		if (found)
		  {
		    defaultname = Kstr(defaultfuncname);
		    (void) KEEP(defaultname);
		  }
	      }
	    /* by default, breakpoints are taken in the current
	       code (illegal if code==NIL) */
	    s = nextWord(s);
bnext:
	    switch (*s++)
	      {
	      case 0:
		break;
	      case ' ': case '\t':
		goto bnext;
	      case '0': case '1': case '2': case '3': case '4':
	      case '5': case '6': case '7': case '8': case '9': /* line number */
		line = atoi(s-1);
#if OS!=MAC_OS
		line--;
#endif
bskipdigs:
		for (; c = *s, c>='0' && c<='9'; s++);
		goto bnext;
	      case ':': /* position number */
		pos = atoi(s);
		goto bskipdigs;
	      case '@': /* function name */
		{
		  CHAR *sbeg;
		  Boolean found;

		  sbeg = (CHAR *)s;
		  s = skipWord(s);
		  defaultname = Kstring(s-(char *)sbeg,sbeg);
		  (void) KEEP(defaultname);
		  code2 = readValueByName(sbeg,s-(char *)sbeg,up,&found);
		  if (!found)
		    {
		      bmsg(message(419) /*"no such function.\n"*/);
		      goto berr;
		    }
		}
		goto bnext;
	      }
	    {
	      ExprPtr code,info,x,cmdline,y;
	      COUNTER l,p;
	      CHAR *msg,*s;
	      ExprPtr fn;
	      INT len;

	      code = getInfo(code2,line,pos,after);
	      if (code==NIL || (line!=0 && line!=MAX_SIGNED_INT &&
				line!=(after?ZONE(DZINFO(code))->line2:
				       ZONE(DZINFO(code))->line1)))
		{
		  bmsg(message(422) /*"no such breakpoint\n"*/);
		  goto berr;
		}
	      info = DZINFO(code);
	      if (after) /* useless, as we can't set any breakpoints on exit */
		{
		  (void) setBrkAfter(code);
		  /*TYPE(info) |= DZ_ACTIVE_ON_EXIT;*/
		  SETCDR(CAR(ARGS(info)),fntrue);
		  l = ZONE(info)->line2;
		  p = ZONE(info)->pos2;
		}
	      else
		{
		  (void) setBrk(code);
		  /*TYPE(info) |= DZ_ACTIVE_ON_ENTRY;*/
		  SETCAR(CAR(ARGS(info)),fntrue);
		  l = ZONE(info)->line1;
		  p = ZONE(info)->pos1;
		}
	      /* now, record breakpoint and return an identifier for it */
	      {
		ExprPtr nest;

		fn = getFnName(code2,defaultname,&nest);
	      }
	      len = (fn==NIL)?0:SLEN(fn);
	      msg = (CHAR *)MALLOC(len+32);
	      sprintf((char *)msg,message(797) /*"[%ld] b%c @"*/,++breakPointCount,after?'a':' ');
	      s = msg+strlen((char *)msg);
	      if (fn==NIL)
		{
		  strcpy(s,message(235) /*"<fn>"*/);
		  s += strlen(s);
		}
	      else
		{
		  memcpy(s,STEXT(fn),len);
		  s += len;
		}
	      sprintf((char *)s,message(798) /*" %d:%d\n"*/,l
#if OS!=MAC_OS
		      +1
#endif
		      ,p);
	      hfputs(msg,hstderr);
	      x = KINT(breakPointCount);
	      (void) KEEP(x);
	      cmdline = Kstr(msg);
	      (void) KEEP(cmdline);
	      MARKFREE(msg);
	      y = Kpair(after?E_TUPLE_2:E_CONS,code,cmdline);
	      (void) KEEP(y);
	      setMapPut(x,y,&breakPoints);
	    }
	  }
berr:
	  break;
	case 'e':
	  s = nextWord(s);
	  if (*s!=0)
	    {
	      INT i;
	      ExprPtr n,code;
	      Boolean found;

	      i = atoi(s);
	      n = KINT(i);
	      code = MapGet(n,breakPoints,&found);
	      if (found)
		{
		  if (code->kind==(CHAR)E_CONS) /* before */
		    clrBrk(CAR(code));
		  else /* after */
		    clrBrkAfter(CAR(code));
		  setMapRemove(n,&breakPoints);
		  hfprintf(hstderr,message(75) /*"breakpoint [%d] deleted.\n"*/,i);
		}
	      else
		{
		  hfputs(message(191) /*"no such breakpoint.\n"*/,hstderr);
		}
	    }
	  break;
	case 'w': /* where (show stack) */
	  s = nextWord(s);
	  {
	    COUNTER mw;

	    if (*s!=0)
	      mw = atoi(s);
	    else mw = maxWhere;
	    showStack(up,mw,code);
	  }
	  break;
	case 'l': /* list */
	  s = nextWord(s);
	  {
	    ExprPtr code2,zlist;

	    code2 = getCurrentCode();
	    /* by default, we list the current code (illegal if code==NIL) */
	    if (*s!=0)
	      {
		char *send;
		Boolean found;
		COUNTER len;

		send = skipWord(s);
		len = send-s;
		code2 = readValueByName(s,len,up,&found);
		if (!found)
		  {
		      bmsg(message(419) /*"no such function.\n"*/);
		      goto lerr;
		  }
		if (defaultfuncname!=NULL)
		  MARKFREE(defaultfuncname);
		defaultfuncname = (char *)MALLOC(len+1);
		memcpy(defaultfuncname,s,len);
		defaultfuncname[len] = 0;
	      }
	    zlist = getInfos(code2);
	    /* get code, and build maximal info from code */
	    code2 = getMaxInfo(zlist,true);
	    if (code2==NIL)
	      bmsg(message(420) /*"no information on this function.\n"*/);
	    else
	      showSource(code2,NIL,0
#ifdef CODEPOINTS
			 ,zlist
#endif
			 ,true,IVAL(REFVAL(rightMarginRef))-9);
	  }
lerr:
	  break;
	case 'h': /* help */
help:
	  hfputs(message(795) /*"commands: h(elp), s(tep), n(ext), c(ont) [result], p(rint) [value],\n  b(reak) [@function][line][:pos], e(rase) [bkpt#], sb(show bkpts),\n  w(here), u(p) [#levels], d(own) [#levels], l(ist) [function], q(uit)\n"*/,hstderr);
	  break;
	default:
	  hfputs(message(794) /*"unrecognized debugger command.\n"*/,hstderr);
	  goto help;
	}
      MARKFREE(buf);
    }
  debugMode = false;
  if (defaultfuncname!=NULL)
    MARKFREE(defaultfuncname);
  theStackFrames = _x.next;
  QUIT;
  return val;
}

DEF_ACTION(break)
{
  return Break(arg);
}
