/* Garbage collector.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "bc.h"
#include "type.h"
#include "compile.h"
#include "gram.h"
#include "expr.h"
#include "special.h"
#include "cons.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "errproto.h"
#include "memproto.h"
#include "stkproto.h"
#ifdef THINK_C
#include <unix.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
/*#include <sys/types.h> already included in timing.h*/
#include <sys/stat.h>
#include <sys/errno.h>
#include <unistd.h>
#else
#include <errno.h>
#endif
#include <fcntl.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include <math.h>
#ifdef HAS_PROTOTYPES
#include "sgc.h"
#endif

#define __SEG__ GC

#undef DEBUG
#undef DBG

#ifdef DEBUG
#if !SPECIAL_MALLOC
#include <malloc.h>
#endif
#endif

#define GENERATIONS

PUBLIC ExprPtr externalKeepList;
PUBLIC gcStackView *theStackFrames=(gcStackView *)0L;
PUBLIC gcStackView *theHeapFrames=(gcStackView *)0L;

PUBLIC long gc_utime=0L,gc_stime=0L,gc_rtime=0L;
PUBLIC long gc_count=0L;
EXTERN Boolean initializing;

PUBLIC SHORT generation_count=0;
EXTERN unsigned long gc_hash_cost_offset[2];

PUBLIC ExprPtr writeBarrier;
/* list (through ELINK) of objects containing fields
   that have been modified since the last generation cutoff;
   end of list is marked by NIL (NOEXPR would introduce a bug
   when trying to put the last item in the barrier into the
   barrier again); actually, to speed up the test against NIL,
   we don't test for NIL but for NOEXPR, ensuring that NIL
   is the last element in the barrier (to do this, ELINK(NIL)==NOEXPR,
   see expr.c/InitExprs()) */
EXTERN COUNTER globLimit; /* the starting index for sweeping the global environment */
PUBLIC Boolean itBarrier=false;
/* true if the ML variable 'it' has been redefined since
   the last generation cutoff */

EXTERN SymPtr lexId,lexPath;
EXTERN ExprPtr currentFile;
EXTERN ExprPtr lexEnv, /* maps ids to precedences as E_INT */
	consEnv; /* sets of constructors */
#if MODULE_SYSTEM & STD_ML_MODULE
EXTERN ExprPtr strEnv,sigEnv;
#endif
EXTERN ExprPtr keepLexEnv,keepConsEnv;
#if MODULE_SYSTEM & STD_ML_MODULE
EXTERN ExprPtr keepStrEnv,keepSigEnv;
#endif
#if FLOAT_FORMAT==IEEE754
EXTERN ExprPtr *NaNs;
#endif
#if DEBUGGING
#if !TRANSLATOR
EXTERN ExprPtr currentCallingEnv;
#endif
#endif
EXTERN ExprPtr stringType,numType,boolType,exnType,dynType,powerType,timeType;
EXTERN SymPtr itSym,it2Sym,
  stringsym,stringSym,listsym,listSym,
  refsym,refSym,funSym,mapSym,arraysym,arraySym,
  tablesym,tableSym,
  boolsym,boolSym,exnsym,exnSym,dynsym,dynSym,
  promisesym,promiseSym,contsym,contSym,
  starSym,unionSym,powerSym,concSym,unitsym,
  outstreamsym,instreamsym,writesym,readsym,
  readlinesym,flushsym,tellsym,seeksym,advancesym,seekendsym;
EXTERN SymPtr alpha,alphaeq,_alpha,beta,_beta,betaeq,gama,gamaeq,
  numalpha,numbeta;
EXTERN ExprPtr num_alpha,num_beta;
EXTERN ExprPtr basisAssum,basisState,basisTable,basisEnv;
EXTERN MapPtr mapgetCache_e,mapgetCache_m,mapgetCache_r;
EXTERN ExprPtr divmodCache_a,divmodCache_d,divmodCache_m,
	divmodCache_dm;
EXTERN ExprPtr parseMarkerStack;
EXTERN ExprPtr specialMarker;
EXTERN ExprPtr theTypes, theBindings;

PUBLIC ExprPtr ProtectedListAgainstGC;
PUBLIC ExprPtr ProtectedTable[NPROTECTED];
PUBLIC ExprPtr *ProtectedTableTop=ProtectedTable;
PUBLIC ExprPtr ProtectedAgainstGC;
PUBLIC Boolean inGC=false;

EXTERN ExprPtr mapStack[];
EXTERN ExprPtr *mapStackTop;

#if PROFILING
EXTERN ExprPtr profileExprs;
#endif

EXTERN unsigned long nSharedKtuple2, nSharedKons, nSharedMapletKons, nSharedMapKons,
  nSharedKcomplex, nSharedKtypeAppl, nSharedKtypeFx, nSharedKrecord1;
EXTERN unsigned long nSharedData, nSharedInts, nSharedReals, nSharedArrays,
  nSharedStrings;

PUBLIC INT liveCells[2][MAX_GENERATIONS],nYoungCells[2];
PUBLIC INT freedStringBytes, freedPatCheckBitBytes, freedStackBytes, freedVectorBytes;
PUBLIC INT freedExternals;
EXTERN INT stackAlloced;
EXTERN INT alloctmpBytes;

EXTERN EXTENDED growFactor;
#if STOP_AND_COPY_VECTORS
EXTERN EXTENDED copySlowdown;
EXTERN EXTENDED scGrowFactor,minScGrowFactor,maxScGrowFactor;
#endif

#if IMPERATIVE_TYPES==EFFECT_TYVAR
EXTERN ExprPtr theFx; /* contains the computed effect of the expression */
#endif

EXTERN ExprPtr nomemException;

#ifdef XLATE
EXTERN MapPtr bcXlateMap;
#endif

#ifdef DEBUG
#if !SPECIAL_MALLOC
struct mallinfo theMi;
#endif

INT stackFreed = 0;
#endif

PUBLIC NORETURN void nomem()
{
  unmarkBucketLists();
  markDataHook = NULL;
  if (nomemException!=NOEXPR)
    {
#ifdef DEBUG
#if SPECIAL_MALLOC
      hfprintf(hstderr,"TotalMem = %ld, MaximumMem = %ld.\n",TotalMem(),MaximumMem());
      hfflush(hstderr);
#else
      theMi = mallinfo();
#endif
#endif
      (void) Error(nomemException); /* ignore the result, since there cannot
				       be any. This is because Error() never enters the
				       debugger on lack of memory */
    }
  berror(message(405));
}

PUBLIC ExprPtr Protect(e) /* called when ProtectedTable[] is full */
IN REGISTER CONST ExprPtr e;
{
  ExprPtr aux;
  VecPtr vp;

  if (!initializing) /* we don't protect anything during initialization */
    {
      ProtectedAgainstGC = e;
      CHECKHETEROFREE();
      scAlloc(*(char **)&vp,VECSIZE(NPROTECTED));
      vp->n = NPROTECTED;
      memcpy(vp->expr,ProtectedTable,sizeof(ProtectedTable));
      MAKEENV1_NOGC(vp,ProtectedListAgainstGC,aux);
      ProtectedListAgainstGC = aux;
      ProtectedTable[0] = e;
      ProtectedTableTop = ProtectedTable+1;
      ProtectedAgainstGC = NIL;
    }
  return e;
}

PUBLIC void EndProtect(save)
IN REGISTER CONST GCMarker *save;
{
  REGISTER ExprPtr aux;
  ExprPtr *pTop;
  REGISTER ExprPtr pList;
  VecPtr vp;

  pList = save->pList;
  pTop = save->pTop;
  if (pList!=ProtectedListAgainstGC)
    {
      /* we don't HITBARRIER(), because ProtectedListAgainstGC
	 is always swept by GC() */
      while (aux = ENEXT(ProtectedListAgainstGC), aux!=pList)
	{
	  ProtectedListAgainstGC = aux;
	}
      vp = EVARS(ProtectedListAgainstGC);
      safememcpy(ProtectedTable,vp->expr,vp->n*sizeof(ExprPtr));
      /*ProtectedTableTop = ProtectedTable+vp->n;*/
      ProtectedListAgainstGC = aux;
    }
  ProtectedTableTop = pTop;
}

PUBLIC void EndProtectExcept(save,expr)
IN REGISTER CONST GCMarker *save;
IN REGISTER CONST ExprPtr expr;
{
  VecPtr vp;
  ExprPtr aux;

  ProtectedAgainstGC = expr;
  EndProtect(save);
  if (!initializing) /* we don't protect anything during initialization */
    {
      if (ProtectedTableTop<ProtectedTable+NPROTECTED)
	{
	  *ProtectedTableTop++ = expr;
	}
      else
	{
	  CHECKHETEROFREE();
	  scAlloc(*(char **)&vp,VECSIZE(NPROTECTED));
	  vp->n = NPROTECTED;
	  memcpy(vp->expr,ProtectedTable,sizeof(ProtectedTable));
	  MAKEENV1_NOGC(vp,ProtectedListAgainstGC,aux);
	  ProtectedListAgainstGC = aux;
	  ProtectedTable[0] = expr;
	  ProtectedTableTop = ProtectedTable+1;
	}
    }
  ProtectedAgainstGC = NIL;
}

PUBLIC ExprPtr getProtectWrapper()
{
  COUNTER n;
  VecPtr vp;
  ExprPtr aux;

  if (ProtectedTable==ProtectedTableTop)
    return ProtectedListAgainstGC;
  n = ProtectedTableTop-ProtectedTable;
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  safememcpy(vp->expr,ProtectedTable,n*sizeof(ExprPtr));
  MAKEENV1_NOGC(vp,ProtectedListAgainstGC,aux);
  ProtectedListAgainstGC = aux;
  ProtectedTableTop = ProtectedTable;
  return aux;
}

PUBLIC void setProtectWrapper(wrap,keep)
IN CONST ExprPtr wrap;
IN CONST ExprPtr keep;
{
  GCMarker dummy;

  dummy.pList = ProtectedListAgainstGC;
  dummy.pTop = ProtectedTable;
  EndProtect(&dummy);
  SETQ(ProtectedListAgainstGC,wrap);
  ProtectedTable[0] = keep;
  ProtectedTableTop = ProtectedTable+1;
}

PRIVATE ExprPtr *gcMarkStack,*gcMarkStackEnd,*gcMarkStackTop;
PRIVATE COUNTER nOverflows=0;

PRIVATE char *gcFlushName(n)
IN COUNTER n;
{
  static char buf[64];
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#define FMT "/tmp/HimMLGC%d%%%d"
#define ID getpid()
#else
#if OS==AMIGA_OS
#define FMT "T:HimMLGC%d%%%d"
#define ID (int)FindTask(NULL)
#else
#define FMT "HimMLGC%d%%%d"
#define ID (int)gcMarkStack
  /* this is correct since no two processes on the mac
      may have data at the same address */
#endif
#endif

  sprintf(buf,FMT,n,ID);
  return buf;
#undef ID
#undef FMT
}

PRIVATE void flushGCMark()
{
  char *name;
  int fd,res,len;

  name = gcFlushName(++nOverflows);
  fd = open(name,O_WRONLY | O_CREAT | O_TRUNC,0666);
  if (fd<0)
    nomem();
  len = GCMARK_STACK_SIZE*sizeof(ExprPtr);
  res = write(fd,(char *)gcMarkStack,len);
  if (res!=len)
    nomem();
  res = close(fd);
  if (res!=0)
    nomem();
  gcMarkStackTop = gcMarkStack;
}

PRIVATE void unflushGCMark()
{
  char *name;
  int fd,res,len;

  name = gcFlushName(nOverflows--);
  fd = open(name,O_RDONLY,0666);
  if (fd<0)
    nomem();
  len = GCMARK_STACK_SIZE*sizeof(ExprPtr);
  res = read(fd,(char *)gcMarkStack,len);
  if (res!=len)
    nomem();
  res = close(fd);
  if (res!=0)
    nomem();
  (void) unlink(name);
  gcMarkStackTop = gcMarkStackEnd;
}

#define pushGCMark(arg) if (gcMarkStackTop==gcMarkStackEnd) flushGCMark(); *gcMarkStackTop++ = (arg)
#define gcMarkEmpty() (gcMarkStackTop==gcMarkStack && nOverflows==0)
#define popGCMark() ((gcMarkStackTop==gcMarkStack)?(unflushGCMark(),*--gcMarkStackTop):(*--gcMarkStackTop))
#define GCMARK1(arg) pushGCMark(arg)

PRIVATE void pushGCVector(ep,epend)
IN REGISTER ExprPtr *ep,*epend;
{
  REGISTER ExprPtr *epp,*eppend,e;

  epp = gcMarkStackTop;
  eppend = gcMarkStackEnd;
  while (ep<epend)
    {
      e = *ep++;
      if (e==NOEXPR)
	continue;
      if (epp==eppend)
	{
	  gcMarkStackTop = epp;
	  flushGCMark();
	  epp = gcMarkStackTop;
	}
      *epp++ = e;
    }
  gcMarkStackTop = epp;
}

PRIVATE void GCmark1 P((IN ExprPtr e));

PRIVATE void markGCVector(n,ep)
IN COUNTER n;
IN CONST REGISTER ExprPtr *ep;
{
  REGISTER ExprPtr *epend;

  epend = ep+n;
  if (STKFULL(n))
    pushGCVector(ep,epend);
  else
    {
      REGISTER ExprPtr e;

      for (; ep<epend; )
	{
	  e = *ep++;
	  if (e==NOEXPR)
	    continue;
	  GCmark1(e);
	}
    }
}

#define GCmarkVector(v) markGCVector(v->n,v->expr)

PRIVATE void GCmarkStackView(sv)
IN CONST gcStackView *sv;
{
  while (sv!=(gcStackView *)0L)
    {
      markGCVector(sv->n,(ExprPtr *)sv->p);
      sv = sv->next;
    }
}

PRIVATE void GCmarkStackViewOff(sv,chunk)
IN CONST gcStackView *sv;
IN CONST StackChunk *chunk;
{
  char *contents;

  if (sv==(gcStackView *)0L)
    return;
  contents = chunk->contents;
  if (contents==(char *)0L || contents==(char *)1L)
    {
      GCmarkStackView(sv);
      return;
    }
  {
    REGISTER INT off;

#ifdef ASCENDING_STACK
    off = contents-chunk->thread->highMark;
#else
    off = contents-chunk->top;
#endif
    while (true)
      {
	sv = (gcStackView *)(((char *)sv) + off);
	markGCVector(sv->n,(ExprPtr *)sv->p);
	sv = sv->next;
	if (sv==(gcStackView *)0L)
	  break;
      }
  }
}

PRIVATE void GCMark1(e)
IN REGISTER ExprPtr e;
{
  GCMARK1(e);
}

PUBLIC doExtDesc doExtMark = {
  X_MARK,
  GCMark1
};

PRIVATE void GCmarkExcHandler(exc)
IN REGISTER CONST excHandler *exc;
{
  for (; exc!=NULL; exc = exc->next)
    switch (exc->tag)
      {
      case X_HANDLER:
	pushGCMark(exc->code);
	/*fallthrough*/
      case X_PROF:
	pushGCMark(exc->data);
	break;
      case X_CONT:
	pushGCMark(exc->data); /* push continuation on GC stack */
	/* We don't mark it directly, becase we would use too much
	   C stack to do this. */
	break;
      }
}

PRIVATE void GCmarkByteCode(bc)
IN CONST byteCodeChunk *bc;
{
  markGCVector(bc->nConsts,bc->consts);
}

PRIVATE void GCmark1(e)
IN REGISTER ExprPtr e;
{
again:
  if (/*e!=NIL && */!GC_MARKED(e))
    switch (e->kind)
      {
      case E_FREE:
	break;
      case E_NIL:
      case E_INT:
      case E_REAL:
	GC_MARK(e);
	break;
      case E_STRING:
	GC_MARK(e);
#if STOP_AND_COPY_VECTORS
	if (SLEN(e)>(sizeof(struct MedCell)))
	  scMARK(STEXT(e),SLEN(e));
	/* note that no string is mutable, so that
	   old generation strings will never be copied
	   in stop-and-copy. */
#endif
	break;
      case E_COMPLEX:
      case E_CONS: case E_UNSHARED_CONS:
      case E_PAT_REVCONS: case E_PATCHOICE:
      case E_ARRAY_2:
      case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
      case E_PAT_REVTUPLE_2:
      case E_DATA:
      case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
      case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
      case E_PAT_RECORD_MAPADD:
      case E_RECORD_1: case E_UNSHARED_RECORD_1:
      case E_MAP: case E_MAPGET:
      case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
      case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      case E_RMU1: case E_UNSHARED_RMU1: case E_RMUX: case E_UNSHARED_RMUX:
      case E_LET:
      case E_TYPE_APPL:
      case E_APPL: case E_GOTO: case E_N_APPL:
      case E_PAT_AS: case E_PAT_MAPADD:
      case E_SET:
      case E_IF: case E_PROG: case E_AND: case E_OR:
      case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
      case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
      case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF:
      case E_MAP_OF: case E_MAP_UNDER_OF: case E_ALL_OF:
      case E_EXISTS_OF: case E_SOME_OF: case E_UNSHARED_SOME_OF:
      case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
      case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
      case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP:
      case E_MAP_MAP: case E_MAP_UNDER_MAP: case E_ALL_MAP:
      case E_EXISTS_MAP: case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
      case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
      case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
      case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE:
      case E_EXISTS_WHILE: case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      case E_PACK: case E_PAT_PACK:
      case E_DEBUGZONE:
      case E_DEBUGNAME:
      case E_TYPEFX:
      case E_OVERWRITE:
      case E_UNDERWRITE:
      case E_EQUAL:
      case E_NEQUAL:
      case E_PROFILE:
	GC_MARK(e);
	pushGCMark(CDR(e));
	e = CAR(e);
	goto again;
      case E_MAPLET:
      case E_MEMO_FN:
      case E_HANDLE:
	GC_MARK(e);
	pushGCMark(CAR(e));
	e = CDR(e);
	goto again;
      case E_VARINTRO:
      case E_VARREF:
	GC_MARK(e);
	break;
      case E_NUMTYPE:
      case E_QUOTE:
      case E_REF:
      case E_GET:
#ifdef OBSOLETE
      case E_ARRAY_1:
#endif
      case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
      case E_RAISE:
      case E_EXCEPTION:
      case E_DELAY:
      case E_COUNTED:
      case E_DEBUGDECL:
      case E_PAT_CARD_EQ:
      case E_PAT_CARD_GEQ:
      case E_POLYFUN:
      case E_N_LAMBDA:
      case E_PAT_CONS:
      case E_DATACON:
      case E_NOT:
	GC_MARK(e);
	e = REFVAL(e);
	goto again;
      case E_ARRAY_GNRL:
      case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
      case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
	GC_MARK(e);
#if STOP_AND_COPY_VECTORS
	if (!(e->mark & MARK_SHARED) && ELINK(e)!=NEWGEN) /* then it is old */
	  ; /* then don't move it */
	else /* it is either shared or new; but if we got there,
		e was not marked, and it could not be in the writeBarrier,
		so it is new: move it. */
#endif
	  scMARK(AVAL(e),VECSIZE(AVAL(e)->n));
	GCmarkVector(AVAL(e));
	e = ASHAPE(e);
	goto again;
      case E_PAT_ANY:
      case E_EMPTY:
#ifdef OBSOLETE
      case E_ARRAY_0:
#endif
      case E_TUPLE_0:
      case E_RECORD_0:
	break;
      case E_ENV:
	GC_MARK(e);
	if (EVARS(e)!=EVARS(GLOBENV))
	  {
#if STOP_AND_COPY_VECTORS
	    if (ELINK(e)!=NEWGEN) /* then it is old */
	      ; /* then don't move it */
	    else /* it is new (envs are never shared): move it. */
#endif
	      scMARK(EVARS(e),VECSIZE(EVARS(e)->n));
	    GCmarkVector(EVARS(e));
	  }
	e = ENEXT(e);
	goto again;
      case E_COMPILED_CLOSURE:
      case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
      case E_MEMO_CLOSURE:
#endif
	GC_MARK(e);
	e = QENV(e);
	goto again;
      case E_PAT_CHECK:
	GC_MARK(e);
#if STOP_AND_COPY_VECTORS
	if (ELINK(e)!=NEWGEN) /* then it is old */
	  ; /* then don't move it */
	else /* it is new (patchecks are never shared): move it. */
#endif
	  scMARK(PATCHECKBITS(e),(E_UNUSED+7) >> 3);
	e = PATCHECKVAL(e);
	goto again;
      case E_STACK:
	GC_MARK(e);
#if STOP_AND_COPY_VECTORS
	if (ELINK(e)!=NEWGEN) /* then it is old */
	  ; /* then don't move it */
	else /* it is new (stacks are never shared): move it. */
#endif
	  {
	    scMARK(SJMPBUF(e),sizeof(stackDesc));
	  }
#if STOP_AND_COPY_VECTORS
	if (ELINK(e)!=NEWGEN) /* then it is old */
	  ; /* then don't sweep through it */
	else /* it is new (stacks are never shared): sweep through it.
	      We don't move stack chunks, since they are always allocated
	      by MALLOC(). */
#endif
	  {
	    GCmarkStackViewOff(SJMPBUF(e)->stack_frames,SCHUNK(e));
	    GCmarkStackView(SJMPBUF(e)->heap_frames);
	    GCmarkExcHandler(SCHUNK(e)->excStack);
#if DEBUGGING!=0
#ifdef OBSOLETE
	    pushGCMark(SCHUNK(e)->applStack);
#endif
#endif
	    e = SCHUNK(e)->protected;
	    goto again;
	  }
	break;
      case E_BYTECODE:
	GC_MARK(e);
	GCmarkByteCode((byteCodeChunk *)e);
	break;
      case E_EXTERNAL:
	GC_MARK(e);
	(*EXTDESC(e)->sweep)(EXTDATA(e),&doExtMark);
	break;
      case TYPE_ASSUM_KIND:
	GC_MARK(((struct type_assum *)e));
	GCMARK1(A_IDS(e));
	GCMARK1(A_IDLOCS(e));
	GCMARK1(A_IDEXTS(e));
	GCMARK1(A_CONS(e));
	GCMARK1(A_TYFCNS(e));
	GCMARK1(A_TYEXTS(e));
#if MODULE_SYSTEM & STD_ML_MODULE
	GCMARK1(A_STRUCTS(e));
	GCMARK1(A_SIGS(e));
#endif
	GCMARK1(A_NONGENS(e));
	GCMARK1(A_UNITS(e));
	GCMARK1(A_TYVARS(e));
	break;
      case TYPE_STATE_KIND:
	GC_MARK(((struct type_state *)e));
	GCMARK1(S_TYVARSUBS(e));
	GCMARK1(S_TYVARSUBCACHE(e));
	GCMARK1(S_CONDS(e));
	GCMARK1(S_EQCONDS(e));
	GCMARK1(S_EQFLAGS(e));
	GCMARK1(S_NONEQS(e));
	GCMARK1(S_TYCONS(e));
	GCMARK1(S_PACKERS(e));
	GCMARK1(S_ANNOTATIONS(e));
	GCMARK1(S_NONGENS(e));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	GCMARK1(S_STRENGTH(e));
	GCMARK1(S_MAXSTRENGTH(e));
#endif
	break;
      case COMPILE_ASSUM_KIND:
	GC_MARK(((struct compile_assum *)e));
	GCMARK1(C_IDS(e));
#if MODULE_SYSTEM & STD_ML_MODULE
	GCMARK1(C_STRUCTS(e));
	GCMARK1(C_SIGS(e));
#endif
	GCMARK1(C_CONS(e));
	GCMARK1(C_TYSTATE(e));
	GCMARK1(C_TYTAGS(e));
	GCMARK1(C_IMPORTS(e));
	GCMARK1(C_FUNDEF(e));
	break;
      case BUILD_HEADER_KIND:
	GC_MARK(((struct build_header *)e));
	e = ((struct build_header *)e)->args;
	goto again;
      default:
	kinderror(message(395),e->kind);
	break;
      }
  if (gcMarkStackTop==gcMarkStack)
    {
      if (nOverflows==0)
	return;
      unflushGCMark();
    }
  e = *--gcMarkStackTop;
  goto again;
}

PRIVATE void PairSlotInsert (ep, ap)
     INOUT ExprPtr ep;
     OUT struct tablext *ap;
{
  ExprPtr elink;
  ExprPtr *epp;

  for (; ep!=NOEXPR; ep=elink)
    {
      elink = ELINK(ep);
      epp = &ap->slot[snd_pair_hashcode1(CAR(ep),CDR(ep),ap->size)];
      ELINK(ep) = *epp;
      *epp = ep;
    }
}

PRIVATE Boolean ExpandPairSlot(epp,stdm)
     INOUT ExprPtr *epp;
     IN COUNTER stdm;
{
  ExprPtr ep;
  struct tablext *ap;
  unsigned long i;
  COUNTER j,m;
  COUNTER minm, maxm;

  if (stdm>= (1 << (8*sizeof (SHORT)-1))-1)
    stdm = (1 << (8*sizeof (SHORT)-1))-1;
  if (stdm<=1)
    return true;
  minm = stdm/2;
  maxm = 2*stdm;
  ep = *epp++;
  if (ep==NOEXPR)
    return;
  if (ep->kind==(CHAR)E_SLOT_BRANCH)
    {
      if (((struct tablext *)ep)->size>=minm && ((struct tablext *)ep)->size<=maxm)
	return;
    }
  else if (ELINK(ep)==NOEXPR)
    return;
  ap = (struct tablext *)SAFEMALLOC(OFFSETOF(struct tablext,slot[stdm]));
  if (ap==NULL)
    return;
  ap->kind = E_SLOT_BRANCH;
  ap->mark = 0;
  ap->size = stdm;
  for (j=0; j<stdm; j++)
    ap->slot[j] = NOEXPR;
  if (ep->kind==(CHAR)E_SLOT_BRANCH)
    {
      for (j=0; j<((struct tablext *)ep)->size; j++)
	PairSlotInsert (((struct tablext *)ep)->slot[j], ap);
      MARKFREE(ep);
    }
  else PairSlotInsert (ep, ap);
  epp[-1] = (ExprPtr)ap;
}

PRIVATE void doPairFree(table,hashsize,nsharedp)
     INOUT ExprPtr *table;
     IN unsigned long hashsize;
     INOUT unsigned long *nsharedp;
{
  ExprPtr *ap,*apend,*app,a;
  unsigned long nshared = *nsharedp;
  COUNTER nPairs, nSurvived;

  if (table!=(ExprPtr *)0L)
    for (ap=table, apend=table+hashsize; ap<apend; ap++)
      {
	nPairs = nSurvived = 0;
	a = *(app = ap);
	if (a==NOEXPR)
	  continue;
	if (a->kind==(CHAR)E_SLOT_BRANCH)
	  {
	    ExprPtr *ap2, *apend2, a2;
#ifdef THEONE
	    ExprPtr theOne = NOEXPR;
	    COUNTER i = 0;
#endif

	    for (ap2=((struct tablext *)a)->slot,
		   apend2 = ap2+((struct tablext *)a)->size;
		 ap2 < apend2; ap2++)
	      {
		a2 = *(app=ap2);
		for (; a2!=NOEXPR; )
		  {
		    nPairs++;
		    if (a2->kind==(CHAR)E_FREE)
		      {
			a2 = *app = ELINK(a2); /* works only because the link is not cleared by EFREE */
			nshared--;
		      }
		    else
		      {
			nSurvived++;
			a2 = *(app = &ELINK(a2));
		      }
		  }
#ifdef THEONE
		if ((a2 = *ap2)!=NOEXPR)
		  switch (i)
		    {
		    case 0:
		      i = 1;
		      theOne = a2;
		      break;
		    case 1:
		      i = 2;
		      break;
		    default: /* i==2 */
		      break;
		    }
#endif
	      }
#ifdef THEONE
	    switch (i)
	      {
	      case 0: /* a is empty. */
		MARKFREE(a);
		*ap = NOEXPR;
		break;
	      case 1: /* a contains just one list */
		MARKFREE(a);
		*ap = theOne;
		break;
	      default: /* a contains more than one list */
		break;
	      }
#endif
	  }
	else
	  for (; (a = *app)!=NOEXPR; )
	    {
	      nPairs++;
	      if (a->kind==(CHAR)E_FREE)
		{
		  *app = ELINK(a); /* works only because the link is not cleared by EFREE */
		  nshared--;
		}
	      else
		{
		  nSurvived++;
		  app = &ELINK(a);
		}
	    }
	{
	  COUNTER i = nPairs;

	  if (i<nSurvived)
	    i = nSurvived;
	  ExpandPairSlot (ap, i);
	}
      }
  *nsharedp = nshared;
}

PRIVATE void PairFree()
{
  doPairFree (Ktuple2Table, ktuple2_hash_size, &nSharedKtuple2);
  doPairFree (KonsTable, kons_hash_size, &nSharedKons);
  doPairFree (MapletKonsTable, mapletkons_hash_size, &nSharedMapletKons);
  doPairFree (MapKonsTable, mapkons_hash_size, &nSharedMapKons);
  doPairFree (KcomplexTable, kcomplex_hash_size, &nSharedKcomplex);
  doPairFree (KtypeApplTable, ktypeappl_hash_size, &nSharedKtypeAppl);
  doPairFree (KtypeFxTable, ktypefx_hash_size, &nSharedKtypeFx);
  doPairFree (Krecord1Table, krecord1_hash_size, &nSharedKrecord1);
}

PRIVATE void IntFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap,*app,a;

  if (IntTable!=(ExprPtr *)0L)
    for (i=0, ap=IntTable; i<int_hash_size; i++, ap++)
      {
	for (app = ap; a = *app, a!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because this field does
				    not interfere with the nextfree field */
		nSharedInts--;
	      }
	    else if (GC_MARKED(a))
	      break; /* reached old generation */
	    else
	      app = &ELINK(a);
	  }
      }
}

#define SlotsFree(epp) \
  { \
    if (*(epp)!=NOEXPR) \
      SlotsFree1(epp); \
  }

PRIVATE void SlotsFree1(epp)
INOUT REGISTER ExprPtr *epp;
{
  REGISTER ExprPtr ep;

  ep = *epp;
  if (ep->kind==(CHAR)E_SLOT_BRANCH)
    {
      SlotsFree(&CAR(ep));
      SlotsFree(&CDR(ep));
      epp = &ELINK(ep);
    }
  for (; ep = *epp, ep!=NOEXPR; )
    {
      if (ep->kind==(CHAR)E_FREE)
	{
	  *epp = ELINK(ep); /* works only because the link is not cleared by EFREE */
	}
      else if (GC_MARKED(ep)) /* reached old generation */
	break;
      else
	epp = &ELINK(ep);
    }
}

PRIVATE void RealFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap,*app,a;

  if (RealTable!=(ExprPtr *)0L)
    for (i=0, ap=RealTable; i<real_hash_size; i++, ap++)
      {
	for (app = ap; a = *app, a!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because this field does
				    not interfere with the nextfree field */
		nSharedReals--;
	      }
	    else if (GC_MARKED(a))
	      break; /* reached old generation */
	    else
	      app = &ELINK(a);
	  }
      }
}

PRIVATE void StringFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap,*app,a;

  if (StringTable!=(ExprPtr *)0L)
    for (i=0, ap=StringTable; i<string_hash_size; i++, ap++)
      {
	for (app=ap; a = *app, a!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because the link is not cleared by EFREE */
		nSharedStrings--;
	      }
	    else if (GC_MARKED(a)) /* reached old generation */
	      break;
	    else
	      app = &ELINK(a);
	  }
      }
}

PRIVATE void VarFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap,*app,a;

  if (VarTable!=(ExprPtr *)0L)
    for (i=0, ap=VarTable; i<VAR_HASH_SIZE; i++, ap++)
      {
	for (app=ap; a = *app, a!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because the link is not cleared by EFREE */
	      }
	    else if (GC_MARKED(a)) /* reached old generation */
	      break;
	    else
	      app = &ELINK(a);
	  }
      }
}

PRIVATE void RefFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap,*app,a;

  if (RefTable!=(ExprPtr *)0L)
    for (i=0, ap=RefTable; i<REF_HASH_SIZE; i++, ap++)
      {
	for (app=ap; a = *app, a!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because the next field
				      does not interfere with the link field */
	      }
	    else if (GC_MARKED(a))
	      break; /* reached old generation */
	    else
	      app = &ELINK(a);
	  }
      }
}

PRIVATE void ExtFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap;

  if (ExtTable!=(ExprPtr *)0L)
    for (i=0, ap=ExtTable; i<EXT_HASH_SIZE; i++, ap++)
      SlotsFree(ap);
}

PRIVATE void ArrayFree()
{
  REGISTER COUNTER i;
  REGISTER ExprPtr *ap, *app, a;

  if (ArrayTable!=(ExprPtr *)0L)
    for (i=0, ap=ArrayTable; i<array_hash_size; i++, ap++)
      for (app=ap; (a = *app)!=NOEXPR; )
	{
	  if (a->kind==(CHAR)E_FREE)
	    {
	      *app = ELINK(a); /* works only because the link is not cleared by EFREE */
	      nSharedArrays--;
	    }
	  else if (GC_MARKED(a))
	    break; /* reached old generation */
	  else
	    app = &ELINK(a);
	}
}

PRIVATE void DataFree()
{
  REGISTER ExprPtr *ap,*apend,*app,a;

  if (DataTable!=(ExprPtr *)0L)
    for (ap=DataTable, apend=DataTable+data_hash_size; ap<apend; ap++)
      {
	for (app=ap; (a = *app)!=NOEXPR; )
	  {
	    if (a->kind==(CHAR)E_FREE)
	      {
		*app = ELINK(a); /* works only because the link is not cleared by EFREE */
		nSharedData--;
	      }
	    else if (GC_MARKED(a))
	      break; /* reached old generation */
	    else
	      app = &ELINK(a);
	  }
      }
}

PRIVATE void AllocFrees()
{
  PairFree();
  IntFree();
  RealFree();
  StringFree();
  VarFree();
  RefFree();
  ExtFree();
  ArrayFree();
  DataFree();
}

PRIVATE void ExpandTables()
{
  COUNTER grow;

#ifdef OBSOLETE
  (void) ExpandKtuple2Table();
  (void) ExpandKonsTable();
  (void) ExpandMapletKonsTable();
  (void) ExpandMapKonsTable();
  (void) ExpandKcomplexTable();
  (void) ExpandKtypeApplTable();
  (void) ExpandKtypeFxTable();
  (void) ExpandKrecord1Table();
#endif
  if (nSharedData>data_hash_size)
    {
      grow = 2; /*(nSharedData + data_hash_size-1) / data_hash_size;*/
      (void) ExpandDataTable(grow*data_hash_size);
    }
  if (nSharedInts>int_hash_size)
    {
      grow = 2; /*(nSharedInts + int_hash_size-1) / int_hash_size;*/
      (void) ExpandIntTable(grow*int_hash_size);
    }
  if (nSharedReals>real_hash_size)
    {
      grow = 2; /*(nSharedReals + real_hash_size-1) / real_hash_size;*/
      (void) ExpandRealTable(grow*real_hash_size);
    }
  if (nSharedArrays>array_hash_size)
    {
      grow = 2; /*(nSharedArrays + array_hash_size-1) / array_hash_size;*/
      (void) ExpandArrayTable(grow*array_hash_size);
    }
  if (nSharedStrings>string_hash_size)
    {
      grow = 2; /*(nSharedStrings + string_hash_size-1) / string_hash_size;*/
      (void) ExpandStringTable(grow*string_hash_size);
    }
}

PUBLIC void GCmark()
{
  ExprPtr *ep;

  /* unmark these specially, so as not to forget to mark their
     fields later on. */
  unmarkOldTypeAssums();
  unmarkOldTypeStates();
  unmarkOldCompileAssums();
  unmarkOldBCs();
  /* now, unmark the objects in the write barrier */
  {
    REGISTER ExprPtr bar;
    
    for (bar=writeBarrier; bar!=NOEXPR; bar=ELINK(bar))
      GC_UNMARK(bar); /* to make GCmark1 look at the fields of bar,
			 that are the ones that have possibly changed */
	/* this will never free the cells in writeBarrier, since they
	   are all in old generations */
  }
  /* deal with GLOBENV specially, because it contains the only
     vector in the system whose size cannot be computed from vp->n
     (at least in the case of STOP_AND_COPY_VECTORS) */
  GC_MARK(GLOBENV);
  scMARK(EVARS(GLOBENV),VECSIZE(MAX_GLOBALS));
  /* should not move actually, because it is larger than MAX_ALLOC_IN_PAGE */
  {
    REGISTER COUNTER i,n;
    REGISTER ExprPtr *ep,e;

    /*hfprintf(hstderr,"GC: globLimit=%d, n=%d\n",globLimit,EVARS(GLOBENV)->n);*/
    for (i=globLimit, n=EVARS(GLOBENV)->n, ep=&EVARS(GLOBENV)->expr[i]; i<n; i++)
      {
	e = *ep++;
	if (e!=NOEXPR)
	  GCmark1(e);
      }
    if (itBarrier)
      {
	e = EVARS(GLOBENV)->expr[ORIGINAL_IT];
	if (e!=NOEXPR)
	  GCmark1(e);
      }
  } /* instead of GCmarkVector(EVARS(GLOBENV)); */
  GCmark1(ENEXT(GLOBENV));
#if DEBUGGING
  {
    REGISTER COUNTER i,n;
    REGISTER ExprPtr *ep;

    for (i=0, n=breakPoints.n, ep=breakPoints.bcs; i<n; i++)
      GCmark1 (*ep++);
  }
#endif
  GCmark1(ProtectedAgainstGC);
  GCmark1(ProtectedListAgainstGC);
  for (ep=ProtectedTable; ep<ProtectedTableTop;)
    {
      GCmark1(*ep++);
    }
  GCmarkStackView(theStackFrames);
  GCmarkStackView(theHeapFrames);
  for (ep=mapStack; ep<mapStackTop;)
    GCmark1(*ep++);
  /* now mark the cells in the writeBarrier. */
  {
    REGISTER ExprPtr bar;
    
    for (bar=writeBarrier; bar!=NOEXPR; bar=ELINK(bar))
      GCmark1(bar);
  }
  if (THEEXPR!=NOEXPR)
    { GCmark1(THEEXPR); }
  GCmark1(moduleName);
  GCmark1(mapgetCache_e);
  GCmark1(mapgetCache_m);
  GCmark1(mapgetCache_r);
  GCmark1(divmodCache_a);
  GCmark1(divmodCache_d);
  GCmark1(divmodCache_m);
  GCmark1(divmodCache_dm);
  GCmark1(lexId);
  GCmark1(lexPath);
  GCmark1(lexEnv);
  GCmark1(consEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  GCmark1(strEnv);
  GCmark1(sigEnv);
#endif
  GCmark1(keepLexEnv);
  GCmark1(keepConsEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  GCmark1(keepStrEnv);
  GCmark1(keepSigEnv);
#endif
  GCmarkExcHandler(excStack);
  GCmark1(theTypes);
  GCmark1(theBindings);
  GCmark1(basisAssum);
  GCmark1(basisState);
  GCmark1(basisTable);
  GCmark1(basisEnv);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  GCmark1(theFx);
#endif
#if DEBUGGING
#if !TRANSLATOR
  if (currentCallingEnv!=NOEXPR)
    GCmark1(currentCallingEnv);
#endif
  GCmark1(monitorExprs);
  if (dbgFile!=NOEXPR)
    GCmark1(dbgFile);
  if (dbgName!=NOEXPR)
    GCmark1(dbgName);
#ifdef OBSOLETE
  GCmark1(currentDebugZone);
  GCmark1(stepFile);
  GCmark1(applStkList);
  for (ep=applStkTable; ep<applStkTop;)
    {
      GCmark1(*ep++);
    }
#endif
#endif
#if PROFILING
  GCmark1(profileExprs);
#endif
  GCmark1(currentFile);
  GCmark1(parseMarkerStack);
  GCmark1(externalKeepList);
  {
    struct cslist *csl;

    for (csl = Checksums; csl!=NULL; csl = csl->next)
      GCmark1(csl->basis);
  }
#ifdef XLATE
  GCmark1(bcXlateMap);
#endif
}

PRIVATE void donothing(e)
IN CONST ExprPtr e;
{
}

PUBLIC doExtDesc doExtFree = {
  X_FREE,
  donothing
};

#define GCFREE(e) \
  (e)->kind = (CHAR)E_FREE

PRIVATE void sweepHomoBucket(buck,endepp)
IN BucketPtr buck;
REGISTER IN CONST ExprPtr *endepp;
{
  REGISTER ExprPtr ep;
  REGISTER ExprPtr *epp,*firstepp;

  for (firstepp = epp = buck->newcells+buck->firstNewCell; epp<endepp; epp++)
    {
      ep = *epp;
      if (GC_MARKED(ep))
	{
	  GC_UNMARK(ep);
	  if (firstepp!=epp) /* firstepp is always equal to epp, or points to the first
				free cell after the firstNewCell position */
	    { /* exchange position of ep with that of the free cell *firstepp */
	      *epp = *firstepp;
	      *firstepp = ep;
	    }
	  firstepp++;
	  /* this all sorts the newcells array, from the firstNewCell position up,
	     so that all addresses of still allocated cells are in an interval
	     [firstNewCell..firstNewCell+buck->occupancy[generation_count] [ */
	}
      else GCFREE(ep);
    }
  liveCells[0][generation_count] += (buck->occupancy[generation_count] =
				     firstepp - (buck->newcells+buck->firstNewCell));
  nYoungCells[0] += BUCKET_SIZE - buck->firstNewCell;
}

PRIVATE void sweepHeteroBucket(buck,endepp)
IN BucketPtr buck;
REGISTER IN CONST ExprPtr *endepp;
{
  REGISTER ExprPtr ep;
  REGISTER ExprPtr *epp,*firstepp;

  for (firstepp = epp = buck->newcells+buck->firstNewCell; epp<endepp; epp++)
    {
      ep = *epp;
      if (GC_MARKED(ep))
	{
	  GC_UNMARK(ep);
	  if (firstepp!=epp) /* firstepp is always equal to epp, or points to the first
				free cell after the firstNewCell position */
	    { /* exchange position of ep with that of the free cell *firstepp */
	      *epp = *firstepp;
	      *firstepp = ep;
	    }
	  firstepp++;
	  /* this all sorts the newcells array, from the firstNewCell position up,
	     so that all addresses of still allocated cells are in an interval
	     [firstNewCell..firstNewCell+buck->occupancy[generation_count] [ */
	}
      else
	{
	  INT size;

	  switch (ep->kind)
	    {
	    case E_STRING:
	      size = SLEN(ep);
	      scFREE(STEXT(ep),size);
	      freedStringBytes += size;
	      break;
	    case E_PAT_CHECK:
	      size = (E_UNUSED+7) >> 3;
	      scFREE(PATCHECKBITS(ep),size);
	      freedPatCheckBitBytes += size;
	      break;
	    case E_STACK:
#ifdef DEBUG
	      stackFreed++;
#endif
#if STOP_AND_COPY_VECTORS
	      scFREE(SJMPBUF(ep),sizeof(stackDesc));
#else
	      MARKFREE(SJMPBUF(ep));
	      /* jmpbufs must be aligned, and cannot be allocated by VecAlloc() */
#endif
	      {
		StackChunk *chunk = SCHUNK(ep);
		struct stackMark *thread;

		if (chunk->contents!=(char *)0L && chunk->contents==(char *)1L
		    && (thread = chunk->thread)!=NULL)
		  {
#ifdef ASCENDING_STACK
		    freedStackBytes += chunk->top-thread->highMark;
#else
		    freedStackBytes += thread->highMark-chunk->top;
#endif
		  }
	      stackChunkFree(chunk);
	      VECFREE(chunk,sizeof(StackChunk));
	      }
	      freedStackBytes += sizeof(stackDesc)+sizeof(StackChunk);
	      break;
	    case E_ENV:
	      if (EVARS(ep)==EVARS(GLOBENV))
		break;
	      /*fallthrough*/
	    case E_ARRAY_GNRL: case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
	    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
	      size = VECSIZE(AVAL(ep)->n);
	      scFREE(AVAL(ep),size);
	      freedVectorBytes += size;
	      break;
	    case E_EXTERNAL:
	      (*EXTDESC(ep)->sweep)(EXTDATA(ep),&doExtFree);
	      freedExternals++;
	      break;
	    case E_FREE:
	      break;
	    default:
	      kinderror(message(396),ep->kind);
	      break;
	    }
	  GCFREE(ep);
	}
    }
  liveCells[1][generation_count] += (buck->occupancy[generation_count] =
				     firstepp - (buck->newcells+buck->firstNewCell));
  nYoungCells[1] += BUCKET_SIZE - buck->firstNewCell;
}
#undef GCFREE

PRIVATE void GCsweep1()
{
  REGISTER BucketPtr buck;

  for (buck=theHomoBucketList; buck!=currentHomoBucket; buck=buck->next)
    sweepHomoBucket(buck,buck->newcells+BUCKET_SIZE);
  if (buck!=(BucketPtr)0L)
    {
      sweepHomoBucket(buck,lastHomoExpr);
      while ((buck = buck->next)!=(BucketPtr)0L)
	sweepHomoBucket(buck,
			buck->newcells+buck->firstNewCell+buck->occupancy[generation_count]);
    }
  for (buck=theHeteroBucketList; buck!=currentHeteroBucket; buck=buck->next)
    sweepHeteroBucket(buck,buck->newcells+BUCKET_SIZE);
  if (buck!=(BucketPtr)0L)
    {
      sweepHeteroBucket(buck,lastHeteroExpr);
      while ((buck = buck->next)!=(BucketPtr)0L)
	sweepHeteroBucket(buck,
			  buck->newcells+buck->firstNewCell+buck->occupancy[generation_count]);
    }
}

#if STOP_AND_COPY_VECTORS
typedef struct PageReserve {
  PagePtr first; /* or (PagePtr)0L if none. */
  PagePtr last; /* or (PagePtr)0L if none. */
  COUNTER nPages;
} PageReserve;

PRIVATE PageReserve oldPages[MAX_GENERATIONS];
	/* list of twin pages, with generation numbers decreasing */

#ifdef GC_SLIDING
PRIVATE void slideOldPages()
{ /* concatenate generation 0 with gen 1, then put gen 2 into gen 1,
     gen 3 into gen 2, and so on. Assume generation_count>=1. */
  PageReserve *old;
  COUNTER i;

  if (pageStart(curPage)!=pageTop(curPage))
    {
      curPage = pageNext(curPage); /* should not be 0 */
      nPages++;
    }
  old = &oldPages[generation_count];
  if (oldPages[0].last==(PagePtr)0L)
    oldPages[0] = oldPages[1];
  else if (oldPages[1].first==(PagePtr)0L)
    ;
  else
    {
      pageNext(oldPages[0].last) = oldPages[1].first;
      pageNext(&oldPages[0].last[1]) = &oldPages[1].first[1];
      oldPages[0].last = oldPages[1].last;
      oldPages[0].nPages += oldPages[1].nPages;
    }
  for (i=1; i<generation_count; i++)
    oldPages[i] = oldPages[i+1];
}
#endif

PRIVATE void cutoffOldPages()
{
  /* cut off all non-empty pages */
  PagePtr lastpp,pp;
  PageReserve *old;

  if (pageStart(curPage)!=pageTop(curPage))
    {
      lastpp = curPage;
      curPage = pageNext(curPage); /* should not be 0 */
      nPages++;
      /* if curPage points to the middle of a page, it means that
	 this page will become old. So we advance curPage to the
	 the first one (or NOEXPR) that is empty. This guarantees
	 curPage to remain new. */
    }
  else
    {
      lastpp = (PagePtr)0L;
      for (pp = twinPageList; pp!=curPage; lastpp = pp, pp = pageNext(pp));
      /* assume that spaceIndex==0 */
    }
  old = &oldPages[generation_count];
  old->first = twinPageList;
  twinPageList = curPage;
  old->last = lastpp;
  old->nPages = nPages;
  allocedPages -= nPages;
  nPages = 0;
  if (curPage==(PagePtr)0L)
    {
      pp = NewPage(newptr_fail);
      if (pp==(PagePtr)0L)
	nomem();
    }
}
#endif

#ifdef GC_SLIDING
PRIVATE void slideOldBuckets1(bucklist)
IN BucketPtr bucklist;
{
  REGISTER BucketPtr bp;
  REGISTER COUNTER i;
  REGISTER SHORT *occ;

  for (bp = bucklist; bp!=(BucketPtr)0L; bp=bp->next)
    {
      bp->occupancy[0] += bp->occupancy[1];
      for (i=1, occ=&bp->occupancy[1]; i<generation_count; i++, occ++)
	occ[0] = occ[1];
      occ[0] = 0;
    }
}

PRIVATE void slideOldBuckets()
{
  slideOldBuckets1(theHomoBucketList);
  slideOldBuckets1(theHeteroBucketList);
}
#endif

PRIVATE void cutoffOldBuckets1(bucklist)
IN BucketPtr bucklist;
{
  BucketPtr bp;
  REGISTER ExprPtr *epp;
  REGISTER ExprPtr ep;
  REGISTER COUNTER i;
  REGISTER COUNTER n;

  for (bp = bucklist; bp!=(BucketPtr)0L; bp=bp->next)
    {
      n = bp->occupancy[generation_count];
      if (n!=0)
	{
	  /* allow for unshared cells to be put in the write barrier
	     when modified (as in Olden(), expr.c) */
	  for (i=0, epp=bp->newcells + bp->firstNewCell; i<n; i++, epp++)
	    {
	      ep = *epp;
	      GC_MARK(ep); /* to speed up future mark phases */
	      if (ELINK(ep)==NEWGEN)
		ELINK(ep) = OLDGEN; /* shared objects are untouched by this */
	    }
	  bp->firstNewCell += n;
	}
      bp->occupancy[generation_count+1] = 0;
      /* for bumpHomoExpr() and bumpHeteroExpr() to be correct */
    }
}

PRIVATE void cutoffOldBuckets()
{
  cutoffOldBuckets1(theHomoBucketList);
  cutoffOldBuckets1(theHeteroBucketList);
}

#ifdef GC_SLIDING
PRIVATE void slideOldGenerations()
{
#if STOP_AND_COPY_VECTORS
  if (spaceIndex!=0)
    return; /* because we cannot refresh if spaceIndex is not the same
	       as when the old generations were formed, we arrange for
	       old generations to be formed only when spaceIndex==0. */
#endif
  if (generation_count<1) /* impossible to slide generations down. */
    return;
  slideOldBuckets();
  slideOldHeaders();
  slideOldTypeAssums();
  slideOldTypeStates();
  slideOldCompileAssums();
  slideOldBCs();
#if STOP_AND_COPY_VECTORS
  slideOldPages();
#endif
  generation_count--;
}
#endif

PRIVATE void cutoffOldGenerations()
{
#if STOP_AND_COPY_VECTORS
  if (spaceIndex!=0)
    return; /* because we cannot refresh if spaceIndex is not the same
	       as when the old generations were formed, we arrange for
	       old generations to be formed only when spaceIndex==0. */
#endif
#ifdef GC_SLIDING
  if (generation_count>=MAX_GENERATIONS-1)
    slideOldGenerations(); /* not recommended: it actually slows down GC. */
#else
  if (generation_count>=MAX_GENERATIONS-1)
    return;
#endif
  cutoffOldBuckets();
  cutoffOldHeaders();
  cutoffOldTypeAssums();
  cutoffOldTypeStates();
  cutoffOldCompileAssums();
  cutoffOldBCs();
#if STOP_AND_COPY_VECTORS
  cutoffOldPages();
#endif
  generation_count++;
}

/* we need a quick random generator, that should not be too bad,
   for estimating when to refresh old generations. */
/* we assume that COUNTERs are at least 16 bits wide (should be no problem),
   and compute in fixed point notation, with the comma 16 bits from
   right. In other words, 1.0 is 65536, and gcRandom() draws numbers
   between 0.0 (0) included and 1.0 excluded (65535 included). */
#define MAX_RAND 65535

PRIVATE unsigned COUNTER gcRandom()
{
  STATIC unsigned COUNTER seed = 1;

  return seed = (9821L*((unsigned long)seed)+13849L) % MAX_RAND;
}

#define gcAlpha ((MAX_RAND*2)/3)
#define gcInvMean (MAX_RAND-gcAlpha)
  /* = inverse of mean for the Poisson distribution */

PRIVATE void refreshGenerationCount(urgency)
IN COUNTER urgency;
{
 /* try to run the garbage collector through some of the more
     recent old generations, by adding them to the new generations
     before the marking phase. */
  /* assume generation_count equals j; then we shall examine
     generations j-1, j-2, ..., j-i,... with probability (1-alpha).alpha^i
     (Poisson distribution, alpha<1), so that very old generations are
     reexamined very rarely, but at least once from time to time
     to give them a chance of being scavenged.
        We take a probability (i.e, sum (1-alpha).alpha^i=1 for i>=1)
    so that the number of generations remains as stable as possible.
    Actually, it is better to do as if j==MAX_GENERATIONS-1, so that
    if j==1, the last (and only) generation has almost no chance of
    being refreshed, but if j==MAX_GENERATIONS-1, the last generation
    is very likely to be refreshed.
  */
  SHORT j;

#if STOP_AND_COPY_VECTORS
  if (spaceIndex!=1)
    return; /* the data in old pages is in spaceIndex 0, so
	     after marking all refreshed data will be in spaceIndex 1,
	     which is valid only is spaceIndex==1. */
#endif
  if (urgency>=10)
    generation_count = 0;
  else
    {
      j = MAX_GENERATIONS-1;
      if (j!=0 && gcRandom()<=(unsigned COUNTER)gcInvMean)
	{
	  j--; /* refresh latest generation with probability (1-alpha) */
	  while (j!=0 && gcRandom()<=(unsigned COUNTER)gcAlpha)
	    { /* then draw with probability (1-alpha).alpha, (1-alpha).alpha^2, etc. */
	      j--;
	    }
	}
      if (j<generation_count)
	generation_count = j;
    }
}

PRIVATE void refreshOldBuckets1(bucklist,oldgen)
IN BucketPtr bucklist;
IN COUNTER oldgen;
{
  REGISTER COUNTER n;
  REGISTER BucketPtr bp;

  for (bp=bucklist; bp!=(Bucket *)0L; bp=bp->next)
    {
      n = bp->firstNewCell;
      {
	REGISTER COUNTER i;
	REGISTER SHORT *occ;

	for (i = generation_count, occ = bp->occupancy+i; i<oldgen; i++)
	  n -= *occ++;
      }
      {
	REGISTER ExprPtr ep,*epp;
	REGISTER COUNTER i;

	for (i = bp->firstNewCell, epp = bp->newcells+i; --i>=n; )
	  {
	    ep = *--epp;
	    GC_UNMARK(ep);
	    if (ELINK(ep)==OLDGEN)
	      ELINK(ep) = NEWGEN;
	    /* for each ep in the writeBarrier which becomes new again,
	       we must avoid to put its link to NEWGEN
	       right away (this would break the writeBarrier).
	       But its mark won't contain MARK_GC, as all data
	       in old generations, so the erasure of MARK_GC also
	       serves to signal ourselves that it has to be removed
	       from the writeBarrier (see below). */
	  }
      }
      bp->firstNewCell = n;
      {
	REGISTER COUNTER i;
	REGISTER SHORT *occ;

	n = 0;
	for (i=generation_count, occ=bp->occupancy+i; i<=oldgen; i++)
	  n += *occ++;
      }
      bp->occupancy[generation_count] = n;
    }
}

PRIVATE void refreshOldBuckets(oldgen)
IN COUNTER oldgen;
{
  refreshOldBuckets1(theHomoBucketList,oldgen);
  refreshOldBuckets1(theHeteroBucketList,oldgen);
}

#if STOP_AND_COPY_VECTORS
PRIVATE void refreshOldPages(oldgen)
IN COUNTER oldgen;
{ /* get back old pages */
  PagePtr first,last;
  PageReserve *pres;
  COUNTER i;

  pres = &oldPages[generation_count];
  first = pres->first;
  last = pres->last;
  nPages += pres->nPages;
  allocedPages += pres->nPages;
  for (i=generation_count; ++i<oldgen;)
    {
      pres++;
      if (last==(PagePtr)0L)
	{
	  first = pres->first;
	  last = pres->last;
	}
      else if (pres->first==(PagePtr)0L)
	;
      else
	{
	  pageNext(last) = pres->first;
	  pageNext(&last[1]) = &pres->first[1];
	  last = pres->last;
	}
      nPages += pres->nPages;
      allocedPages += pres->nPages;
    }
  if (last==(PagePtr)0L)
    ;
  else if (twinPageList==(PagePtr)0L)
    twinPageList = first;
  else
    {
      pageNext(last) = twinPageList;
      pageNext(&last[1]) = &twinPageList[1];
      twinPageList = first;
    }
}
#endif

PRIVATE void refreshWriteBarrier()
{ /* remove all unmarked data from the writeBarrier, after refreshOldBuckets(). */
  REGISTER ExprPtr *epp,ep;

  for (epp = &writeBarrier; (ep = *epp)!=NOEXPR; )
    {
      if (GC_MARKED(ep))
	epp = &ELINK(ep);
      else
	{
	  *epp = ELINK(ep);
	  ELINK(ep) = NEWGEN; /* get it out of the writeBarrier. */
	}
    }
}

PRIVATE void refreshOldGenerations(urgency)
IN COUNTER urgency;
{
  COUNTER oldgen;

  oldgen = generation_count;
  refreshGenerationCount(urgency);
  if (oldgen==generation_count)
    return;
  refreshOldBuckets(oldgen);
  refreshOldHeaders(oldgen);
  refreshOldTypeAssums();
  refreshOldTypeStates();
  refreshOldCompileAssums();
  refreshOldBCs();
#if STOP_AND_COPY_VECTORS
  refreshOldPages(oldgen);
#endif
  refreshWriteBarrier();
}

PRIVATE void resetCurrentBucket()
{
  if (!initializing)
    {
      for (currentHomoBucket=theHomoBucketList;
	   currentHomoBucket!=(BucketPtr)0L &&
	   currentHomoBucket->firstNewCell+currentHomoBucket->occupancy[generation_count]
	   ==BUCKET_SIZE;
	   currentHomoBucket=currentHomoBucket->next);
      for (currentHeteroBucket=theHeteroBucketList;
	   currentHeteroBucket!=(BucketPtr)0L &&
	   currentHeteroBucket->firstNewCell+currentHeteroBucket->occupancy[generation_count]
	   ==BUCKET_SIZE;
	   currentHeteroBucket=currentHeteroBucket->next);
    }
  if (currentHomoBucket==(BucketPtr)0L && theBucketList!=(BucketPtr)0L)
    {
      currentHomoBucket = *endHomoBucketList = theBucketList;
      endHomoBucketList = &currentHomoBucket->next;
      theBucketList = *endHomoBucketList;
      *endHomoBucketList = (BucketPtr)0L;
      if (initializing)
	{
	  nextHomoExpr = currentHomoBucket->newcells;
	  lastHomoExpr = currentHomoBucket->newcells+BUCKET_SIZE;
	}
    }
  if (currentHeteroBucket==(BucketPtr)0L && theBucketList!=(BucketPtr)0L)
    {
      currentHeteroBucket = *endHeteroBucketList = theBucketList;
      endHeteroBucketList = &currentHeteroBucket->next;
      theBucketList = *endHeteroBucketList;
      *endHeteroBucketList = (BucketPtr)0L;
      if (initializing)
	{
	  nextHeteroExpr = currentHeteroBucket->newcells;
	  lastHeteroExpr = currentHeteroBucket->newcells+BUCKET_SIZE;
	}
    }
  if (!initializing)
    {
      if (currentHomoBucket!=(BucketPtr)0L)
	{
	  nextHomoExpr = currentHomoBucket->newcells+currentHomoBucket->firstNewCell+
	    currentHomoBucket->occupancy[generation_count];
	  lastHomoExpr = currentHomoBucket->newcells+BUCKET_SIZE;
	}
      if (currentHeteroBucket!=(BucketPtr)0L)
	{
	  nextHeteroExpr = currentHeteroBucket->newcells+currentHeteroBucket->firstNewCell+
	    currentHeteroBucket->occupancy[generation_count];
	  lastHeteroExpr = currentHeteroBucket->newcells+BUCKET_SIZE;
	}
    }
}

PRIVATE void GCsweep()
{
  liveCells[0][generation_count] = liveCells[1][generation_count] = 0L;
  nYoungCells[0] = nYoungCells[1] = 0L;
  GCsweep1();
  AllocFrees();
  FreeTypeAssums();
  FreeTypeStates();
  FreeCompileAssums();
  FreeBuildHeaders();
  FreeBCs();
  COMMITFREE();
  ExpandTables();
}

PUBLIC int gcTraceFd=-1;
#if CORE_TRACE
EXTERN hFILE *coreTraceFile;
#endif
#if STOP_AND_COPY_VECTORS
PRIVATE COUNTER nCellGCs=0,nPageGCs=0;
#endif

#if STOP_AND_COPY_VECTORS
#define NEWCELL_WEIGHT 0.333
PRIVATE EXTENDED newCellWeight = NEWCELL_WEIGHT;
PRIVATE EXTENDED oldCellWeight = 1.0-NEWCELL_WEIGHT;
#endif

PRIVATE char gcMessage[2048];

PUBLIC void GCgrow(urgency, tag)
IN COUNTER urgency;
IN CONST ExprPtr tag;
{
  Boolean normp[2],norm0p[2];
  COUNTER lowgen = 0;
  INT needed[2];
  INT newLive[2],totalLive[2];
#if STOP_AND_COPY_VECTORS
  INT neededsc;
  Boolean scp,sc0p;
  EXTENDED scGrow;
  COUNTER oldAllocedPages;
#endif
  struct tms before,after;

  inGC = true;
again:
  if (urgency>=20)
    nomem();
  normp[0] = norm0p[0] = (currentHomoBucket==(BucketPtr)0L);
  normp[1] = norm0p[1] = (currentHeteroBucket==(BucketPtr)0L);
#if STOP_AND_COPY_VECTORS
  scp = sc0p = (curPage==(PagePtr)0L);
  if (norm0p[0] || norm0p[1])
    nCellGCs++;
  if (sc0p)
    nPageGCs++;
#endif
  if (!initializing)
    {
      gc_count++;
      if (gcTraceFd!=-1)
	{ /* notice that we must not use hfprintf, which may allocate
	     some memory */
	  char *s = gcMessage;

	  if (tag!=NULL)
	    {
	      strcpy (s, "==["); s += strlen (s);
	      safememcpy (s, STEXT(tag), SLEN(tag)); s += SLEN(tag);
	      *s++ = ']';
	    }
	  strcpy (s, "=========================\nGC..."); s += strlen (s);
	  (void) write(gcTraceFd, gcMessage, s-gcMessage);
	  freedStringBytes = freedPatCheckBitBytes = freedStackBytes = freedVectorBytes = 0;
	  freedExternals = 0;
	}
#if STOP_AND_COPY_VECTORS
      spaceIndex = 1-spaceIndex;
#endif
      (void) times(&before);
      lowgen = generation_count;
#ifdef GENERATIONS
      refreshOldGenerations(urgency);
#endif
#if STOP_AND_COPY_VECTORS
      curPage = &twinPageList[spaceIndex];
      nPages = 0;
#endif
      GCmark();
      GCsweep();
      newLive[0] = liveCells[0][generation_count];
      newLive[1] = liveCells[1][generation_count];
      {
	COUNTER i,j;

	for (j=0; j<2; j++)
	  {
	    totalLive[j] = newLive[j];
	    for (i=0; i<generation_count; )
	      totalLive[j] += liveCells[j][i++];
	    /*needed = (INT)(growFactor*(EXTENDED)totalLive); // grow all data */
	    /*needed = totalLive-newLive + (INT)(growFactor*(EXTENDED)newLive); // grow only new data */
	    /* something intermediate:
	       needed[j] = (INT)(oldCellWeight*(growFactor*(EXTENDED)totalLive[j])+
	       newCellWeight*(((EXTENDED)(totalLive[j]-newLive[j]))+
	       growFactor*(EXTENDED)newLive[j]));
	       */
	    /* our solution: grow only new data, but add a constant to newLive to
	       represent the constant (or so) time that we need to sweep through
	       the hash tables; the goal is to keep the ratio GCtime/computation-time
	       reasonably low: */
	    needed[j] = totalLive[j]-newLive[j] +
	      (INT)(growFactor*(EXTENDED)(newLive[j]+gc_hash_cost_offset[j]));
	  }
      }
#if STOP_AND_COPY_VECTORS
      /* then increase or decrease scGrowFactor to match the filling speed of cells */
      /* we do this by counting the number of times GC was called because of
	 a lack of cells (nCellGCs) and the number of times it was called
	 because pages filled up (nPageGCs). */
      scGrow = 1.0+0.3*(EXTENDED)(nPageGCs-nCellGCs);
      scGrow *= copySlowdown;
      if (scGrow>=3.0)
	{
maxgrow:
	  scGrow = 3.0; /* if greater than that, can be an anomaly */
	}
      scGrowFactor *= ldexp(1.0+scGrow,-1); /* this thing is the mean of the old and the new,
					       to damp the evolution of scGrowFactor */
      if (scGrowFactor<minScGrowFactor)
	scGrowFactor = minScGrowFactor; /* moreover, scGrowFactor should not go
					   below minScGrowFactor */
      else if (scGrowFactor>maxScGrowFactor)
	scGrowFactor = maxScGrowFactor; /* and above maxScGrowFactor */
      {
	COUNTER i;

	oldAllocedPages = 0;
	for (i=0; i<generation_count; )
	  oldAllocedPages += oldPages[i++].nPages;
      }
      neededPages = (INT)(oldCellWeight*(scGrowFactor*(EXTENDED)(oldAllocedPages+allocedPages))+
			  newCellWeight*(((EXTENDED)oldAllocedPages)+
					 scGrowFactor*(EXTENDED)allocedPages));
      neededsc = neededPages;
#endif
    }
  else
    {
#if STOP_AND_COPY_VECTORS
      neededsc = allocedPages+1;
#endif
      nYoungCells[0] = 0;
      needed[0] = nYoungCells[0]+2;
      nYoungCells[1] = 0;
      needed[1] = nYoungCells[1]+2;
    }
#if STOP_AND_COPY_VECTORS
  if (neededsc<allocedPages)
    {
      if (neededsc<minAllocedPages)
	neededsc = minAllocedPages;
      while (neededsc<allocedPages)
	DelPage();
      COMMITFREE();
      scp = false;
    }
#endif
  while (normp[0] || normp[1]
#if STOP_AND_COPY_VECTORS
	 || scp
#endif
	 )
    {
      normp[0] = (normp[0] && needed[0]>nYoungCells[0] &&
		  NewBucket(newptr_fail,false)!=(BucketPtr)0L);
      if (normp)
	needed[0] -= BUCKET_SIZE;
      normp[1] = (normp[1] && needed[1]>nYoungCells[1] &&
		  NewBucket(newptr_fail,false)!=(BucketPtr)0L);
      if (normp)
	needed[1] -= BUCKET_SIZE;
#if STOP_AND_COPY_VECTORS
      scp = (scp && neededsc>allocedPages && NewPage(newptr_fail)!=(PagePtr)0L);
#endif
    }
  /* otherwise, GCgrow was triggered by malloc */
rereset:
  resetCurrentBucket();
  if (norm0p[0] && currentHomoBucket==(BucketPtr)0L)
    {
      if (NewBucket(newptr_fail,true)==(BucketPtr)0L)
	{
bumpurgency:
	  if (gcTraceFd!=-1)
	    { /* notice that we must not use hfprintf, which may allocate
		 some memory */
	      char *s = gcMessage;

	      strcpy(s,"\nfailed at urgency level ");
	      s += strlen(s);
	      strcpy(s,ltoasc((long)urgency));
	      s += strlen(s);
	      strcpy(s,".\n");
	      s += strlen(s);
	      (void) write(gcTraceFd,gcMessage,s-gcMessage);
	    }
	  urgency += 10;
	  goto again;
	}
      goto rereset;
    }
  if (norm0p[1] && currentHeteroBucket==(BucketPtr)0L)
    {
      if (NewBucket(newptr_fail,true)==(BucketPtr)0L)
	goto bumpurgency;
      goto rereset;
    }
#if STOP_AND_COPY_VECTORS
  if (sc0p && curPage==(PagePtr)0L)
    goto bumpurgency;
#endif
  if (!initializing)
    {
#ifdef GENERATIONS
      cutoffOldGenerations();
#endif
      (void) times(&after);
      after.tms_utime -= before.tms_utime;
      gc_utime += after.tms_utime;
      after.tms_stime -= before.tms_stime;
      gc_stime += after.tms_stime;
      gc_rtime += after.tms_utime+after.tms_stime;
      if (gcTraceFd!=-1)
	{ /* notice that we must not use hfprintf, which may allocate
	     some memory */
	  char *s = gcMessage;

	  strcpy(s,"done:\n  total number of allocated memory cells [nCells] = ");
	  s += strlen(s);
	  strcpy(s,ltoasc(nCells)); s += strlen(s);
	  strcpy(s,"\n  allocated homo cells in young generation : ");
	  s += strlen(s);
	  strcpy(s,ltoasc(nYoungCells[0])); s += strlen(s);
	  strcpy(s," (~"); s+= strlen(s);
	  strcpy(s,ltoasc(nYoungCells[0]*sizeof(Expression))); s += strlen(s);
	  strcpy(s," bytes, not counting sharing overhead)\n  allocated hetero cells in young generation : ");
	  s += strlen(s);
	  strcpy(s,ltoasc(nYoungCells[1])); s += strlen(s);
	  strcpy(s," (~"); s+= strlen(s);
	  strcpy(s,ltoasc(nYoungCells[1]*sizeof(Expression))); s += strlen(s);
	  strcpy(s," bytes, not counting sharing and contents overhead)\n  live homo cells in young generation : "); s += strlen(s);
	  strcpy(s,ltoasc(newLive[0])); s += strlen(s);
	  strcpy(s,"\n  live hetero cells in young generation : "); s += strlen(s);
	  strcpy(s,ltoasc(newLive[1])); s += strlen(s);
	  strcpy(s,"\n  total live homo cells : "); s += strlen(s);
	  strcpy(s,ltoasc(totalLive[0])); s += strlen(s);
	  strcpy(s,"\n  total live hetero cells : "); s += strlen(s);
	  strcpy(s,ltoasc(totalLive[1])); s += strlen(s);
#if STOP_AND_COPY_VECTORS
	  strcpy(s,"\n  growth factor of stop-and-copy zone [scGrowFactor] = ");
	  s += strlen(s);
	  strcpy(s,rtoasc(&scGrowFactor)); s += strlen(s);
#endif
	  strcpy(s,"\n  strings freed : "); s += strlen(s);
	  strcpy(s,ltoasc(freedStringBytes)); s += strlen(s);
	  strcpy(s," bytes\n  patcheckbits freed : "); s += strlen(s);
	  strcpy(s,ltoasc(freedPatCheckBitBytes)); s += strlen(s);
	  strcpy(s," bytes\n  stacks freed : "); s += strlen(s);
	  strcpy(s,ltoasc(freedStackBytes)); s += strlen(s);
	  strcpy(s," bytes\n  vectors (environments, arrays, tuples, records) freed : "); s += strlen(s);
	  strcpy(s,ltoasc(freedVectorBytes)); s += strlen(s);
	  strcpy(s," bytes\n  "); s += strlen(s);
	  strcpy(s,ltoasc(freedExternals)); s += strlen(s);
	  strcpy(s," externals freed\n  garbage collection time = "); s += strlen(s);
	  strcpy(s,timetoasc(after.tms_utime)); s += strlen(s);
	  strcpy(s,"s."); s += strlen(s);
#ifdef GENERATIONS
	  strcpy(s,"\n  There were "); s += strlen (s);
	  strcpy(s,ltoasc((long)lowgen)); s += strlen (s);
	  strcpy(s," old generations, plus one new;\n  there are "); s += strlen (s);
	  strcpy(s,ltoasc((long)generation_count)); s += strlen(s);
	  strcpy(s," old generations, plus one new."); s += strlen(s);
#endif
	  strcpy(s,"\n  Number of stacks (threads) allocated since startup: "); s += strlen (s);
	  strcpy(s,ltoasc((long)stackAlloced)); s += strlen (s);
	  strcpy(s,"\n  Number of allocated bytes of temporary (stack) storage: ");
	  s += strlen (s);
	  strcpy(s,ltoasc((long)alloctmpBytes)); s += strlen (s);
	  strcpy(s,"\n  Number of shared ktuple2 cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKtuple2)); s += strlen (s);
	  strcpy (s,", size of ktuple2 hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)ktuple2_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared kons cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKons)); s += strlen (s);
	  strcpy (s,", size of kons hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)kons_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared mapletkons cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedMapletKons)); s += strlen (s);
	  strcpy (s,", size of mapletkons hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)mapletkons_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared mapkons cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedMapKons)); s += strlen (s);
	  strcpy (s,", size of mapkons hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)mapkons_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared kcomplex cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKcomplex)); s += strlen (s);
	  strcpy (s,", size of kcomplex hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)kcomplex_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared ktypeappl cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKtypeAppl)); s += strlen (s);
	  strcpy (s,", size of ktypeappl hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)ktypeappl_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared ktypefx cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKtypeFx)); s += strlen (s);
	  strcpy (s,", size of ktypefx hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)ktypefx_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared krecord1 cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedKrecord1)); s += strlen (s);
	  strcpy (s,", size of krecord1 hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)krecord1_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared data cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedData)); s += strlen (s);
	  strcpy (s,", size of data hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)data_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared array cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedArrays)); s += strlen (s);
	  strcpy (s,", size of array hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)array_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared string cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedStrings)); s += strlen (s);
	  strcpy (s,", size of string hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)string_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared int cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedInts)); s += strlen (s);
	  strcpy (s,", size of int hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)int_hash_size)); s += strlen (s);
	  strcpy(s,"\n  Number of shared real cells: "); s += strlen (s);
	  strcpy (s,ltoasc((long)nSharedReals)); s += strlen (s);
	  strcpy (s,", size of real hash table: "); s += strlen (s);
	  strcpy (s,ltoasc((long)real_hash_size)); s += strlen (s);
	  strcpy (s,"\n"); s += strlen (s);
	  (void) write(gcTraceFd,gcMessage,s-gcMessage);
	}
    }
  inGC = false;
}

PUBLIC void InitGC()
{
  ProtectedAgainstGC = NIL;
  ProtectedListAgainstGC = NIL;
  externalKeepList=NIL;
  writeBarrier=NIL;
  gcMarkStack = (ExprPtr *)SAFEMALLOC(GCMARK_STACK_SIZE*sizeof(ExprPtr));
  gcMarkStackEnd = gcMarkStack+GCMARK_STACK_SIZE;
  gcMarkStackTop = gcMarkStack;
}
