(* Quotient types, example.
   Copyright (C) 1992-1993 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* example of a non-trivial quotient type :
 oriented graphs, here with ordered lists of successors.
 Two graphs are equal if and only if they differ only
 by a renaming of the nodes that are accessible from their roots.
*)
abstype
    graph = GRAPH of unit ref * (unit ref -m> unit ref list)
(* with the invariant that for every GRAPH (root,succ),
 root inset succ andalso
 all elems s subset succ | n => s in map succ end
 nodes are unit refs, because we just want to generate them,
 but they have no particular property.
 *)
with
    eqtype graph {
		  equality =
		  fn (GRAPH (root,succ),GRAPH (root',succ')) =>
		  let exception DoNotMatch
		      fun eq (root,root',mark) =
			  if root inset mark
			      then if ?mark(root)=root'
				       then mark
				   else raise DoNotMatch
			  else eqlist (?succ root,?succ' root',mark ++ {root => root'})
		      and eqlist (nil,nil,mark) = mark
			| eqlist (r::l,r'::l',mark) =
			  eqlist (l,l',eq (r,r',mark))
			| eqlist _ = raise DoNotMatch
		  in
		      (eq (root,root',{});
		       true) handle DoNotMatch => false
		  end,
	      discriminant =
	      fn (GRAPH (root,succ)) => len (?succ(root))}
fun g_succs (GRAPH (_,succ)) node = ?succ(node)
fun g_new () = GRAPH (ref (),{})
fun g_connect (GRAPH (root,succ)) (node,succs) =
    GRAPH (root,succ ++ {node => succs})
(* pre-condition: elems succs subset succ *)
end
