/* Imperative tables.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Although this all works, I've decided not to integrate it
   into HimML, since it consumes much more memory than just
   MapPtrs, and is not integrated with the generation scavenging
   code (so GC time goes up to about 25%, instead of 15%).
*/

#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "ex1proto.h"
#include "expr.h"
#include "table1.h"

PUBLIC unsigned long lowbits = ((unsigned long)
				(((char *)&((struct Bucket *)NULL)->cell[0])-((char *)NULL)))
     & ((1<<NLOWBITS)-1);

EXTERN ExprPtr none,optttm;
EXTERN unsigned long lowbits;

PRIVATE ExprPtr table_get_1 (hn, key)
     IN CONST struct hnode *hn;
     IN CONST ExprPtr key;
{
  REGISTER unsigned long k, ki;
  REGISTER COUNTER i;

  k = ((unsigned long)key) >> NLOWBITS;
  for (i=HN_DEPTH-1; i>0; i--)
    {
      if (hn==NULL)
	return NOEXPR;
      ki = (k >> (HN_BITS*i)) & HN_MASK;
      hn = hn->sons[ki];
    }
  if (hn==NULL)
    return NOEXPR;
  ki = k & HN_MASK;
  return (ExprPtr)hn->sons[ki];
}

PRIVATE void table_put(ext,key,val,behind)
     INOUT ExprPtr ext;
     IN CONST ExprPtr key,val;
     IN int behind;
{
  REGISTER unsigned long k, ki;
  REGISTER COUNTER i;
  REGISTER struct hnode *hn, *e;

  hn = (struct hnode *)EXTDATA(ext);
  if (hn==NULL)
    {
      hn = (struct hnode *)MALLOC(SIZEOF_HNODE_LEAF);
      {
	REGISTER COUNTER j;

	for (j=0; j<(1<<HN_BITS_LEAF); j++)
	  hn->sons[j] = NULL;
	hn->n = 0;
      }
      EXTDATA(ext) = (char *)hn;
      HITBARRIER(ext);
    }
  k = ((unsigned long)key) >> NLOWBITS;
  for (i=HN_DEPTH-1; i>0; i--)
    {
      ki = (k >> (HN_BITS*i)) & HN_MASK;
      e = hn->sons[ki];
      if (e==NULL)
	{
	  e = (struct hnode *)MALLOC(sizeof(struct hnode));
	  {
	    REGISTER COUNTER j;

	    for (j=0; j<(1<<HN_BITS); j++)
	      e->sons[j] = NULL;
	    e->n = 0;
	  }
	  hn->sons[ki] = e;
	  hn->n++;
	}
      hn = e;
    }
  ki = k & HN_MASK;
  e = hn->sons[ki];
  if (e==NULL)
    {
      hn->sons[ki] = (struct hnode *)val;
      hn->n++;
      HITBARRIER(ext);
    }
  else if (behind)
    ;
  else
    {
      hn->sons[ki] = (struct hnode *)val;
      HITBARRIER(ext);
    }
}

PRIVATE void table_remove(ext,key)
     INOUT ExprPtr ext;
     IN CONST ExprPtr key;
{
  REGISTER unsigned long k, ki;
  REGISTER COUNTER i;
  REGISTER struct hnode *hn;
  struct hnode *stack[HN_DEPTH];
  unsigned long kistack[HN_DEPTH];

  hn = (struct hnode *)EXTDATA(ext);
  k = ((unsigned long)key) >> NLOWBITS;
  for (i=HN_DEPTH-1; i>0; i--)
    {
      if (hn==NULL)
	return;
      stack[i] = hn;
      ki = (k >> (HN_BITS*i)) & HN_MASK;
      kistack[i] = ki;
      hn = hn->sons[ki];
    }
  if (hn==NULL)
    return;
  /*stack[0] = hn;*/
  ki = k & HN_MASK;
  /*kistack[0] = ki;*/
  if (hn->sons[ki]==NULL)
    return;
  HITBARRIER(ext);
  for (i=1; i<HN_DEPTH; i++)
    {
      hn->sons[ki] = NULL;
      if (--hn->n != 0)
	return;
      MARKFREE(hn);
      hn = stack[i];
      ki = kistack[i];
    }
  EXTDATA(ext) = NULL;
}

PRIVATE Boolean table_sweep_1 (hn,kprefix,depth,do_f,data)
     IN CONST struct hnode *hn;
     IN unsigned long kprefix;
     IN COUNTER depth;
     IN CONST Boolean (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char *data;
{
  unsigned long i;
  struct hnode *e;

  if (hn==NULL)
    return false;
  if (depth==0)
    {
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  if ((*do_f) ((ExprPtr)(((kprefix | i) << NLOWBITS) | lowbits), (ExprPtr)e, data))
	    return true;
	}
    }
  else
    {
      COUNTER d;

      d = HN_BITS*depth;
      depth--;
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  if (table_sweep_1 (e, kprefix | (i << d), depth, do_f, data))
	    return true;
	}
    }
  return false;
}

PRIVATE Boolean table_sweep (hn,do_f,data)
     IN CONST struct hnode *hn;
     IN CONST Boolean (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char * data;
{
  unsigned long i;
  struct hnode *e;

  if (hn==NULL)
    return false;
  for (i=0; i < (1<<HN_BITS_LEAF); i++)
    {
      e = hn->sons[i];
      if (e==NULL)
	continue;
      if (table_sweep_1 (e, (i << HN_BITS*(HN_DEPTH-1)), HN_DEPTH-2, do_f, data))
	return true;
    }
  return false;
}

PRIVATE Boolean do_table_card (key, val, data)
     IN CONST ExprPtr key;
     IN CONST ExprPtr val;
     INOUT char *data;
{
  ++*(INT *)data;
  return false;
}

PRIVATE INT table_card (hn)
     IN CONST struct hnode *hn;
{
  INT n = 0;

  (void) table_sweep (hn, do_table_card, (char *)&n);
  return n;
}

PRIVATE Boolean do_table_sweep (key, val, data)
     IN CONST ExprPtr key;
     IN CONST ExprPtr val;
     INOUT char *data;
{
  doExtDesc *doext = (doExtDesc *)data;

  (*doext->action)(key);
  if (errno)
    return true;
  (*doext->action)(val);
  if (errno)
    return true;
  return false;
}

PRIVATE void table_free_1 (hn,depth)
     INOUT struct hnode *hn;
     IN COUNTER depth;
{
  COUNTER i;

  if (hn==NULL)
    return;
  if (depth!=0)
    {
      depth--;
      for (i=0; i < (1<<HN_BITS); i++)
	table_free_1 (hn->sons[i], depth);
    }
  MARKFREE(hn);
}

PRIVATE void table_free (hn)
     INOUT struct hnode *hn;
{
  COUNTER i;

  if (hn==NULL)
    return;
  for (i=0; i < (1<<HN_BITS_LEAF); i++)
    table_free_1 (hn->sons[i],HN_DEPTH-2);
  MARKFREE(hn);
}

PRIVATE void tableSweep (extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  struct hnode *hn = (struct hnode *)extdata;

  switch (doext->xtag)
    {
    case X_SAVE:
      outInt (table_card (hn));
      if (errno) return;
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      errno = 0;
      (void) table_sweep (hn, do_table_sweep, (char *)doext);
      break;
    case X_FREE:
      table_free(hn);
      break;
    }
}

PRIVATE ExprPtr tableLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  ExprPtr e,key,val;
  INT n;
  ENTER;

  n = inInt ();
  if (errno) return NIL;
  e = MakeExternal(desc,NULL);
  (void) KEEP(e);
  while (n--!=0)
    {
      key = (*load)(sharedref);
      if (errno) { QUIT; return NIL; }
      val = (*load)(sharedref);
      if (errno) { QUIT; return NIL; }
      table_put(e,key,val,0);
      KEEPONLY(e);
    }
  return e;
}

PUBLIC extDesc tableExtDesc = {
  (CHAR *)"table",
  { 't', 'a', 'b', 'l' }, NULL,
  NULL, /* no hash function, tables are not shared */
  NULL, /* no comparison function, tables are not shared */
  NULL, /* no print function */
  tableSweep,
  tableLoad
};

PRIVATE ExprPtr table_new ()
{
  ExprPtr e;

  e = MakeExternal(&tableExtDesc, NULL);
  return e;
}

DEF_ACTION(table)
{
  return table_new();
}

DEF_PRIVATE_ACTION(q_table_get)
{ /* env is the external with the hnode inside it. */
  ExprPtr e;

  e = table_get_1 ((struct hnode *)EXTDATA(env), arg);
  if (e==NOEXPR)
    return none;
  return Kdata(0,optttm,e); /* SOME e */
}

DEF_ACTION(table_get)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_get),arg);
}

DEF_PRIVATE_ACTION(q_table_put)
{ /* env is the external with the hnode inside it. */
  table_put (env, CAR(arg), CDR(arg), 0);
  return tuple0;
}

DEF_ACTION(table_put)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put),arg);
}

DEF_PRIVATE_ACTION(q_table_put_behind)
{ /* env is the external with the hnode inside it. */
  table_put (env, CAR(arg), CDR(arg), 1);
  return tuple0;
}

DEF_ACTION(table_put_behind)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put_behind),arg);
}

DEF_PRIVATE_ACTION(q_table_remove)
{
  table_remove (env, arg);
  return tuple0;
}

DEF_ACTION(table_remove)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_remove),arg);
}

DEF_ACTION(table_kill)
{
  table_free ((struct hnode *)EXTDATA(arg));
  EXTDATA(arg) = (char *)NULL;
  HITBARRIER(arg);
  return tuple0;
}

Boolean q_table_sweep_do(key,val,data)
IN CONST ExprPtr key,val;
IN CONST char *data;
{
  ExprPtr e, res;
  ENTER;

  e = KONS(key, val);
  res = (*ACTION((ExprPtr)data))(e, QENV ((ExprPtr)data));
  QUIT;
  return (res==trueValue);
}

DEF_PRIVATE_ACTION(q_table_sweep)
{
  Boolean res;

  res = table_sweep ((struct hnode *)EXTDATA(env), q_table_sweep_do, (char *)arg);
  return res?trueValue:falseValue;
}

DEF_ACTION(table_sweep)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_sweep),arg);
}

PRIVATE MapPtr table_collect_1 (hn,kprefix,depth,do_f,data)
     IN CONST struct hnode *hn;
     IN unsigned long kprefix;
     IN COUNTER depth;
     IN CONST MapPtr (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char *data;
{
  unsigned long i;
  struct hnode *e;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  if (hn==NULL)
    return EmptyMap();
  if (depth==0)
    {
      _x.p[0] = NewMap();
      _x.p[1] = NewMap();
      _x.n = 2;
      _x.next = theStackFrames;
      theStackFrames = (gcStackView *)&_x;
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  _x.p[1] = (*do_f) ((ExprPtr)(((kprefix | i) << NLOWBITS) | lowbits), (ExprPtr)e, data);
	  _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
	  _x.p[1] = NewMap();
	}
      theStackFrames = _x.next;
    }
  else
    {
      COUNTER d;

      _x.p[0] = NewMap();
      _x.p[1] = NewMap();
      _x.n = 2;
      _x.next = theStackFrames;
      theStackFrames = (gcStackView *)&_x;
      d = HN_BITS*depth;
      depth--;
      for (i=0; i < (1<<HN_BITS); i++)
	{
	  e = hn->sons[i];
	  if (e==NULL)
	    continue;
	  _x.p[1] = table_collect_1 (e, kprefix | (i << d), depth, do_f, data);
	  _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
	}
      theStackFrames = _x.next;
    }
  return _x.p[0];
}

PRIVATE MapPtr table_collect (hn,do_f,data)
     IN CONST struct hnode *hn;
     IN CONST MapPtr (*do_f) P((IN CONST ExprPtr key, IN CONST ExprPtr val, INOUT char *data));
     INOUT char * data;
{
  unsigned long i;
  struct hnode *e;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

  if (hn==NULL)
    return NewMap();
  _x.p[0] = NewMap();
  _x.p[1] = NewMap();
  _x.n = 2;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  for (i=0; i < (1<<HN_BITS_LEAF); i++)
    {
      e = hn->sons[i];
      if (e==NULL)
	continue;
      _x.p[1] = table_collect_1 (e, (i << HN_BITS*(HN_DEPTH-1)), HN_DEPTH-2, do_f, data);
      _x.p[0] = MapMerge (_x.p[0], _x.p[1]);
      _x.p[1] = NewMap();
    }
  theStackFrames = _x.next;
  return _x.p[0];
}

MapPtr q_table_collect_do(key,val,data)
IN CONST ExprPtr key,val;
IN CONST char *data;
{
  ExprPtr e, res;
  ENTER;

  e = KONS(key,val);
  res = (*ACTION((ExprPtr)data))(e, QENV ((ExprPtr)data));
  QUIT;
  return res;
}

DEF_PRIVATE_ACTION(q_table_collect)
{
  MapPtr res;

  res = table_collect ((struct hnode *)EXTDATA(env), q_table_collect_do, (char *)arg);
  return res;
}

DEF_ACTION(table_collect)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_collect),arg);
}
