(* An implementation of Ada-like rendezvous.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires chan.ml *)

exception EmptySelect

abstype ('a,'b,''env,'rs) rendezvous =
    RDV of ('a -> 'b) * ('a,''env,'rs) chan * ('b,''env,'rs) chan
and ('a,'b,''env,'rs) rdv_arm =
    ARM of (unit -> bool) * ('a,'b,''env,'rs) rendezvous
and ('a,'b,''env,'rs) select =
    SEL of ('a,'b,''env,'rs) rdv_arm list
with
  fun rdv_manager (tm as |[fork,...]| : (''env thread,'rest,'rs) thread_manager,
		   new_sched : unit -> (''env thread,'rs) scheduler) =
      let val |[chan,put,get,get_now,get_wait,...]| =
		chan_manager (tm,new_sched)
	  val |[mutex,condition,signal,with_condition,...]| =
	      mutex_manager (tm,new_sched)
	  fun rendezvous f = RDV (f,chan (),chan ())
	  fun call (RDV (_,out,inm),v) = (put (out,v); get inm)
	  fun accept (SEL nil) = raise EmptySelect
	    | accept (SEL l) =
	      let val unfinished = ref true
		  val finished = condition (mutex ())
		  fun helper (ARM (test,r as RDV (f,inm,out))) =
		      let fun deposit () =
			      if !unfinished
				 then (put (out,f (get_now inm));
				       unfinished := false;
				       signal finished)
			      else ()
		      in
			 if test ()
			    then  (get_wait inm;
				   with_condition (finished, deposit)
				     handle GetNow =>
					helper (ARM (fn () => true,r)))
			 else ()
		      end
	     in
		with_condition (finished,
				fn () => (iterate fork (fn () => helper x)
					   | x in list l
					  end))
	     end
      in
	 |[rendezvous = rendezvous,
	   arm = ARM,
	   select = SEL,
	   accept = accept,
	   call =  call
	   ]|
      end
end
