(* Example of a sharp function in the free monoid of lists over a set.
   (This is the function from which all others can be defined in a monoid.)
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

type 'a monoid = ('a * 'a -> 'a)	(* the internal law of the monoid *)
	* 'a	(* the neutral element *)

fun sharp (times,e) f = (* sharp : 'b monoid * ('a -> 'b) -> ('a list -> 'b) *)
	let fun fsharp nil = e
	| fsharp (a::l) = times(f a,fsharp l)
	in  fsharp
	end

(* examples: *)

val length = sharp (op +,0) (fn _ => 1)
fun member x = sharp (fn (z,y) => z orelse y,false) (fn y => x=y)
val sum = sharp (op +,0) identity
fun maplist f = sharp (op @,nil) (fn x => [f x])
	(* could also define:
	fun maplist f = sharp (op ::,nil) f, but (op ::,nil) is not a monoid
	(although this last definition is valid) *)
