/* Complex arithmetic functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include "eval.h"
#include "expr.h"
#include "ariproto.h"
#include "zproto.h"
#include "ex1proto.h"
#include "rndproto.h"
#include "hakproto.h"
#include "errproto.h"
#ifdef HAS_PROTOTYPES
#include "sarith.h"
#endif

#define __SEG__ ARITH

EXTERN EXTENDED zero,fone,fmone,onepi;

PUBLIC ExprPtr divmodCache_a, divmodCache_b, divmodCache_d,divmodCache_m,
		divmodCache_dm;

#if OS==AMIGA_OS
PUBLIC int __matherr(exc)
struct __exception *exc;
{
  while (true)
    (void) Raise(ARITHEXC); /* cannot get back result,
			       so raise exception ad infinitum;
			       this is not very smart, but what
			       else can we do ? */
  /*NOTREACHED*/
  return 0;
}
#else
PUBLIC int matherr(exc)
struct exception *exc;
{
  while (true)
    (void) Raise(ARITHEXC); /* cannot get back result,
			       so raise exception ad infinitum;
			       this is not very smart, but what
			       else can we do ? */
  /*NOTREACHED*/
  return 0;
}
#endif

#if EXTP
PUBLIC ExprPtr Kext(x)
IN CONST EXTENDED *x;
{
  REAL y;

  y = (REAL)*x;
  return Kreal(&y);
}
#endif

#undef NEED_EXP1
#ifdef THINK_C
# define NEED_EXP1
#else
#if OS!=MAC
# define NEED_EXP1
#endif
#endif

#ifdef NEED_EXP1

EXTENDED exp1(x)
EXTENDED x;
{
  EXTENDED res;

  rexp1(&x,&res);
  return res;
}

EXTENDED log1(x)
EXTENDED x;
{
  EXTENDED res;

  rlog1(&x,&res);
  return res;
}
#endif

PRIVATE ExprPtr Kcomp(re,im)
IN CONST REAL *re,*im;
{
  ExprPtr ex,imx;

  ex = Kreal(re);
  if (!ZEROR(*im))
    { 
      ENTER;

      (void) KEEP(ex);
      imx = Kreal(im);
      (void) KEEP(imx);
      /*if (!ZZEROP(imx))*/
      ex = Kcomplex(ex,imx);
      QUIT;
    }
  return ex;
}

#define RE(z)   XVAL(REPART(z))
#define IM(z)   XVAL(IMPART(z))

PUBLIC ExprPtr z_add(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr val;
  REAL re,im;

  switch (x->kind)
    {
    case E_REAL:
      switch (y->kind)
	{
	case E_REAL:
	  re = XVAL(x)+XVAL(y);
	  val = Kreal(&re);
	  break;
	case E_COMPLEX:
	  re = XVAL(x)+RE(y);
	  val = Kcomp(&re,&IM(y));
	  break;
	default:
	  kinderror (message (914), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    case E_COMPLEX:
      switch (y->kind)
	{
	case E_REAL:
	  re = RE(x)+XVAL(y);
	  val = Kcomp(&re,&IM(x));
	  break;
	case E_COMPLEX:
	  re = RE(x)+RE(y);
	  im = IM(x)+IM(y);
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (914), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    default:
      kinderror (message (914), x->kind);
      val  = NIL;
      break;
    }
  return val;
}

DEF_ACTION_2ARGS(add)
{
  return z_add(arg1,arg2);
}
ACTION_END

PUBLIC ExprPtr z_sub(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr val;
  REAL re,im;

  switch (x->kind)
    {
    case E_REAL:
      switch (y->kind)
	{
	case E_REAL:
	  re = XVAL(x)-XVAL(y);
	  val = Kreal(&re);
	  break;
	case E_COMPLEX:
	  re = XVAL(x)-RE(y);
	  im = -IM(y);
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (915), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    case E_COMPLEX:
      switch (y->kind)
	{
	case E_REAL:
	  re = RE(x)-XVAL(y);
	  val = Kcomp(&re,&IM(x));
	  break;
	case E_COMPLEX:
	  re = RE(x)-RE(y);
	  im = IM(x)-IM(y);
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (915), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    default:
      kinderror (message (915), x->kind);
      val  = NIL;
      break;
    }
  return val;
}

DEF_ACTION_2ARGS(sub)
{
  return z_sub(arg1,arg2);
}
ACTION_END

PUBLIC ExprPtr z_mul(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr val;
  REAL re,im;
  EXTENDED rex,rey,imx,imy;

  switch (x->kind)
    {
    case E_REAL:
      switch (y->kind)
	{
	case E_REAL:
	  re = XVAL(x)*XVAL(y);
	  val = Kreal(&re);
	  break;
	case E_COMPLEX:
	  re = XVAL(x)*RE(y);
	  im = XVAL(x)*IM(y);
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (916), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    case E_COMPLEX:
      switch (y->kind)
	{
	case E_REAL:
	  re = RE(x)*XVAL(y);
	  im = IM(x)*XVAL(y);
	  val = Kcomp(&re,&im);
	  break;
	case E_COMPLEX:
	  rex = RE(x);
	  rey = RE(y);
	  imx = IM(x);
	  imy = IM(y);
	  re = rex*rey - imx*imy;
	  im = rex*imy + imx*rey;
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (916), y->kind);
	  val  = NIL;
	  break;
	}
      break;
    default:
      kinderror (message (916), x->kind);
      val  = NIL;
      break;
    }
  return val;
}

DEF_ACTION_2ARGS(mul)
{
  return z_mul(arg1,arg2);
}
ACTION_END

PUBLIC ExprPtr z_div(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr val;
  REAL re,im;
  EXTENDED rex,imx,rey,imy,r;

  switch (y->kind)
    {
    case E_REAL:
 /*
	switch (classifyD(&XVAL(y)))
	{
	case ZERONUM:	case -ZERONUM:
	case DENORMALNUM:	case -DENORMALNUM:
	return Raise(div0Exception);
	}
	*/
      switch (x->kind)
	{
	case E_REAL:
	  re = XVAL(x)/XVAL(y);
	  val = Kreal(&re);
	  break;
	case E_COMPLEX:
	  re = RE(x)/XVAL(y);
	  im = IM(x)/XVAL(y);
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (917), x->kind);
	  val  = NIL;
	  break;
	}
      break;
    case E_COMPLEX:
      rey = RE(y);
      imy = IM(y);
      r = rey*rey+imy*imy;
 /*
	switch (classifyX(&r))
	{
	case ZERONUM:	case -ZERONUM:
	case DENORMALNUM:	case -DENORMALNUM:
	return Raise(div0Exception);
	}
	*/
      switch (x->kind)
	{
	case E_REAL:
	  rex = XVAL(x);
	  re = (rex*rey)/r;
	  im = (-rex*imy)/r;
	  val = Kcomp(&re,&im);
	  break;
	case E_COMPLEX:
	  rex = RE(x);
	  imx = IM(x);
	  re = (rex*rey+imx*imy)/r;
	  im = (rey*imx-rex*imy)/r;
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (917), x->kind);
	  val  = NIL;
	  break;
	}
      break;
    default:
      kinderror (message (917), y->kind);
      val  = NIL;
      break;
    }
  return val;
}

DEF_ACTION_2ARGS(div)
{
  return z_div(arg1,arg2);
}
ACTION_END

PUBLIC ExprPtr z_pow(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr val;
  REAL re,im;
  EXTENDED aux,piy,xre,xim,dumx,dumy;
#if EXTP
  EXTENDED rex,imx,rey,imy,resx,resy;
#endif

  switch (y->kind)
    {
    case E_REAL:
      if (ZZEROP(y))
	val = one;
      else switch (x->kind)
	{
	case E_REAL:
	  if (ZZEROP(x))
	    val = zzero;
	  else if (NEGR(XVAL(x)))
	    { /* x^y = exp(y*log x) = exp(y*(i.pi+log -x))
		 = (-x)^y*(cos(pi.y)+i.sin(pi.y)) */
	      aux = pow(-(EXTENDED)XVAL(x),XVAL(y));
	      piy = onepi*(EXTENDED)XVAL(y);
	      re = aux*cos(piy);
	      im = aux*sin(piy);
	      val = Kcomp(&re,&im);
	    }
	  else
	    {
	      re = pow(XVAL(x),XVAL(y));
	      val = Kreal(&re);
	    }
	  break;
	case E_COMPLEX:
#if EXTP
	  rex = RE(x);
	  imx = IM(x);
	  rey = XVAL(y);
	  zpow(&rex,&imx,&rey,&zero,&resx,&resy);
	  re = resx;
	  im = resy;
#else
	  zpow(&RE(x),&IM(x),&XVAL(y),&zero,&re,&im);
#endif
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (918), x->kind);
	  val  = NIL;
	  break;
	}
      break;
    case E_COMPLEX:
      switch (x->kind)
	{
	case E_REAL:
	  if (ZZEROP(x))
	    val = zzero;
	  else
	    {
	      if (NEGR(XVAL(x)))
		{ /* x^y = exp(y*log x) = exp(y*(i.pi+log -x))
		     = (-x)^y*(cos(pi.y)+i.sin(pi.y)),
		     (-x)^y = (-x)^(re y)*(cos(im y.log -x)+i.sin(im y.log -x))
		     */
		  aux = -(EXTENDED)XVAL(x);
		  xre = pow(aux,(EXTENDED)RE(y));
		  xim = log(aux)*(EXTENDED)IM(y);
		  dumx = xre*cos(xim);
		  dumy = xre*sin(xim);
		  piy = onepi*(EXTENDED)RE(y);
		  xre = cos(piy);
		  xim = sin(piy);
		  re = dumx*xre-dumy*xim;
		  im = dumx*xim+dumy*xre;
		}
	      else
		{
		  /* x^y = exp(y*log x) = exp((re y+i.im y)*log x)
		     = x^(re y)*(cos(im y.log x)+i.sin(im y.log x)) */
		  aux = (EXTENDED)XVAL(x);
		  xre = pow(aux,(EXTENDED)RE(y));
		  xim = log(aux)*(EXTENDED)IM(y);
		  re = xre*cos(xim);
		  im = xre*sin(xim);
		}
	      val = Kcomp(&re,&im);
	    }
	  break;
	case E_COMPLEX:
#if EXTP
	  rex = RE(x);
	  imx = IM(x);
	  rey = RE(y);
	  imy = IM(y);
	  zpow(&rex,&imx,&rey,&imy,&resx,&resy);
	  re = resx;
	  im = resy;
#else
	  zpow(&RE(x),&IM(x),&RE(y),&IM(y),&re,&im);
#endif
	  val = Kcomp(&re,&im);
	  break;
	default:
	  kinderror (message (918), x->kind);
	  val  = NIL;
	  break;
	}
      break;
    default:
      kinderror (message (918), y->kind);
      val  = NIL;
      break;
    }
  return val;
}

DEF_ACTION_2ARGS(pow)
{
  return z_pow(arg1,arg2);
}
ACTION_END

DEF_ACTION_1ARG(neg)
{
  REAL x,y;

  if (arg->kind==(CHAR)E_REAL)
    {
      x = -XVAL(arg);
      return Kreal(&x);
    }
  x = -RE(arg);
  y = -IM(arg);
  return Kcomp(&x,&y);
}
ACTION_END

DEF_ACTION_1ARG(sqr)
{
  return z_mul(arg,arg);
}
ACTION_END

DEF_ACTION_1ARG(sqrt)
{
  ExprPtr res;
  REAL x,y;
#if EXTP
  EXTENDED rex,rey,resx,resy;
#endif

  if (arg->kind==(CHAR)E_REAL)
    {
      if (classifyD(&XVAL(arg))>=0)
	{
	  x = sqrt(XVAL(arg));
	  res = Kreal(&x);
	}
      else
	{
	  ENTER;

	  x = sqrt(-(EXTENDED)XVAL(arg));
	  arg = Kreal(&x);
	  res = Kcomplex(zzero,KEEP(arg));
	  QUIT;
	}
    }
  else
    {
#if EXTP
      rex = RE(arg);
      rey = IM(arg);
      zsqrt(&rex,&rey,&resx,&resy);
      x = resx;
      y = resy;
#else
      zsqrt(&RE(arg),&IM(arg),&x,&y);
#endif
      res = Kcomp(&x,&y);
    }
  return res;
}
ACTION_END

DEF_ACTION_1ARG(abs)
{
  REAL x;
  EXTENDED rex,rey;

  if (arg->kind==(CHAR)E_REAL)
    {
      if (classifyD(&XVAL(arg))>=0)
	return arg;
      x = -XVAL(arg);
    }
  else
    {
      rex = RE(arg);
      rey = IM(arg);
      x = sqrt(rex*rex+rey*rey);
    }
  return Kreal(&x);
}
ACTION_END

DEF_ACTION_1ARG(re)
{
  if (arg->kind==(CHAR)E_REAL)
    return arg;
  return REPART(arg);
}
ACTION_END

DEF_ACTION_1ARG(im)
{
  if (arg->kind==(CHAR)E_REAL)
    return zzero;
  return IMPART(arg);
}
ACTION_END

DEF_ACTION_1ARG(conj)
{
  if (arg->kind==(CHAR)E_REAL)
    return arg;
  {
    ExprPtr im;
    ENTER;

    im = ACTION_NAME(neg)(env, IMPART(arg)); /* No need to call ENV_CONS(ENV_UNDERLYING(env),1),
						because this is just env. */
    (void) KEEP(im);
    arg = Kcomplex(REPART(arg),im);
    QUIT;
    return arg;
  }
}
ACTION_END

DEF_ACTION_2ARGS(less)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  return (XVAL(arg1)<XVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(leq)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
#if PLDI93_HACK
  return (XVAL(arg1)<=XVAL(arg2))?trueValue:falseValue;
#else
  return (arg1==arg2 || XVAL(arg1)<XVAL(arg2))?trueValue:falseValue;
#endif
}
ACTION_END

DEF_ACTION_2ARGS(greater)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  return (XVAL(arg1)>XVAL(arg2))?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(greq)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
#if PLDI93_HACK
  return (XVAL(arg1)>=XVAL(arg2))?trueValue:falseValue;
#else
  return (arg1==arg2 || XVAL(arg1)>XVAL(arg2))?trueValue:falseValue;
#endif
}
ACTION_END

DEF_ACTION_2ARGS(nummin)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  return (XVAL(arg1)>=XVAL(arg2))?arg2:arg1;
}
ACTION_END

DEF_ACTION_2ARGS(nummax)
{
  if (arg1->kind!=(CHAR)E_REAL || arg2->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  return (XVAL(arg1)>=XVAL(arg2))?arg1:arg2;
}
ACTION_END

#if EXTP
#define MONADIC(name,rfun,zfun) \
  DEF_ACTION_1ARG(name) \
  { \
    EXTENDED rx,ry,resx,resy; \
    REAL x,y; \
    ExprPtr res; \
 \
    if (arg->kind==(CHAR)E_REAL) \
      { \
	x = rfun(XVAL(arg)); \
	res = Kreal(&x); \
      } \
    else \
      { \
	rx = RE(arg); \
	ry = IM(arg); \
	zfun(&rx,&ry,&resx,&resy); \
	x = resx; \
	y = resy; \
	res = Kcomp(&x,&y); \
      } \
    return res; \
  }\
  ACTION_END
#else
#define MONADIC(name,rfun,zfun) \
  DEF_ACTION_1ARG(name) \
  { \
    REAL x,y; \
    ExprPtr res; \
 \
    if (arg->kind==(CHAR)E_REAL) \
      { \
	x = rfun(XVAL(arg)); \
	res = Kreal(&x); \
      } \
    else \
      { \
	zfun(&RE(arg),&IM(arg),&x,&y); \
	res = Kcomp(&x,&y); \
      } \
    return res; \
  }\
  ACTION_END
#endif

MONADIC(exp,exp,zexp)
MONADIC(exp1,exp1,zexp1)
MONADIC(sin,sin,zsin)
MONADIC(cos,cos,zcos)
MONADIC(tan,tan,ztan)
MONADIC(sh,sinh,zsh)
MONADIC(ch,cosh,zch)
MONADIC(th,tanh,zth)
MONADIC(atan,atan,zatan)

DEF_ACTION_1ARG(log)
{
  ExprPtr res;
  REAL x,y;
#if EXTP
  EXTENDED rex,rey,resx,resy;
#endif

  if (arg->kind==(CHAR)E_REAL)
    {
      if (classifyD(&XVAL(arg))>=0)
	{
	  x = log(XVAL(arg));
	  res = Kreal(&x);
	}
      else
	{
	  ENTER;

	  x = log(-(EXTENDED)XVAL(arg));
	  arg = Kreal(&x);
	  res = Kcomplex(KEEP(arg),zpi);
	  QUIT;
	}
    }
  else
    {
#if EXTP
      rex = RE(arg);
      rey = IM(arg);
      zlog(&rex,&rey,&resx,&resy);
      x = resx;
      y = resy;
#else
      zlog(&RE(arg),&IM(arg),&x,&y);
#endif
      res = Kcomp(&x,&y);
    }
  return res;
}
ACTION_END

DEF_ACTION_1ARG(log1)
{
  ExprPtr res;
  REAL x,y;
  EXTENDED xx;
#if EXTP
  EXTENDED rex,rey,resx,resy;
#endif

  if (arg->kind==(CHAR)E_REAL)
    {
      xx = XVAL(arg);
      if (xx>=fmone)
	{
	  x = log1(xx);
	  res = Kreal(&x);
	}
      else
	{
	  ENTER;

	  x = log(fmone-xx);
	  arg = Kreal(&x);
	  res = Kcomplex(KEEP(arg),zpi);
	  QUIT;
	}
    }
  else
    {
#if EXTP
      rex = RE(arg);
      rey = IM(arg);
      zlog1(&rex,&rey,&resx,&resy);
      x = resx;
      y = resy;
#else
      zlog1(&RE(arg),&IM(arg),&x,&y);
#endif
      res = Kcomp(&x,&y);
    }
  return res;
}
ACTION_END

DEF_ACTION_1ARG(asin)
{
  ExprPtr res;
  REAL x,y;
  EXTENDED xx,resx,resy;
#if EXTP
  EXTENDED rex,rey;
#endif

  if (arg->kind==(CHAR)E_REAL)
    {
      xx = XVAL(arg);
      if (xx>=fmone && xx<=fone)
	{
	  x = asin(xx);
	  res = Kreal(&x);
	}
      else
	{
	  zasin(&xx,&zero,&resx,&resy);
#if EXTP
	  x = resx;
	  y = resy;
	  res = Kcomp(&x,&y);
#else
	  res = Kcomp(&resx,&resy);
#endif
	}
    }
  else
    {
#if EXTP
      rex = RE(arg);
      rey = IM(arg);
      zasin(&rex,&rey,&resx,&resy);
      x = resx;
      y = resy;
#else
      zasin(&RE(arg),&IM(arg),&x,&y);
#endif
      res = Kcomp(&x,&y);
    }
  return res;
}
ACTION_END

DEF_ACTION_1ARG(acos)
{
  ExprPtr res;
  REAL x,y;
  EXTENDED xx,resx,resy;
#if EXTP
  EXTENDED rex,rey;
#endif

  if (arg->kind==(CHAR)E_REAL)
    {
      xx = XVAL(arg);
      if (xx>=fmone && xx<=fone)
	{
	  x = acos(xx);
	  res = Kreal(&x);
	}
      else
	{
	  zacos(&xx,&zero,&resx,&resy);
#if EXTP
	  x = resx;
	  y = resy;
	  res = Kcomp(&x,&y);
#else
	  res = Kcomp(&resx,&resy);
#endif
	}
    }
  else
    {
#if EXTP
      rex = RE(arg);
      rey = IM(arg);
      zacos(&rex,&rey,&resx,&resy);
      x = resx;
      y = resy;
#else
      zacos(&RE(arg),&IM(arg),&x,&y);
#endif
      res = Kcomp(&x,&y);
    }
  return res;
}
ACTION_END

#if EXTP
#define MON(name,zfun) \
  DEF_ACTION_1ARG(name) \
  { \
    EXTENDED x,y,resx,resy; \
    REAL rx,ry; \
 \
    if (arg->kind==(CHAR)E_REAL) \
      { \
	x = XVAL(arg); \
	zfun(&x,&zero,&resx,&resy); \
      } \
    else \
      { \
	x = RE(arg); \
	y = IM(arg); \
	zfun(&x,&y,&resx,&resy); \
      } \
    rx = resx; \
    ry = resy; \
    return Kcomp(&rx,&ry); \
  }\
  ACTION_END
#else
#define MON(name,zfun) \
  DEF_ACTION_1ARG(name) \
  { \
    REAL rx,ry; \
 \
    if (arg->kind==(CHAR)E_REAL) \
      { \
	zfun(&XVAL(arg),&zero,&rx,&ry); \
      } \
    else \
      { \
	zfun(&RE(arg),&IM(arg),&rx,&ry); \
      } \
    return Kcomp(&rx,&ry); \
  }\
  ACTION_END
#endif

MON(ash,zash)
MON(ach,zach)
MON(ath,zath)

DEF_ACTION_1ARG(arg)
{
  REAL angle;

  if (arg->kind==(CHAR)E_REAL)
    {
      if (classifyD(&XVAL(arg))>=0)
	return zzero;
      return zpi;
    }
  else
    {
      angle = atan2(IM(arg),RE(arg));
      return Kreal(&angle);
    }
}
ACTION_END

DEF_ACTION_1ARG(real)
{
  return (arg->kind==(CHAR)E_REAL)?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_1ARG(integer)
{
#if EXTP
  EXTENDED x;
#endif

  if (arg->kind!=(CHAR)E_REAL)
    return falseValue;
#if EXTP
  x = XVAL(arg);
  return (floor(x)==x)?trueValue:falseValue;
#else
  return (floor(XVAL(arg))==XVAL(arg))?trueValue:falseValue;
#endif
}
ACTION_END

DEF_ACTION_1ARG(natural)
{
#if EXTP
  EXTENDED x;
#endif

  if (arg->kind!=(CHAR)E_REAL)
    return falseValue;
  if (classifyD(&XVAL(arg))<0)
    return falseValue;
#if EXTP
  x = XVAL(arg);
  return (floor(x)==x)?trueValue:falseValue;
#else
  return (floor(XVAL(arg))==XVAL(arg))?trueValue:falseValue;
#endif
}
ACTION_END

DEF_ACTION_1ARG(floor)
{
  REAL x;

  if (arg->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  x = floor(XVAL(arg));
  return Kreal(&x);
}
ACTION_END

DEF_ACTION_1ARG(ceil)
{
  REAL x;

  if (arg->kind!=(CHAR)E_REAL)
    return Raise(INCOMPEXC);
  x = ceil(XVAL(arg));
  return Kreal(&x);
}
ACTION_END

DEF_ACTION_2ARGS(quotient)
{
  ExprPtr div,aux;
  REAL x;

  if (arg1==divmodCache_a && arg2==divmodCache_b)
    {
      if (divmodCache_d!=NIL)
	return divmodCache_d;
      aux = CAR(divmodCache_dm);
      SETQ(divmodCache_d,aux);
      return aux;
    }
  SETQ(divmodCache_a,arg1);
  SETQ(divmodCache_b,arg2);
#if MAX_ARITY>=2
  div = ACTION_NAME(div)(ENV_CONS(ENV_UNDERLYING(env),2),arg1,arg2);
#else
  div = ACTION_NAME(div)(env,_arg); /* In this case, env==ENV_CONS(ENV_UNDERLYING(env),1) */
#endif
  if (div->kind==(CHAR)E_COMPLEX)
    div = REPART(div);
  x = floor(XVAL(div));
  SETQ(divmodCache_dm,NIL);
  SETQ(divmodCache_m,NIL);
  aux = Kreal(&x);
  SETQ(divmodCache_d,aux);
  return aux;
}
ACTION_END

DEF_ACTION_2ARGS(remainder)
{
  ExprPtr quot,aux;

  if (arg1==divmodCache_a && arg2==divmodCache_b)
    {
      if (divmodCache_m!=NIL)
	return divmodCache_m;
      if (divmodCache_dm!=NIL)
	{
	  aux = CDR(divmodCache_dm);
	  SETQ(divmodCache_m,aux);
	  return aux;
	}
      quot = divmodCache_d;
    }
  else
    {
#if MAX_ARITY>=2
      quot = ACTION_NAME(quotient)(ENV_CONS(ENV_UNDERLYING(env),2),arg1,arg2);
#else
      quot = ACTION_NAME(quotient)(env,_arg); /* In this case, env==ENV_CONS(ENV_UNDERLYING(env),1) */
#endif
    }
  {
    ENTER;

    aux = z_mul(arg2,quot);
    aux = z_sub(arg1,KEEP(aux));
    SETQ(divmodCache_m,aux);
    QUIT;
  }
  return aux;
}
ACTION_END

DEF_ACTION_2ARGS(quotrem)
{
  ExprPtr rem,aux;

  if (arg1==divmodCache_a && arg2==divmodCache_b)
    {
      if (divmodCache_dm!=NIL)
	return divmodCache_dm;
    }
#if MAX_ARITY>=2
  rem = ACTION_NAME(remainder)(ENV_CONS(ENV_UNDERLYING(env),2),arg1,arg2);
#else
  rem = ACTION_NAME(remainder)(env,_arg); /* In this case, env==ENV_CONS(ENV_UNDERLYING(env),1) */
#endif
  aux = Ktuple2(divmodCache_d,rem);
  SETQ(divmodCache_dm,aux);
  return aux;
}
ACTION_END

DEF_ACTION_0ARG(rrand)
{
  REAL x;

  randReal(&x);
  return Kreal(&x);
}
ACTION_END

DEF_ACTION_0ARG(zrand)
{
  REAL x,y;

  randReal(&x);
  randReal(&y);
  return Kcomp(&x,&y);
}
ACTION_END

DEF_ACTION_1ARG(denormal)
{
  if (arg->kind==(CHAR)E_COMPLEX)
    {
      switch (classifyD(&XVAL(CAR(arg))))
	{
	case C_ZERONUM: case -C_ZERONUM:
	case C_DENORMALNUM: case -C_DENORMALNUM:
	  switch (classifyD(&XVAL(CDR(arg))))
	    {
	    case C_ZERONUM: case -C_ZERONUM:
	    case C_DENORMALNUM: case -C_DENORMALNUM:
	      return trueValue;
	    }
	}
      return falseValue;
    }
  switch (classifyD(&XVAL(arg)))
    {
    case C_ZERONUM: case -C_ZERONUM:
    case C_DENORMALNUM: case -C_DENORMALNUM:
      return trueValue;  
    }
  return falseValue;
}
ACTION_END

DEF_ACTION_1ARG(rotten)
{
  if (arg->kind==(CHAR)E_COMPLEX)
    {
      switch (classifyD(&XVAL(CAR(arg))))
	{
	case C_QNAN:
	case C_SNAN: case -C_SNAN:
	case C_INFINITE: case -C_INFINITE:
	  return trueValue;
	}
      switch (classifyD(&XVAL(CDR(arg))))
	{
	case C_QNAN:
	case C_SNAN: case -C_SNAN:
	case C_INFINITE: case -C_INFINITE:
	  return trueValue;
	}
      return falseValue;
    }
  switch (classifyD(&XVAL(arg)))
    {
    case C_QNAN:
    case C_SNAN: case -C_SNAN:
    case C_INFINITE: case -C_INFINITE:
      return trueValue;  
    }
  return falseValue;
}
ACTION_END

DEF_ACTION_2ARGS(ldexp)
{
  COUNTER exp = IVAL (arg2);
  ExprPtr res;
  REAL rex, rey;

  if (arg1->kind==(CHAR)E_REAL)
    {
      rex = ldexp (XVAL (arg1), exp);
      res = Kreal (&rex);
    }
  else
    {
      rex = ldexp (RE (arg1), exp);
      rey = ldexp (IM (arg1), exp);
      res = Kcomp (&rex, &rey);
    }
  return res;
}
ACTION_END

DEF_ACTION_1ARG(frexp)
{
  int exp;
  REAL x, y;
  ExprPtr res, xp;
  ENTER;

  if (arg->kind==(CHAR)E_REAL)
    {
      x = frexp (XVAL (arg), &exp);
      res = Kreal (&x);
    }
  else
    {
      x = frexp (RE (arg), &exp);
      y = ldexp (IM (arg), -exp);
      res = Kcomp (&x, &y);
    }
  (void) KEEP (res);
  xp = Kint (exp);
  (void) KEEP (xp);
  res = Ktuple2 (res, xp);
  QUIT;
  return res;
}
ACTION_END

DEF_ACTION_1ARG(modf)
{
  double intpart;
  REAL ipart, fpart;
  ExprPtr ipx, fpx, res;
  ENTER;

  if (arg->kind!=(CHAR)E_REAL)
    return Raise (INCOMPEXC);
  fpart = modf (XVAL (arg), &intpart);
  ipart = intpart;
  ipx = Kreal (&ipart);
  (void) KEEP (ipx);
  fpx = Kreal (&fpart);
  (void) KEEP (fpx);
  res = Ktuple2 (ipx, fpx);
  QUIT;
  return res;
}
ACTION_END

#undef __SEG__
#define __SEG__ INIT

PUBLIC void InitArith()
{
  divmodCache_a=divmodCache_b=divmodCache_d=divmodCache_m=divmodCache_dm=NIL;
}
