/* Compiler to interpreted form.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "compile.h"
#include "type.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "regexp.h"
#include "grmproto.h"
#include "evlproto.h"
#include "cmpproto.h"
#include "errproto.h"
#include "prtproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rcompile.h"
#ifdef HAS_PROTOTYPES
#include "scompile.h"
#endif

#define __SEG__ COMPILE

#undef DEBUG
#undef DEBUGMOD
#undef DBGINLINE

#define EP(expr) valPrint(expr,alpha,do_printfile,(char *)hstderr,0)

#ifdef OLDDATA
PUBLIC ExprPtr nilTags,stringTags,intTags,listTags,refTags,
	       arrayTags,mapTags,boolTags,exnTags,
	       promiseTags,contTags,dynTags,numTags,
	       funTags,consTags,record0Tags,record1Tags,
	       recordNTags,record1NTags,recordTags,
	       tuple0Tags,tupleTags,tuple1Tags,tuple12NTags,
	       tuple2Tags,tuple2NTags,tupleNTags
;
#endif

EXTERN SymPtr equalitySym,discrSym;
EXTERN ExprPtr basisTable,basisEnv;
#if DEBUGGING
EXTERN ExprPtr currentFile;
EXTERN ExprPtr fntrue;
EXTERN ExprPtr benignExcRegexpRef;
EXTERN Boolean debugEnabled;
#endif
#if PROFILING
PUBLIC ExprPtr profileExprs;
#endif

EXTERN SymPtr itSym;

EXTERN ExprPtr maxInlineRef;

EXTERN CHAR *yyfilename;
PUBLIC COUNTER minDebugLevel=0;

IMPORT ExprPtr MkQuote P((ExprPtr e));

PRIVATE ExprPtr compile P((IN CONST ExprPtr expr,IN CONST ExprPtr assum,IN COUNTER gto));

#define CP(expr,assum,gto) compile(expr,assum,gto)

PRIVATE ExprPtr Cons(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return z;
}

#ifdef OBSOLETE
PRIVATE ExprPtr MapletCons(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKEMAPLET(x,y,z);
  return z;
}
#endif

PRIVATE ExprPtr CONS(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return KEEP(z);
}

#ifdef OBSOLETE
PRIVATE ExprPtr MAPLETCONS(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKEMAPLET(x,y,z);
  return KEEP(z);
}
#endif

	/*****  compilation *****/

PRIVATE ExprPtr cpLiteral(expr,assum,gto) /* A_STRING, A_NUMBER, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return CAR(ARGS(expr));
}

PUBLIC ExprPtr getTopEnv()
{
  ExprPtr env;

  env = basisEnv;
  while (ENEXT(env)!=NIL)
    env = ENEXT(env);
  return env;
}

PUBLIC COUNTER itIndex()
{
  return ORIGINAL_IT;
}

PRIVATE ExprPtr getType(e,assum,foundp)
IN CONST ExprPtr e;
IN CONST ExprPtr assum;
OUT Boolean *foundp;
{
  ExprPtr type;

  type = NGet(e,S_ANNOTATIONS(C_TYSTATE(assum)),foundp);
  if (*foundp && type->kind==(CHAR)E_BIAS)
    type = CAR(type);
  return type;
}

PRIVATE ExprPtr getTypeSure(e,assum)
IN CONST ExprPtr e;
IN CONST ExprPtr assum;
{
  ExprPtr type;
  Boolean found;

  type = getType(e,assum,&found);
  if (!found)
    berror(message(72));
  return type;
}

PRIVATE COUNTER funArity(expr,assum) /* return 0 if function is 0-ary, 1 if unary,
					etc., -1 if do not know. */
     IN CONST ExprPtr expr;
     IN CONST ExprPtr assum;
{
  Boolean found;
  ExprPtr type,aux;

  type = getType (expr,assum,&found);
  if (!found)
    return -1;
  if (type->kind!=(CHAR)E_TYPE_APPL)
    return -1;
  aux = APPLFUN(type);
  if (fxSym(aux)!=funSym)
    return -1;
  type = CAR(APPLARG(type));
  switch (type->kind)
    {
    case E_VAR: return -1; /* polymorphic: may fail to be unary. */
    case E_NIL: return 0;
    case E_CONS:
      {
	COUNTER n;

	for (n=0; type->kind==(CHAR)E_CONS; type=CDR(type))
	  n++;
	if (type->kind==(CHAR)E_NIL)
	  return n;
	return -1; /* a tuple type with an ellipsis. */
      }
    default: return 1;
    }
}

#if DEBUGGING!=0

PRIVATE ExprPtr wrapDebugZone(expr,ex,cnt,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr ex;
IN COUNTER cnt;
IN CONST ExprPtr assum;
{
  ExprPtr info,dz;
  ExprPtr args,type;
  Boolean found;
  ENTER;

  if (C_FLAGS(assum) & C_NODEBUG)
    return ex; /* in (*$D-*) mode, don't put debug zone headers */
  switch (ex->kind)
    {
    case E_DEBUGZONE:
      return ex; /* do not put double debug zone headers */
    }
  type = NGet(expr,S_ANNOTATIONS(C_TYSTATE(assum)),&found);
  if (!found)
    berror(message(428));
  args = KONS(type,KONS(currentFile,NIL));
  /* We use KONS so as to share all KONS(file,NIL)
     which have the same file, and among them to share all those that have the
     same type; as a consequence, these cells must not be patched... */
  info = ZBuild(ZONE(expr),A_CODEPOINT,args);
  dz = MakeDebugZone(ex,KEEP(info));
  KEEPONLY(dz);
  return dz;
}

PRIVATE ExprPtr wrapDebugName1(type,name,ex,assum,consp,excp)
IN CONST ExprPtr type;
IN CONST ExprPtr name;
IN CONST ExprPtr ex;
IN CONST ExprPtr assum;
IN Boolean consp,excp;
{
  ExprPtr wrap,aux;
  ENTER;

  aux = MakePair(excp?E_GOTO:(consp?E_APPL:E_LET),name,type);
  (void) KEEP(aux);
  aux = MakeCounted(E_COUNTED,aux,(INT)VARRANK(ex)
#ifdef OLDDEBUG
		    +1
#endif
		    );
  wrap = MakeDebugName(ex,KEEP(aux));
  KEEPONLY(wrap);
  return wrap;
}

PRIVATE ExprPtr wrapDebugName(expr,name,ex,assum,consp,excp)
IN CONST ExprPtr expr;
IN CONST ExprPtr name;
IN CONST ExprPtr ex;
IN CONST ExprPtr assum;
IN Boolean consp,excp;
{
  ExprPtr type;
  Boolean found;

  if (STEXT(name)[0]=='.') /* a temporary name generated for instance by freshVarName() */
    return ex;
  type = NGet(expr,S_ANNOTATIONS(C_TYSTATE(assum)),&found);
  if (!found)
    berror(message(429));
  if (consp) /* if what we got was a constructor map */
    {
      if (type->kind==(CHAR)E_BIAS)
	type = CAR(type);
      type = MapGetSure(name,type); /* get the type of the constructor */
    }
  return wrapDebugName1(type,name,ex,assum,consp,excp);
}

#endif

PRIVATE Boolean noSharingSure(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr type;
  Boolean found;

  type = NGet(expr,S_ANNOTATIONS(C_TYSTATE(assum)),&found);
  if (!found)
    berror(message(672));
  if (type->kind==(CHAR)E_BIAS)
    return CDR(CDR(type))==falseValue;
  return false;
#ifdef OBSOLETE
  ExprPtr type;

  type = getTypeSure(expr,assum);
  return eqTypeTdg(type,C_TYSTATE(assum),true)==falseValue;
#endif
}

#define noSharingPossibly noSharingSure

#ifdef OBSOLETE
/* I used this once when there was a bug and some annotations
   (even those not at toplevel or with sharingp!=trueValue)
   did not have any E_BIAS.  So I had to be conservative.
   Normally, not having a bias means that this is a toplevel
   construction and sharing is true, so we return false as
   in noSharingSure. */
PRIVATE Boolean noSharingPossibly(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr type;
  Boolean found;

  type = NGet(expr,S_ANNOTATIONS(C_TYSTATE(assum)),&found);
  if (!found)
    berror(message(672));
  if (type->kind==(CHAR)E_BIAS)
    return CDR(CDR(type))==falseValue;
  return true;
}
#endif

PRIVATE Boolean cstExpr(e,val)
IN CONST ExprPtr e;
OUT ExprPtr *val;
{
again:
  switch (e->kind)
    {
    case E_NIL: case E_TUPLE_0: case E_RECORD_0: case E_EMPTY:
    case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      *val = e;
      return true;
    case E_QUOTE:
      *val = REFVAL(e);
      return true;
    case E_DEBUGZONE:
      e = DZCODE(e);
      goto again;
    default:
      return false;
    }
}

PRIVATE ExprPtr cpNonSense P((IN CONST ExprPtr expr,IN CONST ExprPtr assum,IN CONST COUNTER gto));

PRIVATE Boolean simpleConstant1 P((IN CONST ExprPtr e));

PRIVATE MapPtr simpleConstantRef = NOEXPR;

DEF_PRIVATE_ACTION_1ARG(simpleconstant)
{
  return simpleConstant1(ENV_UNDERLYING(env))?trueValue:falseValue;
}
ACTION_END

PRIVATE Boolean simpleConstant1(e)
IN CONST ExprPtr e;
{
  if (STKFULL(e))
    return CallCC(ACTION_NAME(simpleconstant),e,true)==trueValue;
again:
  {
    Boolean found;

    (void) NGet(e, REFVAL(simpleConstantRef), &found);
    if (found)
      return true; /* break cycles */
    setNPut(e,UNIT,&REFVAL(simpleConstantRef),simpleConstantRef);
    /* remember that we have already started to examine e. */
  }
  switch (e->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0: case E_DATACON:
    case E_QUOTE:
#ifdef OBSOLETE
    case E_ARRAY_0:
#endif
      return true;
    case E_TUPLE_1: case E_EXCEPTION: case E_REF:
#ifdef OBSOLETE
    case E_ARRAY_1:
#endif
    case E_COUNTED:
      e = REFVAL(e);
      goto again;
    case E_RECORD_1:
    case E_DATA:
      e = CDR(e);
      goto again;
    case E_CONS: case E_MAPLET: case E_MAP: case E_TUPLE_2: case E_ARRAY_2:
      if (!simpleConstant1(CAR(e)))
	return false;
      e = CDR(e);
      goto again;
    case E_TUPLE_GNRL: case E_RECORD_GNRL: case E_ARRAY_GNRL:
      {
	COUNTER i,n;

	for (i=0, n = AVAL(e)->n; i<n; i++)
	  if (!simpleConstant1(AVAL(e)->expr[i]))
	    return false;
	return true;
      }
    }
  return false;
}

PRIVATE Boolean simpleConstant(e)
IN CONST ExprPtr e;
{
  Boolean ok;
  ENTER;

  simpleConstantRef = MakeSing(E_REF,NIL);
  (void) KEEP(simpleConstantRef);
  ok = simpleConstant1(e);
  simpleConstantRef = NOEXPR;
  QUIT;
  return ok;
}

PRIVATE ExprPtr cpVar(expr,assum,gto,checkinline)
     /* A_VARIABLE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
IN CONST COUNTER checkinline;
{
  ExprPtr i_l,name,aux,con;
  COUNTER i,l,level;
  Boolean found;

  name = CAR(ARGS(expr));
  i_l = MapGetSure(name,C_IDS(assum));
  /* in the A_VARIABLE case, tc() should ensure that such an i_l is found;
     in the A_EXT_VAL case, if it is not found, then it is a value to be
     resolved at link time. */
  if (i_l->kind==(CHAR)E_CONS)
    {
      i = IVAL(CAR(i_l));
      l = IVAL(CDR(i_l));
    }
  else
    {
      i = IVAL(i_l);
      l = 0;
    }
  level = C_LEVEL(assum);
  if (l==0)
    {
      ExprPtr env,val;

      env = getTopEnv();
      if (i==ORIGINAL_IT)
	{
	  if (minDebugLevel!=0) /* if we are compiling a module */
	    { /* then using 'it' is illegal. */
	      type_soft_error(ZONE(expr),(CHAR *)message(812));
	      return cpNonSense(expr,assum,gto);
	    }
	  aux = MkQuote(EVARS(env)->expr[i]);
	  return KEEP(aux);
	}
      if (i<EVARS(env)->n)
	{
	  val = EVARS(env)->expr[i];
	  if (simpleConstant(val))
	    {
	      aux = MkQuote(val);
	      return KEEP(aux);
	    }
	}
    }
  level -= l;
  aux = Kvarref(i,level);
  (void) KEEP(aux);
  if (checkinline)
    {
      con = MapGet(i_l,C_CONS(assum),&found);
      if (found)
	switch (con->kind)
	  {
	  case E_COUNTED: /* possibly inlineable function */
#if DEBUGGING
	    if (debugEnabled)
	      {
		if (dbgVetoInline (name, i_l))
		  break;
	      }
#endif
	    {
	      COUNTER ll = C_LEVEL (assum) - COUNTEDN (con);
	      con = COUNTEDVAL (con);
	      con = shiftExpr (con, ll);
	    do_inline:
	      aux = MakePair(E_PAT_AS,aux,con);
	      (void) KEEP(aux);
	    }
	    break;
	  case E_DATACON: /* datatype constructor taking an argument, always inlineable */
	    goto do_inline;
	  }
    }
  return aux;
}

PRIVATE ExprPtr cpVariable(expr,assum,gto)
     /* A_VARIABLE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpVar(expr, assum, gto, 1);
}

#define cpString cpLiteral
#define cpNumber cpLiteral

PRIVATE ExprPtr cpError(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  crash_error(ZONE(expr),(CHAR *)message(425));
  return NIL;
}

	/* the following should be patched out by tc() */
#define cpUnitPower cpError
#define cpUnitStar cpError
#define cpSetUnit cpError

PRIVATE ExprPtr cpLabel(expr,assum,gto) /* A_LABEL, return label (string) */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  SymPtr e;

  e = CAR(ARGS(expr));
  return e;
}

PRIVATE ExprPtr cpMaplet(expr,assum,gto)
	/* A_RECORD_MAPLET, A_PAT_RECORD_MAPLET, A_MAPLET,
	   return couple (label,expr), (label,pat) or (expr,expr) */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,lbl,ex;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  lbl = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  ex = CP(e,assum,-1);
  ex = Cons(lbl,ex);
  KEEPONLY(ex);
  return ex;
}

#define cpRecordMaplet cpMaplet

PRIVATE ExprPtr MkMaplet(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr vx,vy;

  if (cstExpr(x,&vx) && cstExpr(y,&vy))
    return MkQuote(MapSingleton(vx,vy));
  MAKEMAPLET(x,y,vx);
  return vx;
}

PRIVATE ExprPtr MkMerge(x,y,revp)
IN CONST ExprPtr x,y;
IN Boolean revp;
{
  ExprPtr vx,vy;

  if (cstExpr(x,&vx) && cstExpr(y,&vy))
    return MkQuote(revp?MapMerge(vy,vx):MapMerge(vx,vy));
  return MakePair(revp?E_UNDERWRITE:E_OVERWRITE,x,y);
}

PRIVATE ExprPtr cpMap1(argsp,len,assum,revp)
INOUT ExprPtr *argsp;
IN CONST INT len;
IN CONST ExprPtr assum;
IN Boolean revp;
{
  ExprPtr e,maplet,ex1,ex2;
  INT sublen;

  switch (len)
    {
      case 0L:
	return NewMap();
      case 1L:
	e = CAR(*argsp);
	*argsp = CDR(*argsp);
	maplet = CP(e,assum,-1);
	e = MkMaplet(CAR(maplet),CDR(maplet));
	return KEEP(e);
      default:
	sublen = len>>1;
	ex1 = cpMap1(argsp,sublen,assum,revp);
	ex2 = cpMap1(argsp,len-sublen,assum,revp);
	e = MkMerge(ex1,ex2,revp);
	return KEEP(e);
    }
}

PRIVATE ExprPtr cpMap(expr,assum,gto)
	/* A_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr ex,args;
  ENTER;

  args = ARGS(expr);
  ex = cpMap1(&args,ListLength(args),assum,false);
#if DEBUGGING
  (void) KEEP(ex);
  ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum);
#endif
  KEEPONLY(ex);
  return ex;
}

MAPSWEEPFUNC(do_getfields)
{
  field **fp;

  fp = MAPDATA(field **);
  (*fp)->label = x;
  (*fp)->val = y;
  (*fp)++;
  return false;
}

PRIVATE MapPtr getShape(ordermap) /* also used when ordermap is a record type */
IN CONST ExprPtr ordermap;
{
  COUNTER n;
  MapPtr shape;
  ENTER;

  n = MapCard(ordermap);
  {
    field *fp;
    field *fields;

    fields = (field *)alloctmp(n*sizeof(field));
    fp = fields;
    (void) MapSweep(ordermap,do_getfields,&fp);
    sortFieldArray(n,fields);
    shape = fieldsToShape(n,fields);
  }
  KEEPONLY(shape);
  return shape;
}

PRIVATE ExprPtr MkRecord1(x,y,nosharing)
IN CONST ExprPtr x,y;
IN Boolean nosharing;
{
  ExprPtr vy;

  if (cstExpr(y,&vy))
    {
      vy = Krecord1(x,vy);
      (void) KEEP(vy);
      return MkQuote(vy);
    }
  return MakePair(nosharing?E_UNSHARED_RECORD_1:E_RECORD_1,x,y);
}

PRIVATE ExprPtr MkRecordGnrl(shape,len,fields,nosharing)
IN CONST ExprPtr shape;
IN INT len;
IN CONST ExprPtr *fields;
IN Boolean nosharing;
{
  ExprPtr *vals = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
  ExprPtr *fp;
  ExprPtr *index;
  INT i;

  index = AVAL(shape)->expr;
  for (i=0, fp = fields; i<len; i++)
    if (!cstExpr(*fp++, &vals[IVAL(*index++)-1]))
      break;
  if (i==len)
    {
      shape = KrecordGnrl(ASHAPE(shape),len,vals);
      (void) KEEP(shape);
      return MkQuote(shape);
    }
  return MakeRecord(nosharing?E_UNSHARED_RECORD_GNRL:E_RECORD_GNRL,shape,(COUNTER)len,fields);
}

PRIVATE ExprPtr cpRecord0(args,assum,len,compileswaps,nosharing)
	/* A_RECORD, return expr */
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
IN INT len;
IN Boolean compileswaps,nosharing;
{
  ExprPtr e;
  ENTER;

  switch (len)
    {
    case 0:
      e = record0;
      break;
    case 1:
      {
	ExprPtr aux;

	aux = CP(CAR(args),assum,-1);
	e = MkRecord1(CAR(aux),CDR(aux),nosharing);
      }
      break;
    default:
      {
	COUNTER i;
	field *fp;
	ExprPtr maplet,shape;
	field *fields;

	fields = (field *)alloctmp_anchored(len*sizeof(field), (char *)&fields);
	for (i=0, fp=fields; i<len; i++, fp++)
	  {
	    maplet = CP(CAR(args),assum,-1);
	    args = CDR(args);
	    fp->label = CAR(maplet);
	    fp->val = CDR(maplet);
	  }
	if (compileswaps)
	  {
	    field *fp2;
	    field *fields2;

	    fields2 = (field *)alloctmp_anchored(len*sizeof(field), (char *)&fields);
	    /* due to the fact that execution ordering matters, we must for example
	       compile |[b = e1, a = e2]| by evaluating e1 then e2, not the converse.
	       Then, we have to put the a field first, and the b field first.
	       This is why instead of building a regular shape (map labels -> indices),
	       we build a record of shape relocations (i.e, nth expression -> its position
	       in the final record). */
	    for (i=0, fp=fields, fp2=fields2; i<len;i++,  fp++, fp2++)
	      fp2->label = fp->label;
	    sortFieldArray((COUNTER)len,fields2);
	    shape = fieldsToShape((COUNTER)len,fields2);
	    for (i=0, fp=fields, fp2=fields2; i<len; i++, fp++, fp2++)
	      fp2->val = MapGetSure(fp->label,shape);
	    fieldsToExprs((ExprPtr *)fields2,fields2,len); /* a bit hackish, but should work. */
	    shape = KrecordGnrl(shape,len,(ExprPtr *)fields2);
	  }
	else
	  {
	    sortFieldArray((COUNTER)len,fields);
	    shape = fieldsToShape((COUNTER)len,fields);
	  }
	fieldsToExprs((ExprPtr *)fields,fields,len); /* a bit hackish, but should work. */
	e = MkRecordGnrl(shape,len,(ExprPtr *)fields,nosharing);
      }
      break;
    }
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpRecord(expr,assum,gto)
	/* A_RECORD, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e,args;
  ENTER;

  args = ARGS(expr);
  e = cpRecord0(args,assum,ListLength(args),true,noSharingSure(expr,assum));
#ifdef OBSOLETE
#if DEBUGGING
  (void) KEEP(e);
  e = wrapDebugZone(expr,e,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr mkIL(i,l)
IN COUNTER i,l;
{
  ExprPtr i_l;
  ENTER;

  i_l = KINT(i);
  if (l!=0)
    i_l = KONS(i_l,KINT(l));
  KEEPONLY(i_l);
  return i_l;
}

#define cpFixity cpError

PRIVATE ExprPtr MkTuple1(x,nosharing)
IN CONST ExprPtr x;
IN Boolean nosharing;
{
  ExprPtr vx;

  if (cstExpr(x,&vx))
    {
      vx = Ktuple1(vx);
      (void) KEEP(vx);
      return MkQuote(vx);
    }
  return MakeSing(nosharing?E_UNSHARED_TUPLE_1:E_TUPLE_1,x);
}

PRIVATE ExprPtr MkTuple2(x,y,nosharing)
IN CONST ExprPtr x,y;
IN Boolean nosharing;
{
  ExprPtr vx,vy;

  if (cstExpr(x,&vx) && cstExpr(y,&vy))
    {
      vy = Ktuple2(vx,vy);
      (void) KEEP(vy);
      return MkQuote(vy);
    }
  return MakePair(nosharing?E_UNSHARED_TUPLE_2:E_TUPLE_2,x,y);
}

PRIVATE ExprPtr MkTupleGnrl(len,fields,nosharing)
IN INT len;
IN CONST ExprPtr *fields;
IN Boolean nosharing;
{
  ExprPtr *vals = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
  ExprPtr *fp,*vp;
  ExprPtr e;
  INT i;

  for (i=0, fp = fields, vp = vals; i<len; i++)
    if (!cstExpr(*fp++, vp++))
      break;
  if (i==len)
    {
      e = KtupleGnrl(len,vals);
      (void) KEEP(e);
      return MkQuote(e);
    }
  return MakeRecord(nosharing?E_UNSHARED_TUPLE_GNRL:E_TUPLE_GNRL,NIL,(COUNTER)len,fields);
}

PRIVATE ExprPtr cpTuple0(args,assum,len,nosharing)
	/* A_TUPLE, return expr */
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
IN INT len;
IN Boolean nosharing;
{
  ExprPtr e,e1,e2,aux;
  COUNTER i;
  ENTER;

  switch (len)
    {
    case 0:
      e = tuple0;
      break;
    case 1:
      aux = CAR(args);
      e1 = CP(aux,assum,-1);
      e = MkTuple1(e1,nosharing);
      break;
    case 2:
      aux = CAR(args);
      e1 = CP(aux,assum,-1);
      aux = CAR(CDR(args));
      e2 = CP(aux,assum,-1);
      e = MkTuple2(e1,e2,nosharing);
      break;
    default:
      {
	ExprPtr *ep;
	ExprPtr *exprs;

	exprs = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
	for (i=0, ep=exprs; i<len; i++, ep++, args = CDR(args))
	  {
	    *ep = CP(CAR(args),assum,-1);
	  }
	e = MkTupleGnrl(len,exprs,nosharing);
	break;
      }
    }
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpTuple(expr,assum,gto)
	/* A_TUPLE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args;
  ENTER;

  args = ARGS(expr);
  args = cpTuple0(args,assum,ListLength(args),noSharingSure(expr,assum));
#ifdef OBSOLETE
#if DEBUGGING!=0
  (void) KEEP(args);
  args = wrapDebugZone(expr,args,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(args);
  return args;
}

PRIVATE ExprPtr cpProgn(expr,assum,gto)
	/* A_PROGN, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args;
  ExprPtr *exprs,*ep,ex,e,aux;
  COUNTER i,n;
  ENTER;

  args = ARGS(expr);
  n = ListLength(args)-1;
  exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args))
    *ep++ = CP(CAR(args),assum,-1);
  ex = CP(CAR(args),assum,gto);
  BEGINQUICKKEEP();
  while (n--!=0)
    {
      e = *--ep;
      MAKEPROG(E_PROG,e,ex,aux);
      ex = aux;
      QUICKKEEP(ex);
    }
  ENDQUICKKEEP();
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpList1(args,assum)
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
{
  ExprPtr *exprs,*ep,ex,e,aux;
  COUNTER i,n;
  Boolean cst;
  ENTER;

  n = ListLength(args);
  exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args))
    *ep++ = CP(CAR(args),assum,-1);
  ex = NIL;
  cst = true;
  BEGINQUICKKEEP();
  while (n--!=0)
    {
      e = *--ep;
      if (cst)
	{
	  if (cstExpr(e,&aux))
	    {
	      ex = Kons(aux,ex);
	      QUICKKEEP(ex);
	      continue;
	    }
	  cst = false;
	  if (ex!=NIL)
	    {
	      ex = MkQuote(ex);
	      QUICKKEEP(ex);
	    }
	}
      MAKECONS(e,ex,aux);
      ex = aux;
      QUICKKEEP(ex);
    }
  if (cst)
    ex = MkQuote(ex);
  ENDQUICKKEEP();
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpList(expr,assum,gto)
	/* A_LIST, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr ex;
  ENTER;

  ex = cpList1(ARGS(expr),assum);
#ifdef OBSOLETE
#if DEBUGGING
  (void) KEEP(ex);
  ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(ex);
  return ex;
}

PRIVATE MapPtr getLocalVars P((IN CONST MapPtr map,IN CONST ExprPtr pat));

PRIVATE MapPtr getLabeledVars(map,maplet)
IN CONST MapPtr map;
IN CONST ExprPtr maplet;
{
  switch (TYPE(maplet))
    {
    case A_PAT_RECORD_ELLIPSIS:
      return NewMap();
    default:
      return getLocalVars(map,CAR(CDR(ARGS(maplet))));
    }
}

PRIVATE MapPtr getPatternRowVars(map,row,acc)
IN CONST MapPtr map;
IN CONST ExprPtr row;
IN CONST MapPtr acc;
{
  while (row!=NIL)
    {
      acc = MapMerge(acc,getLabeledVars(map,CAR(row)));
      row = CDR(row);
    }
  return acc;
}

PRIVATE MapPtr getPatternListVars(map,row,acc)
IN CONST MapPtr map;
IN CONST ExprPtr row;
IN CONST MapPtr acc;
{
  while (row!=NIL)
    {
      acc = MapMerge(acc,getLocalVars(map,CAR(row)));
      row = CDR(row);
    }
  return acc;
}

DEF_PRIVATE_ACTION_1ARG(getlocalvars)
{
  struct getlocalvars *data;

  data = (struct getlocalvars *)ENV_UNDERLYING(env);
  return getLocalVars(data->map,data->pat);
}
ACTION_END

PRIVATE MapPtr getLocalVars1(map,pat)
IN CONST MapPtr map;
IN CONST ExprPtr pat;
{
  struct getlocalvars *data = (struct getlocalvars *)alloctmp(sizeof(struct getlocalvars));

  data->map = map;
  data->pat = pat;
  return CallCC(ACTION_NAME(getlocalvars),(ExprPtr)data,true);
}

PRIVATE MapPtr getLocalVars(map,pat)
IN CONST MapPtr map;
IN CONST ExprPtr pat;
{
  if (STKFULL(map))
    return getLocalVars1(map,pat);
again:
  switch (TYPE(pat))
    {
    case A_PAT_ANY:
    case A_PAT_CONST:
    case A_PAT_VAR:
    case A_PAT_TUPLE_ELLIPSIS:
    case A_PAT_NIL:
    case A_PAT_ANYLIST:
    case A_PAT_EMPTY:
    case A_PAT_ANYMAP:
      return NewMap();
    case A_REAL_VAR:
      {
	Boolean found;
	ExprPtr i_l;

	i_l = MapGet(CAR(ARGS(pat)),map,&found);
	return found?MapSingleton(CAR(ARGS(pat)),i_l):NewMap();
      }
    case A_PAT_RECORD:
      return getPatternRowVars(map,ARGS(pat),NewMap());
    case A_PAT_TUPLE:
    case A_PAT_MAPADD:
    case A_PAT_MAPLET:
    case A_PAT_AS:
      return getPatternListVars(map,ARGS(pat),NewMap());
    case A_PAT_APPLY:
      pat = CAR(CDR(ARGS(pat)));
      goto again;
    case A_PAT_CAST:
    case A_PAT_PACK:
      pat = CAR(ARGS(pat));
      goto again;
    default:
      kinderror(message(624),(int)TYPE(pat));
      return NIL;
    }
}

PRIVATE void addLocalVars (assum, local_assum, pat)
INOUT ExprPtr assum;
IN CONST ExprPtr local_assum;
IN CONST ExprPtr pat;
{
  MapPtr locvars, cons, i_ls;
  ENTER;

  locvars = getLocalVars(C_IDS(local_assum),pat);
  setMapMerge(&C_IDS(assum),locvars);
  i_ls = Rng0 (locvars);
  (void) KEEP (i_ls);
  cons = MapDomRestrTo (C_CONS(local_assum), i_ls);
  setMapMerge(&C_CONS(assum), cons);
  QUIT;
}

PRIVATE ExprPtr cpDeclList1 P((IN CONST ExprPtr args,INOUT ExprPtr assum));

PRIVATE ExprPtr getDomains1(args,assum,declsp)
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
OUT ExprPtr *declsp;
{
  ExprPtr e,desc,a;
  ExprPtr descs;
  struct data {
    ExprPtr decl;
    ExprPtr e;
    ExprPtr local_assum;
  } *exprs,*ep;
  CHAR *tags,*tp;
  COUNTER i,n;
  ENTER;

  for (n=0, a=args;
       a!=NIL && (e = CAR(a),
		  TYPE(e)==A_IN_SET || TYPE(e)==A_IN_LIST || TYPE(e)==A_SUB_MAP);
       n++, a = CDR(a));
  *declsp = a;
  exprs = (struct data *)alloctmp_anchored(n*sizeof(struct data), (char *)&exprs);
  tags = (CHAR *)alloctmp_anchored(n, (char *)&exprs);
  for (i=0, ep=exprs, tp=tags; i<n; i++, args=CDR(args), ep++, tp++)
    {
      ep->local_assum = DupCompAssum(assum);
#if DEBUGGING
      if (C_LEVEL(assum)==C_ILEVEL(assum))
	C_DBGCNT(ep->local_assum) = 1;
#endif
      ep->e = e = CAR(args);
      desc = CP(e,ep->local_assum,-1);
      C_NLOCALS(assum) = C_NLOCALS(ep->local_assum);
      ep->decl = CDR(desc);
      switch (IVAL(CAR(desc)))
	{
	case A_IN_LIST: *tp = E_PAT_RECORD_GNRL_CONS; break;
	case A_IN_SET: *tp = E_PAT_REVRECORD_GNRL_CONS; break;
	case A_SUB_MAP: *tp = E_PAT_RECORD_MAPADD; break;
	default:
	  berror(message(427),IVAL(CAR(desc)));
	}
#if DEBUGGING
      C_DBGCNT(assum) = C_DBGCNT(ep->local_assum);
#endif
#ifdef OBSOLETE
      SETQ(C_TYTAGS(assum),C_TYTAGS(ep->local_assum)); /* because tytags are global */
#endif
    }
  descs = NIL;
  for (; --ep, --tp, --i>=0; )
    {
      descs = MakePair((ExprKind)*tp,ep->decl,descs);
      (void) KEEP(descs);
      e = CAR(ARGS(ep->e));
      addLocalVars (assum, ep->local_assum, e);
    }
  KEEPONLY(descs);
  return descs;
}

PRIVATE ExprPtr getDomains(args,assum)
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
{
  ExprPtr decls,descs;
  COUNTER savelevel;

  descs = getDomains1(args,assum,&decls);
  if (decls!=NIL)
    {
      savelevel = C_LEVEL(assum);
      C_LEVEL(assum) = C_ILEVEL(assum);
      decls = MakeTuple1(cpDeclList1(decls,assum));
      C_LEVEL(assum) = savelevel;
      (void) KEEP(decls);
      descs = Nappend(descs,decls);
    }
  return descs;
}

PRIVATE ExprPtr LocalCompAssum(assum)
IN CONST ExprPtr assum;
{
  ExprPtr local_assum;

  local_assum = DupCompAssum(assum);
  if (C_FLAGS(assum) & C_NODEBUG)
    C_FLAGS(local_assum) |= C_NODBGNAME;
  C_NLOCALS(local_assum) = 0;
  C_ILEVEL(local_assum) = C_LEVEL(local_assum)+1;
  return local_assum;
}

PRIVATE ExprPtr cpOfOrMap1(kind,expr,assum)
	/* A_*_OF, A_*_MAP, return expr */
IN enum ExprKind kind;
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr ef,eP,f,p,domains,local_assum,args;
  ENTER;

  args = ARGS(expr);
  ef = CAR(args);
  args = CDR(args);
  eP = CAR(args);
  args = CDR(args);
  local_assum = LocalCompAssum(assum);
  domains = getDomains(args,local_assum);
  domains = CONS(KINT(C_NLOCALS(local_assum)),domains);
  C_LEVEL(local_assum)++;
  C_ILEVEL(local_assum) = C_LEVEL(local_assum);
  f = CP(ef,local_assum,-1);
  p = CP(eP,local_assum,-1);
  f = MakeOfOrMap(kind,CONS(f,p),domains);
  SETQ(C_TYTAGS(assum),C_TYTAGS(local_assum)); /* because tytags are global */
#if DEBUGGING
  C_DBGCNT(assum) = C_DBGCNT(local_assum);
  (void) KEEP(f);
  f = wrapDebugZone(expr,f,C_DBGCNT(assum)++,assum);
#endif
  KEEPONLY(f);
  return f;
}

PRIVATE ExprPtr cpListOf(expr,assum,gto)
	/* A_LIST_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1((noSharingSure(expr,assum)?E_UNSHARED_LIST_OF:E_LIST_OF),expr,assum);
}

PRIVATE ExprPtr cpListMap(expr,assum,gto)
	/* A_LIST_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1((noSharingSure(expr,assum)?E_UNSHARED_LIST_MAP:E_LIST_MAP),expr,assum);
}

PRIVATE ExprPtr cpIn(expr,assum,gto)
	/* A_IN_SET, A_IN_LIST, A_SUB_MAP, return typex::pat::ex */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e1,e2,pat,ex;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  args = CDR(args);
  e2 = CAR(args);
  ex = CP(e2,assum,-1);
  pat = CP(e1,assum,-1);
  pat = optPattern(pat);
  ex = Cons(KINT((INT)TYPE(expr)),CONS(pat,ex));
  KEEPONLY(ex);
  return ex;
}

#define cpInSet cpIn
#define cpInList cpIn
#define cpSubMap cpIn

PRIVATE ExprPtr cpMapOfOrMap1(mapkind,setkind,expr,assum)
	/* A_MAP_OF, A_MAP_MAP, return expr */
IN enum ExprKind mapkind,setkind;
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr args,ef,eP,f,p,domains,local_assum;
  ENTER;

  args = ARGS(expr);
  ef = CAR(args);
  args = CDR(args);
  eP = CAR(args);
  args = CDR(args);
  local_assum = LocalCompAssum(assum);
  domains = getDomains(args,local_assum);
  domains = CONS(KINT(C_NLOCALS(local_assum)),domains);
  C_LEVEL(local_assum)++;
  C_ILEVEL(local_assum) = C_LEVEL(local_assum);
  f = CP(ef,local_assum,-1);
  p = CP(eP,local_assum,-1);
  if (CDR(f)==tuple0)
    f = MakeOfOrMap(setkind,CONS(CAR(f),p),domains);
  else
    {
      f = MakePair(E_MAPLET,CAR(f),CDR(f));
      (void) KEEP(f);
      f = MakeOfOrMap(mapkind,CONS(f,p),domains);
    }
  SETQ(C_TYTAGS(assum),C_TYTAGS(local_assum)); /* because tytags are global */
#if DEBUGGING
  C_DBGCNT(assum) = C_DBGCNT(local_assum);
  (void) KEEP(f);
  f = wrapDebugZone(expr,f,C_DBGCNT(assum)++,assum);
#endif
  KEEPONLY(f);
  return f;
}

PRIVATE ExprPtr cpMapOf(expr,assum,gto)
	/* A_MAP_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapOfOrMap1(E_MAP_OF,E_SET_OF,expr,assum);
}

PRIVATE ExprPtr cpMapMap(expr,assum,gto)
	/* A_MAP_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapOfOrMap1(E_MAP_MAP,E_SET_MAP,expr,assum);
}

PRIVATE ExprPtr cpMapUnder(expr,assum,gto)
	/* A_MAP_UNDER, return couple expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args;
  ENTER;

  args = ARGS(expr);
  expr = cpMap1(&args,ListLength(args),assum,true);
  KEEPONLY(expr);
  return expr;
}

PRIVATE ExprPtr cpMapUnderOf(expr,assum,gto)
	/* A_MAP_UNDER_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapOfOrMap1(E_MAP_UNDER_OF,E_SET_OF,expr,assum);
}

PRIVATE ExprPtr cpMapUnderMap(expr,assum,gto)
	/* A_MAP_UNDER_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapOfOrMap1(E_MAP_UNDER_MAP,E_SET_MAP,expr,assum);
}

PRIVATE ExprPtr cpProgOf(expr,assum,gto)
	/* A_PROG_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_PROG_OF,expr,assum);
}

PRIVATE ExprPtr cpProgMap(expr,assum,gto)
	/* A_PROG_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_PROG_MAP,expr,assum);
}

PRIVATE ExprPtr cpAllOf(expr,assum,gto)
	/* A_ALL_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_ALL_OF,expr,assum);
}

PRIVATE ExprPtr cpAllMap(expr,assum,gto)
	/* A_ALL_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_ALL_MAP,expr,assum);
}

PRIVATE ExprPtr cpExistsOf(expr,assum,gto)
	/* A_EXISTS_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_EXISTS_OF,expr,assum);
}

PRIVATE ExprPtr cpExistsMap(expr,assum,gto)
	/* A_EXISTS_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1(E_EXISTS_MAP,expr,assum);
}

PRIVATE ExprPtr cpSomeOf(expr,assum,gto)
	/* A_SOME_OF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1((noSharingSure(expr,assum)?E_UNSHARED_SOME_OF:E_SOME_OF),expr,assum);
}

PRIVATE ExprPtr cpSomeMap(expr,assum,gto)
	/* A_SOME_MAP, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpOfOrMap1((noSharingSure(expr,assum)?E_UNSHARED_SOME_MAP:E_SOME_MAP),expr,assum);
}

PRIVATE MapPtr MapDiff(m1,m2)
IN CONST MapPtr m1,m2;
{
  MapPtr m;
  ENTER;

  m = MapInter(m1,m2);
  m = MapDomRestrBy(m1,m);
  KEEPONLY(m);
  return m;
}

MAPCOLLECTFUNC(do_chopinline)
{
  switch (y->kind)
    {
    case E_COUNTED: /* case E_POLYFUN: case E_PROFILE: */
      return MapletKons(x,y);
    }
  return NewMap();
}

PRIVATE MapPtr ChopInline(map) /* kill all new inline definitions */
IN CONST MapPtr map;
{
  ExprPtr chopped;
  ENTER;

  chopped = MapCollect(map,do_chopinline,NULL);
  chopped = MapDiff(map,chopped);
  KEEPONLY(chopped);
  return chopped;
}

PRIVATE ExprPtr cpLet(expr,assum,gto)
	/* A_LET, return expr (also used by A_LOCAL and A_ABSTYPE) */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,declist,ex,local_assum;
  COUNTER nlocals;
  ENTER;

#ifdef OLDLET
  args = ARGS(expr);
  e = CAR(args);
  local_assum = LocalCompAssum(assum);
  C_LEVEL(local_assum) = C_ILEVEL(local_assum);
  declist = CP(e,local_assum,-1); /* local bindings */
  nlocals = C_NLOCALS(local_assum);
  args = CDR(args);
  e = CAR(args);
  if (TYPE(expr)==A_LET)
    ex = CP(e,local_assum,(gto<0)?(-1):(gto+1));
  /* if terminal recursion is forbidden for let x=e in f,
     then it is also forbidden for f;
     if it is allowed, then applications in f must pop
     one more level of the current environment, hence the gto+1
     */
  else
    { /* A_LOCAL, A_ABSTYPE : create new bindings in the calling environment */
      MapPtr oldids,oldcons;

      C_ILEVEL(local_assum) = C_ILEVEL(assum);
      C_NLOCALS(local_assum) = C_NLOCALS(assum);
      oldids = KEEP(C_IDS(local_assum));
      oldcons = KEEP(C_CONS(local_assum));
      ex = CP(e,local_assum,-1);
      C_NLOCALS(assum) = C_NLOCALS(local_assum);
      setMapMerge(&C_IDS(assum),MapDiff(C_IDS(local_assum),oldids));
      setMapMerge(&C_CONS(assum),
		  ChopInline (MapDiff(C_CONS(local_assum),oldcons))
		  );
      /* The ChopInline() is necessary to prevent this from dumping core:
	 local fun l1 a  = a+1 in fun l a = l1 a end; l 0;
	 Otherwise, l gets inlined but in an environment (toplevel) that is not
	 the right one (the environment inside the local declaration).
	 This is sad, and is rather similar to the problem of not inlining
	 imported definitions in do_export (below). */
    }
  ex = MakeLet(CONS(KINT(nlocals),declist),ex); /* inefficient when nlocals==0,
						   but so what. */
  SETQ(C_TYTAGS(assum),C_TYTAGS(local_assum)); /* because tytags are global */
  SETQ(C_IMPORTS(assum),C_IMPORTS(local_assum)); /* because imports are global */
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
  if (TYPE(expr)==A_LET) /* don't add debug info on A_LOCAL and A_ABSTYPE */
    {
      C_DBGCNT(assum) = C_DBGCNT(local_assum);
      (void) KEEP(ex);
      ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum);
    }
#endif
#endif
  KEEPONLY(ex);
  return ex;
#else
  args = ARGS(expr);
  e = CAR(args);
  if (TYPE(expr)==A_LET)
    {
      local_assum = LocalCompAssum(assum);
      C_LEVEL(local_assum) = C_ILEVEL(local_assum);
      declist = CP(e,local_assum,-1); /* local bindings */
      nlocals = C_NLOCALS(local_assum);
      args = CDR(args);
      e = CAR(args);
      ex = CP(e,local_assum,(gto<0)?(-1):(gto+1));
      /* if terminal recursion is forbidden for let x=e in f,
	 then it is also forbidden for f;
	 if it is allowed, then applications in f must pop
	 one more level of the current environment, hence the gto+1
      */
      ex = MakeLet(CONS(KINT(nlocals),declist),ex); /* inefficient when nlocals==0,
						       but so what. */
      SETQ(C_TYTAGS(assum),C_TYTAGS(local_assum)); /* because tytags are global */
      SETQ(C_IMPORTS(assum),C_IMPORTS(local_assum)); /* because imports are global */
#ifdef BREAK_ON_CONTROL
#if DEBUGGING!=0
      if (TYPE(expr)==A_LET) /* don't add debug info on A_LOCAL and A_ABSTYPE */
	{
	  C_DBGCNT(assum) = C_DBGCNT(local_assum);
	  (void) KEEP(ex);
	  ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum);
	}
#endif
#endif
    }
  else
    {
      kinderror (message(949), TYPE(expr));
    }
  KEEPONLY(ex);
  return ex;
#endif
}

PRIVATE ExprPtr getRealLabel(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr type,shape,lbl;

  lbl = CAR(ARGS(expr));
  if (/*lbl!=NIL && */lbl->kind!=(CHAR)E_INT)
    {
      Boolean found;

      type = getTypeSure(expr,assum);
      type = CAR(APPLARG(type));
      /* type is now |[labi : typei, .. with a possible REST_TAG]| */
      if ((void) MapGet(REST_TAG,type,&found), !found)
	{
	  shape = getShape(type);
	  lbl = MapGetSure(lbl,shape);
	  /* in this case, we have been able to determine statically
	     the position in the record of the field */
	}
    }
  return lbl;
}

PRIVATE ExprPtr getLabel(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  while (TYPE(expr)==A_CAST)
    expr = CAR(ARGS(expr));
  if (TYPE(expr)==A_SELECT_FN)
    return getRealLabel(expr,assum);
  return NIL;
}

PRIVATE ExprPtr forgetExtra(expr)
IN CONST ExprPtr expr;
{
  while (true)
    {
      /*if (expr==NIL)
	return expr;*/
      switch (expr->kind)
	{
	case E_DEBUGZONE:
	  expr = DZCODE(expr);
	  break;
	case E_DEBUGNAME:
	  expr = DNPAT(expr);
	  break;
	case E_PROFILE:
	  expr = CAR(expr);
	  break;
	default:
	  return expr;
	}
    }
}

PRIVATE ExprPtr makeSelect(lab,rec,rectype)
IN CONST ExprPtr lab,rec;
IN CONST ExprPtr rectype;
{
  ExprPtr code,ll,val;
  Boolean found;
  ExprKind kind;

  /*if (rectype==NIL)
	code = MakePair((lab->kind==(CHAR)E_INT)?E_TSELECT:E_RSELECT,lab,rec);
  else*/
  switch (rectype->kind)
    {
      case E_CONS: /* selection from a tuple */
	for (ll=rectype; ISCONS(ll); ll=CDR(ll));
	if (ll==NIL) /* no ellipsis */
	  {
	    if (CDR(rectype)==NIL) /* 1-tuple (should not happen) */
	      kind = E_TSEL1;
	    else if (CDR(CDR(rectype))==NIL) /* 2-tuple */
	      kind = E_TSEL2;
	    else
	      kind = E_TSELGNRL; /* other tuple */
	  }
	else /* has an ellipsis */
	  kind = E_TSELECT;
	break;
      case E_MAPLET:
      case E_MAP:
	(void) MapGet(REST_TAG,rectype,&found);
	if (found) /* has an ellipsis */
	  kind = E_RSELECT;
	else
	  {
	    if (rectype->kind==(CHAR)E_MAPLET)
	      kind = E_RSEL1;
	    else
	      kind = E_RSELGNRL;
	  }
	break;
      default: /* case E_NIL: no known type */
	kind = (lab->kind==(CHAR)E_INT)?E_TSELECT:E_RSELECT;
	break;
    }
  if (cstExpr(rec,&val))
    {
      if (lab->kind==(CHAR)E_INT) /* should always be the case, since if val is constant,
				     its type is well-known, so its offsets are, too. */
	{
	  switch (val->kind)
	    {
	    case E_TUPLE_1: val = REFVAL(val);
cst:
	      val = MkQuote(val);
	      return KEEP(val);
	    case E_TUPLE_2: val = (lab==ione)?CAR(val):CDR(val); goto cst;
	    case E_RECORD_1: val = CDR(val); goto cst;
	    case E_TUPLE_GNRL: case E_RECORD_GNRL:
	      val = AVAL(val)->expr[IVAL(lab)-1]; goto cst;
	    }
	}
    }
  code = MakePair(kind,lab,rec);
  return KEEP(code);
}

PRIVATE ExprPtr maplet2before(e)
IN CONST ExprPtr e;
{ /* transform (e1 => e2) into e1 before e2, i.e, car(e1,e2) */
  ExprPtr ee;

  if (MAPLETY(e)==UNIT)
    return MAPLETX(e); /* very common case */
  ee = MakePair(E_UNSHARED_CONS,MAPLETX(e),MAPLETY(e));
  (void) KEEP(ee);
  ee = MakePair(E_TSEL2,ione,ee);
  return KEEP(ee);
}

PRIVATE ExprPtr cpApply(expr,assum,gto)
	/* A_APPLY, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e1,e2,f,ff,arg,i_l,con,label;
  COUNTER i,l;
  Boolean found;
  /* it is important that we put debugzones around all applications,
     not only because most breakpoints are best put around applications,
     but mostly because of Raise() calls in primitives; when an
     uncaught exception is raised by a primitive, it will allow
     DoBreak() to get the correct DEBUGZONE, to constrain the
     type of return : 'b -> 'a, and to allow bcEval() to resume
     the computation */
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  label = getLabel(e1,assum);
  e2 = CAR(CDR(args));
  arg = CP(e2,assum,-1);
  if (label!=NIL) /* #lab x */
    {
      ExprPtr type,aux;

      type = getType(e1,assum,&found);
	/* found may be true of false; if false, type is (not necessarily) NIL. */
      if (found)
	{
	  if (type->kind==(CHAR)E_TYPE_APPL &&
	      (aux = APPLFUN(type), fxSym(aux)==funSym))
	    type = CAR(APPLARG(type));
	  else type = NIL;
	}
      else type = NIL;
      e1 = makeSelect(label,arg,type);
#if DEBUGGING
      (void) KEEP(e1);
      e1 = wrapDebugZone(expr,e1,C_DBGCNT(assum)++,assum);
#endif
      KEEPONLY(e1);
      return e1;
    }
  f = CP(e1,assum,-1);
  ff = forgetExtra(f);
  if (/*ff!=NIL && */ff->kind==(CHAR)E_VARREF)
    {
      i = VARRANK(ff);
      l = C_LEVEL(assum)-VARLEVEL(ff);
      i_l = mkIL(i,l);
      con = MapGet(i_l,C_CONS(assum),&found);
      if (found)
	{
	  switch (con->kind)
	    {
	    case E_EXCEPTION:
	      /* cannot cons the exception name with its argument, since
		 the exception name is not known until run-time */
	      break;
#ifdef OLDDATA
	    case E_EMPTY:
	    case E_MAPLET:
	    case E_MAP: /* datatype constructor with no tag */
	      {
retarg:
#if DEBUGGING
		(void) KEEP(arg);
		arg = wrapDebugZone(expr,arg,C_DBGCNT(assum)++,assum);
#endif
		KEEPONLY(arg);
		return arg;
	      }
	    case E_STRING: /* datatype constructor with tag */
	      if (cstExpr(arg,&f))
		arg = MkQuote(KONS(con,f));
	      else
		arg = MakePair(noSharingSure(expr,assum)?E_UNSHARED_CONS:E_CONS,con,arg);
	      goto retarg;
#else
	    case E_DATACON: /* datatype constructor */
	      {
		ExprPtr c;

		c = COUNTEDVAL (con);
		if (cstExpr (arg, &f))
		  {
		    if ((COUNTEDN (con) & 1)==0) /* shared */
		      arg = Kdata(c->magic,CAR(c),f);
		    else
		      arg = MakeData(c->magic,CAR(c),f);
		    (void) KEEP(arg);
qretarg:
		    arg = MkQuote(arg);
retarg:
#if DEBUGGING
		    (void) KEEP(arg);
		    arg = wrapDebugZone(expr,arg,C_DBGCNT(assum)++,assum);
#endif
		    KEEPONLY(arg);
		    return arg;
		  }
		/* otherwise, let bc.c/compileApply1() do the work,
		   i.e. make constructor inlineable (PAT_AS (f, con)
		   means that f can be inlined as function con. */
		f = MakePair (E_PAT_AS, f, con);
		(void) KEEP (f);
		goto appl;
	      }
	    case E_DATA: /* datatype constructor */
	      if (cstExpr(arg,&f))
		{
		  if (con->mark & MARK_SHARED)
		    arg = Kdata(con->magic,CAR(con),f);
		  else
		    arg = MakeData(con->magic,CAR(con),f);
		  (void) KEEP(arg);
		  goto qretarg;
		}
	      /* otherwise, let bc.c/compileApply1() do the work,
		 i.e., compile f applied to arg. */
	      break;
#endif
	    }
	}
      if (l==0)
	{
	  /* special case identity a when identity has not been redefined */
	  switch (VARRANK(ff))
	    {
	    case ORIGINAL_ID:
	      goto retarg;
	    case ORIGINAL_CONS:
	      if (cstExpr(arg,&ff))
		{
		  arg = MkQuote(KONS(CAR(ff),CDR(ff)));
		  goto retarg;
		}
	      ff = forgetExtra(arg);
	      if (/*ff!=NIL && */(ff->kind==(CHAR)E_TUPLE_2 ||
			ff->kind==(CHAR)E_UNSHARED_TUPLE_2))
		{
		  arg = CONS(CAR(ff),CDR(ff));
		  if (noSharingSure(expr,assum))
		    arg->kind = (CHAR)E_UNSHARED_CONS;
		  goto retarg;
		}
	      break;
	    case ORIGINAL_GET:
	      ff = forgetExtra(arg);
	      arg = MakeGet(ff);
	      goto retarg;
	    case ORIGINAL_SET:
	      if (cstExpr(arg,&ff))
		{
		  ExprPtr e1,e2;

		  e1 = MkQuote(CAR(ff)); (void) KEEP(e1);
		  e2 = MkQuote(CDR(ff)); (void) KEEP(e2);
		  arg = MakeSet(e1,e2);
		  goto retarg;
		}
	      ff = forgetExtra(arg);
	      if (/*ff!=NIL && */(ff->kind==(CHAR)E_TUPLE_2 ||
			ff->kind==(CHAR)E_UNSHARED_TUPLE_2))
		{
		  arg = MakeSet(CAR(ff),CDR(ff));
		  goto retarg;
		}
	      break;
	    case ORIGINAL_FORCE:
	      f = arg;
	      arg = tuple0; /* force arg is identical to arg () */
	      goto appl;
	    case ORIGINAL_REF:
	      MAKEREF(arg,f);
#if DEBUGGING
	      (void) KEEP(f);
	      f = wrapDebugZone(expr,f,C_DBGCNT(assum)++,assum);
#endif
	      KEEPONLY(f);
	      return f;
	    case ORIGINAL_NOT:
	      if (cstExpr(arg,&ff))
		{
		  arg = (ff==trueValue)?falseValue:trueValue;
		  goto qretarg;
		}
	      arg = MakeSing(E_NOT,arg);
	      goto retarg;
	    case ORIGINAL_EQUAL:
	      if (cstExpr(arg,&ff))
		{
		  arg = (CAR(ff)==CDR(ff))?trueValue:falseValue;
		  goto qretarg;
		}
	      ff = forgetExtra(arg);
	      if (/*ff!=NIL && */(ff->kind==(CHAR)E_TUPLE_2 ||
			ff->kind==(CHAR)E_UNSHARED_TUPLE_2))
		{
		  arg = MakePair(E_EQUAL,CAR(ff),CDR(ff));
		  goto retarg;
		}
	      break;
	    case ORIGINAL_NEQUAL:
	      if (cstExpr(arg,&ff))
		{
		  arg = (CAR(ff)!=CDR(ff))?trueValue:falseValue;
		  goto qretarg;
		}
	      ff = forgetExtra(arg);
	      if (/*ff!=NIL && */(ff->kind==(CHAR)E_TUPLE_2 ||
			ff->kind==(CHAR)E_UNSHARED_TUPLE_2))
		{
		  arg = MakePair(E_NEQUAL,CAR(ff),CDR(ff));
		  goto retarg;
		}
	      break;
	    }
	  if (arg!=NIL)
	    switch (VARRANK(ff))
	      {
	      case ORIGINAL_APPEND_OF:
		ff = forgetExtra(arg);
		{
		  ExprKind new;

		  new = E_FREE;
		  switch (ff->kind)
		    {
		    case E_LIST_OF: new = E_APPEND_OF; break;
		    case E_UNSHARED_LIST_OF: new = E_UNSHARED_APPEND_OF; break;
		    case E_LIST_MAP: new = E_APPEND_MAP; break;
		    case E_UNSHARED_LIST_MAP: new = E_UNSHARED_APPEND_MAP; break;
		    case E_LIST_WHILE: new = E_APPEND_WHILE; break;
		    case E_UNSHARED_LIST_WHILE: new = E_UNSHARED_APPEND_WHILE; break;
		    }
		  if (new!=E_FREE)
		    {
		      arg = MakeOfOrMap(new,CAR(ff),CDR(ff));
		      goto retarg;
		    }
		}
		break;
	      case ORIGINAL_UNION_OF:
	      case ORIGINAL_MERGE_OF:
		ff = forgetExtra(arg);
		{
		  ExprKind new;
		  ExprPtr fp;

		  new = E_FREE;
		  fp = CAR(ff);
		  switch (ff->kind)
		    {
		    case E_LIST_OF: case E_UNSHARED_LIST_OF:
		    case E_SET_OF:
		      new = E_UNION_OF; break;
		    case E_MAP_OF: case E_MAP_UNDER_OF:
		      new = E_UNION_OF; fp = maplet2before(fp); break;
		    case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
		    case E_SET_MAP:
		      new = E_UNION_MAP; break;
		    case E_MAP_MAP: case E_MAP_UNDER_MAP:
		      new = E_UNION_MAP; fp = maplet2before(fp); break;
		    case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
		    case E_SET_WHILE:
		      new = E_UNION_WHILE; break;
		    case E_MAP_WHILE: case E_MAP_UNDER_WHILE:
		      new = E_UNION_WHILE; fp = maplet2before(fp); break;
		    }
		  if (new!=E_FREE)
		    {
		      arg = MakeOfOrMap(new,fp,CDR(ff));
		      goto retarg;
		    }
		}
		break;
	      case ORIGINAL_MERGE_UNDER_OF:
		ff = forgetExtra(arg);
		{
		  ExprKind new;

		  new = E_FREE;
		  switch (ff->kind)
		    {
		    case E_LIST_OF: case E_UNSHARED_LIST_OF:
		      new = E_UNION_UNDER_OF; break;
		    case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
		      new = E_UNION_UNDER_MAP; break;
		    case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
		      new = E_UNION_UNDER_WHILE; break;
		    }
		  if (new!=E_FREE)
		    {
		      arg = MakeOfOrMap(new,CAR(ff),CDR(ff));
		      goto retarg;
		    }
		}
		break;
	      case ORIGINAL_INTER_OF:
		ff = forgetExtra(arg);
		{
		  ExprKind new;
		  ExprPtr fp;

		  new = E_FREE;
		  fp = CAR(ff);
		  switch (ff->kind)
		    {
		    case E_SET_OF: new = E_INTER_OF; break;
		    case E_MAP_OF: case E_MAP_UNDER_OF:
		      new = E_INTER_OF; fp = maplet2before(fp); break;
		    case E_SET_MAP: new = E_INTER_MAP; break;
		    case E_MAP_MAP: case E_MAP_UNDER_MAP:
		      new = E_INTER_MAP; fp = maplet2before(fp); break;
		    case E_SET_WHILE: new = E_INTER_WHILE; break;
		    case E_MAP_WHILE: case E_MAP_UNDER_WHILE:
		      new = E_INTER_WHILE; fp = maplet2before(fp); break;
		    }
		  if (new!=E_FREE)
		    {
		      arg = MakeOfOrMap(new,fp,CDR(ff));
		      goto retarg;
		    }
		}
		break;
	      }
	}
    }
  else
    switch (ff->kind)
      {
      case E_APPL:
      case E_GOTO:
	{ /* treat cases like ?m x and throw cont val */
	  ExprPtr fff;

	  fff = APPLFUN(ff);
	  if (/*fff!=NIL && */fff->kind==(CHAR)E_VARREF)
	    {
	      i = VARRANK(fff);
	      l = C_LEVEL(assum)-VARLEVEL(fff);
	      if (l==0)
		switch (i)
		  {
		  case ORIGINAL_MAPAPP:
		    arg = MakeMapget(APPLARG(ff),arg);
		    goto retarg;
		  }
	    }
	  break;
	}
#if TRANSLATOR
      case E_PROFILE:
	if (CAR(ff)->kind!=(CHAR)E_POLYFUN)
	  break;
	/*fallthrough*/
#endif
      case E_POLYFUN: /* (fn ...) a, i.e. case a of ...: then don't put
			 any debugging information around the case */
	{
	  if (gto<0)
	    arg = MakeAppl(ff,arg);
	  else
	    arg = MakeGoto((CHAR)E_GOTO,ff,arg);
	  KEEPONLY(arg);
	  return arg;
	}
      }
appl:
  if (gto<0)
    arg = MakeAppl(f,arg);
  else
    arg = MakeGoto((CHAR)E_GOTO,f,arg);
  goto retarg;
}

PRIVATE ExprPtr cpCast(expr,assum,gto)
	/* A_CAST, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e;

  e = CAR(ARGS(expr));
  return CP(e,assum,gto);
}

PRIVATE ExprPtr cpAnd1(kind,expr,assum,gto)
	/* A_AND, A_OR, return expr */
IN COUNTER kind;
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,ex1,ex2;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  ex1 = CP(e,assum,-1);
  e = CAR(CDR(args));
  ex2 = CP(e,assum,gto);
  if ((kind==E_AND && ex1->kind==(CHAR)E_QUOTE && REFVAL(ex1)==trueValue)
      || (kind==E_OR && ex1->kind==(CHAR)E_QUOTE && REFVAL(ex1)==falseValue))
    e = ex2;
  else if ((kind==E_AND && ex1->kind==(CHAR)E_QUOTE && REFVAL(ex1)==falseValue)
	   || (kind==E_OR && ex1->kind==(CHAR)E_QUOTE && REFVAL(ex1)==trueValue))
    e = ex1;
  else
    {
      MAKEPROG(kind,ex1,ex2,e);
    }
#ifdef BREAK_ON_CONTROL
#if DEBUGGING
  (void) KEEP(e);
  e = wrapDebugZone(expr,e,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpAnd(expr,assum,gto)
	/* A_AND, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpAnd1(E_AND,expr,assum,gto);
}

PRIVATE ExprPtr cpOr(expr,assum,gto)
	/* A_AND, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpAnd1(E_OR,expr,assum,gto);
}

PRIVATE ExprPtr cpHandle(expr,assum,gto)
	/* A_HANDLE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,ex1,ex2;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  ex1 = CP(e,assum,-1); /* -1 since in a handle b, we still have to pop
			   and exception frame once a has been computed,
			   even though a may be a tail call. */
  e = CAR(CDR(args));
  ex2 = CP(e,assum,gto); /* here on the other hand, terminal recursion is preserved */
  ex1 = MakeHandle(ex1,ex2);
#if DEBUGGING
  (void) KEEP(ex1);
  ex1 = wrapDebugZone(expr,ex1,C_DBGCNT(assum)++,assum);
#endif
  KEEPONLY(ex1);
  return ex1;
}

PRIVATE ExprPtr cpRaise(expr,assum,gto)
	/* A_RAISE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e,ex;
  ENTER;

  e = CAR(ARGS(expr));
  ex = CP(e,assum,-1);
  ex = MakeRaise(ex);
#if DEBUGGING
  (void) KEEP(ex);
  ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum); /* kept to allow user to
  	resume computation after raising an uncaught exception */
#endif
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpMatch(expr,assum,gto)
	/* A_MATCH, A_FN, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,ex,patbody,val,aux;
  INT n;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  val = CP(e,assum,gto); /* return E_POLYFUN (pattern,body), n */
  patbody = COUNTEDVAL(val);
  n = COUNTEDN(val);
  for (; args=CDR(args), args!=NIL; )
    {
      e = CAR(args);
      ex = CP(e,assum,gto); /* return E_POLYFUN (pattern,body), n */
      aux = MakeSing(E_REF,NewMap());
      (void) KEEP(aux);
      SETCAR(patbody,
	     tackPattern(CAR(patbody),CAR(COUNTEDVAL(ex)),CDR(COUNTEDVAL(ex)),aux,
			 &COUNTEDN(ex)));
      if (n<COUNTEDN(ex))
	n = COUNTEDN(ex);
    }
  SETCOUNTEDN(val,n);
  KEEPONLY(val);
  return val;
}

#if TRANSLATOR
PRIVATE ExprPtr nameFn1(expr,ex,fundef,profile)
IN CONST ExprPtr expr;
IN CONST ExprPtr ex;
IN CONST ExprPtr fundef;
IN Boolean profile;
{
  if (fundef!=NIL) /* compile a [CLR_]PROFILE around the POLYFUN
			       even when profiling was disabled. */
    {
      ExprPtr prof,e;
      ENTER;

      e = CAR(fundef);
      if (profile)
	prof = profAlloc();
      else prof = NIL;
      e = CONS(prof,e);
      e = CONS(CDR(fundef),e);
      e = ZBuild(ZONE(expr),profile?EA_PROFILE:EA_CLR_PROFILE,e);
      (void) KEEP(e);
      if (profile)
	profileExprs = CONS(e,profileExprs);
      ex = MakePair(E_PROFILE,ex,e);
      KEEPONLY(ex);
    }
  return ex;
}

PRIVATE ExprPtr nameFn(expr,ex,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr ex;
IN CONST ExprPtr assum;
{
  return nameFn1(expr,ex,C_FUNDEF(assum),(C_FLAGS(assum) & C_PROFILE)!=0);
}

PRIVATE ExprPtr cpFn(expr,assum,gto)
     /* A_FN, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER gto;
{
  ExprPtr ex;
  COUNTER arity;

  ex = cpMatch(expr,assum,gto);
  arity = funArity (expr, assum);
  if (arity>=0 && arity==(COUNTER)(SHORT)arity && ex->kind==(CHAR)E_POLYFUN)
    SETFUNARITY(ex,arity); /* inform bc.c of actual arity of function. */
  ex = nameFn(expr,ex,assum);
  return ex;
}
#else
#define cpFn cpMatch
#endif

PRIVATE ExprPtr cpLambda1(expr,assum,gto,dbgbodyp)
	/* A_LAMBDA, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
IN CONST COUNTER dbgbodyp;
{
  ExprPtr args,e,pat,ex,local_assum,lam;
  ENTER;

  args = ARGS(expr);
  local_assum = LocalCompAssum(assum);
  e = CAR(args);
  pat = CP(e,local_assum,-1);
  C_LEVEL(local_assum)++;
  args = CDR(args);
  e = CAR(args);
  ex = CP(e,local_assum,gto+1); /* this is the one spot where
				   gto is incremented, that is, where
				   terminal recursion comes into view */
  ex = forgetExtra(ex);
#if PROFILING
  if ((C_FLAGS(assum) & C_PROFILE)!=0 && C_FUNDEF(assum)!=NIL)
    ex = nameFn(expr,ex,assum);
#endif
#if DEBUGGING!=0
  if (dbgbodyp)
    ex = wrapDebugZone(e,ex,C_DBGCNT(local_assum)++,local_assum);
#endif

  pat = optPattern(pat);
  /* eta-conversion does not work yet. For example, write
     fn x => card x : card inside refers to the global card,
     so will have level 1 when compiled, though its eta-converted
     (which is simply card) has level 0. Moreover, it is difficult
     to use when constructing functions by cases, and of dubious
     value. */
  ex = CONS(pat,ex);
  lam = MakePolyfun(-1, ex,(INT)C_NLOCALS(local_assum));
  SETQ(C_TYTAGS(assum),C_TYTAGS(local_assum)); /* because tytags are global */
#if DEBUGGING
  C_DBGCNT(assum) = C_DBGCNT(local_assum);
#endif
  KEEPONLY(lam);
  return lam;
}

PRIVATE ExprPtr cpLambda(expr,assum,gto)
	/* A_LAMBDA, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpLambda1(expr,assum,gto,1);
}

PRIVATE ExprPtr cpIf(expr,assum,gto)
	/* A_IF, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,cond,th,el;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  cond = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  th = CP(e,assum,gto);
  e = CAR(CDR(args));
  el = CP(e,assum,gto);
  if (cond->kind==E_QUOTE && REFVAL(cond)==trueValue)
    el = th;
  else if (cond->kind==E_QUOTE && REFVAL(cond)==falseValue)
    ;
  else
    el = MakeTest(cond,CONS(th,el));
#ifdef BREAK_ON_CONTROL
#if DEBUGGING
  (void) KEEP(el);
  el = wrapDebugZone(expr,el,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(el);
  return el;
}

PRIVATE ExprPtr cpWhile1(kind,expr,assum)
	/* A_WHILE*, return expr */
IN CONST ExprKind kind;
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr args,e,cond,body,suchthat;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  cond = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  body = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  suchthat = CP(e,assum,-1);
  body = MakeWhile(kind,CONS(cond,suchthat),body);
#ifdef BREAK_ON_CONTROL
#if DEBUGGING
  (void) KEEP(body);
  body = wrapDebugZone(expr,body,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(body);
  return body;
}

PRIVATE ExprPtr cpMapWhile1(mapkind,setkind,expr,assum)
	/* A_MAP_WHILE*, return expr */
IN CONST ExprKind mapkind,setkind;
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr args,e,cond,body,suchthat;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  cond = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  body = CP(e,assum,-1);
  args = CDR(args);
  e = CAR(args);
  suchthat = CP(e,assum,-1);
  if (CDR(body)==tuple0)
    body = MakeWhile(setkind,CONS(cond,suchthat),CAR(body));
  else
    {
      body = MakePair(E_MAPLET,CAR(body),CDR(body));
      (void) KEEP(body);
      body = MakeWhile(mapkind,CONS(cond,suchthat),body);
    }
#ifdef BREAK_ON_CONTROL
#if DEBUGGING
  (void) KEEP(body);
  body = wrapDebugZone(expr,body,C_DBGCNT(assum)++,assum);
#endif
#endif
  KEEPONLY(body);
  return body;
}

PRIVATE ExprPtr cpWhile(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpWhile1(E_PROG_WHILE,expr,assum);
}

PRIVATE ExprPtr cpWhileList(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpWhile1((noSharingSure(expr,assum)?E_UNSHARED_LIST_WHILE:E_LIST_WHILE),expr,assum);
}

PRIVATE ExprPtr cpWhileMap(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapWhile1(E_MAP_WHILE,E_SET_WHILE,expr,assum);
}

PRIVATE ExprPtr cpWhileMapUnder(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpMapWhile1(E_MAP_UNDER_WHILE,E_SET_WHILE,expr,assum);
}

PRIVATE ExprPtr cpWhileAll(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpWhile1(E_ALL_WHILE,expr,assum);
}

PRIVATE ExprPtr cpWhileExists(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpWhile1(E_EXISTS_WHILE,expr,assum);
}

PRIVATE ExprPtr cpWhileSome(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return cpWhile1((noSharingSure(expr,assum)?E_UNSHARED_SOME_WHILE:E_SOME_WHILE),expr,assum);
}

PRIVATE ExprPtr cpDeclList1(args,assum)
IN CONST ExprPtr args;
INOUT ExprPtr assum;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  REGISTER ExprPtr e,b,last;

  _x.p[0] = NIL;
  _x.n = 1;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  last = NIL;
  for (; args!=NIL; args=CDR(args))
    {
      e = CAR(args);
      b = CP(e,assum,-1);
      if (b!=NIL)
	{
	  if (last!=NIL)
	    {
	      SETCDR(last,b);
	    }
	  else
	    _x.p[0] = b;
	  while (last=b, b=CDR(b), b!=NIL);
#if DEBUGGING!=0
	  if (C_FLAGS(assum) & C_NODEBUG)
	    ;
	  else
	    {
	      ExprPtr info;
	      ENTER;

	      info = KONS(NIL,KONS(currentFile,NIL));
	      info = ZBuild(ZONE(e),A_CODEPOINT,info);
	      (void) KEEP(info);
	      info = Kref(E_DEBUGDECL, info);
	      (void) KEEP(info);
	      b = MakePair(E_CONS,info,NIL);
	      SETCDR(last,b);
	      last = b;
	      QUIT;
	      /* New debugging scheme, not understood by eval.c */
	    }
#endif
	}
    }
  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE ExprPtr cpDeclList(expr,assum,gto)
	/* A_DECL_LIST, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr ex;
  ENTER;

  ex = cpDeclList1(ARGS(expr),assum);
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpLocal(expr,assum,gto)
	/* A_LOCAL, return bindings; or also A_ABSTYPE */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
#ifdef OLDLET
  expr = cpLet(expr,assum,gto);
  return CONS(expr,NIL);
#else
  ExprPtr args,e,ex,declist;
  MapPtr ids0,ids1,cons0,cons1;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  /* To hide the local ids, here is what we do:
     Before we compile local bindings,
     ids0 = C_IDS(assum);
     After we have compiled local bindings,
     ids1 = C_IDS(assum);
     The new ones are 
     After we have compiled the other bindings,
     ids2 = C_IDS(assum);
     The new bindings after all that, which remain visible,
     are those in MapDiff (ids2, ids1).
  */
  ids0 = KEEP(C_IDS(assum));
  cons0 = KEEP(C_CONS(assum));
  args = ARGS(expr);
  e = CAR(args);
  declist = CP(e,assum,-1); /* local bindings */
  ids1 = KEEP(C_IDS(assum));
  cons1 = KEEP(C_CONS(assum));
  args = CDR(args);
  e = CAR(args);
  ex = CP(e,assum,-1);
  ids0 = MapOverwrite(ids0, MapDiff (C_IDS(assum),ids1));
  SETQ(C_IDS(assum), ids0);
  cons0 = MapOverwrite(cons0, MapDiff (C_CONS(assum), cons1));
  SETQ(C_CONS(assum), cons0);
  ex = Append(declist, ex);
  KEEPONLY(ex);
  return ex;
#endif
}

PRIVATE ExprPtr cpVal(expr,assum,gto)
	/* A_VAL, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,bindings;
  struct data {
    ExprPtr e;
    ExprPtr binding;
    ExprPtr local_assum;
  } *exprs,*ep;
  COUNTER i,n;
  ENTER;

  args = CDR(ARGS(expr));
  n = ListLength(args);
  exprs = (struct data *)alloctmp(n*sizeof(struct data));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args), ep++)
    {
      ep->local_assum = DupCompAssum(assum);
#if DEBUGGING
      if (C_LEVEL(assum)==C_ILEVEL(assum))
	C_DBGCNT(ep->local_assum) = 1;
#endif
      ep->e = e = CAR(args);
      ep->binding = CP(e,ep->local_assum,-1);
      C_NLOCALS(assum) = C_NLOCALS(ep->local_assum);
#if DEBUGGING
      C_DBGCNT(assum) = C_DBGCNT(ep->local_assum);
#endif
    }
  bindings = NIL;
  for (; --ep, --i>=0; )
    {
      e = CAR(ARGS(ep->e));
      addLocalVars (assum, ep->local_assum, e);
      bindings = Append(ep->binding,bindings);
    }
  KEEPONLY(bindings);
  return bindings;
}

PRIVATE Boolean unsafeValRec P((IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				IN COUNTER level));

#ifdef HAS_PROTOTYPES
PRIVATE Boolean unsafeValRecVec(IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				IN COUNTER level)
#else
PRIVATE Boolean unsafeValRecVec(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
#endif
{
  VecPtr vp;
  COUNTER i,n;

  vp = AVAL(expr);
  n = vp->n;
  for (i=0; i<n; i++)
    if (unsafeValRec(vp->expr[i],assum,level))
      return true;
  return false;
}

#ifdef HAS_PROTOTYPES
PRIVATE Boolean unsafeValRecBinding(IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				    IN COUNTER level)
#else
PRIVATE Boolean unsafeValRecBinding(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
#endif
{
  ExprPtr bind;

  while (expr!=NIL)
    {
      bind = CAR(expr);
      switch (bind->kind)
	{
	case E_LET: /* added Feb 5, 1998 (jg); this was not there before,
		       so I assume it was a bug. */
	  level++;
	  if (unsafeValRecBinding(CDR(LETBIND(bind)),assum,level) ||
	      unsafeValRecBinding(LETBODY(bind),assum,level))
	    return true;
	  level--;
	  break;
	case E_DEBUGDECL:
	  break;
	default:
	  if (unsafeValRec(CDR(CAR(expr)),assum,level))
	    return true;
	  break;
	}
      expr = CDR(expr);
    }
  return false;
}

PRIVATE Boolean unsafeValRecPattern P((IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				       IN COUNTER level));

DEF_PRIVATE_ACTION_1ARG(unsafevalrecpat)
{
  struct unsafevalrec *data;

  data = (struct unsafevalrec *)ENV_UNDERLYING(env);
  data->res = unsafeValRecPattern(data->expr,data->assum,data->level);
  return NIL;
}
ACTION_END

PRIVATE Boolean unsafeValRecPat1(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
{
  struct unsafevalrec *data = (struct unsafevalrec *)alloctmp(sizeof(struct unsafevalrec));

  data->expr = expr;
  data->assum = assum;
  data->level = level;
  (void) CallCC(ACTION_NAME(unsafevalrecpat),(ExprPtr)data,true);
  return data->res;
}

#ifdef HAS_PROTOTYPES
PRIVATE Boolean unsafeValRecPattern(IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				    IN COUNTER level)
#else
PRIVATE Boolean unsafeValRecPattern(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
#endif
{
  if (STKFULL(expr))
    return unsafeValRecPat1(expr,assum,level);
again:
  switch (expr->kind)
    {
    case E_POLYFUN:
      /*case E_N_LAMBDA:*/
      expr = COUNTEDVAL(expr);
      if (unsafeValRec(CDR(expr),assum,level))
	return true;
      expr = CAR(expr);
      goto again;
    case E_MEMO_FN:
      expr = MEMOFUN(expr);
      goto again;
      /* the other cases search for the secondary bodies of polyfuns */
    case E_VARINTRO:
    case E_NIL: case E_TUPLE_0: case E_RECORD_0:
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_QUOTE: case E_VARREF: case E_PAT_ANY:
      return false;
    case E_DEBUGNAME:
      expr = DNPAT(expr);
      goto again;
    case E_CONS: case E_UNSHARED_CONS:
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_PAT_AS: case E_APPL: case E_MAPLET:
      if (unsafeValRecPattern(CAR(expr),assum,level))
	return true;
      /*fallthrough*/
    case E_DATA:
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
      expr = CDR(expr);
      goto again;
    case E_PAT_REVCONS: case E_PAT_REVTUPLE_2:
      if (unsafeValRecPattern(CDR(expr),assum,level))
	return true;
      expr = CAR(expr);
      goto again;
#ifdef DATACONPAT
    case E_DATACON:
#endif
    case E_PAT_CONS:
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
    case E_REF:
      expr = REFVAL(expr);
      goto again;
    case E_PAT_TUPLE_GNRL_CONS: case E_PAT_REVTUPLE_GNRL_CONS:
    case E_PAT_RECORD_GNRL_CONS: case E_PAT_REVRECORD_GNRL_CONS:
      {
	VecPtr vp;
	COUNTER i,n;

	vp = AVAL(expr);
	n = vp->n;
	for (i=0; i<n; i++)
	  if (unsafeValRecPattern(vp->expr[i],assum,level))
	    return true;
      }
      return false;
    case E_PAT_CHECK:
      expr = PATCHECKVAL(expr);
      goto again;
    case E_PAT_PACK:
      expr = PACKVAL(expr);
      goto again;
    case E_PAT_CARD_EQ: case E_PAT_CARD_GEQ:
      expr = COUNTEDVAL(expr);
      goto again;
    case E_PATCHOICE:
      /* a choice is (primary pattern,(secondary pattern,secondary body)) */
      expr = CDR(expr);
      if (unsafeValRec(CDR(expr),assum,level))
	return true;
      expr = CAR(expr);
      goto again;
    case E_PAT_RECORD_MAPADD:
      if (unsafeValRecPattern(CDR(CAR(expr)),assum,level))
	return true;
      expr = CDR(expr);
      goto again;
    case E_PAT_MAPADD:
      if (unsafeValRecPattern(CAR(PATMAPADDXY(expr)),assum,level) ||
	  unsafeValRecPattern(CDR(PATMAPADDXY(expr)),assum,level))
	return true;
      expr = PATMAPADDREST(expr);
      goto again;
    case E_PROFILE:
      expr = CAR(expr);
      goto again;
    }
  return false;
}

#ifdef HAS_PROTOTYPES
PRIVATE Boolean unsafeValRecDom(IN CONST ExprPtr expr,IN CONST ExprPtr assum,
				IN COUNTER level)
#else
PRIVATE Boolean unsafeValRecDom(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
#endif
{
again:
  switch (expr->kind)
    {
    case E_NIL:
      return false;
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      if (unsafeValRec(CDR(CAR(expr)),assum,level))
	return true;
      expr = CDR(expr);
      goto again;
    case E_TUPLE_1: /* list of local bindings */
      return unsafeValRec(REFVAL(expr),assum,level+1);
    default:
      kinderror(message(504),expr->kind);
      /*NOTREACHED*/
      return false;
    }
}

DEF_PRIVATE_ACTION_1ARG(unsafevalrec)
{
  struct unsafevalrec *data;

  data = (struct unsafevalrec *)ENV_UNDERLYING(env);
  data->res = unsafeValRec(data->expr,data->assum,data->level);
  return NIL;
}
ACTION_END

PRIVATE Boolean unsafeValRec1(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
{
  struct unsafevalrec *data = (struct unsafevalrec *)alloctmp(sizeof(struct unsafevalrec));

  data->expr = expr;
  data->assum = assum;
  data->level = level;
  (void) CallCC(ACTION_NAME(unsafevalrec),(ExprPtr)data,true);
  return data->res;
}

PRIVATE Boolean isConstructor(assum, f)
IN CONST ExprPtr assum;
IN CONST ExprPtr f;
{
  ExprPtr ff;
  COUNTER i,l;
  ExprPtr i_l;
  Boolean found;

  ff = forgetExtra(f);
  if (ff->kind==(CHAR)E_VARREF)
    {
      ExprPtr con;

      i = VARRANK(ff);
      l = C_LEVEL(assum)-VARLEVEL(ff);
      i_l = mkIL(i,l);
      con = MapGet(i_l,C_CONS(assum),&found);
      if (found && (con->kind==E_DATA || con->kind==E_DATACON || con->kind==E_EXCEPTION || con->kind==E_STRING))
	return true;
    }
  return false;
}

#ifdef HAS_PROTOTYPES
PRIVATE Boolean unsafeValRec(IN CONST ExprPtr expr,IN CONST ExprPtr assum,
			     IN COUNTER level)
#else
PRIVATE Boolean unsafeValRec(expr,assum,level)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER level;
#endif
{
  if (STKFULL(expr))
    return unsafeValRec1(expr,assum,level);
again:
  switch (expr->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV:
    case E_STACK:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE:
#ifdef OBSOLETE
    case E_MEMO_CLOSURE:
#endif
    case E_DATACON: case E_QUOTE: case E_DELAY:
    case E_POLYFUN: case E_MEMO_FN:
      /*case E_N_LAMBDA:*/
      return false;
    case E_REF: /* should be a safe val rec, by making a cyclic structure */
      /* same thing on arrays (should recognize applications of makearray
	 and arrayoflist */
    case E_TUPLE_1: case E_UNSHARED_TUPLE_1:
      expr = REFVAL(expr);
      goto again;
    case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
    case E_CONS: case E_UNSHARED_CONS:
    case E_MAPLET: case E_OVERWRITE: case E_UNDERWRITE:
    case E_MAPGET:
    case E_RMU1: case E_UNSHARED_RMU1:
    case E_RMUX: case E_UNSHARED_RMUX:
    case E_IF: case E_PROG: case E_AND: case E_OR:
      if (unsafeValRec(CAR(expr),assum,level))
	return true;
      /*fallthrough*/
    case E_RECORD_1: case E_UNSHARED_RECORD_1:
    case E_TSELECT: case E_TSEL1: case E_TSEL2: case E_TSELGNRL:
    case E_RSELECT: case E_RSEL1: case E_RSELGNRL:
      expr = CDR(expr);
      goto again;
    case E_TUPLE_GNRL: case E_UNSHARED_TUPLE_GNRL:
    case E_RECORD_GNRL: case E_UNSHARED_RECORD_GNRL:
      return unsafeValRecVec(expr,assum,level);
    case E_PAT_AS:
    case E_PROFILE:
      expr = CAR(expr);
      goto again;
    case E_VARREF:
      return VARLEVEL(expr)==level; /* safe if the variable was defined on
				       a previous level or next level (see E_LET) */
    case E_LET:
      level++;
      if (unsafeValRecBinding(CDR(LETBIND(expr)),assum,level))
	return true;
      expr = LETBODY(expr);
      goto again;
    case E_APPL: case E_GOTO:
      /*case E_N_APPL:*/
      /* application is safe only when function part is a constructor,
	 or a label selection function, or in general any case where cpApply
	 has optimized E_APPL or E_GOTO out. There remains the case where
	 the function is a E_POLYFUN (case expression). In fact, we also have
	 to deal with the cases where the function part is a constructor. */
      if (isConstructor(assum, APPLFUN(expr)))
	;
      else if ((APPLFUN(expr)->kind!=E_POLYFUN
		/*&& APPLFUN(expr)->kind!=E_N_LAMBDA*/
		)
	  || unsafeValRecPattern(APPLFUN(expr),assum,level+1))
	return true;
      expr = APPLARG(expr);
      goto again;
    case E_GET:
      expr = GETREF(expr);
      goto again;
    case E_SET:
      if (unsafeValRec(SETREF(expr),assum,level))
	return true;
      expr = SETVAL(expr);
      goto again;
    case E_PROG_OF: case E_LIST_OF: case E_UNSHARED_LIST_OF:
    case E_SET_OF: case E_APPEND_OF: case E_UNSHARED_APPEND_OF:
    case E_UNION_OF: case E_UNION_UNDER_OF: case E_INTER_OF: case E_MAP_OF:
    case E_MAP_UNDER_OF: case E_ALL_OF: case E_EXISTS_OF:
    case E_SOME_OF: case E_UNSHARED_SOME_OF:
    case E_PROG_MAP: case E_LIST_MAP: case E_UNSHARED_LIST_MAP:
    case E_SET_MAP: case E_APPEND_MAP: case E_UNSHARED_APPEND_MAP:
    case E_UNION_MAP: case E_UNION_UNDER_MAP: case E_INTER_MAP: case E_MAP_MAP:
    case E_MAP_UNDER_MAP: case E_ALL_MAP: case E_EXISTS_MAP:
    case E_SOME_MAP: case E_UNSHARED_SOME_MAP:
      if (unsafeValRecDom(CDR(OFDOM(expr)),assum,level))
	return true;
      level++;
      expr = OFFP(expr);
      if (unsafeValRec(CDR(expr),assum,level))
	return true;
      expr = CAR(expr);
      goto again;
    case E_PROG_WHILE: case E_LIST_WHILE: case E_UNSHARED_LIST_WHILE:
    case E_SET_WHILE: case E_APPEND_WHILE: case E_UNSHARED_APPEND_WHILE:
    case E_UNION_WHILE: case E_UNION_UNDER_WHILE: case E_INTER_WHILE:
    case E_MAP_WHILE: case E_MAP_UNDER_WHILE: case E_ALL_WHILE: case E_EXISTS_WHILE:
    case E_SOME_WHILE: case E_UNSHARED_SOME_WHILE:
      if (unsafeValRec(CAR(WCOND(expr)),assum,level) ||
	  unsafeValRec(CDR(WCOND(expr)),assum,level))
	return true;
      expr = WBODY(expr);
      goto again;
    case E_HANDLE:
      if (unsafeValRecPattern(HHANDLERS(expr),assum,level+1))
	return true;
      expr = HBODY(expr);
      goto again;
    case E_RAISE: /* raising an exception is safe: it aborts the definition */
    case E_EXCEPTION:
      return false;
    case E_PACK:
      expr = PACKVAL(expr);
      goto again;
    case E_DEBUGZONE:
      expr = DZCODE(expr);
      goto again;
    }
  return false;
}

#if TRANSLATOR
PRIVATE ExprPtr recordInline(pat,ex,assum,bindings)
IN CONST ExprPtr pat,ex;
INOUT ExprPtr assum;
IN CONST ExprPtr bindings;
{
  INT maxinline;
  ExprPtr e,pat0,ex0;
  ENTER;

  pat0 = pat;
  ex0 = ex;
patlab:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      pat = DNPAT(pat);
      goto patlab;
    case E_VARINTRO:
      e = ex;
exlab:
      switch (e->kind)
	{
	case E_PROFILE:
	  e = CAR(e);
	  goto exlab;
	case E_POLYFUN:
	  /*case E_N_LAMBDA:*/
	  maxinline = IVAL(REFVAL(maxInlineRef));
	  if (funSize(e,maxinline)<=maxinline)
	    {
	      ex = Kcounted (E_COUNTED, ex, C_LEVEL (assum));
	      (void) KEEP (ex);
	      setMapPut(mkIL(VARRANK(pat),C_ILEVEL(assum)),ex,&C_CONS(assum));
	    }
	  break;
	}
      break;
    }
  bindings = CONS(CONS(pat0,ex0),bindings);
  KEEPONLY(bindings);
  return bindings;
}
#else
#define recordInline(pat,e,assum) CONS(CONS(pat,e),bindings)
#endif

PRIVATE ExprPtr BuildFunDef(bind)
IN CONST ExprPtr bind; /* A_FUNBIND, return (file name, function name);
			may also be used on regular A_BIND. */
{
  ExprPtr pat;

  pat = CAR(ARGS(bind));
 again:
  switch (TYPE(pat))
    {
    case A_REAL_VAR:
      return KONS(currentFile, CAR(ARGS(pat)));
    case A_PAT_AS:
    case A_PAT_CAST:
      pat = CAR(ARGS(pat));
      goto again;
    }
  return NIL;
}

PRIVATE ExprPtr cpValRec1(args,assum)
IN CONST ExprPtr args;
INOUT ExprPtr assum;
{
  ExprPtr e,epat,bindings,ex;
  struct data {
    ExprPtr e;
    ExprPtr pat;
    ExprPtr local_assum;
  } *exprs,*ep;
  COUNTER i,n;
  ExprPtr savefundef;
  ENTER;

  n = ListLength(args);
  exprs = (struct data *)alloctmp(n*sizeof(struct data));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args), ep++)
    {
      ep->local_assum = DupCompAssum(assum);
#if DEBUGGING
      if (C_LEVEL(assum)==C_ILEVEL(assum))
	C_DBGCNT(ep->local_assum) = 1;
#endif
      ep->e = e = CAR(args);
      {
	ExprPtr fundef;

	fundef = BuildFunDef(e);
	if (fundef!=NIL)
	  C_FUNDEF(ep->local_assum) = fundef;
      }
      epat = CAR(ARGS(e));
      ep->pat = CP(epat,ep->local_assum,-1);
      C_NLOCALS(assum) = C_NLOCALS(ep->local_assum);
#if DEBUGGING
      C_DBGCNT(assum) = C_DBGCNT(ep->local_assum);
#endif
#ifdef OBSOLETE
      SETQ(C_TYTAGS(assum),C_TYTAGS(ep->local_assum)); /* because tytags are global */
#endif
    }
  for (; --ep, --i>=0; )
    {
      e = CAR(ARGS(ep->e));
      addLocalVars (assum, ep->local_assum, e);
    }
  bindings = NIL;
  savefundef = C_FUNDEF(assum);
  (void) KEEP(savefundef);
  for (ep=exprs+n, i=n; --ep, --i>=0; )
    {
      e = CAR(CDR(ARGS(ep->e)));
      C_FUNDEF(assum) = C_FUNDEF(ep->local_assum);
      ex = CP(e,assum,-1);
      if (unsafeValRec(ex,assum,C_LEVEL(assum)))
	type_error(ZONE(e),(CHAR *)message(3));
      bindings = recordInline(ep->pat,ex,assum,bindings);
    }
  C_FUNDEF(assum) = savefundef;
  KEEPONLY(bindings);
  return bindings;
}

PRIVATE ExprPtr cpValRec(expr,assum,gto)
	/* A_VALREC, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,ex;
  ENTER;

  args = CDR(ARGS(expr));
  ex = cpValRec1(args,assum);
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpBind(expr,assum,gto)
	/* A_BIND, return bindings, i.e., lists of pat::expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,pat,ex,savefundef,fundef,bindings;
  ENTER;

  savefundef = C_FUNDEF(assum);
  fundef = BuildFunDef(expr);
  if (fundef!=NIL)
    C_FUNDEF(assum) = fundef;
  args = ARGS(expr);
  e = CAR(CDR(args));
  ex = CP(e,assum,-1);
  e = CAR(args);
  pat = CP(e,assum,-1);
  pat = optPattern(pat);
  bindings = recordInline(pat,ex,assum,NIL);
  C_FUNDEF(assum) = savefundef;
  KEEPONLY(bindings);
  return bindings;
}

PRIVATE ExprPtr cpFunBind(expr,assum,gto)
	/* A_FUNBIND, return bindings, i.e., lists of pat::expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr savefundef;
  ExprPtr aux;
  ENTER;

  savefundef = C_FUNDEF(assum);
  (void) KEEP(savefundef);
  C_FUNDEF(assum) = BuildFunDef(expr);
  aux = cpBind(expr,assum,gto);
  C_FUNDEF(assum) = savefundef;
  KEEPONLY(aux);
  return aux;
}

/* Replacing patterns that match only one value by E_QUOTE(this value)
   is not sure to work, since such patterns must match not only
   the shared value, but also any unshared and structurally identical value.
   We therefore do this only when the pattern has an equality type.
*/
PRIVATE Boolean cstPattern(pat,val)
IN CONST ExprPtr pat;
OUT ExprPtr *val;
{
  switch (pat->kind)
    {
    case E_NIL: case E_TUPLE_0: case E_RECORD_0: case E_EMPTY:
    case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
      *val = pat;
      return true;
    case E_QUOTE:
      *val = REFVAL(pat);
      return true;
    default:
      return false;
    }
}

PRIVATE ExprPtr cpPatTupleGnrl(args,assum,len,tail,nocst)
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
IN INT len;
IN CONST ExprPtr tail;
IN Boolean nocst;
{
  ExprPtr e;
  ExprPtr *exprs,*ep,*cstp;
  COUNTER i;
  ENTER;

  if (len==0)
    return tail;
  nocst |= (tail==patany);
  exprs = (ExprPtr *)alloctmp(2*len*sizeof(ExprPtr));
  for (i=0, ep=exprs, cstp=exprs+len; i<len; i++, ep++, cstp++, args=CDR(args))
    {
      e = CAR(args);
      *ep = CP(e,assum,-1);
      if (!nocst)
	{
	  nocst = !cstPattern(*ep,cstp);
	}
    }
  if (!nocst)
    {
      tail = KtupleGnrl(len,exprs+len);
      (void) KEEP(tail);
      tail = MkQuote(tail);
    }
  else
    for (; --i>=0; )
      {
	e = *--ep;
	if (e==patany && tail==patany)
	  ; /* simplify (a,b,c,_,...) into (a,b,c,...) */
	else  /* this also simplifies (a,_,...) into (a,...), and (_,...) into _ */
	  {
	    tail = MakePatTupleCons(e,tail);
	    (void) KEEP(tail);
	  }
      }
  KEEPONLY(tail);
  return tail;
}

PRIVATE ExprPtr cpPatTuple0(args,assum,len,ellipsis,nosharing)
	/* A_TUPLE, return pattern */
IN CONST ExprPtr args;
IN CONST ExprPtr assum;
IN INT len;
IN Boolean ellipsis,nosharing;
{
  ExprPtr e,e1,e2,aux,tail;
  ENTER;

  tail = ellipsis?patany:tuple0;
  switch (len)
    {
    case 0:
      e = tail;
      break;
    case 1: /* there are no 1-tuple patterns, so there must be an ellipsis,
	       so tail is actually E_PAT_ANY. */
      aux = CAR(args);
      e1 = CP(aux,assum,-1);
      if (e1==patany)
        e = patany; /* simplify (_,...) into _ */
      else
	{
	  if (!ellipsis && cstPattern(e1,&aux) && !nosharing)
	    {
	      e = Ktuple1(aux);
	      (void) KEEP(e);
	      e = MkQuote(e);
	    }
	  else
	    e = MakePatTupleCons(e1,tail);
	}
      break;
    case 2:
      if (!ellipsis)
	{
	  aux = CAR(args);
	  e1 = CP(aux,assum,-1);
	  aux = CAR(CDR(args));
	  e2 = CP(aux,assum,-1);
	  if (e1==patany && e2==patany)
	    e = patany; /* simplify (_,_) into _ */
	  else if (cstPattern(e1,&aux) && cstPattern(e2,&tail) && !nosharing)
	    {
	      e = Ktuple2(aux,tail);
	      (void) KEEP(e);
	      e = MkQuote(e);
	    }
	  else
	    {
	      MAKEPAIR(E_TUPLE_2,e1,e2,e);
	    }
	  break;
	}
      /*fallthrough*/ /* if ellipsis, do as for general tuples */
    default:
      e = cpPatTupleGnrl(args,assum,len,tail,nosharing);
      break;
    }
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpPatTuple(expr,assum,gto)
	/* A_PAT_TUPLE, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,ll;
  INT len;
  ENTER;

  args = ARGS(expr);
  for (ll=args, len=0; ll!=NIL; ll=CDR(ll), len++)
    {
      if (TYPE(CAR(ll))==A_PAT_TUPLE_ELLIPSIS)
	break;
    }
  args = cpPatTuple0(args,assum,len,(ll!=NIL),noSharingPossibly(expr,assum));
  KEEPONLY(args);
  return args;
}

PRIVATE ExprPtr cpNullDecl(expr,assum,gto)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  return NIL;
}

#define cpQuantity cpNullDecl
#define cpNewQuan cpError
#define cpUnit cpNullDecl
#define cpNewUnit cpError
#define cpType cpNullDecl
#define cpTypeBind cpError

PRIVATE ExprPtr cpDataType(expr,assum,gto)
	/* A_DATATYPE, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e;
  ENTER;

  e = CAR(ARGS(expr));
  e = CP(e,assum,-1);
  KEEPONLY(e);
  return e;
}

#define cpDataTypeBind cpError

#define cpAbsType cpLocal
#define cpTypeVarSeq cpError
#define cpConstructor cpError

PRIVATE ExprPtr cpException(expr,assum,gto)
	/* A_EXCEPTION, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,bind,bindings;
  ExprPtr *exprs,*ep;
  COUNTER i,n;
  ENTER;

  args = ARGS(expr);
  n = ListLength(args);
  exprs = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (i=0, ep=exprs; i<n; i++, args=CDR(args))
    {
      e = CAR(args);
      bind = CP(e,assum,-1);
      if (bind!=NIL)
	*ep++ = bind;
    }
  bindings = NIL;
  for (; ep!=exprs; )
    bindings = CONS(*--ep,bindings);
  KEEPONLY(bindings);
  return bindings;
}

#if DEBUGGING
PRIVATE Boolean isBenignExc (id)
IN CONST ExprPtr id;
{
  regexp *re;
  CHAR *res;

  re = regcomp (REFVAL (benignExcRegexpRef));
  if (re==NULL)
    return false; /* if matching fails, then no exception is benign. */
  res = regexec (re, id, 0, SLEN (id));
  MARKFREE (re);
  if (regexperr!=RE_NOERR)
    return false; /* if matching fails, then no exception is benign. */
  return (res!=NULL);
}
#else
#define isBenignExc(id) false
#endif

PRIVATE ExprPtr cpNewExc(expr,assum,gto)
	/* A_NEWEXC, A_EXT_EXC, return binding */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,con,id,ex,aux;
  COUNTER i,l,flags;
  ExprPtr i_l;
  ENTER;

  args = ARGS(expr);
  id = CAR(ARGS(CAR(args)));
  flags = isBenignExc(id)?EXC_BENIGN:EXC_CHECKED;
  if (TYPE(expr)==A_EXT_EXC)
    con = Kcounted(E_EXCEPTION,id,flags); /* do not generate a new exception */
  else
    con = MakeException(id,flags);
  (void) KEEP(con);
  i = C_NLOCALS(assum)++;
  l = C_ILEVEL(assum);
  i_l = mkIL(i,l);
  setMapPut(id,i_l,&C_IDS(assum));
  setMapPut(i_l,con,&C_CONS(assum)); /* do not allow this to be inlined; if we wanted it to,
					we should move this line down to just before ex = Kvarintro(i)
					below. */
  /* in the case 'con', con => con
      in the case 'con of type', con => fn x => con::x (i.e, E_DATACON(con))
  */
  if (CDR(args)!=NIL)
    {
      if (TYPE(expr)!=A_EXT_EXC) /* normal case */
	aux = MakeCounted(E_DATACON,con,(INT)2); /* generative */
      else
	aux = MakeCounted(E_DATACON,con,(INT)0); /* *not* generative */
      con = KEEP(aux);
    }
  ex = Kvarintro(i);
#if DEBUGGING
  if (l!=0 && (C_FLAGS(assum) & C_NODBGNAME)==0)
    {
#ifdef OLDDEBUG
      C_NLOCALS(assum)++; /* to leave a slot free to store the name and type info at run-time */
#endif
      ex = wrapDebugName(expr,id,KEEP(ex),assum,false,true);
    }
#endif
  ex = CONS(KEEP(ex),con);
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpExcBind(expr,assum,gto)
	/* A_EXC_BIND, return NIL */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,i_l;

  args = ARGS(expr);
  i_l = MapGetSure(CAR(ARGS(CAR(CDR(args)))),C_IDS(assum));
  setMapPut(CAR(ARGS(CAR(args))),i_l,&C_IDS(assum));
  return NIL;
}

#define cpMapletType cpError
#define cpRecordType cpError
#define cpTypeApply cpError
#define cpTypePower cpError
#define cpTypeStar cpError
#define cpTupleType cpError
#define cpFunctionType cpError
#define cpMapType cpError

PRIVATE ExprPtr cpPatAny(expr,assum,gto)
	/* A_PAT_ANY, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  return patany;
}

#define cpPatRecordEllipsis cpError
#define cpPatTupleEllipsis cpError
#define cpPatRecordMaplet cpMaplet

PRIVATE ExprPtr getLabeledPats(args,assum,len,ellp)
	/* produce map from labels to non-patany patterns. */
IN CONST ExprPtr args;
INOUT ExprPtr assum;
IN INT len;
INOUT Boolean *ellp;
{
  ExprPtr e,map;
  ENTER;

  map = NewMap();
  while (len--!=0)
    {
      e = CAR(args);
      args = CDR(args);
      e = CP(e,assum,-1);
      if (CDR(e)!=patany)
	map = MapPut(CAR(e),CDR(e),map);
      else *ellp = true; /* replace _ by ... */
    }
  KEEPONLY(map);
  return map;
}

MAPCOLLECTFUNC(do_compileindex)
{
  MapPtr shape;
  ExprPtr xx;

  shape = MAPDATA(MapPtr);
  xx = MapGetSure(x,shape);
  return MapSingleton(xx,y);
}

MAPSWEEPFUNC(do_makerecpattern)
{
  struct makerecpattern *data;
  ExprPtr cons;
  ENTER;

  data = MAPDATA(struct makerecpattern *);
  cons = CONS(x,y);
  data->tail = MakePatRecordMapAdd(cons,data->tail);
  KEEPONLY(data->tail);
  return false;
}

MAPSWEEPFUNC(do_setrecpattern)
{
  ExprPtr *pats;

  pats = MAPDATA(ExprPtr *);
  pats[IVAL(x)-1] = y;
  return false;
}

MAPSWEEPFUNC(do_denserec)
{
  struct denserec *data;

  data = MAPDATA(struct denserec *);
  data->n++;
  if (IVAL(x)>data->maxn)
    data->maxn = IVAL(x);
  return false;
}

PRIVATE Boolean denseRec(map,maxn)
IN CONST MapPtr map;
OUT INT *maxn;
{
  struct denserec *data = (struct denserec *)alloctmp(sizeof(struct denserec));

  data->maxn = data->n = 0;
  (void) MapSweep(map,do_denserec,data);
  *maxn = data->maxn;
  return data->maxn<=data->n+(data->n>>2)+1;
}

PRIVATE ExprPtr makeRecPattern(map,ellipsis)
IN CONST MapPtr map;
IN Boolean ellipsis;
{
  struct makerecpattern *data = (struct makerecpattern *)alloctmp_anchored(sizeof(struct makerecpattern),
									   (char *)&map);
  INT maxn;
  Boolean dense,imprecise;

  data->tail = ellipsis?patany:record0;
  imprecise = (ChooseInDom(map)->kind!=(CHAR)E_INT);
  if (imprecise)
    {
defaultrec:
      (void) MapSweep(map,do_makerecpattern,data);
    }
  else /* otherwise, we know the indexes, and all the fields are present */
    {
      INT i;

      dense = denseRec(map,&maxn);
      if (ellipsis && !dense)
	goto defaultrec;
      {
	ExprPtr *pats;

	pats = (ExprPtr *)alloctmp_anchored(maxn*sizeof(ExprPtr), (char *)&map);
	for (i=0; i<maxn; i++)
	  pats[i] = patany;
	(void) MapSweep(map,do_setrecpattern,pats);
	while (maxn--!=0)
	  {
	    data->tail = MakePatRecordCons(pats[maxn],data->tail);
	    (void) KEEP(data->tail);
	  }
      }
    }
  return data->tail;
}

MAPSWEEPFUNC(do_cstpatrecord)
{
  ExprPtr pat,val;
  Boolean found;
  struct cstpatrecord *data;

  data = MAPDATA(struct cstpatrecord *);
  pat = MapGet(x,data->labpats,&found);
  if (!found || !cstPattern(pat,&val))
    return true;
  data->cpr_x.p[IVAL(y)-1] = val;
  return false;
}

PRIVATE ExprPtr cstPatRecord(shape,labpats)
IN CONST MapPtr shape;
IN CONST MapPtr labpats;
{
  struct cstpatrecord *data;
  ExprPtr pat;
  INT i,n;

  n = MapCard(shape);
  data = (struct cstpatrecord *)alloctmp(VARSIZE(struct cstpatrecord,cpr_x.p,n));
  data->labpats = labpats;

  data->cpr_x.n = n;
  data->cpr_x.next = theHeapFrames;
  for (i=0; i<n; i++)
    data->cpr_x.p[i] = NOEXPR;
  theHeapFrames = (gcStackView *)&data->cpr_x;

  if (MapSweep(shape,do_cstpatrecord,data))
    pat = NOEXPR;
  else
    {
      pat = KrecordGnrl(shape,n,data->cpr_x.p);
      BEGINQUICKKEEP();
      QUICKKEEP(pat);
      pat = MkQuote(pat);
      ENDQUICKKEEP();
    }

  theHeapFrames = data->cpr_x.next;
  return pat;
}

PRIVATE ExprPtr cpPatRecord(expr,assum,gto)
	/* A_PAT_RECORD, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e,args,ll,type,shape,map;
  INT len;
  Boolean found,ellipsis;
  ENTER;

  args = ARGS(expr);
  for (ll=args, len=0; ll!=NIL; ll=CDR(ll), len++)
    {
      if (TYPE(CAR(ll))==A_PAT_RECORD_ELLIPSIS)
	break;
    }
  ellipsis = (ll!=NIL);
  map = getLabeledPats(args,assum,len,&ellipsis);
  if (map==NewMap())
    e = patany;
  else
  {
    type = getTypeSure(expr,assum);
    if ((void) MapGet(REST_TAG,type,&found), !found)
      {
  	/* in this case, we have been able to determine statically
	   the positions in the record of the fields */
	shape = getShape(type);
	switch (shape->kind)
	  {
	    case E_EMPTY:
	      e = record0;
	      break;
	    case E_MAPLET: /* map must be a maplet */
	      shape = MAPLETX(map);
	      ll = MAPLETY(map);
	      if (cstPattern(ll,&e) && !noSharingPossibly(expr,assum))
		{
		  e = Krecord1(shape,e);
		  (void) KEEP(e);
		  e = MkQuote(e);
		}
	      else
		{
		  MAKERECORD1(shape,ll,e);
		}
	      break;
	    default: /* record with at least two fields */
	      if (noSharingPossibly(expr,assum) || (e = cstPatRecord(shape,map))==NOEXPR)
		{
		  map = MapCollect(map,do_compileindex,shape);
		  e = makeRecPattern(map,ellipsis);
		}
	      break;
	  }
      }
    else
      e = makeRecPattern(map,ellipsis);
  }
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpPatAs(expr,assum,gto)
	/* A_PAT_AS, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,pat1,pat2;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  pat1 = CP(e,assum,-1);
  e = CAR(CDR(args));
  pat2 = CP(e,assum,-1);
  e = MakePatAs(pat1,pat2);
  /* E_PAT_AS patterns are considered as never constant. */
  KEEPONLY(e);
  return e;
}

#define cpPatCast cpCast

PRIVATE ExprPtr cpPatNil(expr,assum,gto)
	/* A_PAT_NIL, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  return NIL;
}

#define cpPatAnyList cpPatAny

#ifdef OLDDATA
MAPSWEEPFUNC(do_setbits)
{
  CHAR *bits;
  COUNTER i;

  bits= MAPDATA(CHAR *);
  i = (COUNTER)IVAL(x);
  bits[i >> 3] |= (1 << (i & 7));
  return false;
}

PRIVATE ExprPtr BuildPatCheck(pat,tagset)
IN CONST ExprPtr pat;
IN CONST MapPtr tagset;
{
  char *bits;
  long n;
  ExprPtr e;

  n = (E_UNUSED+7) >> 3;
  CHECKHETEROFREE();
  scAlloc(bits,n);
  if (tagset==NIL) /* all bits */
    memset(bits,0xff,n);
  else
    {
      memset(bits,0,n);
      (void) MapSweep(tagset,do_setbits,bits);
    }
  EALLOC_HETERO_NOGC(e);
  e->kind = (CHAR)E_PAT_CHECK;
  PATCHECKVAL(e) = pat;
  PATCHECKBITS(e) = bits;
  (void) KEEP(e);
  return e;
}
#endif

PRIVATE ExprPtr cpPatApply(expr,assum,gto)
	/* A_PAT_APPLY, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,f,ff,arg,i_l,con;
  COUNTER i,l;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  f = CP(e,assum,-1);
  e = CAR(CDR(args));
  arg = CP(e,assum,-1);
  ff = forgetExtra(f);
  if (/*ff!=NIL && */ff->kind==(CHAR)E_VARREF)
    {
      i = VARRANK(ff);
      l = C_LEVEL(assum)-VARLEVEL(ff);
      if (l==0)
	{
	  if (i==ORIGINAL_CONS)
	    { /* compile pat1::pat2 patterns */
	      ff = forgetExtra(arg);
	      switch (ff->kind)
		{
		case E_QUOTE:
		  ff = REFVAL(ff);
		  arg = KONS(CAR(ff),CDR(ff)); /* no need to check that it has an equality
						  type: it has one, since ff was compiled
						  as a constant already. */
		  arg = MkQuote(arg);
		  goto retarg;
		case E_PAT_ANY: /* op :: _, or _::_ */
		  arg = Cons(patany,patany); /* check length>=1 */
		  goto retarg;
		case E_PAT_TUPLE_GNRL_CONS:
		case E_PAT_REVTUPLE_GNRL_CONS:
		  /* op :: (pat,...), or op :: (pat1,pat2,...) */
		  arg = forgetExtra(CDR(ff));
		  if (arg==patany) /* cannot be tuple0, because of types */
		    arg = Cons(CAR(ff),patany);
		  else /* arg is a pat_tuple_cons */
		    arg = Cons(CAR(ff),CAR(arg));
		  goto retarg;
		case E_TUPLE_2: case E_UNSHARED_TUPLE_2:
		case E_PAT_REVTUPLE_2: /* pat1::pat2, normal case */
		  arg = Cons(CAR(ff),CDR(ff));
		  goto retarg;
		default: /* op :: pat, where pat is typically a variable */
		  arg = MakePatCons(arg);
		  goto retarg;
		}
	    }
	  else if (i==ORIGINAL_REF)
	    {
	      MAKEREF(arg,f);
	      KEEPONLY(f);
	      return f;
	    }
	}
      i_l = mkIL(i,l);
      con = MapGetSure(i_l,C_CONS(assum));
      switch (con->kind)
	{
#ifdef OLDDATA
	case E_EMPTY:
	case E_MAPLET:
	case E_MAP: /* anonymous datatype constructor */
	  arg = BuildPatCheck(arg,con);
	  goto retarg;
	case E_STRING: /* tagged datatype constructor */
	  if (cstPattern(arg,&f) && !noSharingPossibly(expr,assum))
	    {
	      arg = KONS(con,f);
	      arg = MkQuote(arg);
	    }
	  else
	    arg = Cons(con,arg);
	  goto retarg;
#else
	case E_DATACON: /* datatype constructor */
	  con = COUNTEDVAL (con);
	  /*FALLTHROUGH*/
	case E_DATA: /* datatype constructor */
	  if (cstPattern(arg,&f) && !noSharingPossibly(expr,assum))
	    {
	      arg = Kdata(con->magic,CAR(con),f);
	      (void) KEEP(arg);
	      arg = MkQuote(arg);
	    }
	  else
	    {
	      arg = MakeData(con->magic,CAR(con),arg);
	    }
	  goto retarg;
	case E_STRING: /* omitted constructor */
	  goto retarg;
#endif
	case E_EXCEPTION:
	  /* we don't know the exception constructor until run-time,
	     so we cannot precompile the pattern. */
	  arg = MakeAppl(f,arg);
	  goto retarg;
	}
    }
  if (cstPattern(f,&ff) && cstPattern(arg,&con) && !noSharingPossibly(expr,assum))
    {
      arg = KONS(ff,con);
      arg = MkQuote(arg);
    }
  else
    arg = Cons(f,arg);
retarg:
  KEEPONLY(arg);
  return arg;
}

PRIVATE ExprPtr cpPatEmpty(expr,assum,gto)
	/* A_PAT_EMPTY, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  return NewMap();
}

#define cpPatAnyMap cpPatAny
#define cpPatMaplet cpMaplet

PRIVATE ExprPtr cpPatMapAdd(expr,assum,gto)
	/* A_PAT_MAPADD, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,xy,rest,xval,yval,rval;
  ENTER;

  args = ARGS(expr);
  e = CAR(args);
  xy = CP(e,assum,-1);
  e = CAR(CDR(args));
  rest = CP(e,assum,-1);
  if (cstPattern(CAR(xy),&xval) && cstPattern(CDR(xy),&yval) &&
      cstPattern(rest,&rval) && !noSharingPossibly(expr,assum))
    {
      e = MapPut(xval,yval,rval);
      e = MkQuote(e);
    }
  else
    e = MakePatMapAdd(xy,rest);
  KEEPONLY(e);
  return e;
}

#ifdef OLDDATA
	/* the following functions build a map from type names to tytag maps.
		A tytag map is a map from datatype constructors to tytags.
		A tytag is : the set k1 of possible ExprKinds of the data
		inside the branch, and
		the set k2 of possible ExprKinds of the constructed data,
		and a flag, which is true if we need to compile the
		constructor.
			tytag = E_CONS(k1,k2) if flag=true,
				E_TUPLE_2(k1,k2) if flag=false
		Instead of sets, we may find NIL, meaning all ExprKinds.
		(the case for optimized abstract data types, for which
		we don't know the set of ExprKinds; or for inner data
		described by a type variable)
		The rules are:
		initially:
		- k2 = {E_STRING} for a 0-ary constructor (k1 = {}, flag = true)
		- k2 = {E_CONS} for a n-ary constructor (and k1 is the k2
		  of the inner data; flag = true)
		optimization:
		- k2 := k1 when no other branch has a k2 intersecting k1
		  (and flag := false)
	*/

MAPFOLDFUNC(do_gettytagk2)
{
  return CDR(y);
}

MAPFOLDOP(do_sloppyunion)
{
  if (e1==NIL || e2==NIL)
    return NIL;
  return MapOverwrite(e1,e2);
}

PRIVATE MapPtr getTyTagK1(type,tytags)
IN CONST ExprPtr type;
IN CONST MapPtr tytags;
{
  ExprPtr f,tytagmap;
  Boolean found;
  INT len;

  /*if (type==NIL)
    return nilTags;*/
  switch (type->kind)
    {
    case E_NIL:
      return nilTags;
    case E_VAR: /* type variable; implementation may be anything */
      return NIL;
    case E_TYPE_APPL:
      f = APPLFUN(type);
      f = fxSym(f);
      if (f==stringSym)
	return stringTags;
      if (f==intSym)
	return intTags;
      if (f==listSym)
	return listTags;
      if (f==refSym)
	return refTags;
      if (f==arraySym)
	return arrayTags;
      if (f==tableSym)
	return externalTags;
      if (f==funSym)
	return funTags;
      if (f==mapSym)
	return mapTags;
      if (f==boolSym)
	return boolTags;
      if (f==exnSym)
	return exnTags;
      if (f==promiseSym)
	return promiseTags;
      if (f==contSym)
	return contTags;
      if (f==dynSym)
	return dynTags;
      tytagmap = MapGet(f,tytags,&found);
      if (found)
	{
	  /* take the union of the k2 of constructors in tytagmap */
	  return MapFold(tytagmap,do_gettytagk2,(char *)0,
			 NewMap(),do_sloppyunion);
	}
      return NIL; /* if not found, return all possible tags */
    case E_EMPTY: /* record types */
      return record0Tags;
    case E_MAPLET: /* record types */
      return (MAPLETX(type)==REST_TAG)?recordTags:record1Tags;
    case E_MAP: /* record types */
      (void) MapGet(REST_TAG,type,&found);
      if (found)
	{
	  if (MapCardLeq(type,(INT)2L))
	    return record1NTags;
	}
      return recordNTags;
    case E_NUMTYPE: /* numerical types */
      return numTags;
    case E_CONS: /* cartesian products */
      for (len = 0; ISCONS(type); type=CDR(type), len++);
      switch (len)
	{
	case 0:
	  return (type==NIL)?tuple0Tags:tupleTags;
	case 1:
	  return (type==NIL)?tuple1Tags:tuple12NTags;
	case 2:
	  return (type==NIL)?tuple2Tags:tuple2NTags;
	default:
	  return tupleNTags;
	}
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    case E_TYPEFX: /* a set of effects cannot be implemented */
#endif
    default:
      kinderror(message(431),type->kind);
      /*NOTREACHED*/
      return NIL;
    }
}

MAPCOLLECTFUNC(do_inittytag1)
{
  struct inittytag1 *data;
  ExprPtr k1,k2,oldaux,tytag,aux;
  Boolean found;
  ENTER;

  data = MAPDATA(struct inittytag1 *);
  found = false;
  oldaux = NIL;
  if (data->tytagmap!=NIL)
    oldaux = MapGet(x,data->tytagmap,&found);
  if (y->kind==(CHAR)E_TYPE_APPL &&
      (aux = APPLFUN(y), fxSym(aux)==funSym))
    {
      y = CAR(APPLARG(y));
      k2 = consTags;
      k1 = getTyTagK1(y,C_TYTAGS(data->assum));
    }
  else
    {
      k2 = stringTags;
      k1 = NewMap(); /* if we don't compile the constructor, there will remain (),
			but we also need to recognize the constructor without inner
			data */
    }
  if (!found || k1!=CAR(oldaux))
    {
      MAKECONS(k1,k2,aux);
      (void) KEEP(aux);
    }
  else
    aux = oldaux;
  tytag = MapSingleton(x,aux);
  KEEPONLY(tytag);
  return tytag;
}

PRIVATE MapPtr InitTyTag1(cons,assum,tytagmap)
IN CONST MapPtr cons;
IN CONST ExprPtr assum;
IN CONST MapPtr tytagmap;
{
  struct inittytag1 *data = (struct inittytag1 *)alloctmp(sizeof(struct inittytag1));
  MapPtr tytags;
  ENTER;

  data->assum = assum;
  data->tytagmap = tytagmap;
  tytags = MapCollect(cons,do_inittytag1,data);
  KEEPONLY(tytags);
  return tytags;
}

MAPCOLLECTFUNC(do_inittytag)
{
  ExprPtr assum;
  Boolean found;

  assum = MAPDATA(ExprPtr);
  return MapSingleton(x,InitTyTag1(y,assum,MapGet(x,C_TYTAGS(assum),&found)));
}

#ifdef OBSOLETE
MAPSWEEPFUNC(do_findtytaginter)
     /* x is a constructor, y a tytag */
{
  struct findtytaginter *data;

  data = MAPDATA(struct findtytaginter *);
  if (x==data->con)
    return false;
  return CDR(y)==NIL || data->k1==NIL || SetInterp(CDR(y),data->k1);
}

MAPSWEEPFUNC(do_findsimpletytags1)
     /* x is a constructor, y a tytag */
{
  struct findsimpletytags *data;
  struct findtytaginter *inter = (struct findtytaginter *)alloctmp(sizeof(struct findtytaginter));

  data = MAPDATA(struct findsimpletytags *);
  if (y->kind==(CHAR)E_CONS) /* if flag=true */
    {
      inter->con = x;
      inter->k1 = (CAR(y)==NewMap())?nilTags:CAR(y);
      if (MapSweep(data->cons,do_findtytaginter,inter))
	;
      else
	{
	  CDR(y) = inter->k1;
	  y->kind = (CHAR)E_TUPLE_2;
	  data->modified++;
	}
    }
  return false;
}

MAPSWEEPFUNC(do_findsimpletytags)
{
  struct findsimpletytags *data;

  data = MAPDATA(struct findsimpletytags *);
  data->cons = y;
  return MapSweep(y,do_findsimpletytags1,data);
}
#endif

PRIVATE void InitTyTag(tykons,assum)
IN CONST MapPtr tykons;
IN CONST ExprPtr assum;
{
  MapPtr newtytags;
#if DEBUGGING && (!OPT_DEBUG_DATATYPE)
#define NO_OPT_DEBUG
  /* in case we have a debugger, and if OPT_DEBUG_DATATYPE
     is undefined (which is better by default for debugging),
     then tytag maps are trivial, and constructors are always
     compiled: this way, data with unsafe types under the debugger
     are shown with their constructors (usually, constructors
     not compiled are inferred by the printing routines from
     the types of the arguments) */
#endif
#ifndef NO_OPT_DEBUG
  struct findsimpletytags *data = (struct findsimpletytags *)
    alloctmp(sizeof(struct findsimpletytags));
#endif
  ENTER;

#ifndef NO_OPT_DEBUG
  do {
#endif
    newtytags = MapCollect(tykons,do_inittytag,assum);
    setMapMerge(&C_TYTAGS(assum),newtytags);
#ifndef NO_OPT_DEBUG
    data->assum = assum;
    data->modified = 0;
    (void) MapSweep(newtytags,do_findsimpletytags,data);
  } while (data->modified>0);
#endif
  QUIT;
#undef NO_OPT_DEBUG
}
#else
MAPSWEEPFUNC(do_countconsts)
{
  INT *mp;
  ExprPtr aux;

  mp = MAPDATA(INT *);
  if (y->kind==(CHAR)E_TYPE_APPL &&
      (aux = APPLFUN(y), fxSym(aux)==funSym))
    ;
  else ++(*mp);
  return false;
}

MAPSWEEPFUNC(do_makettm)
{
  struct inittytag *data = MAPDATA(struct inittytag *);
  ExprPtr aux;

  if (y->kind==(CHAR)E_TYPE_APPL &&
      (aux = APPLFUN(y), fxSym(aux)==funSym))
    data->shape[data->m++] = x;
  else
    data->shape[data->p++] = x;
  return false;
}

MAPSWEEPFUNC(do_inittytag)
{
  ExprPtr assum;
  INT card;
  struct inittytag *data = (struct inittytag *)alloctmp(sizeof(struct inittytag));
  ExprPtr ttm;
  ENTER;

  assum = MAPDATA(ExprPtr);
  card = MapCard(y);
  data->m = 0;
  (void) MapSweep(y,do_countconsts,&data->m);
  if (card==1 && data->m==0) /* datatype with only one constructor, which is non-constant. */
    ttm = ChooseInDom(y); /* then create the fake tytag 'con'. */
  else
    {
      data->p = 0;
      data->shape = (ExprPtr *)alloctmp(card * sizeof(ExprPtr));
      (void) MapSweep(y,do_makettm,data);
      ttm = Karray(E_ARRAY_GNRL,NIL,card,data->shape);
      (void) KEEP(ttm);
      ttm = Kdata(-data->p,ttm,NIL);
      (void) KEEP(ttm);
    }
  setMapPut(x,ttm,&C_TYTAGS(assum));
  QUIT;
  return false;
}

PRIVATE void InitTyTag(tykons,assum)
IN CONST MapPtr tykons;
IN CONST ExprPtr assum;
{
  (void) MapSweep(tykons,do_inittytag,assum);
}
#endif

#ifdef OLDDATA
MAPSWEEPFUNC(do_databind1)
{
  struct databind *data;
  ExprPtr i_l,con,f,aux;
  COUNTER i,l;

  data = MAPDATA(struct databind *);
  i = C_NLOCALS(data->assum)++;
  l = C_ILEVEL(data->assum);
  i_l = mkIL(i,l);
  setMapPut(x,i_l,&C_IDS(data->assum));
  if (y->kind==(CHAR)E_CONS) /* flag = true, compile constructor */
    {
      con = x;
      /* in the case 'con', con => con
	 in the case 'con of type', con => fn x => con::x
	 */
      if (CAR(y)==NewMap()) /* case 'con'	*/
	f = con;
      else /* case 'con of type'	*/
	{
	  aux = MakeCounted(E_DATACON,con,(INT)!data->sharing); /* shared/non-shared + non-generative */
	  f = KEEP(aux);
	}
    }
  else
    {
      con = CAR(y);
      if (CAR(y)==NewMap()) /* case 'con'	*/
	f = NIL;
      else /* case 'con of type'	*/
	{
	  f = Kvarref(ORIGINAL_ID,C_LEVEL(data->assum));
	  /* this is the identity function */
	  (void) KEEP(f);
	}
    }
  setMapPut(i_l,(f->kind==(CHAR)E_DATACON)?f:con,&C_CONS(data->assum));
  /* put f instead of con into C_CONS map when f is a DATACON: this will
     allow us to inline the constructor when possible. */
  aux = Kvarintro(i);
#if DEBUGGING
  if (l!=0 && (C_FLAGS(data->assum) & C_NODBGNAME)==0)
    {
#ifdef OLDDEBUG
      C_NLOCALS(data->assum)++; /* to leave a slot free to store the name and type info at run-time */
#endif
      aux = wrapDebugName(data->subdecl,x,KEEP(aux),data->assum,true,false);
    }
#endif
  (void) KEEP(aux);
  data->bindings = CONS(CONS(aux,f),data->bindings);
  return false;
}
#endif

MAPSWEEPFUNC(do_databind)
{
  struct databind *data;
  ExprPtr ttm,type;
  COUNTER j,n,magic,m;
  ENTER;

  data = MAPDATA(struct databind *);
  ttm = MapGetSure(x,C_TYTAGS(data->assum));
#if DEBUGGING
  data->subdecl = MapGetSure(x,data->subdecls);
#endif
  if (ttm->kind==(CHAR)E_STRING) /* only one, non-constant constructor. */
    {
      ExprPtr i_l, aux, f, varx, x;
      COUNTER i, l;

      i = C_NLOCALS(data->assum)++;
      l = C_ILEVEL(data->assum);
      i_l = mkIL(i,l);
      setMapPut (ttm, i_l, &C_IDS(data->assum));
      setMapPut (i_l, ttm, &C_CONS(data->assum));
      aux = Kvarintro (i);
      (void)KEEP(aux);
      varx = Kvarintro (0);
      (void) KEEP (varx);
      x = Kvarref(0, 0);
      (void) KEEP (x);
      f = MakePolyfun (-1, CONS (varx, x),(INT)1); /* the identity function. */
      (void) KEEP (f);
      data->bindings = CONS (CONS (aux, f), data->bindings);
    }
  else
    {
      magic = ttm->magic;
      for (j=0, n=AVAL(CAR(ttm))->n; j<n; j++)
	{
	  ExprPtr i_l,con,id,f,aux;
	  COUNTER i,l;

	  i = C_NLOCALS(data->assum)++;
	  l = C_ILEVEL(data->assum);
	  i_l = mkIL(i,l);
	  id = AVAL(CAR(ttm))->expr[j];
	  type = MapGetSure(id,y);
	  if (fxSym(applFun(type,C_TYSTATE(data->assum)))==funSym)
	    {
	      type = CAR(CDR(APPLARG(type)));
	      (void) MapGet(fxSym(applFun(type,C_TYSTATE(data->assum))),
			    data->tyeqs, &data->sharing);
	    }
	  else data->sharing = true;
#ifdef OBSOLETE
	  if (fxSym(applFun(type,C_TYSTATE(data->assum)))==funSym)
	    data->sharing = (eqTypeTdg(CAR(CDR(APPLARG(type))),
				       C_TYSTATE(data->assum),true)!=falseValue);
	  else data->sharing = true;
#endif
	  setMapPut(id,i_l,&C_IDS(data->assum));
	  /* in the case 'con', con => E_DATA(j+magic,ttm,NIL)
	     in the case 'con of type', con => fn x => E_DATA(j+magic,ttm,x);
	     this distinction will be done in bc.c/compileDataCon()
	  */
	  m = j+magic;
	  if (data->sharing)
	    con = Kdata(m,ttm,NIL);
	  else
	    con = MakeData(m,ttm,NIL);
	  (void) KEEP(con);
	  f = MakeCounted(E_DATACON,con,(INT)0); /* 0 = non-generative */
	  (void) KEEP(f);
	  setMapPut(i_l,f,&C_CONS(data->assum)); /* put f, not con, into C_CONS: this will allow
						    the constructor to be inlined. */
	  aux = Kvarintro(i);
#if DEBUGGING
	  if (l!=0 && (C_FLAGS(data->assum) & C_NODBGNAME)==0)
	    {
#ifdef OLDDEBUG
	      C_NLOCALS(data->assum)++; /* to leave a slot free to store the name and type info at run-time */
#endif
	      aux = wrapDebugName(data->subdecl,id,KEEP(aux),data->assum,true,false);
	    }
#endif
	  (void) KEEP(aux);
	  data->bindings = CONS(CONS(aux,f),data->bindings);
	}
    }
  KEEPONLY(data->bindings);
  return false;
}

PRIVATE ExprPtr cpPrimeDataBind(expr,assum,gto)
	/* A_PRIME_DATABIND, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e,tykons,args,ciargs;
  struct databind *data = (struct databind *)alloctmp(sizeof(struct databind));
  ENTER;

  args = ARGS(expr);
  ciargs = ARGS(CAR(args));
#if DEBUGGING
  data->expr = expr;
  data->subdecls = CAR(ciargs);
  ciargs = CDR(ciargs);
#endif
  tykons = CAR(ciargs);
  InitTyTag(tykons,assum);
  data->assum = assum;
  data->bindings = NIL;
  data->tyeqs = CAR(CDR(ciargs));
  (void) MapSweep(tykons,do_databind,data);
  e = Nreverse(data->bindings);
  KEEPONLY(e);
  return e;
}

#define cpUnitName cpError
#define cpTypeVar cpError

PRIVATE ExprPtr cpRealVar(expr,assum,gto)
	/* A_REAL_VAR, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr sym,ex;
  COUNTER i,l;
  ExprPtr i_l;
  ENTER;

  sym = CAR(ARGS(expr));
  l = C_ILEVEL(assum);
  if (sym==itSym && l==0)
    {
      ex = Kvarintro(ORIGINAL_IT);
      KEEPONLY(ex);
      return ex;
    }
  i = C_NLOCALS(assum)++;
  i_l = mkIL(i,l);
  setMapPut(sym,i_l,&C_IDS(assum));
  ex = Kvarintro(i);
#if DEBUGGING
  if (l!=0 && (C_FLAGS(assum) & C_NODBGNAME)==0)
    {
#ifdef OLDDEBUG
      C_NLOCALS(assum)++; /* to leave a slot free to store the name and type info at run-time */
#endif
      ex = wrapDebugName(expr,sym,KEEP(ex),assum,false,false);
    }
#endif
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpPatVar(expr,assum,gto)
	/* A_PAT_VAR, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr i_l,con;
  Boolean found;

  i_l = MapGetSure(CAR(ARGS(expr)),C_IDS(assum));
  con = MapGet(i_l,C_CONS(assum),&found);
#ifdef OLDDATA
  if (found)
    switch (con->kind)
      {
      case E_EMPTY: /* anonymous O-ary datatype constructor */
	return NIL;
      case E_STRING: /* tagged datatype constructor */
	return con;
      }
#else
  if (found)
    switch (con->kind)
      {
      case E_DATACON:
	con = COUNTEDVAL(con);
	/*FALLTHROUGH*/
      case E_DATA: /* datatype constructor */
	if (con->magic<0) /* with no argument */
	  {
#ifdef DATACONPAT
	    con = MakeCounted(E_DATACON,con,(INT)0);
	    (void) KEEP(con);
#endif
	    return con;
	  }
      }
#endif
  return cpVar(expr,assum,gto,0);
}

PRIVATE ExprPtr cpPatConst(expr,assum,gto)
	/* A_PAT_CONST, return pat */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr e;

  e = CAR(ARGS(expr));
  return CP(e,assum,-1);
}

PRIVATE ExprPtr cpSelectFn(expr,assum,gto)
	/* A_SELECT_FN, return expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr label,fun,type,x,xpat;
  ENTER;

  label = getRealLabel(expr,assum);
  type = getTypeSure(expr,assum); /* this is the type of the selection function, */
  type = CAR(APPLARG(type)); /* and we get the type of the record that we select on. */
  xpat = Kvarintro(0);
  (void) KEEP(xpat);
  x = Kvarref(0,0);
  (void) KEEP(x);
  fun = MakePolyfun(-1,CONS(xpat,makeSelect(label,x,type)),(INT)1);
	/* fn x => #lab x */
  /*fun = MkQuote(KEEP(fun));*/
  KEEPONLY(fun);
  return fun;
}

PRIVATE unsigned long varNameGen=0L;

PRIVATE SymPtr freshVarName()
{
  SymPtr sym;
  CHAR buf[32];

  sprintf((char *)buf,".%ld",varNameGen++);
  sym = Ksymbol(buf);
  return KEEP(sym);
}

MAPSWEEPFUNC(do_get_mu)
{
  struct get_mu_data *data;
  ExprPtr var,lvar,patvar,usevar,label,patmaplet,usemaplet;

  data = MAPDATA(struct get_mu_data *);
  label = ZBuild(data->z1,A_LABEL,CONS(x,NIL)); /* zone data->z1 is dummy */
  (void) KEEP(label);
  var = freshVarName();
  lvar = CONS(var,NIL);
  patvar = ZBuild(data->z1,A_REAL_VAR,lvar);
  (void) KEEP(patvar);
  usevar = ZBuild(data->z2,A_VARIABLE,lvar);
  patmaplet = ZBuild(data->z1,A_PAT_RECORD_MAPLET,CONS(label,CONS(patvar,NIL)));
  (void) KEEP(patmaplet);
  usemaplet = ZBuild(data->z2,A_RECORD_MAPLET,CONS(label,CONS(usevar,NIL)));
  (void) KEEP(usemaplet);
  data->pat = CONS(patmaplet,data->pat);
  data->rec = CONS(usemaplet,data->rec);
  return false;
}

PRIVATE ExprPtr cpMu(expr,assum,gto)
	/* A_MU, return expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,e1,rec,mod,type,typeall,rectype,modtype;
  Boolean found;
  COUNTER fnlevel;
  ExprPtr sharingp;
  ENTER;

  args = ARGS(expr);
  e1 = CAR(args);
  rec = CP(e1,assum,-1);
  e = CAR(CDR(args));
  type = NGet(e1,S_ANNOTATIONS(C_TYSTATE(assum)),&found); /* found should be true */
  fnlevel = 0;
  sharingp = trueValue;
  if (type->kind==(CHAR)E_BIAS)
    {
      fnlevel = IVAL(CAR(CDR(type)));
      sharingp = CDR(CDR(type));
      type = CAR(type);
    }
  typeall = getTypeSure(expr,assum); /* its fnlevel should be the same */
  modtype = getTypeSure(e,assum);
  rectype = MapDomRestrBy(typeall,modtype); /* get extra fields in rec */
  if (rectype==NewMap()) /* if all fields of rec are hidden */
    { /* then compile rec;mod */
      mod = CP(e,assum,-1);
      MAKEPROG(E_PROG,rec,mod,e);
    }
  else if ((void) MapGet(REST_TAG,rectype,&found), !found)
    { /* if record type is closed, we can compile indexes directly,
       by (fn |[l1=x1,-,lm=xm,...]| => |[l1=x1,-,lm=xm,lm+1=e1,-,ln=en]|) rec */
      struct get_mu_data *data = (struct get_mu_data *)alloctmp(sizeof(struct get_mu_data));
      ExprPtr oldstate,state;

      oldstate = C_TYSTATE(assum); (void) KEEP(oldstate);
      state = DupTypeState(oldstate);
      SETQ(C_TYSTATE(assum),state);
      data->z1 = ZONE(e1);
      data->z2 = ZONE(e);
      data->pat = ZBuild(ZONE(e1),A_PAT_RECORD_ELLIPSIS,NIL);
      data->pat = CONS(KEEP(data->pat),NIL);
      data->rec = ARGS(e);
      (void) MapSweep(rectype,do_get_mu,data);
      data->pat = ZBuild(data->z1,A_PAT_RECORD,data->pat);
      (void) KEEP(data->pat);
      addAnnotation1(data->pat,type,fnlevel,sharingp,state); /* to compile the pattern precisely */
      data->rec = ZBuild(data->z2,A_RECORD,data->rec);
      (void) KEEP(data->rec);
      addAnnotation1(data->rec,typeall,fnlevel+1,sharingp,state); /* to compile the record precisely */
      e = ZBuild(ZONE(expr),A_LAMBDA,CONS(data->pat,CONS(data->rec,NIL)));
      (void) KEEP(e);
      e = ZBuild(ZONE(expr),A_FN,CONS(e,NIL));
      (void) KEEP(e);
      e = CP(e,assum,-1);
      switch (e->kind)
	{
	case E_PROFILE:
	  e = CAR(e);
	  break;
	}
      if (gto<0)
	e = MakeAppl(e,rec);
      else
	e = MakeGoto((CHAR)E_GOTO,e,rec);
      SETQ(C_TYSTATE(assum),oldstate);
    }
  else
    { /* otherwise, rec has a record type with ellipsis */
      mod = CP(e,assum,-1);
      e1 = (mod->kind==(CHAR)E_QUOTE)?REFVAL(mod):mod;
      switch (e1->kind)
	{
	case E_RECORD_0: /* should not happen */
	  e = rec;
	  break;
	case E_RECORD_1:
	case E_UNSHARED_RECORD_1:
	  if (mod->kind==(CHAR)E_QUOTE)
	    {
	      mod = MakeQuote(CDR(e1));
	      (void) KEEP(mod);
	      mod = MakePair((ExprKind)e1->kind,CAR(e1),mod);
	      (void) KEEP(mod);
	    }
	  e = MakePair(noSharingSure(expr,assum)?E_UNSHARED_RMU1:E_RMU1,rec,mod);
	  break;
	case E_RECORD_GNRL:
	case E_UNSHARED_RECORD_GNRL:
	  e = MakePair(noSharingSure(expr,assum)?E_UNSHARED_RMUX:E_RMUX,rec,mod);
	  break;
	}
    }
#if DEBUGGING
  (void) KEEP(e);
  e = wrapDebugZone(expr,e,C_DBGCNT(assum)++,assum);
#endif
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpDelay(expr,assum,gto)
	/* A_DELAY, return expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr pat,e;
  ENTER;

  e = CAR(ARGS(expr));
  pat = ZBuild(ZONE(expr),A_PAT_ANY,NIL);
  (void) KEEP(pat);
  e = ZBuild(ZONE(expr),A_LAMBDA,CONS(pat,CONS(e,NIL))),
  (void) KEEP(e); /* compile fn _ => expr, and put in inside an E_DELAY node */
  e = cpLambda1(e,assum,-1,0);
  e = MakeDelay(e);
  KEEPONLY(e);
  return e;
}

PRIVATE ExprPtr cpNonSense(expr,assum,gto)
	/* A_NONSENSE, return expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr exc;
  COUNTER level;

  level = C_LEVEL(assum);
  /* compile as in cpVariable */
  exc = Kvarref(ORIGINAL_NONSENSE,level);
  expr = MakeRaise(KEEP(exc));
  return KEEP(expr);
}

PRIVATE ExprPtr cpPack1(expr,assum,kind)
	/* A_PACK, A_PAT_PACK return expr */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN ExprKind kind;
{
  ExprPtr e,ex,type;
  Boolean found;
  ENTER;

  e = CAR(ARGS(expr));
  ex = CP(e,assum,-1);
  type = NGet(expr,S_PACKERS(C_TYSTATE(assum)),&found); /* found should be true */
  if (type->kind==(CHAR)E_BIAS)
    type = CAR(type);
  ex = MakePack(kind,ex,type);
  KEEPONLY(ex);
  return ex;
}

PRIVATE ExprPtr cpPack(expr,assum,gto)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN COUNTER gto;
{
  ExprPtr ex,val;

  ex = cpPack1(expr,assum,E_PACK);
  if (cstExpr(PACKVAL(ex),&val))
    {
      ex = MakePack(E_PACK,val,PACKTYPE(ex));
      (void) KEEP(ex);
      ex = MkQuote(ex);
      (void) KEEP(ex);
    }
#if DEBUGGING
  ex = wrapDebugZone(expr,ex,C_DBGCNT(assum)++,assum);
  (void) KEEP(ex);
#endif
  return ex;
}

PRIVATE ExprPtr cpPatPack(expr,assum,gto)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN COUNTER gto;
{
  return cpPack1(expr,assum,E_PAT_PACK);
}

PRIVATE ExprPtr cpInteger(expr,assum,gto)
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN COUNTER gto;
{
  INT n;
  REAL x;

  expr = CAR(ARGS(expr));
  switch (expr->kind)
    {
    case E_REAL:
      x = XVAL(expr);
      n = (INT)x;
      if ((REAL)n!=x)
	type_error(ZONE(expr),(CHAR *)message(506));
      break;
    case E_INT:
      n = IVAL(expr);
      break;
    default:
      n = 0; /* should not happen. */
      break;
    }
  return KINT(n);
}

#define cpTypeEllipsis cpError

#define cpExtDatatype cpNullDecl
/* no need to compute any tytag map, as this implies that the tags of data
   of the extern datatype will be unknown */

#define cpExtEqtype cpNullDecl
#define cpExtQuantity cpNullDecl

PRIVATE void cpExtVal1(sym,expr,assum,excp,type)
IN CONST ExprPtr sym;
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN Boolean excp;
IN CONST ExprPtr type;
{
  ExprPtr i_l,ex,aux;
  COUNTER i,l;
  Boolean found;
  ENTER;

  (void) MapGet(sym,C_IDS(assum),&found);
  if (found) /* already known externs are resolved at compile-time */
    return;
  /*otherwise, reserve a slot for the variable, and create a binding x = ?, where ? is a dummy. */
  l = C_ILEVEL(assum);
  i = C_NLOCALS(assum)++;
  i_l = mkIL(i,l);
  setMapPut(sym,i_l,&C_IDS(assum));
  ex = Kvarintro(i);
#if DEBUGGING
  if (l!=0 && (C_FLAGS(assum) & C_NODBGNAME)==0)
    {
#ifdef OLDDEBUG
      C_NLOCALS(assum)++; /* to leave a slot free to store the name and type info at run-time */
#endif
      if (type==NOEXPR)
	ex = wrapDebugName(expr,sym,KEEP(ex),assum,false,excp);
      else
	ex = wrapDebugName1(type,sym,KEEP(ex),assum,false,excp);
    }
#endif
  (void) KEEP(ex);
  ex = CONS(ex,sym); /* sym is a dummy, to be patched at link-time */
  aux = Cons(ex,C_IMPORTS(assum));
  SETQ(C_IMPORTS(assum),aux);
  QUIT;
}

PRIVATE ExprPtr cpExtVal(expr,assum,gto)
	/* A_EXT_VAL, A_EXT_EXC, return bindings */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr args,e,sym;

  for (args = ARGS(expr); args!=NIL; args = CDR(args))
    {
      e = CAR(ARGS(CAR(args))); /* get A_VARIABLE node */
      sym = CAR(ARGS(e));
      cpExtVal1(sym,e,assum,false /*(TYPE(expr)==A_EXT_EXC)*/,NOEXPR);
    }
  return NIL;
}

#define cpExtExc cpNewExc

PRIVATE ExprPtr cpMemoFn(expr,assum,gto)
	/* A_MEMO_FN, return expr */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  ExprPtr aux,fn;
  ENTER;

  fn = cpMatch(expr,assum,gto);
  MAKEMEMO(NewMap(),fn,aux);
#if TRANSLATOR
  (void) KEEP(aux);
  aux = nameFn(expr,aux,assum);
#endif
  KEEPONLY(aux);
  return aux;
}

#define cpConsInfo cpError

PRIVATE ExprPtr cpEqtypeInfo(expr,assum,gto)
	/* A_EQTYPE_INFO, return environment */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  return NIL;
}

#ifdef OBSOLETE
PRIVATE void untrimTable(assum,local_assum)
IN CONST ExprPtr assum;
INOUT ExprPtr local_assum;
{ /* !!! this is wrong (do the linking!) */
  ExprPtr aux;
  ENTER;

  aux = MapMerge(C_IDS(assum),C_IDS(local_assum));
  SETQ(C_IDS(assum),aux);
#if MODULE_SYSTEM & STD_ML_MODULE
  aux = MapMerge(C_STRUCTS(assum),C_STRUCTS(local_assum));
  SETQ(C_STRUCTS(assum),aux);
  aux = MapMerge(C_SIGS(assum),C_SIGS(local_assum));
  SETQ(C_SIGS(assum),aux);
#endif
  aux = MapMerge(C_CONS(assum),C_CONS(local_assum));
  SETQ(C_CONS(assum),aux);
  aux = MapMerge(C_TYTAGS(local_assum),C_TYTAGS(assum));
  SETQ(C_TYTAGS(assum),aux);
  QUIT;
}
#endif

PRIVATE ExprPtr resolveImports(zone,imports,assum) /* return reversed list of bindings */
IN CONST ExprPtr zone;
INOUT ExprPtr imports;
IN CONST ExprPtr assum;
{
  ExprPtr binding,i_l,pat,bindings,bind,ll;
  Boolean found,excp;
  COUNTER i,l,level;
  ENTER;

  bindings = NIL;
  level = C_LEVEL(assum)+2; /* the (local) module environment is allocated just below
			       a (fake) copy of GLOBENV, which is again just below the
			       current environment. */
  for (ll = imports; ll!=NIL; ll = CDR(ll))
    {
      binding = CAR(ll);
      i_l = MapGet(CDR(binding),C_IDS(assum),&found);
      if (!found) /* then insert an implicit 'extern x:type' declaration */
	{
	  /* we have to estimate whether is has to be treated as an exception
	     constructor by the debugger, if we have one. We can only know
	     this if the module was compiled with debug options, i.e. when
	     the pattern contains a E_DEBUGNAME cell. */
	  l = C_ILEVEL(assum);
	  i = C_NLOCALS(assum);
	  pat = CAR(binding);
	  excp = (pat->kind==(CHAR)E_DEBUGNAME && DNINFO(pat)->kind==(CHAR)E_GOTO);
	  cpExtVal1(CDR(binding),zone,assum,excp,LETBODY(COUNTEDVAL(DNINFO(pat))));
	}
      else
	{
	  if (i_l->kind==(CHAR)E_CONS)
	    {
	      i = IVAL(CAR(i_l));
	      l = IVAL(CDR(i_l));
	    }
	  else
	    {
	      i = IVAL(i_l);
	      l = 0;
	    }
	}
      bind = Kvarref(i,level-l);
      (void) KEEP(bind);
      bind = CONS(CAR(binding),bind);
      bindings = Cons(bind,bindings);
      KEEPONLY(bindings);
    }
  KEEPONLY(bindings);
  return bindings;
}

MAPSWEEPFUNC(do_export)
{
  struct export *data = MAPDATA(struct export *);
  ExprPtr var,val,i_l,binding,bindings,cons;
  COUNTER i,l,maxl;
  Boolean found,consp,excp;
  ENTER;

  if (y->kind==(CHAR)E_CONS)
    {
      i = IVAL(CAR(y));
      l = IVAL(CDR(y));
    }
  else
    {
      i = IVAL(y);
      l = 0;
    }
  maxl = C_LEVEL(data->basis->basisTable);
  if (l<=maxl) /* C_LEVEL should be 1, variables with higher levels
		  are locals (and it is probably a bug if l>1). */
    {
      val = Kvarref(i,maxl-l);
      (void) KEEP(val);
      l = C_ILEVEL(data->assum);
      i = C_NLOCALS(data->assum)++;
      i_l = mkIL(i,l);
      setMapPut(x,i_l,&C_IDS(data->assum));
      cons = MapGet(y,C_CONS(data->basis->basisTable),&found);
      if (found)
	{
	  switch (cons->kind)
	    {
	    case E_COUNTED: /* case E_POLYFUN: case E_PROFILE: */
	      /* !!! We should shiftExpr(cons,IVAL(CDR(y)) - l) here,
		 however this means shiftExpr(cons,-1):
		 all references to varref(rank, level = 0) should
		 become references to where resolveImports have relocated
		 them, all references to varref(rank, level = 1) should
		 become references to toplevel primitives, and all other
		 references should be shifted down by 1.
		 For now, I simply disallow inlining of imported functions,
		 this will be simpler (but it prevents me from inlining
		 functions that are explicitly defined in header files, sob.) */
	      break;
	    default:
	      setMapPut(i_l,cons,&C_CONS(data->assum));
	    }
	}
      var = Kvarintro(i);
#if DEBUGGING
      if (l!=0 && (C_FLAGS(data->assum) & C_NODBGNAME)==0)
	{
#ifdef OLDDEBUG
	  C_NLOCALS(data->assum)++;
#endif
	  consp = found && (cons->kind!=(CHAR)E_EXCEPTION);
	  excp = found && (cons->kind==(CHAR)E_EXCEPTION);
	  var = wrapDebugName1(MapGetSure(x,A_IDS(data->basis->basisAssum)),
			       x,KEEP(var),data->assum,consp,excp);
	}
#endif
      (void) KEEP(var);
      binding = CONS(var,val);
      bindings = Cons(binding,REFVAL(data->bindings));
      SETREFVAL(data->bindings,bindings);
    }
  QUIT;
  return false;
}

PRIVATE ExprPtr cpBasis(expr,assum,gto)
	/* A_BASIS, return bindings */
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN CONST COUNTER gto;
{
  Basis *basis;
  ExprPtr bindings,def;
  struct export *data;
  ENTER;

  basis = (Basis *)EXTDATA(CAR(ARGS(expr)));
  bindings = resolveImports(expr,C_IMPORTS(basis->basisTable),assum);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(bindings);
    hfprintf(hstderr,"%d imports, ",n);
  }
#endif
#if TRANSLATOR
  def = MakeAppl(basis->bindings,NIL);
  (void) KEEP(def);
  def = CONS(CONS(patany,def),NIL);
#else
  def = basis->bindings;
#endif
  bindings = Reverse1(bindings,def);
  (void) KEEP(bindings);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(def);
    hfprintf(hstderr,"%d defined, ",n);
  }
#endif
  data = (struct export *)alloctmp(sizeof(struct export));
  data->assum = assum;
  data->basis = basis;
  data->bindings = MakeSing(E_REF,NIL);
  (void) KEEP(data->bindings);
  (void) MapSweep(MapDomRestrBy(C_IDS(basis->basisTable),A_IDEXTS(basis->basisAssum)),
		  do_export,(char *)data);
#ifdef DEBUGMOD
  {
    INT n;

    n = ListLength(REFVAL(data->bindings));
    hfprintf(hstderr,"%d exported.\n",n);
    hfflush(hstderr);
  }
#endif
  bindings = MakeLet(CONS(KINT(-1-C_NLOCALS(basis->basisTable)),bindings),
		     REFVAL(data->bindings));
  /* the number of local variables is set to a negative number, so that eval knows
     what to do (allocate a fake copy of GLOBENV before allocating the local environment) */
  (void) KEEP(bindings);
#if MODULE_SYSTEM & STD_ML_MODULE
  setMapMerge(&C_STRUCTS(assum),C_STRUCTS(basis->basisTable));
  setMapMerge(&C_SIGS(assum),C_SIGS(basis->basisTable));
#endif
  setMapMerge(&C_TYTAGS(assum),C_TYTAGS(basis->basisTable)); /* because tytags are global */
  bindings = Cons(bindings,NIL);
  KEEPONLY(bindings);
  return bindings;
}

PRIVATE ExprPtr cpEnterFile(expr,assum,gto)
     /* A_ENTER_FILE, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{ /* this is for error reporting reasons, so that errors occurring in
     an included file indeed get reported as errors from that file. */
  ExprPtr arg;

  if (yyfilename!=NULL)
    MARKFREE(yyfilename);
  arg = CAR(ARGS(expr));
  yyfilename = (CHAR *)MALLOC(SLEN(arg)+1);
  safememcpy(yyfilename,STEXT(arg),SLEN(arg));
  yyfilename[SLEN(arg)] = 0;
  setCurrentFile();
  return NIL;
}

PRIVATE ExprPtr cpExitFile(expr,assum,gto)
     /* A_EXIT_FILE, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{ /* this is for error reporting reasons, so that errors occurring in
     an included file indeed get reported as errors from that file. */
  if (yyfilename!=NULL)
    MARKFREE(yyfilename);
  yyfilename = NULL;
  setCurrentFile();
  return NIL;
}

PRIVATE ExprPtr cpPragma(expr,assum,gto)
     /* A_PRAGMA, return bindings */
IN CONST ExprPtr expr;
INOUT ExprPtr assum;
IN CONST COUNTER gto;
{
  CHAR *pragma;

  pragma = STEXT(CAR(ARGS(expr)));
  switch (pragma[3])
    {
    case 'D': /* debug pragma */
      if (pragma[4]=='-')
	C_FLAGS(assum) |= C_NODEBUG;
      else C_FLAGS(assum) &= ~C_NODEBUG;
      break;
    case 'P': /* profiler pragma */
      if (pragma[4]=='-')
	C_FLAGS(assum) &= ~C_PROFILE;
      else C_FLAGS(assum) |= C_PROFILE;
      break;
    }
  return NIL;
}

PRIVATE ExprPtr (*cpAction[]) P((ExprPtr,ExprPtr,COUNTER)) = {
  cpString, cpNumber, cpVariable,
  cpUnitPower, cpUnitStar, cpSetUnit,
  cpLabel, cpRecordMaplet, cpRecord,
  cpFixity, cpTuple, cpProgn,
  cpList, cpListOf, cpListMap,
  cpInSet, cpInList,
  cpMaplet, cpMap, cpMapOf, cpMapMap,
  cpMapUnder, cpMapUnderOf, cpMapUnderMap,
  cpProgOf, cpProgMap,
  cpAllOf, cpAllMap, cpExistsOf, cpExistsMap, cpSomeOf, cpSomeMap,
  cpLet, cpApply, cpCast,
  cpAnd, cpOr,
  cpHandle, cpRaise, cpFn, cpLambda,
  cpIf, cpWhile, cpWhileList, cpWhileMap, cpWhileMapUnder,
  cpWhileAll, cpWhileExists, cpWhileSome,
  cpDeclList, cpLocal, cpVal, cpValRec, cpBind,
  cpPatTuple,
  cpQuantity, cpNewQuan, cpUnit, cpNewUnit,
  cpType, cpTypeBind, cpDataType, cpDataTypeBind,
  cpAbsType, cpTypeVarSeq, cpConstructor,
  cpException, cpNewExc, cpExcBind,
  cpMapletType, cpRecordType, cpTypeApply,
  cpTypePower, cpTypeStar, cpTupleType,
  cpFunctionType, cpMapType,
  cpPatAny, cpPatRecordEllipsis, cpPatTupleEllipsis,
  cpPatRecordMaplet, cpPatRecord, cpPatAs, cpPatCast,
  cpPatNil, cpPatAnyList, cpPatApply, cpPatEmpty,
  cpPatAnyMap, cpPatMaplet, cpPatMapAdd,
  cpPrimeDataBind, cpUnitName, cpTypeVar,
  cpRealVar, cpPatVar, cpPatConst,
  cpSelectFn, cpMatch, cpMu, cpSubMap, cpDelay, cpNonSense,
  cpPack, cpPatPack, cpInteger,
  cpTypeEllipsis, cpExtDatatype, cpExtEqtype, cpExtQuantity, cpExtVal,
  cpMemoFn, cpConsInfo, cpEqtypeInfo, cpExtExc, cpBasis,
  cpEnterFile, cpExitFile, cpPragma, cpFunBind
};

DEF_PRIVATE_ACTION_1ARG(compile)
{
  struct compile *data;

  data = (struct compile *)ENV_UNDERLYING(env);
  return (*cpAction[TYPE(data->expr)])(data->expr,data->assum,data->gto);
}
ACTION_END

PRIVATE ExprPtr compile1(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER gto;
{
  struct compile *data = (struct compile *)alloctmp(sizeof(struct compile));

  data->expr = expr;
  data->assum = assum;
  data->gto = gto;
  return CallCC(ACTION_NAME(compile),(ExprPtr)data,true);
}

PRIVATE ExprPtr compile(expr,assum,gto)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
IN COUNTER gto;
{
  ExprPtr ex;
  ENTER;

  if (STKFULL(ex))
    return compile1(expr,assum,gto);
  ex = (*cpAction[TYPE(expr)])(expr,assum,gto);
  KEEPONLY(ex);
  return ex;
}

PUBLIC ExprPtr cp(expr,assum)
IN CONST ExprPtr expr;
IN CONST ExprPtr assum;
{
  ExprPtr ex;
  ENTER;

  ex = (*cpAction[TYPE(expr)])(expr,assum,(COUNTER)-1);
  KEEPONLY(ex);
  return ex;
}
