/* Initialization of compiler to interpreted form.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "compile.h"
#include "expr.h"
#include "ex2proto.h"
#include "regproto.h"
#include "cmpproto.h"
#include "evlproto.h"
#ifdef HAS_STDARG
#include <stdarg.h>
#else
#ifdef HAS_VARARGS
#include <varargs.h>
#endif
#endif
#ifdef HAS_PROTOTYPES
#include "scmpinit.h"
#endif

#define __SEG__ INIT

/* during initialization, there is no garbage collection,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KINT
#define KINT Kint

#ifdef HIMMLX
EXTERN CHAR *funNames[];
#endif

PUBLIC ExprPtr basisTable,basisEnv;

EXTERN ExprPtr
	nomemException,
	stackException,noninvException,
	nthException,rangeException,
	asciiException,stringnthException,
	nonsenseException,catchException,
	subscriptException,ioException,
	ufindException,futureException,reException;
EXTERN SymPtr nilSym,consSym,refsym,boolSym,
	derefSym,assignSym,forceSym,
	bindSym,matchSym,emptySym,mapgetSym,
	parsweepSym,nomemSym,
	stackSym,incompSym,noninvSym,nthSym,
	rangeSym,asciiSym,stringnthSym,
	subscriptSym,arithSym,ioSym,reSym,
	ufindSym,nonsenseSym,futureSym,
	appendSym,dunionSym,overwriteSym,
	underwriteSym,interSym,identitySym,
	mergeSym,mapappSym,catchSym;
#ifdef OLDDATA
EXTERN ExprPtr nilTags,stringTags,intTags,listTags,refTags,
	       arrayTags,mapTags,boolTags,exnTags,
	       promiseTags,contTags,dynTags,numTags,
	       funTags,consTags,record0Tags,record1Tags,
	       recordNTags,record1NTags,recordTags,
	       tuple0Tags,tupleTags,tuple1Tags,tuple12NTags,
	       tuple2Tags,tuple2NTags,tupleNTags
;
#endif
EXTERN SymPtr itSym;
EXTERN SymPtr falseSym,trueSym;
EXTERN ExprPtr currentFile;

#if PROFILING
EXTERN ExprPtr profileExprs;
#endif

PUBLIC void InitCompile()
{
  VecPtr vp;
  ExprPtr aux;
  ENTER;

#if PROFILING
  profileExprs = NIL;
#endif
  currentFile = NIL;

#if STOP_AND_COPY_VECTORS
  if (scRealSize(VECSIZE(MAX_GLOBALS))<=MAXALLOC_IN_PAGE)
    berror(message(69),MAX_GLOBALS,VECSIZE(MAX_GLOBALS),MAXALLOC_IN_PAGE);
#endif
  vp = (VecPtr)SAFEMALLOC(VECSIZE(MAX_GLOBALS));
    /* correct even if STOP_AND_COPY_VECTORS!=0, because VECSIZE(MAX_GLOBALS)
       is greater than MAXALLOC_IN_PAGE; correct in the other case because
       it will never be freed. */
  vp->n = NORIGINALS;
#ifdef HIMMLX
#define PRIM(i,name) vp->expr[i] = makePrimitive(name); funNames[i] = (CHAR *)#name
#else
#define PRIM(i,name) vp->expr[i] = makePrimitive(name)
#endif

  PRIM(ORIGINAL_CONS,cons);
  PRIM(ORIGINAL_REF, ref);
  PRIM(ORIGINAL_GET, deref);
  PRIM(ORIGINAL_SET, assign);
  PRIM(ORIGINAL_APPEND_OF, appendof);
  PRIM(ORIGINAL_UNION_OF, unionof);
  PRIM(ORIGINAL_MERGE_OF, mergeof);
  PRIM(ORIGINAL_MERGE_UNDER_OF, mergeunderof);
  PRIM(ORIGINAL_INTER_OF, interof);
  PRIM(ORIGINAL_ID, id);
  vp->expr[ORIGINAL_IT] = tuple0;
  vp->expr[ORIGINAL_TRUE] = trueValue;
  vp->expr[ORIGINAL_FALSE] = falseValue;
  vp->expr[ORIGINAL_NIL] = NIL;
  vp->expr[ORIGINAL_BIND_EXC] = BINDEXC;
  vp->expr[ORIGINAL_MATCH_EXC] = MATCHEXC;
  vp->expr[ORIGINAL_EMPTY_EXC] = EMPTYEXC;
  vp->expr[ORIGINAL_MAPGET_EXC] = MAPGETEXC;
  vp->expr[ORIGINAL_MAP_EXC] = MAPEXC;
  vp->expr[ORIGINAL_NOMEM_EXC] = nomemException;
  vp->expr[ORIGINAL_STACK_EXC] = stackException;
  vp->expr[ORIGINAL_INCOMP_EXC] = INCOMPEXC;
  vp->expr[ORIGINAL_NTH_EXC] = nthException;
  vp->expr[ORIGINAL_RANGE_EXC] = rangeException;
  vp->expr[ORIGINAL_ASCII_EXC] = asciiException;
  vp->expr[ORIGINAL_STRNTH_EXC] = stringnthException;
  vp->expr[ORIGINAL_NONINV_EXC] = noninvException;
  vp->expr[ORIGINAL_SUBSCRIPT_EXC] = subscriptException;
  vp->expr[ORIGINAL_ARITH_EXC] = ARITHEXC;
  vp->expr[ORIGINAL_IO_EXC] = MakeCompiled(E_COMPILED_CLOSURE,
					   ACTION_NAME(make_io_exc),
					   ioException);
#ifdef HIMMLX
  funNames[ORIGINAL_IO_EXC] = (CHAR *)"make_io_excAction";
#endif
  vp->expr[ORIGINAL_RE_EXC] = MakeCompiled(E_COMPILED_CLOSURE,
					   ACTION_NAME(make_io_exc),
					   reException);
#ifdef HIMMLX
  funNames[ORIGINAL_RE_EXC] = (CHAR *)"make_io_excAction";
#endif
  vp->expr[ORIGINAL_UFIND_EXC] = ufindException;
  vp->expr[ORIGINAL_CATCH_EXC] = catchException;
  PRIM(ORIGINAL_OVERWRITE, overwrite);
  PRIM(ORIGINAL_FORCE, force);
  PRIM(ORIGINAL_MAPAPP, mapapp);
  vp->expr[ORIGINAL_NONSENSE] = nonsenseException;
  vp->expr[ORIGINAL_FUTURE] = futureException;
  PRIM(ORIGINAL_EQUAL, equal);
  PRIM(ORIGINAL_NEQUAL, nequal);
  PRIM(ORIGINAL_NOT, not);
  MAKEENV(vp,NIL,aux);
  SETQ(basisEnv,aux);
  GLOBENV = basisEnv;

  aux = NewCompAssum();
  SETQ(basisTable,aux);
  aux = MapMerge(MapMerge(MapMerge(MapPair(consSym,KINT(ORIGINAL_CONS),
					   forceSym,KINT(ORIGINAL_FORCE)),
				   MapPair(refsym,KINT(ORIGINAL_REF),
					   appendSym,KINT(ORIGINAL_APPEND_OF))),
			  MapMerge(MapPair(dunionSym,KINT(ORIGINAL_UNION_OF),
					   overwriteSym,KINT(ORIGINAL_MERGE_OF)),
				   MapPair(underwriteSym,KINT(ORIGINAL_MERGE_UNDER_OF),
					   interSym,KINT(ORIGINAL_INTER_OF)))),
		 MapMerge(MapMerge(MapPair(identitySym,KINT(ORIGINAL_ID),
					   nonsenseSym,KINT(ORIGINAL_NONSENSE)),
				   MapPair(mapappSym,KINT(ORIGINAL_MAPAPP),
					   itSym,KINT(ORIGINAL_IT))),
			  MapMerge(MapPair(trueSym,KINT(ORIGINAL_TRUE),
					   falseSym,KINT(ORIGINAL_FALSE)),
				   MapPair(nilSym,KINT(ORIGINAL_NIL),
					   bindSym,KINT(ORIGINAL_BIND_EXC)))));
  aux = MapMerge(MapMerge(MapMerge(MapPair(ioSym,KINT(ORIGINAL_IO_EXC),
					   futureSym,KINT(ORIGINAL_FUTURE)),
				   aux),
			  MapPair(derefSym,KINT(ORIGINAL_GET),
					   assignSym,KINT(ORIGINAL_SET))),
		 MapMerge(MapMerge(MapMerge(MapPair(matchSym,KINT(ORIGINAL_MATCH_EXC),
						    emptySym,KINT(ORIGINAL_EMPTY_EXC)),
					    MapPair(mapgetSym,KINT(ORIGINAL_MAPGET_EXC),
						    parsweepSym,KINT(ORIGINAL_MAP_EXC))),
				   MapMerge(MapPair(nomemSym,KINT(ORIGINAL_NOMEM_EXC),
						    stackSym,KINT(ORIGINAL_STACK_EXC)),
					    MapPair(incompSym,KINT(ORIGINAL_INCOMP_EXC),
						    noninvSym,KINT(ORIGINAL_NONINV_EXC)))),
			  MapMerge(MapMerge(MapPair(nthSym,KINT(ORIGINAL_NTH_EXC),
						    rangeSym,KINT(ORIGINAL_RANGE_EXC)),
					    MapPair(asciiSym,KINT(ORIGINAL_ASCII_EXC),
						    stringnthSym,KINT(ORIGINAL_STRNTH_EXC))),
				   MapPair(subscriptSym,KINT(ORIGINAL_SUBSCRIPT_EXC),
					   arithSym,KINT(ORIGINAL_ARITH_EXC)))));
  aux = MapMerge(MapMerge(MapMerge(MapPair(ufindSym,KINT(ORIGINAL_UFIND_EXC),
					   reSym,KINT(ORIGINAL_RE_EXC)),
				   MapPair(mergeSym,KINT(ORIGINAL_OVERWRITE),
					   catchSym,KINT(ORIGINAL_CATCH_EXC))),
			  MapPut(Ksymbol(message(275)),KINT(ORIGINAL_NOT),
				 MapPair(Ksymbol(message(276)),KINT(ORIGINAL_EQUAL),
					 Ksymbol(message(277)),KINT(ORIGINAL_NEQUAL)))),
		 aux);
  SETQ(C_IDS(basisTable),aux);
  C_NLOCALS(basisTable) = NORIGINALS;
#define X(i) KINT(i),vp->expr[i]
  aux = MapMerge(MapMerge(MapMerge(MapMerge(MapPair(X(ORIGINAL_BIND_EXC),
						    KINT(ORIGINAL_IO_EXC),ioException),
					    MapPair(X(ORIGINAL_NONSENSE),
						    X(ORIGINAL_MATCH_EXC))),
				   MapMerge(MapPair(X(ORIGINAL_EMPTY_EXC),
						    X(ORIGINAL_MAPGET_EXC)),
					    MapPair(X(ORIGINAL_MAP_EXC),
						    X(ORIGINAL_NOMEM_EXC)))),
			  MapMerge(MapMerge(MapPair(X(ORIGINAL_STACK_EXC),
						    X(ORIGINAL_INCOMP_EXC)),
					    MapPair(X(ORIGINAL_NONINV_EXC),
						    X(ORIGINAL_NTH_EXC))),
				   MapMerge(MapPair(X(ORIGINAL_RANGE_EXC),
						    X(ORIGINAL_ASCII_EXC)),
					    MapPair(X(ORIGINAL_STRNTH_EXC),
						    X(ORIGINAL_SUBSCRIPT_EXC))))),
		 MapMerge(MapPut(KINT(ORIGINAL_RE_EXC),reException,
				 MapPair(KINT(ORIGINAL_FALSE),falseValue,
					 KINT(ORIGINAL_TRUE),trueValue)),
			  MapMerge(MapPair(X(ORIGINAL_ARITH_EXC),
					   X(ORIGINAL_UFIND_EXC)),
				   MapPair(X(ORIGINAL_TRUE),
					   X(ORIGINAL_FALSE)))));
  SETQ(C_CONS(basisTable),aux);
  aux = MapSingleton(boolSym,CAR(trueValue));
  SETQ(C_TYTAGS(basisTable),aux); /* install tytagmap for Booleans. */
#undef X

#ifdef OLDDATA
  nilTags = Singleton(KINT(E_NIL));
  stringTags = Singleton(KINT(E_STRING));
  intTags = Singleton(KINT(E_INT));
  consTags = Singleton(KINT(E_CONS));
  listTags = SetUnion(consTags,nilTags);
  refTags = Singleton(KINT(E_REF));
#ifdef OBSOLETE
  arrayTags = SetUnion(Pair(KINT(E_ARRAY_0),KINT(E_ARRAY_1)),
		       Pair(KINT(E_ARRAY_2),KINT(E_ARRAY_GNRL)));
#else
  arrayTags = Pair(KINT(E_ARRAY_2),KINT(E_ARRAY_GNRL));
#endif
  mapTags = SetAdd(KINT(E_EMPTY),
		   Pair(KINT(E_MAPLET),KINT(E_MAP)));
  boolTags = stringTags;
  exnTags = SetAdd(KINT(E_CONS),stringTags);
  promiseTags = Singleton(KINT(E_COMPILED_CLOSURE));
  contTags = Singleton(KINT(E_STACK));
  dynTags = Singleton(KINT(E_PACK));
  numTags = SetAdd(KINT(E_INT),
		   Pair(KINT(E_REAL),KINT(E_COMPLEX)));
  record0Tags = Singleton(KINT(E_RECORD_0));
  record1Tags = Singleton(KINT(E_RECORD_1));
  recordNTags = Singleton(KINT(E_RECORD_GNRL));
  record1NTags = SetUnion(record1Tags,recordNTags);
  recordTags = SetUnion(record0Tags,record1NTags);
  tuple0Tags = Singleton(KINT(E_TUPLE_0));
  tuple1Tags = Singleton(KINT(E_TUPLE_1));
  tuple2Tags = Singleton(KINT(E_TUPLE_2));
  tupleNTags = Singleton(KINT(E_TUPLE_GNRL));
  tuple2NTags = SetUnion(tuple2Tags,tupleNTags);
  tuple12NTags = SetUnion(tuple1Tags,tuple2NTags);
  tupleTags = SetUnion(tuple0Tags,tuple12NTags);
  funTags =
#ifdef OBSOLETE
    SetAdd(KINT(E_MEMO_CLOSURE),
	   Pair(KINT(E_COMPILED_CLOSURE),KINT(E_INTERPRETED_CLOSURE)));
#else
  Pair(KINT(E_COMPILED_CLOSURE),KINT(E_INTERPRETED_CLOSURE));
#endif
#endif
  QUIT;
}

#ifdef OLDDATA
PUBLIC void addConstructor(tyname,id,con,innerkinds)
IN CONST ExprPtr tyname;
IN CONST ExprPtr id;
IN CONST ExprPtr con;
IN CONST MapPtr innerkinds;
{
  ExprPtr i_l,e;
  COUNTER i;
  VecPtr vp;
  ExprPtr aux;
  MapPtr tytags;
  Boolean found;
  ENTER;

  i = C_NLOCALS(basisTable)++;
  i_l = KINT(i);
  setMapPut(id,i_l,&C_IDS(basisTable));
  setMapPut(i_l,con,&C_CONS(basisTable));
  if (innerkinds!=NewMap())
    {
      e = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(datacon_shared),con);
      (void) KEEP(e);
    }
  else
    e = con;
  vp = EVARS(basisEnv);
  vp->expr[i] = e;
  vp->n = i+1;
  tytags = MapGet(tyname,C_TYTAGS(basisTable),&found);
  e = (innerkinds==NewMap())?stringTags:consTags;
  MAKECONS(innerkinds,e,aux);
  (void) KEEP(aux);
  if (!found)
    tytags = MapSingleton(con,aux);
  else
    tytags = MapPut(con,aux,tytags);
  setMapPut(tyname,tytags,&C_TYTAGS(basisTable));
  QUIT;
}
#else
#ifdef HAS_STDARG
PUBLIC ExprPtr addConstructors(IN CONST ExprPtr tyname,
			       IN COUNTER minmagic,
			       IN COUNTER maxmagic,...)
{
  va_list args;
#define va_begin(a,m) va_start(a,m)
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC ExprPtr addConstructors(tyname, minmagic, maxmagic, args)
IN CONST ExprPtr tyname;
IN COUNTER minmagic;
IN COUNTER maxmagic;
va_dcl
{
#define va_begin(a,m) va_start(a)
#else
PUBLIC ExprPtr addConstructors(tyname, minmagic, maxmagic, args)
IN CONST ExprPtr tyname;
IN COUNTER minmagic;
IN COUNTER maxmagic;
IN CONST CHAR *args;
{
#define va_begin(a,m)
#define va_arg(a,b) (*((b *) ((a +=  ((sizeof(b)+sizeof(int)-1) & ~(sizeof(int)-1))   ) -  ((sizeof(b)+sizeof(int)-1) & ~(sizeof(int)-1)) ))
)
#define va_end(a)  
#endif
#endif
  ExprPtr i_l,e,id,ttm,data;
  COUNTER i,n,magic;
  VecPtr vp;
  ExprPtr *epp;
  ExprPtr *shape;
  ENTER;

  va_begin(args,maxmagic);
  n = maxmagic-minmagic+1;
  shape = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
  for (magic=minmagic, epp=shape; magic<=maxmagic; magic++)
    {
      id = va_arg(args,ExprPtr);
      *epp++ = id;
    }
  va_end(args);
  ttm = Karray(E_ARRAY_GNRL,NIL,n,shape);
  (void) KEEP(ttm);
  ttm = Kdata(minmagic,ttm,NIL);
  (void) KEEP(ttm);
  for (magic=minmagic; magic<=maxmagic; magic++)
    {
      id = shape[magic-minmagic];
      data = Kdata(magic,ttm,NIL);
      (void) KEEP(data);
      if (magic>=0)
	{
	  e = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(datacon_shared),data);
	  (void) KEEP(e);
	}
      else e = data;
      i = C_NLOCALS(basisTable)++;
      i_l = KINT(i);
      setMapPut(id,i_l,&C_IDS(basisTable));
      if (magic>=0)
	{
	  data = MakeCounted (E_DATACON, data, (INT)0L); /* allow one to inline constructor */
	}
      setMapPut(i_l,data,&C_CONS(basisTable));
      vp = EVARS(basisEnv);
      vp->expr[i] = e;
#ifdef HIMMLX
      if (magic>=0)
	funNames[i] = (CHAR *)"datacon_sharedAction";
#endif
      vp->n = i+1;
    }
  QUIT;
  return ttm;
}
#endif

PUBLIC void addVal(name,val,xcon)
IN CONST ExprPtr name;
IN CONST ExprPtr val;
IN CONST ExprPtr xcon;
{
  ExprPtr i_l;
  COUNTER i;
  VecPtr vp;
  ENTER;

  i = C_NLOCALS(basisTable)++;
  i_l = KINT(i);
  setMapPut(name,i_l,&C_IDS(basisTable));
  if (xcon!=NOEXPR)
    setMapPut(i_l,xcon,&C_CONS(basisTable));
  vp = EVARS(basisEnv);
  vp->expr[i] = val;
  vp->n = i+1;
  QUIT;
}
