/* Debugger initialization.
   Copyright (C) 1998-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "type.h"
#include "bc.h"
#include "bcaux.h"
#include "dbgproto.h"
#include "typproto.h"
#include "cmpproto.h"
#include "grmproto.h"
#include "exeproto.h"
#ifdef HAS_PROTOTYPES
#include "sdbginit.h"
#endif

#define __SEG__	DEBUGINIT

#if DEBUGGING!=0
#ifdef OBSOLETE
PUBLIC ExprPtr fntrue;
EXTERN EXTENDED zero;
EXTERN ExprPtr alpha;
#endif
PUBLIC ExprPtr benignExcRegexpRef;

DECL_ACTION(apply_bytecode_closure)

EXTERN ExprPtr stringType;

PUBLIC void InitDebug()
{
  ENTER;
#define BKP_SIZE 128
#define BKP_SIZEBREAKS 16

  breakPoints.n = 0;
  breakPoints.size = BKP_SIZE;
  breakPoints.pcs = (OPCODE **)MALLOC(BKP_SIZE*sizeof(OPCODE *));
  breakPoints.savedops = (OPCODE *)MALLOC(BKP_SIZE*sizeof(OPCODE));
  breakPoints.bcs = (ExprPtr *)MALLOC(BKP_SIZE*sizeof(ExprPtr));
  breakPoints.nbreaks = 0;
  breakPoints.sizebreaks = BKP_SIZEBREAKS;
  breakPoints.breaks = (COUNTER *)MALLOC(BKP_SIZEBREAKS*sizeof(COUNTER));
  monitorExprs = NIL;
  lastPrintExpr = NIL;
  lastPrintVals = NewMap();
  lastPrintTypes = NewMap();
  dbgFile = NOEXPR;
  dbgName = NOEXPR;
  {
    ExprPtr s;

    s = Kstr ((CHAR *)message (739));
    benignExcRegexpRef = addTypeRef(737, 736, stringType, s);
    QUIT;
  }
  {
    ExprPtr sym, type;
    ExprPtr code, e;
    bcFile *f;
    TZone zone;
    INT i, pc;

    sym = Kstr ((CHAR *)message (741));
    type = funType (NIL, NIL);
    addIdType(sym,type);
    f = bcOpen ();
    e = MakePair (E_CONS, NIL, NIL);
    e = MakePair (E_CONS, sym, e);
    zone.line1 = zone.line2 = zone.pos1 = zone.pos2 = 0;
    e = ZBuild(&zone, EA_CLR_PROFILE, e);
    bccPut (e, f);
    i = getConstant (f, sym);
    bcdPut(DBG_NAME,f);
    bcdPut((OPCODE)i,f);
    pc = bcTell(f);
    bcdPut(DBG_WHERE,f);
    bcdPut((OPCODE)pc,f);
    bcdPut((OPCODE)2, f); /* stackDepth */
    bcdPut((OPCODE)0, f); /* mapStackDepth */
    bcdPut((OPCODE)0, f); /* excStackDepth (obsolete) */
    bcPut (BC_BREAK, f);
    bcPut (BC_RETURN, f);
    bcPut (1, f);
    e = bcClose (f);
    code = MakeCompiled(E_INTERPRETED_CLOSURE,ACTION_NAME(apply_bytecode_closure),e);
    addVal(sym,code,NOEXPR);
  }
#ifdef OBSOLETE
  currentDebugZone = NIL;
  stepFile = NIL;
  initApplStk();
  {
    ExprPtr aux;

    MAKECONS(patany,trueValue,aux);
    (void) KEEP(aux);
    fntrue = MakePolyfun(-1,aux,(INT)0);
  }
  ADDFUN(379,funType(alpha,alpha),break);
#endif
  QUIT;

}
#endif

#if PROFILING
PUBLIC ExprPtr ncallsSym,properSym,totalSym,ngcSym,gctimeSym,timeSym,locSym;
PUBLIC ExprPtr defaultProf;
EXTERN ExprPtr intType,stringType,timeType;
EXTERN extDesc profExtDesc;

PUBLIC void InitProfile()
{
  ExprPtr type,t2,e,sym;
  TZone zone;
  ENTER;

  registerExtDesc(&profExtDesc);
  e = MakePair (E_CONS, NIL, NIL);
  sym = Ksymbol(message(931));
  e = MakePair (E_CONS, sym, e);
  zone.line1 = zone.line2 = zone.pos1 = zone.pos2 = 0;
  defaultProf = ZBuild(&zone, EA_CLR_PROFILE, e);
  ncallsSym = Ksymbol(message(596));
  properSym = Ksymbol(message(597));
  totalSym = Ksymbol(message(598));
  ngcSym = Ksymbol(message(599));
  gctimeSym = Ksymbol(message(600));
  timeSym = Ksymbol(message(601));
  locSym = Ksymbol(message(605));
  type = funType(NIL,NIL);
  ADDFUN(602,type,reset_profiles);
  ADDFUN(603,type,clear_profiles);
  QUIT;
  type = coupleType(timeType,timeType);
  type = MapPut(ngcSym,intType,
		MapPair(gctimeSym,type,
			timeSym,type));
  t2 = coupleType(intType,intType);
  t2 = Kons(stringType,Kons(stringType,Kons(t2,Kons(t2,NIL))));
  type = MapMerge(MapPair(locSym,t2,
			  ncallsSym,intType),
		  MapPair(properSym,type,
			  totalSym,type));
  type = funType(NIL,setType(type));
  ADDFUN(604,type,report_profiles);
}
#endif
