/* Definitions for basic structure allocation functions (unshared).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <setjmp.h>

#define BUCKET_SIZE 512

typedef struct Bucket  {
  struct Bucket *next;
  SHORT firstNewCell; /* index of the first new cell in the newcell array below */
  SHORT occupancy[MAX_GENERATIONS];
  /* occupancy[i] is the number of allocated cells in the bucket
     that are in the ith generation. */
#if BUCKET_GAP!=0
  char filler[BUCKET_GAP];
#endif
  Expression cell[BUCKET_SIZE];
  Expression *newcells[BUCKET_SIZE]; /* used by GCsweep() */
  /* this is the list of pointers to entries in the cell array,
     ordered by increasing generations; the first entry corresponding
     to the current generation is firstNewCell */
} Bucket,*BucketPtr;

EXTERN BucketPtr theBucketList;
EXTERN BucketPtr theHomoBucketList,*endHomoBucketList;
EXTERN BucketPtr theHeteroBucketList,*endHeteroBucketList;
EXTERN ExprPtr *nextHomoExpr,*nextHeteroExpr;
EXTERN ExprPtr *lastHomoExpr,*lastHeteroExpr;
EXTERN BucketPtr currentHomoBucket,currentHeteroBucket;
EXTERN BucketPtr oldHomoBucketList,oldHeteroBucketList;
IMPORT void bumpHomoExpr P((void));
IMPORT void bumpHeteroExpr P((void));

EXTERN BucketPtr theSlotBucketList;
EXTERN ExprPtr theAvailSlotList;
EXTERN ExprPtr reserveSlot;

typedef struct VecBucket {
  struct VecBucket *next;
  struct VecBucket *contents[1]; /* actually variable-sized */
} VecBucket,*VecBucketPtr;

#define VECBN_N 13

EXTERN COUNTER VecBucketN[VECBN_N];
EXTERN VecBucketPtr theVecBucketList[VECBN_N];
EXTERN VecBucketPtr theAvailVecBuckets[VECBN_N];

EXTERN INT nCells;

typedef CONST char *(*allocFunction) P((long size));

EXTERN BucketPtr NewBucket P((allocFunction new,Boolean force)) _MALLOC;
EXTERN BucketPtr NewSlotBucket P((allocFunction new)) _MALLOC;
EXTERN VecBucketPtr NewVecBucket P((allocFunction new,IN INT n)) _MALLOC;

EXTERN VOIDPTR VecAlloc P((IN INT size)) _MALLOC;
EXTERN void VecFree P((IN VOIDPTR p,IN INT size));

EXTERN char dummy;

#define VECALLOC(p,size) \
{ \
  REGISTER INT _V_size=size, \
    _V_n = (_V_size + (sizeof(VecBucketPtr)-1))/sizeof(VecBucketPtr); \
  if (_V_n==0) p = (VOIDPTR)&dummy; \
  else if (_V_n>=VECBN_N) p = (VOIDPTR)MALLOC(_V_size); \
  else { \
    REGISTER VecBucketPtr *_V_vp = theAvailVecBuckets+_V_n; \
    if (*_V_vp==(VecBucketPtr)0L) (void) NewVecBucket(newptr_gc,_V_n); \
    p = (VOIDPTR)*_V_vp; \
    *_V_vp = *(VecBucketPtr *)*_V_vp; \
  } \
}

#define VECFREE(p,size) \
{ \
  REGISTER INT _V_n = (size + (sizeof(VecBucketPtr)-1))/sizeof(VecBucketPtr); \
  if (_V_n!=0) { \
    if (_V_n>=VECBN_N) { MARKFREE(p); } \
    else { \
      REGISTER VecBucketPtr *_V_vp = theAvailVecBuckets+_V_n; \
      *(VecBucketPtr *)p = *_V_vp; \
      *_V_vp = (VecBucketPtr)p; \
    } \
  } \
}

#define CHECKHOMOFREE() \
  if (nextHomoExpr!=lastHomoExpr) ; else \
       bumpHomoExpr()

#define CHECKHETEROFREE() \
  if (nextHeteroExpr!=lastHeteroExpr) ; else \
       bumpHeteroExpr()

#define EALLOC_HOMO_NOGC(e) \
     e = *nextHomoExpr++; \
     (e)->mark = 0; \
     ELINK(e) = NEWGEN

#define EALLOC_HOMO(e) \
     CHECKHOMOFREE(); \
     EALLOC_HOMO_NOGC(e)

#define EALLOC_HETERO_NOGC(e) \
     e = *nextHeteroExpr++; \
     (e)->mark = 0; \
     ELINK(e) = NEWGEN

#define EALLOC_HETERO(e) \
     CHECKHETEROFREE(); \
     EALLOC_HETERO_NOGC(e)

#define QALLOC_HOMO(e) \
     e = *nextHomoExpr++

#define QALLOC_HETERO(e) \
     e = *nextHeteroExpr++

#define EFREE(e)

#define SLOTALLOC(e) \
	e = theAvailSlotList; \
	theAvailSlotList = NEXTFREE(e)

#define MAKEINT(ii,e) \
{   \
	REGISTER ExprPtr tmpxIz;	\
	\
	EALLOC_HOMO(tmpxIz); \
	tmpxIz->kind = (CHAR)E_INT;	\
	IVAL(tmpxIz) = ii; \
	INEXT(tmpxIz) = NOEXPR;  \
	e = tmpxIz; \
}

#define MAKEREAL(xx,e)  \
	EALLOC_HOMO(e);  \
	(e)->kind = (CHAR)E_REAL;   \
	XVAL(e) = xx

#define MAKESTRING(l,t,e)   \
	EALLOC_HETERO(e);  \
	(e)->kind = (CHAR)E_STRING; \
	SLEN(e) = l;  \
	STEXT(e) = t

#define MAKEPAIR1(xkind,xcar,xcdr,e)   \
	EALLOC_HOMO(e);  \
	(e)->kind = (CHAR)xkind;   \
	CAR(e) = xcar;   \
	CDR(e) = xcdr

#define MAKEPAIR(xkind,xcar,xcdr,e)   \
	EALLOC_HOMO(e);  \
	(e)->kind = (CHAR)xkind;   \
	CAR(e) = xcar;   \
	CDR(e) = xcdr

#define MAKEDATA(xmagic,xcar,xcdr,e) \
	EALLOC_HOMO(e); \
	(e)->kind = (CHAR)E_DATA; \
	(e)->magic = xmagic; \
	CAR(e) = xcar; \
	CDR(e) = xcdr

#define MAKECONS1(xcar,xcdr,e) MAKEPAIR1(E_CONS,xcar,xcdr,e)
#define MAKECONS(xcar,xcdr,e) MAKEPAIR(E_CONS,xcar,xcdr,e)
#define MAKEUNSHAREDCONS(xcar,xcdr,e) MAKEPAIR(E_UNSHARED_CONS,xcar,xcdr,e)
#define	MAKETUPLE2(xcar,xcdr,e) MAKEPAIR(E_TUPLE_2,xcar,xcdr,e)
#define	MAKEUNSHAREDTUPLE2(xcar,xcdr,e) MAKEPAIR(E_UNSHARED_TUPLE_2,xcar,xcdr,e)
#define	MAKERECORD1(label,val,e) MAKEPAIR(E_RECORD_1,label,val,e)
#define	MAKEMAPLET(x,y,e) MAKEPAIR(E_MAPLET,x,y,e)
#define MAKEMEMO(rem,f,e) MAKEPAIR(E_MEMO_FN,rem,f,e)
#define MAKEPROG(k,e1,e2,e) MAKEPAIR(k,e1,e2,e)

#define	MAKEKTUPLE(xkind,xcdr,e) \
	EALLOC_HETERO(e); \
	(e)->kind = (CHAR)xkind; \
	ASHAPE(e) = NIL; \
	AVAL(e) = xcdr
#define	MAKETUPLE(xcdr,e) MAKEKTUPLE(E_TUPLE_GNRL,xcdr,e)

#define	MAKEKTUPLE_NOGC(xkind,xcdr,e) \
	EALLOC_HETERO_NOGC(e); \
	(e)->kind = (CHAR)xkind; \
	ASHAPE(e) = NIL; \
	AVAL(e) = xcdr
#define	MAKETUPLE_NOGC(xcdr,e) MAKEKTUPLE_NOGC(E_TUPLE_GNRL,xcdr,e)

#define	MAKEKRECORD(xkind,xcar,xcdr,e) \
	EALLOC_HETERO(e); \
	(e)->kind = (CHAR)xkind; \
	ASHAPE(e) = xcar; \
	AVAL(e) = xcdr
#define	MAKERECORD(xcar,xcdr,e) MAKEKRECORD(E_RECORD_GNRL,xcar,xcdr,e)

#define	MAKEKRECORD_NOGC(xkind,xcar,xcdr,e)	\
	EALLOC_HETERO_NOGC(e);	\
	(e)->kind = (CHAR)xkind;	\
	ASHAPE(e) = xcar;   \
	AVAL(e) = xcdr
#define	MAKERECORD_NOGC(xcar,xcdr,e) MAKEKRECORD_NOGC(E_RECORD_GNRL,xcar,xcdr,e)

#define MAKESING(k,expr,e) \
	EALLOC_HOMO(e);  \
	(e)->kind = (CHAR)k;	\
	REFVAL(e) = expr

#define	MAKEREF(expr,e)	MAKESING(E_REF,expr,e)

#define MAKEENV1(v,n,e)  \
	EALLOC_HETERO(e); \
	(e)->kind = (CHAR)E_ENV; \
	EVARS(e) = v; \
	ENEXT(e) = n

#define MAKEENV1_NOGC(v,n,e) \
	EALLOC_HETERO_NOGC(e); \
	(e)->kind = (CHAR)E_ENV; \
	EVARS(e) = v; \
	ENEXT(e) = n

#define MAKEENV(v,n,e)  \
	MAKEENV1(v,n,e)

#define MAKEENV_NOGC(v,n,e)  \
	MAKEENV1_NOGC(v,n,e)

IMPORT ExprPtr MakePair P((IN ExprKind kind,IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _MALLOC;
#define	MakeLet(bind,body) MakePair(E_LET,bind,body)
#define	MakeGoto(kind,f,arg) MakePair(kind,f,arg)
#define	MakeAppl(f,arg) MakeGoto(E_APPL,f,arg)
#define	MakePatAs(pat1,pat2) MakePair(E_PAT_AS,pat1,pat2)
#define	MakePatMapAdd(xy,rest) MakePair(E_PAT_MAPADD,xy,rest)
#define MakeSet(ref,val) MakePair(E_SET,ref,val)
#define	MakeTest(cond,thenelse) MakePair(E_IF,cond,thenelse)
#define	MakeOfOrMap(kind,fP,dom) MakePair(kind,fP,dom)
#define	MakeWhile(kind,cond,body) MakePair(kind,cond,body)
#define	MakeHandle(body,handlers) MakePair(E_HANDLE,body,handlers)
#define	MakePack(kind,expr,type) MakePair(kind,expr,type)
#define MakeDebugZone(code,info) MakePair(E_DEBUGZONE,code,info)
#define MakeDebugName(code,info) MakePair(E_DEBUGNAME,code,info)
#define MakeMapget(m,s) MakePair(E_MAPGET,m,s)
#define makeMu(kind,rec,mod) MakePair(kind,rec,mod)

#define MakePatTupleCons(pat,tail) MakePair(E_PAT_TUPLE_GNRL_CONS,pat,tail)
#define MakePatRecordCons(pat,tail) MakePair(E_PAT_RECORD_GNRL_CONS,pat,tail)
#define MakePatRecordMapAdd(pat,tail) MakePair(E_PAT_RECORD_MAPADD,pat,tail)
#define MakePatCons(pat) MakeSing(E_PAT_CONS,pat)

IMPORT ExprPtr MakeData P((IN SHORT magic,IN CONST ExprPtr car,IN CONST ExprPtr cdr)) _MALLOC;

IMPORT ExprPtr MakeCounted P((IN ExprKind kind,IN CONST ExprPtr expr,IN INT n)) _MALLOC;
IMPORT ExprPtr MakePolyfun P((IN SHORT magic,IN CONST ExprPtr expr,IN INT n)) _MALLOC;
#define MakeSing(kind,expr) MakeCounted(kind,expr,0L)
#define MakeQuote(expr) MakeSing(E_QUOTE,expr)

#define	MakeRaise(exc) MakeSing(E_RAISE,exc)
#define EXC_BENIGN 0x1
#define EXC_CHECKED 0x0
#define	MakeException(exc,flags) MakeCounted(E_EXCEPTION,exc,flags)
#define MakeGet(ref) MakeSing(E_GET,ref)
#define	MakeDelay(code) MakeSing(E_DELAY,code)
#define MakeTuple1(e) MakeSing(E_TUPLE_1,e)

IMPORT ExprPtr MakePromise P((IN CONST ExprPtr fun)) _MALLOC;
IMPORT ExprPtr MakeStack P((IN CONST stackDesc *jb,IN CONST StackChunk *chunk)) _MALLOC;
IMPORT ExprPtr MakeRecord P((IN ExprKind kind,IN CONST ExprPtr shape,IN INT n,IN ExprPtr *exprs)) _MALLOC;
IMPORT ExprPtr MakeArray P((COUNTER n,ExprPtr e)) _MALLOC;
IMPORT ExprPtr ListToArray0 P((ExprPtr l)) _MALLOC;
IMPORT ExprPtr MakeExternal P((IN CONST extDescPtr desc,IN CONST char *data)) _MALLOC;
IMPORT ExprPtr MakeCompiled P((IN ExprKind kind,
			       IN compiledAction action,
			       IN CONST ExprPtr env)) _MALLOC;
#define makePrimitive(name) MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(name),NIL)

typedef struct field {
  ExprPtr label;
  ExprPtr val;
} field;

IMPORT int CompareStrings P((IN CONST ExprPtr s1,IN CONST ExprPtr s2));
IMPORT void sortFieldArray P((IN COUNTER n,INOUT field *array));
IMPORT ExprPtr fieldsToShape1 P((IN COUNTER n,IN COUNTER n0,IN CONST field *array));
#define fieldsToShape(n,array) fieldsToShape1(n,1,array)
IMPORT void fieldsToExprs P((OUT ExprPtr *exprs,IN CONST field *array,IN COUNTER n));

IMPORT ExprPtr MakeRecordGnrl P((IN COUNTER n,IN CONST field *array,
				 IN CONST ExprPtr shape)) _MALLOC;


IMPORT extDesc *getExtDesc P((IN CONST CHAR *id));
IMPORT void registerExtDesc P((INOUT extDesc *desc));

IMPORT INT idiv P((INT a,INT b));
IMPORT INT imod P((INT a,INT b));
