/* Initialization of Arjen Lenstra's large integer package.
   Copyright (C) 2000-2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "expr.h"
#include "exeproto.h"
#include "ex2proto.h"
#include "typproto.h"
#include "cmpproto.h"
#include "grmproto.h"
#include "type.h"
#include "lipact.h"

IMPORT void InitLIP P((void));

#define __SEG__ LIPINIT

EXTERN extDesc lipExtDesc;
EXTERN extDesc spgExtDesc;

EXTERN ExprPtr lipErrExc;

EXTERN ExprPtr intType, boolType, stringType, exnType, numType;

IMPORT void zstart P((void));
IMPORT void zpInit P((void));

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return KEEP(sym);
}

/* there is no garbage collection during initialization,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KONS
#define KONS Kons

PUBLIC void InitLIP ()
{
  ExprPtr type, aux;
  SymPtr lipErrSym, Intsym, IntSym;
  ExprPtr IntType, listIntType, listintType;
  ENTER;

  zstart();
  zpInit();
  registerExtDesc(&lipExtDesc);
  registerExtDesc(&spgExtDesc);
  lipErrSym = KSYMBOL(765);
  setConstructor (lipErrSym);
  lipErrExc = MakeException(lipErrSym,EXC_BENIGN);
  type = funType (intType, exnType);
  addIdType (lipErrSym, type);
  aux = MakeCompiled (E_COMPILED_CLOSURE,
		      ACTION_NAME (make_io_exc),
		      lipErrExc);
  addVal (lipErrSym, aux, lipErrExc);
  QUIT;
  Intsym = KSYMBOL(766);
  IntSym = KSYMBOL(767);
  IntType = KtypeAppl (IntSym, NIL);
  aux = KONS(NIL,NewMap());
  setMapPut (Intsym, KONS (aux, IntType), &A_TYFCNS(basisAssum));
  QUIT;
  listintType = listType (intType);
  listIntType = listType (IntType);
  type = funType (KONS(intType, coupleType(intType, intType)), intType);
  ADDFUN(768, type, zmulmods);
  ADDFUN(862, type, zexpmods);
  QUIT;
  type = funType (IntType, numType);
  ADDFUN(769, type, zdoub);
  ADDFUN(881, type, zln);
  QUIT;
  type = funType (intType, IntType);
  ADDFUN(770, type, zintoz);
  ADDFUN(903, type, zrandoml1);
  QUIT;
  type = funType (IntType, intType);
  ADDFUN(771, type, ztoint);
  ADDFUN(877, type, zsign);
  ADDFUN(880, type, z2log);
  ADDFUN(883, type, zweight);
  QUIT;
  ADDFUN(772, funType (coupleType (intType, listintType), IntType), zsbastoz);
  QUIT;
  ADDFUN(773, funType (coupleType (IntType, listIntType), IntType), zbastoz);
  QUIT;
  type = funType (coupleType (intType, IntType), listintType);
  ADDFUN(774, type, zstobas);
  ADDFUN(775, type, zstosymbas);
  QUIT;
  type = funType (coupleType (IntType, IntType), listIntType);
  ADDFUN(776, type, ztobas);
  ADDFUN(777, type, ztosymbas);
  QUIT;
  type = funType (coupleType (IntType, IntType), intType);
  ADDFUN(778, type, zcompare);
  ADDFUN(897, type, zjacobi);
  QUIT;
  type = funType (IntType, IntType);
  ADDFUN(779, type, znegate);
  ADDFUN(805, type, zsqr);
  ADDFUN(843, type, ztom);
  ADDFUN(844, type, zmtoz);
  ADDFUN(849, type, zmontsqr);
  ADDFUN(851, type, zmontinv);
  ADDFUN(865, type, z2mul);
  ADDFUN(866, type, z2div);
  ADDFUN(878, type, zabs);
  ADDFUN(884, type, znot);
  ADDFUN(904, type, zrand1);
  QUIT;
  type = funType (coupleType (IntType, intType), IntType);
  ADDFUN(794, type, zsadd);
  ADDFUN(798, type, zsmul);
  ADDFUN(847, type, zsmontmul);
  ADDFUN(860, type, zsexp);
  ADDFUN(868, type, zlshift);
  ADDFUN(869, type, zrshift);
  ADDFUN(873, type, zroot);
  ADDFUN(891, type, zlowbits);
  ADDFUN(890, type, zhighbits);
  ADDFUN(893, type, zsetbit);
  ADDFUN(894, type, zclrbit);
  ADDFUN(896, type, zreverse);
  QUIT;
  type = funType (coupleType (IntType, IntType), IntType);
  ADDFUN(796, type, zadd);
  ADDFUN(797, type, zsub);
  ADDFUN(800, type, zmul);
  ADDFUN(845, type, zmontadd);
  ADDFUN(846, type, zmontsub);
  ADDFUN(848, type, zmontmul);
  ADDFUN(834, type, zmod);
  ADDFUN(839, type, zsqmod);
  ADDFUN(841, type, zinvmod);
  ADDFUN(850, type, zmontdiv);
  ADDFUN(854, type, zmontexp);
  ADDFUN(857, type, z2expmod);
  ADDFUN(859, type, zmontexp_m_ary);
  ADDFUN(861, type, zexp);
  ADDFUN(874, type, zgcd);
  ADDFUN(875, type, zgcdeucl);
  ADDFUN(885, type, zand);
  ADDFUN(888, type, zxor);
  ADDFUN(889, type, zor);
  QUIT;
  type = coupleType (IntType, intType);
  type = funType (type, type);
  ADDFUN(810, type, zsdiv);
  QUIT;
  type = funType (coupleType (IntType, intType), intType);
  ADDFUN(818, type, zsmod);
  QUIT;
  type = coupleType (IntType, IntType);
  type = funType (type, type);
  ADDFUN(823, type, zdiv);
  QUIT;
  type = funType (KONS (IntType, coupleType (IntType, IntType)), IntType);
  ADDFUN(835, type, zaddmod);
  ADDFUN(836, type, zsubmod);
  ADDFUN(838, type, zmulmod);
  ADDFUN(840, type, zdivmod);
  ADDFUN(856, type, zexpmod);
  ADDFUN(858, type, zexpmod_m_ary);
  QUIT;
  type = funType (KONS (IntType, coupleType (intType, IntType)), IntType);
  ADDFUN(837, type, zsmulmod);
  ADDFUN(855, type, zsexpmod);
  QUIT;
  type = funType (IntType, NIL);
  ADDFUN(842, type, zmstart);
  QUIT;
  type = coupleType (IntType, IntType);
  type = funType (type, coupleType (type, IntType));
  ADDFUN(852, type, zexteucl);
  QUIT;
  type = funType (KONS (IntType, KONS (IntType, coupleType (IntType, IntType))), IntType);
  ADDFUN(853, type, zchirem);
  ADDFUN(864, type, zmontexp_doub);
  QUIT;
  type = funType (KONS (IntType, KONS (IntType,
				       KONS (IntType, coupleType (IntType, IntType)))),
		  IntType);
  ADDFUN(863, type, zexpmod_doub);
  QUIT;
  type = funType (IntType, boolType);
  ADDFUN(867, type, z2mod);
  QUIT;
  type = funType (IntType, coupleType (intType, IntType));
  ADDFUN(870, type, z_makeodd);
  QUIT;
  type = funType (intType, intType);
  ADDFUN(871, type, zsqrts);
  /*ADDFUN(764, type, prime);*/
  ADDFUN(879, type, z2logs);
  ADDFUN(882, type, zweights);
  ADDFUN(900, type, zsrand);
  QUIT;
  type = funType (IntType, coupleType (IntType, IntType));
  ADDFUN(872, type, zsqrt);
  QUIT;
  type = funType (intType, funType (NIL, intType));
  ADDFUN(763, type, nextprime);
  QUIT;
  type = funType (coupleType (IntType, intType),
		  KtypeAppl (optSym, Kons (coupleType (IntType, IntType), NIL)));
  ADDFUN(764, type, zpollardrho);
  QUIT;
  type = funType (KONS (IntType, coupleType (intType, intType)),
		  KtypeAppl (optSym, Kons (coupleType (intType, IntType), NIL)));
  ADDFUN(876, type, ztridiv);
  QUIT;
  type = funType (IntType, setType(intType));
  ADDFUN(886, type, IntToSet);
  QUIT;
  type = funType (mapType(intType, alpha), IntType);
  ADDFUN(887, type, SetToInt);
  QUIT;
  type = funType (coupleType (IntType, intType), boolType);
  ADDFUN(892, type, zbit);
  ADDFUN(899, type, zprobprime);
  QUIT;
  type = funType (coupleType (intType, intType), intType);
  ADDFUN(895, type, zreverses);
  ADDFUN(898, type, zjacobis);
  QUIT;
  type = funType (coupleType (intType, funType (intType, intType)), IntType);
  ADDFUN(901, type, zrandoml);
  QUIT;
  type = funType (coupleType (IntType, funType (intType, intType)), IntType);
  ADDFUN(902, type, z_rand);
  QUIT;
  type = funType (Kons (intType, coupleType (intType, funType (intType, intType))),
		  IntType);
  ADDFUN(905, type, zrandomprime);
  QUIT;
  type = funType (coupleType (intType, intType), IntType);
  ADDFUN(906, type, zrandomprime1);
  QUIT;
  type = funType (Kons (intType, Kons (intType, coupleType (IntType, funType (intType, intType)))),
		  coupleType (IntType, IntType));
  ADDFUN(907, type, zrandompprime);
  ADDFUN(909, type, zrandomfprime);
  QUIT;
  type = funType (Kons (intType, coupleType (intType, IntType)),
		  coupleType (IntType, IntType));
  ADDFUN(908, type, zrandompprime1);
  ADDFUN(910, type, zrandomfprime1);
  QUIT;
  type = funType (Kons (intType, Kons (intType, coupleType (boolType, funType (intType, intType)))),
		  Kons (IntType, coupleType (IntType, IntType)));
  ADDFUN(911, type, zrandomgprime);
  QUIT;
  type = funType (Kons (intType, coupleType (intType, boolType)),
		  Kons (IntType, coupleType (IntType, IntType)));
  ADDFUN(912, type, zrandomgprime1);
  QUIT;
  type = funType (Kons (IntType, Kons (intType, coupleType (intType, intType))),
		  KtypeAppl (optSym, Kons (coupleType (IntType, boolType), NIL)));
  ADDFUN(913, type, zecm);
  QUIT;
}
