/* Making config.h from OPTIONS file.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>

#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#define P(l) l
#else
#define P(l) ()
extern int strcmp();
extern void strcpy();
extern int strlen();
#endif

#define MAXC 1024

int himml_lightp=0;
int stop_and_copy=0;
int translator=0;

void finish P((void));
void defact P((char *var,char *def));
void himml_light P((char *var,char *def));
void stop_and_copy_record P((char *var,char *def));
void translator_record P((char *var,char *def));
void imperative_types P((char *var,char *def));
char *strcomma P((char *s));
void continuation_model P((char *var,char *def));
void debugging P((char *var,char *def));

void finish()
{
  if (himml_lightp)
    {
      printf("\n#if HIMML_LIGHT\n#undef STOP_AND_COPY_VECTORS\n#define STOP_AND_COPY_VECTORS 0\n#endif\n");
    }
  printf("\n#include \"perks.h\"\n");
}

void defact(var,def)
char *var,*def;
{
  printf("#define %s %s\n",var,def);
}

void himml_light(var,def)
char *var,*def;
{
  defact(var,def);
  himml_lightp = 1;
}

void stop_and_copy_record(var,def)
char *var,*def;
{
  (void) sscanf(def,"%d",&stop_and_copy);
}

void translator_record(var,def)
char *var,*def;
{
  (void) sscanf(def,"%d",&translator);
}

void imperative_types(var,def)
char *var,*def;
{
  if (strcmp(def,"\"imperative tyvars\"")==0)
    printf("#define IMPERATIVE_TYPES STD_ML_TYVAR\n");
  else if (strcmp(def,"\"weak tyvars\"")==0)
    printf("#define IMPERATIVE_TYPES STD_ML_NJ_TYVAR\n");
  else fprintf(stderr,"Value of IMPERATIVE_TYPES option not recognized (%s).\n",def);
  printf("#define IMPERATIVE_TYPES_STRING %s\n",def);
}

char *strcomma(s)
register char *s;
{
  register char c;
  register int str;

  str = 0;
  while (c=*s++)
    {
      if (c=='\\')
	{
	  if (*s++==0)
	    break;
	}
      else if (c=='\"')
	str = !str;
      else if (c==',' && !str)
	break;
    }
  s--;
  return s;
}

void continuation_model(var,def)
char *var,*def;
{
  int comma;
  char *s;
  char model[128],*m;
  char modelset[512],modelset2[512],*ms,*oms,*xms;

  m = model;
  strcpy(m,"0"); m += strlen(m);
  ms = modelset; oms = modelset2;
  strcpy(ms,"NewMap()");
  while (*def!=0)
    {
      s = strcomma(def);
      comma = (*s==',');
      *s = 0;
      if (strcmp(def,"\"callcc\"")==0)
	{
	  strcpy(m,"|CALLCC"); m+=strlen(m);
rec:
	  sprintf(oms,"SetAdd(KSTR(%s),%s)",def,ms);
	  xms = oms; oms = ms; ms = xms;
	}
      else if (strcmp(def,"\"catch\"")==0)
	{
	  strcpy(m,"|CATCH"); m+=strlen(m);
	  goto rec;
	}
      else fprintf(stderr,"Value of CONTINUATION_MODEL option not recognized (%s).\n",def);
      def = s;
      if (comma)
	{
	  char c;

	  while (c=*++def, c==' ' || c=='\t');
	}
    }
  printf("#define CONTINUATION_MODEL (%s)\n",model);
  printf("#define CONTINUATION_MODEL_SET %s\n",ms);
}

void debugging(var,def)
char *var,*def;
{
  int comma;
  char *s;
  char model[128],*m;
  char modelset[512],modelset2[512],*ms,*oms,*xms;

  m = model;
  strcpy(m,"0"); m += strlen(m);
  ms = modelset; oms = modelset2;
  strcpy(ms,"NewMap()");
  while (*def!=0)
    {
      s = strcomma(def);
      comma = (*s==',');
      *s = 0;
      if (strcmp(def,"\"standard\"")==0)
	{
	  strcpy(m,"|STD_DEBUGGING"); m+=strlen(m);
rec:
	  sprintf(oms,"SetAdd(KSTR(%s),%s)",def,ms);
	  xms = oms; oms = ms; ms = xms;
	}
      else fprintf(stderr,"Value of DEBUGGING option not recognized (%s).\n",def);
      def = s;
      if (comma)
	{
	  char c;

	  while (c=*++def, c==' ' || c=='\t');
	}
    }
  printf("#define DEBUGGING (%s)\n",model);
  printf("#define DEBUGGING_SET %s\n",ms);
}

struct entry {
  char *name;
  void (*action) P((char *var,char *def));
};

struct entry options[] =
{
  { "HIMML_LIGHT", himml_light },
  { "IMPERATIVE_TYPES", imperative_types },
  { "CONTINUATION_MODEL", continuation_model },
  { "DEBUGGING", debugging },
  { NULL, NULL }
};

main(argc,argv)
int argc;
char *argv[];
{
  int c,i,lineno;
  char line[MAXC];
  char *varbeg,*varend,*defbeg,*defend;
  struct entry *e;
  int skip;

  puts("/* Configuration file for HimML, automatically generated from OPTIONS by makeconfig */\n");
  puts("#include \"general.h\"\n");
  lineno = 1;
  skip = 1;
  while ((c=getchar())!=-1)
    {
      if (c=='#') /* comment at start of line */
	{
	  c = getchar();
	  if (c=='#')
	    skip = 0; /* reached the special ###### line */
	  if (c!=-1 && c!='\n')
	    while (c=getchar(), c!=-1 && c!='\n');
	  if (c==-1)
	    break;
	  lineno++;
	}
      else if (c=='\n')
	lineno++;
      else
	{
	  line[0] = c;
	  for (i=1; c=getchar(), c!=-1 && c!='\n' && i<MAXC-1;)
	    line[i++] = c;
	  line[i++] = 0;
	  if (i>=MAXC)
	    fprintf(stderr,"Line %d too long, truncated to %d characters.\n",
		    lineno,MAXC-1);
	  if (c==-1)
	    break;
	  lineno++;
	  if (skip)
	    continue;
	  for (varbeg=line; c=*varbeg++, c==' ' || c=='\t';);
	  varbeg--;
	  for (varend=varbeg; c=*varend++, c!=' ' && c!='\t' && c!='=' && c!=0;);
	  varend--;
	  for (defbeg=varend; c=*defbeg++, c==' ' || c=='\t';);
	  defbeg--;
	  if (c==0)
	    {
	      fprintf(stderr,"Missing '=' on line %d, skipping.\n",lineno);
	    }
	  else
	    {
	      for (defbeg++; c=*defbeg++, c==' ' || c=='\t';);
	      defbeg--;
	      defend = line+i-1; /* points to the '\0' at end of line */
	      while (defend>=defbeg && (c=*--defend, c==' ' || c=='\t'));
	      *varend = 0;
	      *++defend = 0;
	      for (e=options; e->name!=NULL; e++)
		{
		  if (strcmp(e->name,varbeg)==0)
		    {
		      (*e->action)(varbeg,defbeg);
		      break;
		    }
		}
	      if (e->name==NULL)
		defact(varbeg,defbeg);
	    }
	}
    }
  finish();
  exit(0);
}
