/* Imperative tables.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Although this all works, I've decided not to integrate it
   into HimML, since it consumes much more memory than just
   MapPtrs, and is not integrated with the generation scavenging
   code (so GC time goes up to about 25%, instead of 15%).
*/

#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "portable.h"
#include "eval.h"
#include "ex1proto.h"
#include "expr.h"
#include "table.h"

EXTERN ExprPtr none,optttm;

PRIVATE void tableSweep (extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  switch (doext->xtag)
    {
    case X_SAVE:
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)((ExprPtr)extdata);
      break;
    case X_FREE:
      break;
    }
}

PRIVATE ExprPtr tableLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  ExprPtr map,e;

  map = (*load)(sharedref);
  if (errno)
    return NIL;
  e = MakeExternal(desc,(char *)map);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc tableExtDesc = {
  (CHAR *)"table",
  { 't', 'a', 'b', 'l' }, NULL,
  NULL, /* no hash function, tables are not shared */
  NULL, /* no comparison function, tables are not shared */
  NULL, /* no print function */
  tableSweep,
  tableLoad
};

DEF_ACTION_0ARG(table)
{
  return MakeExternal(&tableExtDesc,(char *)NIL);
}
ACTION_END

DEF_PRIVATE_ACTION_COLLECT(q_table_get)
{
  Boolean found;
  ExprPtr e;

  e = NGet (arg, (ExprPtr)EXTDATA(ENV_UNDERLYING(env)), &found);
  if (found)
    res = Kdata(0,optttm,e); /* SOME e */
  else res = none;
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(table_get)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_get),arg);
}
ACTION_END

DEF_PRIVATE_ACTION_2ARGS(q_table_put)
{ /* env is the external with the nmap inside it. */
  ExprPtr uenv = ENV_UNDERLYING(env);

  setNPut (arg1, arg2, (ExprPtr *)&EXTDATA(uenv), uenv);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(table_put)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put),arg);
}
ACTION_END

DEF_PRIVATE_ACTION_2ARGS(q_table_put_behind)
{ /* env is the external with the nmap inside it. */
  ExprPtr uenv = ENV_UNDERLYING(env);

  setNPutBehind (arg1, arg2, (ExprPtr *)&EXTDATA(uenv), uenv);
  return tuple0;
}
ACTION_END

DEF_ACTION_1ARG(table_put_behind)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_put_behind),arg);
}
ACTION_END

DEF_PRIVATE_ACTION_COLLECT(q_table_remove)
{
  ExprPtr uenv = ENV_UNDERLYING(env);

  setNRemove (arg, (ExprPtr *)&EXTDATA(uenv), uenv);
  res = tuple0;
}
ACTION_END_COLLECT

DEF_ACTION_1ARG(table_remove)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_remove),arg);
}
ACTION_END

DEF_ACTION_1ARG(table_kill)
{
  EXTDATA(arg) = (char *)NIL;
  HITBARRIER(arg);
  return tuple0;
}
ACTION_END

Boolean q_table_sweep_do(loc,data)
IN CONST ExprPtr loc;
IN CONST char *data;
{
  ExprPtr res;

#if MAX_ARITY>=2
  res = (*ACTION((ExprPtr)data))(ENV_CONS(QENV ((ExprPtr)data), 2),
				 MAPLETX(loc), MAPLETY(loc));
#else
  {
    ExprPtr e;
    ENTER;

    e = Ktuple2 (MAPLETX(loc), MAPLETY(loc));
    (void) KEEP (e);
    res = (*ACTION((ExprPtr)data))(ENV_CONS(QENV ((ExprPtr)data), 1), e);
    QUIT;
  }
#endif
  return (res==trueValue);
}

DEF_PRIVATE_ACTION_1ARG(q_table_sweep)
{
  Boolean res;

  res = NSweep ((ExprPtr)EXTDATA(ENV_UNDERLYING(env)), q_table_sweep_do, (char *)arg);
  return res?trueValue:falseValue;
}
ACTION_END

DEF_ACTION_1ARG(table_sweep)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_sweep),arg);
}
ACTION_END

MapPtr q_table_collect_do(loc,data)
IN CONST ExprPtr loc;
IN CONST char *data;
{
  ExprPtr e, res;
  ENTER;

#if MAX_ARITY>=2
  res = (*ACTION((ExprPtr)data))(ENV_CONS(QENV ((ExprPtr)data), 2),
				 MAPLETX(loc), MAPLETY(loc));
#else
  {
    ExprPtr e;

    e = Ktuple2 (MAPLETX(loc), MAPLETY(loc));
    (void) KEEP (e);
    res = (*ACTION((ExprPtr)data))(ENV_CONS(QENV ((ExprPtr)data), 1), e);
  }
#endif
  KEEPONLY(res);
  return res;
}

DEF_PRIVATE_ACTION_1ARG(q_table_collect)
{
  MapPtr res;

  res = NCollect ((ExprPtr)EXTDATA(ENV_UNDERLYING(env)), q_table_collect_do, (char *)arg);
  return res;
}
ACTION_END

DEF_ACTION_1ARG(table_collect)
{
  return MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_table_collect),arg);
}
ACTION_END
