/* Definitions for the type checker.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

IMPORT void TypePrint P((IN CONST ExprPtr type,
			 IN printFunc doprint,
			 INOUT char *printdata,
			 IN CONST ExprPtr state));
EXTERN COUNTER crtab_maxlines;
EXTERN ExprPtr maxLinesRef;
IMPORT Boolean TypePretty P((IN CONST ExprPtr type,
			     IN printFunc doprint,
			     INOUT char *printdata,
			     IN CONST ExprPtr state,
			     IN COUNTER lmargin,
			     IN COUNTER rmargin));

EXTERN ExprPtr basisAssum,basisState;
EXTERN ExprPtr moduleName;

EXTERN SymPtr stringsym,stringSym,intsym,intSym,
  listsym,listSym,arraysym,arraySym,iarraysym,iarraySym,
  refsym,refSym,funSym,mapSym,optsym,optSym,tablesym,tableSym,
  boolsym,boolSym,exnsym,exnSym,dynsym,dynSym,
  promisesym,promiseSym,ufindsym,ufindSym,contsym,contSym,unitsym,
  outstreamsym,instreamsym,writesym,readsym,
  readlinesym,flushsym,tellsym,seeksym,advancesym,seekendsym,
  suspendsym,truncatesym,convertsym,killsym,rematchsym,resubstsym,
  rematchfullsym,rematchpsym;
EXTERN SymPtr alpha,alphaeq,_alpha,_alphaeq,beta,betaeq,_beta,gama,gamaeq;
EXTERN SymPtr nilSym,consSym,derefSym,assignSym,forceSym,
  bindSym,matchSym,emptySym,mapgetSym,
  parsweepSym,nomemSym,nonsenseSym,catchSym,
  stackSym,incompSym,noninvSym,nthSym,
  rangeSym,asciiSym,stringnthSym,
#ifdef OBSOLETE
  undefSym,
#endif
  subscriptSym,arithSym,ioSym,reSym,ufindSym,futureSym,
  appendSym,dunionSym,overwriteSym,
  underwriteSym,interSym,identitySym,
  mergeSym,powerSym,throwSym,callccSym,mapappSym;
EXTERN SymPtr falseSym,trueSym;
EXTERN SymPtr ellSym,elleqSym,ellimpSym,elleqimpSym;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
EXTERN ExprPtr fxInitSym,fxReadSym,fxWriteSym,fxExnSym,fxFromSym,fxGotoSym,fxTopSym;
EXTERN ExprPtr rgnSym,effectSym;
EXTERN ExprPtr pureGenFx,genRgn;
EXTERN MapPtr autoDecorations,autoPatFx;
#endif

EXTERN ExprPtr stringType,intType,numType,boolType,exnType,powerType,dynType,timeType,
  iarrayType;

EXTERN SymPtr itSym;
EXTERN ExprPtr unitset;

IMPORT ExprPtr tc P((ExprPtr expr,ExprPtr assum,ExprPtr state));
#define TC tc

#define REST_TAG NIL

     /* types may be:
	E_VAR : type variables
	E_TYPE_APPL : application of a type name to a list of arguments
	E_EMPTY, E_MAPLET, E_MAP : record type (maps from symbols [or REST_TAG] to types)
	E_NUMTYPE :   numerical type (in argument, a map from
	type variables or dimension types (TypeAppl(quan,NIL)) to non-zero exponents)
	E_CONS : cartesian products
 		 and in the EFFECT_TYVAR case, an APPLFUN (what is applied in
		 an E_TYPE_APPL) may be not only a name (free or rigid), or
		 a type function (a cons (header,type)), but also a:
	E_TYPEFX : (pointer to) sets of effects; all effects have the
		   form name(region,args...), or are effect variables;
		   second component is the real name of the type function
		   (always constant).
		   */
#define	E_BIAS E_TUPLE_2
#define Kbias(a,b) Ktuple2(a,b)
     /*
       fnlevel : number of abstractions or delays above current expression
		 (counted from toplevel)
       ids : map from variables (including constructors) to their types
       idlocs : map from variables to their zones
       idexts : map with the same structure as ids, expresses constraints on
                external vars (except that variables used but not defined are
		mapped to UNIT; see below for an explanation).
       cons : map from datatype constructors to their types.
       tyfcns : map from type identifiers to type functions
	        (couples (tyhdr,type), where tyhdr=(tyvarseq,eqconds), and
		eqconds is a map from tyvars to an eqcond, that is a TDG
		involving type names and nonref vars whose truth implies
		that the corresponding tyvar can only be replaced by types
		that admit equality)
       tyexts : map with the same structure as tyfcns, expressing constraints
                on external datatypes or dimensions (as idexts with ids).
       structs : map from structure and functor identifiers
		 to (name,assum) where name is a structure
		 or functor name, and assum is the most general signature
		 (or the coercing one)
       sigs : map from signature identifiers to assums representing signatures
       nongens : set of non-generic (free) type variables
		 (and also free region and effect variables, as E_TYPEFX
		 nodes, in the EFFECT_TYVAR case)
       units : map from unit identifiers to couples (scale,dimension type)
	       scale = tuple2(coeff,map from unit names to exponents)
       tyvars : map from quantified explicit type variables to temporary non-generics.

       ids and idexts are used to record what variables are imported, exported,
       or simply used in a CaML-like module. A variable may be regular (defined,
       then possibly used), imported (declared extern, then used at least once,
       never redefined), or singular (declared extern, not used, not defined).
       By default, an extern variable is singular. Then, the status of a variable
       is determined by whether ids and idexts contain the id of this variable.

       singular: IDEXTS: x => type, not in IDS
       imported: IDEXTS: x => UNIT, IDS: x => type
       regular:  not in IDEXTS, IDS: x => type

       tyfcns and tyexts are used to record the same thing on datatypes and
       dimensions (not their constructors, just the type names, arities, and so on).
       But there is no singular state, as an extern declaration serves as a
       definition, so singular is merged with imported.

       imported: TYEXTS: id => tyfcn, TYFCNS: id => tyfcn
       regular:  not in TYEXTS, TYFCNS: id => tyfcn
       */

#define TA_REGULAR 0
#define TA_PRECOMPILE 1

#define A_NOEQ_DATATYPES (0x1)
#define A_NOWARN_NOBIND (0x2)

struct type_assum {
  CHAR kind;   /*  = TYPE_ASSUM_KIND   */
  CHAR mark;
  CHAR ta_kind;
  CHAR flags;
  COUNTER fnlevel;
  struct type_assum *next;
  ExprPtr ids;
  ExprPtr idlocs;
  ExprPtr idexts;
  ExprPtr cons;
  ExprPtr tyfcns;
  ExprPtr tyexts;
#if MODULE_SYSTEM & STD_ML_MODULE
  ExprPtr structs;
  ExprPtr sigs;
#endif
  ExprPtr nongens;
  ExprPtr units;
  ExprPtr tyvars;
};

#if MODULE_SYSTEM & STD_ML_MODULE
#define MAKE_ASSUM(fnlevel,ids,idlocs,idexts,cons,tyfcns,tyexts,structs,sigs,nongens,units,tyvars)	\
	MakeTypeAssum(fnlevel,ids,idlocs,idexts,cons,tyfcns,tyexts,structs,sigs,nongens,units,tyvars)
#else
#define MAKE_ASSUM(fnlevel,ids,idlocs,idexts,cons,tyfcns,tyexts,structs,sigs,nongens,units,tyvars)	\
	MakeTypeAssum(fnlevel,ids,idlocs,idexts,cons,tyfcns,tyexts,nongens,units,tyvars)
#endif

#define	A_FNLEVEL(assum)	((struct type_assum *)assum)->fnlevel
#define A_KIND(assum) ((struct type_assum *)assum)->ta_kind
#define A_FLAGS(assum) ((struct type_assum *)assum)->flags
#define A_IDS(assum)	((struct type_assum *)assum)->ids
#define A_IDLOCS(assum) ((struct type_assum *)assum)->idlocs
#define A_IDEXTS(assum)	((struct type_assum *)assum)->idexts
#define A_CONS(assum)   ((struct type_assum *)assum)->cons
#define A_TYFCNS(assum) ((struct type_assum *)assum)->tyfcns
#define A_TYEXTS(assum) ((struct type_assum *)assum)->tyexts
#if MODULE_SYSTEM & STD_ML_MODULE
#define A_STRUCTS(assum) ((struct type_assum *)assum)->structs
#define A_SIGS(assum) ((struct type_assum *)assum)->sigs
#else
#define A_STRUCTS(assum) NIL
#define A_SIGS(assum) NIL
#endif
#define A_NONGENS(assum)	((struct type_assum *)assum)->nongens
#define A_UNITS(assum)  ((struct type_assum *)assum)->units
#define A_TYVARS(assum) ((struct type_assum *)assum)->tyvars

  /* tyvarsubs : substitution, mapping non-generic type variables to types
     conds : map f => set of (type1,type2), where f is the first flexible
	     symbol of type1 and type1 is not trivial
	     (a 2nd order pattern is appl(f,...) where f is flexible
	     it is trivial if it is appl(f,x1,...,xn) where xi are
	     generic and pairwise distinct)
     eqconds : TDG (as in noneqs) describing conjunction of mandatory
	       equality conditions
     eqflags : map from type variables to TDGs describing if this variable
	       admits equality (no entry means no equality, except if
	       there are two quotes in the name of the variable)
     nexgen : next number for creating the next new generic variable
     noneqs : set of type names that don't preserve equality
	      (actually map from propositional variables to TDGs [except trueValue]
	      with, as propositional variables:
	      - type names [meaning: this type name respects equality]
	      - shared E_REF(type name) [meaning: this type name is *not*
	        a 'ref' or an 'array'])
     tycons : map from type names to couples (tyhdr,cons) for datatype type names,
	      or to the default unit identifier for a dimension,
	      or from unit names to scales,
	      or from structure names to structure signatures,
	      or from functor names to functor signatures.
     packers : map from abstract syntax expressions (only for dynamics)
	       to the types of packed things, coded as a tree of maplets
	       (this is because maps may only have ExprPtrs in their domains,
	       and because it will be clear that there is no repetition)
     annotations : map from abstract syntax expressions to their types,
		   coded as a tree of maplets (same reason as for packers). Not all
		   expressions are annotated, only those needed by the debugger or
		   the compiler.
     nongens : set of (imperative) type variables that have not been generalized
     */

struct  type_state  {
  CHAR kind; /* = TYPE_STATE_KIND */
  CHAR mark;
  struct type_state *next;
  ExprPtr conds;
  ExprPtr eqconds;
  ExprPtr eqflags;
  COUNTER nexgen;
  ExprPtr noneqs;
  ExprPtr tycons;
  ExprPtr packers;
  ExprPtr annotations;
  MapPtr nongens;
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
  MapPtr strength,maxstrength;
#endif
  ExprPtr tyvarsubs,tyvarsubcache;
};

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
#define MAKE_STATE(conds,eqconds,eqflags,nexgen,noneqs,tycons,packers,annotations,nongens,strength,maxstrength)   \
	MakeTypeState(conds,eqconds,eqflags,nexgen,noneqs,tycons,packers,annotations,nongens,strength,maxstrength)
#else
#define MAKE_STATE(conds,eqconds,eqflags,nexgen,noneqs,tycons,packers,annotations,nongens)	\
	MakeTypeState(conds,eqconds,eqflags,nexgen,noneqs,tycons,packers,annotations,nongens)
#endif
#define S_TYVARSUBS(state) ((struct type_state *)state)->tyvarsubs
#define S_TYVARSUBCACHE(state) ((struct type_state *)state)->tyvarsubcache
#define	S_CONDS(state) ((struct type_state *)state)->conds
#define	S_EQCONDS(state) ((struct type_state *)state)->eqconds
#define	S_EQFLAGS(state) ((struct type_state *)state)->eqflags
#define S_NEXGEN(state)	((struct type_state *)state)->nexgen
#define S_NONEQS(state)	((struct type_state *)state)->noneqs
#define S_TYCONS(state) ((struct type_state *)state)->tycons
#define S_PACKERS(state) ((struct type_state *)state)->packers
#define S_ANNOTATIONS(state) ((struct type_state *)state)->annotations
#define S_NONGENS(state) ((struct type_state *)state)->nongens
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
#define S_STRENGTH(state) ((struct type_state *)state)->strength
#define S_MAXSTRENGTH(state) ((struct type_state *)state)->maxstrength
#endif

#if MODULE_SYSTEM & STD_ML_MODULE
IMPORT ExprPtr MakeTypeAssum P((COUNTER fnlevel,ExprPtr ids,ExprPtr idlocs,
				ExprPtr idexts,ExprPtr cons,ExprPtr tyfcns,
				ExprPtr tyexts,
				ExprPtr structs,ExprPtr sigs,
				ExprPtr nongens,ExprPtr units,ExprPtr tyvars));
#else
IMPORT ExprPtr MakeTypeAssum P((COUNTER fnlevel,ExprPtr ids,ExprPtr idlocs,
				ExprPtr idexts,ExprPtr cons,ExprPtr tyfcns,
				ExprPtr tyexts,
				ExprPtr nongens,ExprPtr units,ExprPtr tyvars));
#endif
IMPORT ExprPtr NewTypeAssum P((void));
IMPORT ExprPtr DupTypeAssum P((IN CONST ExprPtr assum));
IMPORT void AddTypeAssum P((IN CONST ExprPtr assum,IN CONST ExprPtr new));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IMPORT ExprPtr MakeTypeState P((ExprPtr conds,ExprPtr eqconds,ExprPtr eqflags,
				COUNTER nexgen,ExprPtr noneqs,ExprPtr tycons,MapPtr packers,
				MapPtr annotations,MapPtr nongens,
				MapPtr strength,MapPtr maxstrength
				));
#else
IMPORT ExprPtr MakeTypeState P((ExprPtr conds,ExprPtr eqconds,ExprPtr eqflags,
				COUNTER nexgen,ExprPtr noneqs,ExprPtr tycons,ExprPtr packers,
				ExprPtr annotations,MapPtr nongens
				));
#endif
IMPORT ExprPtr NewTypeState P((void));
IMPORT ExprPtr DupTypeState P((IN CONST ExprPtr state));
IMPORT void overwriteTypeState P((INOUT ExprPtr newstate,IN CONST ExprPtr state));

#if IMPERATIVE_TYPES==EFFECT_TYVAR
IMPORT void resetFx P((void));
IMPORT ExprPtr fxRealSigma P((IN CONST ExprPtr sigma,IN CONST ExprPtr state));
IMPORT ExprPtr freshFxVar P((IN CONST MapPtr constr,IN CONST ExprPtr sort,INOUT ExprPtr assum,INOUT ExprPtr state));
IMPORT ExprPtr funTypeFx P((IN CONST ExprPtr a,IN CONST ExprPtr b,IN CONST ExprPtr sigma));
IMPORT ExprPtr promiseTypeFx P((IN CONST ExprPtr a,IN CONST ExprPtr sigma));
IMPORT ExprPtr exnTypeFx P((IN CONST ExprPtr rho));
IMPORT ExprPtr initFx P((IN CONST ExprPtr rho,IN CONST ExprPtr type));
IMPORT ExprPtr exnFx P((IN CONST ExprPtr rho,IN CONST ExprPtr type));
IMPORT ExprPtr fromFx P((IN CONST ExprPtr rho));
IMPORT ExprPtr gotoFx P((IN CONST ExprPtr rho));
IMPORT ExprPtr readFx P((IN CONST ExprPtr rho));
IMPORT ExprPtr writeFx P((IN CONST ExprPtr rho));
#define fxSym(f) (((f)->kind==(CHAR)E_TYPEFX)?CAR(f):(f))
IMPORT void addTopFx P((INOUT ExprPtr type,INOUT ExprPtr assum,IN CONST ExprPtr state));
IMPORT MapPtr obsFx P((IN CONST MapPtr constr,IN CONST MapPtr vars,IN CONST ExprPtr state));
IMPORT MapPtr convertPatFx P((IN CONST MapPtr fx));
#else
#define resetFx()
#define funTypeFx(a,b,sigma) funType(a,b)
IMPORT ExprPtr promiseType P((IN CONST ExprPtr a));
#define promiseTypeFx(a,sigma) promiseType(a)
#define fxSym(f) f
#endif
IMPORT ExprPtr coupleType P((IN CONST ExprPtr a,IN CONST ExprPtr b));
IMPORT ExprPtr funType P((IN CONST ExprPtr a,IN CONST ExprPtr b));
IMPORT ExprPtr mapType P((IN CONST ExprPtr a,IN CONST ExprPtr b));
#define setType(a)  mapType(a,NIL)
IMPORT ExprPtr listType P((IN CONST ExprPtr a));

IMPORT  Boolean genericp P((ExprPtr tyvar,ExprPtr assum));
#define	rigidp(name,assum)	genericp(name,assum)

	/*  attribute flags */
#define F_EQ 1
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
#define F_IMP 2
#endif
#define F_NUM 4

#define IN_FLAGS(f1,f2) (((f1) & ~(f2))==0)

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
# define INF_STRENGTH MAX_SIGNED_LONG
#endif

IMPORT void getNumber P((IN COUNTER n,OUT CHAR *buf));
IMPORT int varFlags P((SymPtr tyvar));
IMPORT ExprPtr varEqTdg P((ExprPtr var,ExprPtr state));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IMPORT INT naturalStrength P((SymPtr tyvar));
IMPORT INT tyvarStrength P((SymPtr tyvar,ExprPtr state));
IMPORT INT tyvarMaxStrength P((SymPtr tyvar,ExprPtr state));
#endif
IMPORT SymPtr freshInstfromFlags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	P((SHORT flags,ExprPtr eqtdg,ExprPtr assum,ExprPtr state,IN INT strength,IN INT maxstrength));
#else
	P((SHORT flags,ExprPtr eqtdg,ExprPtr assum,ExprPtr state));
#endif
IMPORT SymPtr _freshInstfromFlags
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	P((SHORT flags,ExprPtr eqtdg,ExprPtr assum,ExprPtr state,IN INT strength,IN INT maxstrength));
#else
	P((SHORT flags,ExprPtr eqtdg,ExprPtr assum,ExprPtr state));
#endif

IMPORT ExprPtr nongenValue P((IN CONST ExprPtr tyvar,IN CONST ExprPtr state,OUT Boolean *foundp));
#define setNongenValue(typevar,val,state) setMapPut(typevar,val,&S_TYVARSUBCACHE(state))
IMPORT ExprPtr _nongenValue P((SymPtr typevar,ExprPtr state));
IMPORT void flushTyVarSubCache P((INOUT ExprPtr state));

#if IMPERATIVE_TYPES==EFFECT_TYVAR
IMPORT ExprPtr newTypeInst P((ExprPtr type,ExprPtr assum,ExprPtr state));
IMPORT ExprPtr newTypeInst1 P((ExprPtr type,ExprPtr assum,ExprPtr state,INOUT MapPtr *rename));
#define developType(type,state) newTypeInst(type,NIL,state)
#else
IMPORT ExprPtr newTypeInst P((ExprPtr type,ExprPtr assum,ExprPtr state,Boolean noimpp));
IMPORT ExprPtr newTypeInst1 P((ExprPtr type,ExprPtr assum,ExprPtr state,Boolean noimpp,INOUT MapPtr *rename));
#define developType(type,state) newTypeInst(type,NIL,state,false)
#endif
IMPORT ExprPtr eqTypeTdg P((ExprPtr type,ExprPtr state,Boolean eqp));
#if IMPERATIVE_TYPES==STD_ML_TYVAR
IMPORT Boolean impTypep P((ExprPtr type,ExprPtr state));
#endif
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IMPORT Boolean impTypep P((ExprPtr type,ExprPtr state,INT strength));
#endif
IMPORT Boolean numTypep P((ExprPtr type,ExprPtr state));

IMPORT MapPtr typeMul1 P((MapPtr map1,MapPtr map2));
IMPORT MapPtr typePower1 P((MapPtr map,ExprPtr exp));
IMPORT ExprPtr typeMul P((IN CONST TZone *zone1,IN CONST TZone *zone2,IN CONST ExprPtr type1,IN CONST ExprPtr type2,IN CONST ExprPtr state));
IMPORT ExprPtr typePower P((IN CONST TZone *zone,IN CONST ExprPtr type,IN CONST ExprPtr exp,IN CONST ExprPtr state));
IMPORT ExprPtr scaleMul P((IN CONST ExprPtr scale1,IN CONST ExprPtr scale2));
IMPORT ExprPtr scalePower P((IN CONST ExprPtr scale,IN CONST ExprPtr exp));
IMPORT ExprPtr instScale P((IN CONST ExprPtr scale,IN CONST ExprPtr state));

IMPORT MapPtr instNumType P((MapPtr map,MapPtr subst,ExprPtr state,OUT Boolean *finep));

IMPORT void incompatibleTypes P((IN CONST TZone *zone,
				 IN CONST ExprPtr type1,IN CONST ExprPtr type2,
				 IN CONST CHAR *msg1, IN CONST CHAR *msg2,
				 INOUT ExprPtr assum,INOUT ExprPtr state));
IMPORT void typeError P((TZone *zone,CHAR *msg,ExprPtr type,ExprPtr state));
IMPORT char *varMsg P((IN CONST char *msg,IN CONST SymPtr var));
IMPORT char *nameMsg P((IN CONST char *msg,IN CONST SymPtr var));

IMPORT MapPtr overrideMapType P((IN CONST MapPtr type1,IN CONST MapPtr type2,
				 IN CONST ExprPtr assum,IN CONST ExprPtr state));

IMPORT void coerceFlagTypes
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	P((TZone *zone,int flags,ExprPtr eqtdg,INT strength,INT maxstrength,ExprPtr type,ExprPtr assum,ExprPtr state));
#else
	P((TZone *zone,int flags,ExprPtr eqtdg,ExprPtr type,ExprPtr assum,ExprPtr state));
#endif
IMPORT void coerceTyFcn P((IN CONST TZone *zone,IN CONST ExprPtr tyfcn1,IN CONST ExprPtr tyfcn2,INOUT ExprPtr assum,INOUT ExprPtr state));

#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IMPORT ExprPtr Weaken P((ExprPtr type,ExprPtr assum,ExprPtr state,INT delta));
#endif
IMPORT void coerceScales P((IN CONST TZone *zone,IN CONST ExprPtr scale1,IN CONST ExprPtr scale2,INOUT ExprPtr assum,INOUT ExprPtr state));

IMPORT MapPtr getTypeSubst P((TZone *zone,ExprPtr tyhdr,ExprPtr args,ExprPtr assum,ExprPtr state));
IMPORT ExprPtr typeSubst P((TZone *zone,IN CONST ExprPtr type,IN CONST MapPtr subst,IN CONST ExprPtr assum,INOUT ExprPtr state));
IMPORT void maximizeEquality P((IN CONST MapPtr tycons,IN CONST MapPtr tyfcns,INOUT ExprPtr state,IN CONST MapPtr sureeq,IN CONST MapPtr spec,IN Boolean noeq,IN CONST TZone *zone));
IMPORT SymPtr freshTyName P((IN CONST SymPtr name,IN CONST ExprPtr assum,INOUT ExprPtr state));
IMPORT SymPtr copyTyName P((IN CONST SymPtr name,INOUT ExprPtr state));
IMPORT MapPtr InitAbsEq P((IN CONST MapPtr tyfcns,IN CONST ExprPtr state));
IMPORT ExprPtr funEqTdg P((IN CONST SymPtr f,IN CONST ExprPtr state));
IMPORT ExprPtr funNotRefTdg P((IN CONST SymPtr f,IN CONST ExprPtr state));
IMPORT ExprPtr realName P((IN CONST SymPtr name,IN CONST ExprPtr state));
IMPORT ExprPtr applFun P((IN CONST ExprPtr type,IN CONST ExprPtr state));
IMPORT ExprPtr correctTypeAppl P((IN CONST ExprPtr type,IN CONST ExprPtr state));
IMPORT MapPtr getSubst1 P((IN CONST ExprPtr varlist,IN CONST ExprPtr arglist));
IMPORT SymPtr getCarTyName P((IN CONST SymPtr name,INOUT ExprPtr state));
IMPORT SymPtr getCdrTyName P((IN CONST SymPtr name,INOUT ExprPtr state));
IMPORT ExprPtr getTyHdr P((IN CONST ExprPtr tyvarseq,IN CONST ExprPtr xx,IN CONST ExprPtr state));
IMPORT void getTyHdrSubsts P((IN CONST ExprPtr tyhdr1,IN CONST ExprPtr tyhdr2,OUT MapPtr *s1p,OUT MapPtr *s2p,INOUT ExprPtr state));

#if IMPERATIVE_TYPES==STD_ML_TYVAR
IMPORT Boolean expansivep P((IN CONST ExprPtr expr));
#endif
IMPORT MapPtr valTyVars P((IN CONST ExprPtr tyvarlist,INOUT ExprPtr assum,INOUT ExprPtr state));
IMPORT MapPtr tyVars
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
	P((IN CONST ExprPtr type,
	IN CONST ExprPtr assum,IN CONST ExprPtr state,
	SHORT flags,INT strength,INOUT ExprPtr memoref));
#else
	P((IN CONST ExprPtr type,
	IN CONST ExprPtr assum,IN CONST ExprPtr state,
	SHORT flags,INOUT ExprPtr memoref));
#endif
IMPORT MapPtr unquant_x P((IN CONST MapPtr exps,
			   IN CONST ExprPtr assum,IN CONST ExprPtr state,
			   INOUT ExprPtr memoref));
#if IMPERATIVE_TYPES!=EFFECT_TYVAR
IMPORT MapPtr unquant_y P((IN CONST MapPtr exps,
			   IN CONST ExprPtr assum,IN CONST ExprPtr state,
			   INOUT ExprPtr memoref));
#endif
IMPORT MapPtr unquant_packers P((IN CONST ExprPtr assum,
				 IN CONST ExprPtr state,INOUT ExprPtr memoref));
#if IMPERATIVE_TYPES==STD_ML_NJ_TYVAR
IMPORT MapPtr getStrengths P((IN CONST MapPtr tyvars,IN CONST ExprPtr state));
IMPORT void setStrengths P((IN CONST MapPtr tyvars,INOUT ExprPtr state,IN INT strength));
IMPORT void checkStrengths P((IN CONST TZone *zone,IN CONST MapPtr tyvars,IN CONST ExprPtr state));
#endif
IMPORT ExprPtr checkExplicits P((IN CONST TZone *zone,IN CONST MapPtr tyvarmap,IN CONST ExprPtr assum,IN CONST ExprPtr state));
IMPORT MapPtr getSureeq P((IN CONST ExprPtr eqlist));
IMPORT void addPacker P((IN CONST ExprPtr expr,IN CONST ExprPtr type,IN COUNTER fnlevel,INOUT ExprPtr state));
IMPORT void addAnnotation1 P((IN CONST ExprPtr expr,IN CONST ExprPtr type,IN COUNTER fnlevel,IN CONST ExprPtr sharingp,INOUT ExprPtr state));
IMPORT void addAnnotation P((IN CONST ExprPtr expr,IN CONST ExprPtr type,IN COUNTER fnlevel,INOUT ExprPtr state));

#define TDG_OR_HASH_SIZE 137
#define TDG_EQUIV_HASH_SIZE 37
EXTERN ExprPtr tdgOrHashTable,tdgEquivHashTable;

IMPORT Boolean tdgVarLess P((IN CONST ExprPtr x,IN CONST ExprPtr y));
IMPORT ExprPtr tdgNot P((IN CONST ExprPtr t));
IMPORT ExprPtr nameToTdg P((IN CONST ExprPtr x));
IMPORT ExprPtr tdgOr P((IN CONST ExprPtr t1,IN CONST ExprPtr t2));
IMPORT ExprPtr tdgAnd P((IN CONST ExprPtr t1,IN CONST ExprPtr t2));
IMPORT ExprPtr tdgImply P((IN CONST ExprPtr t1,IN CONST ExprPtr t2));
IMPORT ExprPtr tdgEquiv P((IN CONST ExprPtr t1,IN CONST ExprPtr t2));
IMPORT ExprPtr tdgInst P((IN CONST ExprPtr t,IN CONST ExprPtr subst,IN Boolean defselfp));
IMPORT ExprPtr tdgGetEquivVar P((IN CONST ExprPtr t));

IMPORT Boolean unify1 P((IN CONST ExprPtr type1,IN CONST ExprPtr type2,
			 INOUT ExprPtr assum,INOUT ExprPtr state,IN Boolean higherp));

#define TYPE_HASH_SIZE 137
#define TYPE_HASH(t) (((unsigned long)t) % TYPE_HASH_SIZE)
IMPORT ExprPtr newMemoRef P((void));

#if MODULE_SYSTEM & STD_ML_MODULE
	/* A signature is a type_assum, summing up all declarations in the sig, with
				nongens=the set of all type, structure, signature and functor names
					in the sig.
				A signature is not completely specified, and all conditions on names
			and bindings of names are held in the current state.
		A structure is coded as a signature, too, with
				nongens containing no type, structure, signature or functor names.
		A functor is a triple (locals,sig,struct), where locals are the flexible names in sig.
		A functor signature is a couple (locals,sig,sig).
	*/

#define TYPE_FUNCTOR_KIND E_TUPLE_GNRL

#define FUNCTOR_LOCALS(f) AVAL(f)->expr[0]
#define FUNCTOR_IN(f) AVAL(f)->expr[1]
#define FUNCTOR_OUT(f) AVAL(f)->expr[2]

IMPORT ExprPtr makeFunctor P((IN CONST MapPtr locals,IN CONST ExprPtr in,IN CONST ExprPtr out));
#define	MAKE_FUNCTOR(locals,in,out)	MakeLet(in,out)

IMPORT Boolean setTyFcn P((IN CONST ExprPtr f,IN CONST ExprPtr tyfcn,
			   IN CONST ExprPtr assum,INOUT ExprPtr state));
IMPORT Boolean recoverUnify2 P((IN CONST ExprPtr type1,IN CONST ExprPtr type2,
				IN CONST ExprPtr assum,INOUT ExprPtr state));
IMPORT ExprPtr instSig P((IN CONST ExprPtr sig,IN CONST ExprPtr state));
IMPORT void getLongStruct P((IN CONST TZone *zone,IN CONST ExprPtr path,INOUT ExprPtr *assump,IN CONST ExprPtr state));
IMPORT Boolean unify2 P((IN CONST ExprPtr type1,IN CONST ExprPtr type2,IN CONST ExprPtr assum,INOUT ExprPtr state));
#else
#define getLongStruct(zone,path,assump,state)
#endif

#ifdef OLDEXT
IMPORT void checkExports P((IN CONST TZone *zone,IN CONST ExprPtr assum1,
			    IN CONST ExprPtr assum2));
#endif
