/* Compiler to C auxiliary functions.
   Copyright (C) 2004,2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "expr.h"
#include "cons.h"
#include "bc.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "cxaux.h"

#define __SEG__ BYTECODE

#undef DEBUG
#undef DEBUG_ANLZ
/*#define DEBUG_ANLZ*/
#undef DEBUG_OPT

#define OPT

#undef EMIT_ASSERT

EXTERN CHAR *funNames[];

#ifdef OBSOLETE
EXTERN ExprPtr primCompileMap;

PUBLIC cxFile *cxOpen()
{
  ExprPtr decls, actions;
  cxFile *f;

  decls = outstringAction (voidString, NIL);
  (void) KEEP(decls);
  actions = outstringAction (voidString, NIL);
  (void) KEEP(actions);
  f = (cxFile *)MALLOC(sizeof(cxFile));
  f->nargs = 1;
  f->stackDepth = f->maxStackDepth = 2; /* room for the calling environment + one argument */
  f->mapStackDepth = f->maxMapStackDepth = 0;
  f->cStackDepth = f->maxCStackDepth = 0;
  f->ncvars = 0;
  f->dead = 0;
  f->decls = NULL;
  f->lastdeclp = &f->decls;
  f->actions = NULL;
  f->lastactionp = &f->actions;
  return f;
}

PUBLIC void cxOutDecl(f,s)
     INOUT cxFile *f;
     IN CONST CHAR *s;
{
  INT len;
  INT size;
  cxBuffer *buf;

  len = strlen (s);
  size = VARSIZE(cxBuffer,c,len);
  buf = (cxBuffer *)MALLOC(size);
  buf->next = NULL;
  strcpy (buf->c, s);
  *f->lastdeclp = buf;
  f->lastdeclp = &buf->next;
}

PUBLIC void cxOutAction(f,s)
     INOUT cxFile *f;
     IN CONST CHAR *s;
{
  INT len;
  INT size;
  cxBuffer *buf;

  len = strlen (s);
  size = VARSIZE(cxBuffer,c,len);
  buf = (cxBuffer *)MALLOC(size);
  buf->next = NULL;
  strcpy (buf->c, s);
  *f->lastactionp = buf;
  f->lastactionp = &buf->next;
}

PUBLIC void cxPush(amount,f)
IN COUNTER amount;
IN cxFile *f;
{
  f->stackDepth += amount;
  if (f->dead)
    return;
  if (f->stackDepth>f->maxStackDepth)
    {
      f->maxStackDepth = f->stackDepth;
      if (f->maxStackDepth>BC_STACK_SIZE)
	{
	  type_error((TZone *)0L,message(647));
	}
    }
}

PUBLIC void cxMapPush(amount,f)
IN COUNTER amount;
IN cxFile *f;
{
  f->mapStackDepth += amount;
  if (f->dead)
    return;
  if (f->mapStackDepth>f->maxMapStackDepth)
    f->maxMapStackDepth = f->mapStackDepth;
}

PUBLIC void cxLet (f)
     INOUT cxFile *f;
{
  COUNTER i;
  CHAR buf[32];

  cxPush (1,f);
  if (f->dead)
    return;
  i = ++f->ncvars;
  f->cvars[f->stackDepth-1] = -i;
  sprintf (buf, "ExprPtr _x%d;", i);
  cxOutDecl (f,buf);
}

PUBLIC void cxLetProtect (f)
     INOUT cxFile *f;
{
  COUNTER i;
  CHAR buf[32];

  cxPush (1,f);
  if (f->dead)
    return;
  i = f->cStackDepth++;
  if (f->cStackDepth>f->maxCStackDepth)
    f->maxCStackDepth = f->cStackDepth;
  f->cvars[f->stackDepth-1] = i;
  sprintf (buf, "_x.n = %d;", f->cStackDepth);
  cxOutAction (f,buf);
}

PUBLIC CHAR *cxVarName(f,i)
     IN CONST cxFile *f;
     IN COUNTER i;
{
  COUNTER k;
  static CHAR buf[32];

  k = f->cvars[i];
  if (k>=0)
    sprintf (buf, "_x.p[%d]", k);
  else sprintf (buf, "_x%d", -k);
  return buf;
}
#endif

/*
  bcAnalyze(), among other things, computes relations between variables
  at each program counter.
  A variable is a pair (i, pc) where i is the level of the variable (from 0 to
  current stack level - 1), and pc is the current program counter.
  Such pairs are encoded as i + pc * stackSize.
 */

#define M_CLOS0 (-6)
#define M_VARREF (-5)
#define M_INP (-4)
#define M_CGLOB (-3)
#define M_GLOB (-2)
#define M_CST (-1)
#define M_VAR 0
#define M_OP 1
#define M_KDATA 2
#define M_DATA 3
#define M_KRECORD1 4
#define M_RECORD1 5
#define M_CLOS 6
#define M_CONSENV 7
#define M_NTH 8
#define M_SELECT 9
#define M_KMU1 10
#define M_MU1 11

PRIVATE ExprPtr bcVar (i, pc)
     IN COUNTER i, pc;
{
  ExprPtr e, ei, epc;
  ENTER;

  ei = KINT (i);
  epc = KINT (pc);
  e = Kdata (M_VAR, ei, epc);
  KEEPONLY(e);
  return e;
}

#define M_ISVAR(e) (e->kind==(CHAR)E_DATA && e->magic==M_VAR)
#define M_VAR_OFF(e) IVAL(CDR(e))
#define M_VAR_I(e) IVAL(CAR(e))

/*
  Let us say var is the type of variables (i, pc).
  We compute maps:
  mean : var -m> var expr
    taking each variable to the expression it denotes;
    we use a set of expressions to encode phi functions (merge nodes).
  where
  datatype 'var expr = VAR of 'var (another variable)
                     | OP<i> of 'var * ... * 'var (operator # <i>)
		     | KDATA of shape * magic * 'var
		     | DATA of shape * magic * 'var
		     | KRECORD1 of shape * 'var
		     | RECORD1 of shape * 'var
		     | KMU1<i> of 'var * 'var
		     | MU1<i> of 'var * 'var
		     | CST<i> (constant from bc->consts[])
		     | GLOB<i> (constant from CGlobals[])
		     | INP<i> (input parameter #i)
		     | VARREF<l,j> of 'var
		     | CLOS<i> of 'var
		     | CLOS0<i>
		     | CONSENV<i> of 'var
		     | NTH<i> of 'var
		     | SELECT<i> of 'var
		     | NOEXPR
		     | VOLATILE (unknown, and may change from call to call; a raised exception,
		                 or the result of a call; OP(BC_REF, ...) and OP(BC_EXCEPTION, ...)
				 should be considered
				 volatile, too; similarly for OP(BC_STARTMAPSWEEP, ...),
				 OP(BC_STARTLISTSWEEP, ...), OP(BC_STARTSUBMAPSWEEP, ...)
				 )
		     | 'var tree (merge nodes, built with E_CONS nodes)
  Variables VAR (i, pc) are Kdata[magic=M_VAR](i, pc)
  OP<i> = Kdata[magic=M_OP](<i>,args)
  KDATA (shape, magic, arg) = Kdata[magic=M_KDATA](<shape>, E_COUNTED (magic, arg))
  DATA (shape, magic, arg) = Kdata[magic=M_DATA](<shape>, E_COUNTED (magic, arg))
    where shape is the index inside bc->consts[]
  KRECORD1 (shape, arg) = Kdata[magic=M_KRECORD1](<shape>, arg)
  RECORD1 (shape, arg) = Kdata[magic=M_RECORD1](<shape>, arg)
  KMU1<i> (rec, v) = Kdata[magic=M_KMU1](rec, E_COUNTER (i, v)) (label is consts[i])
  MU1<i> (rec, v) = Kdata[magic=M_MU1](rec, E_COUNTER (i, v)) (label is consts[i])
  CST<i> = Kdata[magic=M_CST](<i>,NIL)
  CGLOB<i> = Kdata[magic=M_CGLOB](<i>,NIL)
  GLOB<i> = Kdata[magic=M_GLOB](<i>,NIL)
  INP<i> = Kdata[magic=M_INP](<i>,NIL)
  VARREF<l,j> var = Kdata[magic=M_VARREF]((l,j), var)
  CLOS<i> var = Kdata[magic=M_CLOS](<i>, var)
  CLOS0<i> = Kdata[magic=M_CLOS0](<i>, NIL)
  CONSENV<i> var = Kdata[magic=M_CONSENV](<i>, var)
    where i=-1 to represent the effect of BC_CONSMEMO, i>=0 for BC_CONSENV
  NTH<i> var = Kdata[magic=M_NTH](<i>, var)
  SELECT<i> var = Kdata[magic=M_SELECT](<i>, var) (selects field consts[i] from record)
  NOEXPR = NIL
  VOLATILE = tuple0
 */

#define A_VOLATILE tuple0
#define A_NOEXPR NIL

PRIVATE Boolean m_volatile(e)
     IN CONST ExprPtr e;
{
  if (e==A_VOLATILE)
    return true;
  if (e==NOEXPR)
    return false;
  if (e->kind!=(CHAR)E_DATA)
    return false;
  if (e->magic!=M_OP)
    return false;
  {
    INT i = IVAL(CAR(e));

    return (i==BC_REF || i==BC_EXCEPTION || i==BC_STARTMAPSWEEP || i==BC_STARTLISTSWEEP || i==BC_STARTSUBMAPSWEEP);
  }
}

#ifdef OBSOLETE
PRIVATE MapPtr bcDrop (info, olddepth, newdepth)
     IN CONST MapPtr info;
     IN COUNTER olddepth, newdepth;
{
  COUNTER i;
  ExprPtr ei;
  ENTER;

  for (i=newdepth; i<olddepth; i++)
    {
      ei = KINT (i);
      info = MapRemove (ei, info);
    }
  KEEPONLY(info);
  return info;
}
#endif

PRIVATE MapPtr bcNextInfo (depth, off)
     IN COUNTER depth, off;
{
  COUNTER i;
  ExprPtr ni, ei, e;
  ENTER;

  ni = NewMap ();
  for (i=0; i<depth; i++)
    {
      ei = KINT (i);
      e = KINT (off);
      e = Kdata (M_VAR, ei, e);
      (void) KEEP (e);
      ni = MapPut (ei, e, ni);
    }
  KEEPONLY(ni);
  return ni;
}

PRIVATE MapPtr bcSet (i, e, info)
     IN COUNTER i;
     IN CONST ExprPtr e;
     IN CONST MapPtr info;
{
  ExprPtr ei;
  ENTER;

  ei = KINT (i);
  info = MapAdd (ei, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetCGlob (i, j, info)
     IN COUNTER i;
     IN COUNTER j; /* C global number */
     IN CONST MapPtr info;
{
  ExprPtr e;
  ENTER;

  e = KINT (j);
  e = Kdata (M_CGLOB, e, NIL);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetGlob (i, j, info)
     IN COUNTER i;
     IN COUNTER j; /* globVarRef number */
     IN CONST MapPtr info;
{
  ExprPtr e;
  ENTER;

  e = KINT (j);
  e = Kdata (M_GLOB, e, NIL);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

#define M_ISGLOB(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && e->magic==M_GLOB)
#define M_GLOB_I(e) IVAL(CAR(e))

PRIVATE MapPtr bcSetVarRef (i, l, j, var, info)
     IN COUNTER i;
     IN COUNTER l, j, var;
     IN CONST MapPtr info;
{
  ExprPtr el, ej, e, evar;
  ENTER;

  el = KINT (l);
  ej = KINT (j);
  e = Ktuple2 (el, ej);
  (void) KEEP (e);
  evar = KINT (var);
  evar = MapGetSure (evar, info);
  e = Kdata (M_VARREF, e, evar);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

#define M_ISCST(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && e->magic==M_CST)
#define M_CST_I(e) IVAL(CAR(e))
#define M_ISINP(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && e->magic==M_INP)
#define M_INP_I(e) IVAL(CAR(e))
#define M_ISVARREF(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && e->magic==M_VARREF)
#define M_VARREF_L(e) IVAL(CAR(CAR(e)))
#define M_VARREF_J(e) IVAL(CDR(CAR(e)))

PRIVATE MapPtr bcSetCst (i, j, info)
     IN COUNTER i;
     IN COUNTER j; /* const number */
     IN CONST MapPtr info;
{
  ExprPtr e;
  ENTER;

  e = KINT (j);
  e = Kdata (M_CST, e, NIL);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetStk (i, j, info)
     IN COUNTER i;
     IN COUNTER j; /* stack index */
     IN CONST MapPtr info;
{
  ExprPtr e, ej;
  ENTER;

  ej = KINT (j);
  e = MapGetSure (ej, info);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetSwap (i, j, info)
     IN COUNTER i;
     IN COUNTER j; /* stack index */
     IN CONST MapPtr info;
{
  ExprPtr ei, ej;
  ENTER;

  ei = KINT (i);
  ei = MapGetSure (ei, info);
  ej = KINT (j);
  ej = MapGetSure (ej, info);
  info = bcSet (i, ej, info);
  info = bcSet (j, ei, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetOp1 (i, op, j, info)
     IN COUNTER i;
     IN COUNTER op;
     IN COUNTER j;
     IN CONST MapPtr info;
{
  ExprPtr ej, e, eop;
  ENTER;

  eop = KINT (op);
  ej = KINT (j);
  e = MapGetSure (ej, info);
  e = Kdata (M_OP, eop, e);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

#define M_ISOP2(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && e->magic==M_OP && CDR(e)->kind==(CHAR)E_TUPLE_2)
#define M_OP2_I(e) IVAL(CAR(e))
#define M_OP2_CAR(e) CAR(CDR(e))
#define M_OP2_CDR(e) CDR(CDR(e))

PRIVATE MapPtr bcSetOp2 (i, op, j1, j2, info)
     IN COUNTER i;
     IN COUNTER op;
     IN COUNTER j1, j2;
     IN CONST MapPtr info;
{
  ExprPtr ej1, ej2, e, eop;
  ENTER;

  eop = KINT (op);
  ej1 = KINT (j1);
  ej1 = MapGetSure (ej1, info);
  ej2 = KINT (j2);
  ej2 = MapGetSure (ej2, info);
  e = Ktuple2 (ej1, ej2);
  (void) KEEP (e);
  e = Kdata (M_OP, eop, e);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetOp3 (i, op, j1, j2, j3, info)
     IN COUNTER i;
     IN COUNTER op;
     IN COUNTER j1, j2, j3;
     IN CONST MapPtr info;
{
  ExprPtr ej[3], e, eop;
  ENTER;

  eop = KINT (op);
  ej[0] = KINT (j1);
  ej[0] = MapGetSure (ej[0], info);
  ej[1] = KINT (j2);
  ej[1] = MapGetSure (ej[1], info);
  ej[2] = KINT (j3);
  ej[2] = MapGetSure (ej[2], info);
  e = KtupleGnrl (3, ej);
  (void) KEEP (e);
  e = Kdata (M_OP, eop, e);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetKData (i, shape_index, magic, j, info)
     IN COUNTER i;
     IN COUNTER shape_index, magic, j;
     IN CONST MapPtr info;
{
  ExprPtr e, ej;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  ej = Kcounted (E_COUNTED, ej, magic);
  (void) KEEP (ej);
  e = KINT (shape_index);
  e = Kdata (M_KDATA, e, ej);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetData (i, shape_index, magic, j, info)
     IN COUNTER i;
     IN COUNTER shape_index, magic, j;
     IN CONST MapPtr info;
{
  ExprPtr e, ej;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  ej = Kcounted (E_COUNTED, ej, magic);
  (void) KEEP (ej);
  e = KINT (shape_index);
  e = Kdata (M_DATA, e, ej);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetKrecord1 (i, shape_index, j, info)
     IN COUNTER i;
     IN COUNTER shape_index, j;
     IN CONST MapPtr info;
{
  ExprPtr e, ej;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  e = KINT (shape_index);
  e = Kdata (M_KRECORD1, e, ej);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetRecord1 (i, shape_index, j, info)
     IN COUNTER i;
     IN COUNTER shape_index, j;
     IN CONST MapPtr info;
{
  ExprPtr e, ej;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  e = KINT (shape_index);
  e = Kdata (M_RECORD1, e, ej);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetKmu1 (i, lab_index, j, k, info)
     IN COUNTER i;
     IN COUNTER lab_index, j, k;
     IN CONST MapPtr info;
{
  ExprPtr e, ej, ek;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  ek = KINT (k);
  ek = MapGetSure (ek, info);
  ek = Kcounted (E_COUNTED, ek, lab_index);
  (void) KEEP (ek);
  e = Kdata (M_KMU1, ej, ek);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetMu1 (i, lab_index, j, k, info)
     IN COUNTER i;
     IN COUNTER lab_index, j, k;
     IN CONST MapPtr info;
{
  ExprPtr e, ej, ek;
  ENTER;

  ej = KINT (j);
  ej = MapGetSure (ej, info);
  ek = KINT (k);
  ek = MapGetSure (ek, info);
  ek = Kcounted (E_COUNTED, ek, lab_index);
  (void) KEEP (ek);
  e = Kdata (M_MU1, ej, ek);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetTupleGnrl (i, op, shape_index, n, j, info)
     IN COUNTER i;
     IN COUNTER op;
     IN COUNTER shape_index; /* or -1 if tuple. */
     IN COUNTER n, j; /* build a tuple from x(j-n) to x(j-1) */
     IN CONST MapPtr info;
{
  ExprPtr ej, e, es;
  gcStackView *sv;
  COUNTER k;
  ENTER;

  sv = (gcStackView *) alloctmp (OFFSETOF(gcStackView,p[n]));
  sv->n = 0;
  sv->next = theHeapFrames;
  theHeapFrames = sv;
  for (k=0; k<n; k++)
    {
      ej = Kint (j-n+k);
      ej = MapGetSure (ej, info);
      sv->p[k] = ej;
      sv->n++;
    }
  if (shape_index>=0)
    es = KINT (shape_index);
  else es = NIL;
  e = KrecordGnrl (es, n, sv->p);
  (void) KEEP (e);
  ej = KINT (op);
  e = Kdata (M_OP, ej, e);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  theHeapFrames = sv->next;
  return info;
}

PRIVATE MapPtr bcSetClos (i, j, var, info)
     IN COUNTER i;
     IN COUNTER j, var;
     IN CONST MapPtr info;
{
  ExprPtr ej, e, evar;
  ENTER;

  ej = KINT (j);
  evar = KINT (var);
  evar = MapGetSure (evar, info);
  e = Kdata (M_CLOS, ej, evar);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetClos0 (i, j, info)
     IN COUNTER i;
     IN COUNTER j;
     IN CONST MapPtr info;
{
  ExprPtr ej, e;
  ENTER;

  ej = KINT (j);
  e = Kdata (M_CLOS0, ej, NIL);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

#define M_ISCLOS(e) (e!=NOEXPR && e->kind==(CHAR)E_DATA && (e->magic==M_CLOS || e->magic==M_CLOS0))
#define M_CLOS_I(e) IVAL(CAR(e))

PRIVATE MapPtr bcSetConsEnv (i, n, var, info)
     IN COUNTER i;
     IN COUNTER n, var;
     IN CONST MapPtr info;
{
  ExprPtr en, e, evar;
  ENTER;

  en = KINT (n);
  evar = KINT (var);
  evar = MapGetSure (evar, info);
  e = Kdata (M_CONSENV, en, evar);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetNth (i, n, var, info)
     IN COUNTER i;
     IN COUNTER n, var;
     IN CONST MapPtr info;
{
  ExprPtr en, e, evar;
  ENTER;

  en = KINT (n);
  evar = KINT (var);
  evar = MapGetSure (evar, info);
  e = Kdata (M_NTH, en, evar);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

PRIVATE MapPtr bcSetSelect (i, n, var, info)
     IN COUNTER i;
     IN COUNTER n, var;
     IN CONST MapPtr info;
{
  ExprPtr en, e, evar;
  ENTER;

  en = KINT (n);
  evar = KINT (var);
  evar = MapGetSure (evar, info);
  e = Kdata (M_SELECT, en, evar);
  (void) KEEP (e);
  info = bcSet (i, e, info);
  QUIT;
  return info;
}

#define CASE(op) case op:

#ifdef DEBUG_ANLZ
PRIVATE void bcPrtExpr (out, e)
     IN CONST hFILE *out;
     IN CONST ExprPtr e;
{
  switch (e->kind)
    {
    case E_TUPLE_0: /* A_VOLATILE */
      hfputs ("?", out);
      break;
    case E_NIL: /* NOEXPR */
      hfputs ("NOEXPR", out);
      break;
    case E_DATA:
      switch (e->magic)
	{
	case M_VAR:
	  hfprintf (out, "x%d@%d", IVAL(CAR(e)), IVAL(CDR(e)));
	  break;
	case M_OP:
	  switch (IVAL(CAR(e)))
	    {
	    case BC_REF: hfputs ("BC_REF", out); break;
	    case BC_KTUPLE1: hfputs ("BC_KTUPLE1", out); break;
	    case BC_MAKETUPLE1: hfputs ("BC_MAKETUPLE1", out); break;
	    case BC_KONS: hfputs ("BC_KONS", out); break;
	    case BC_CONS: hfputs ("BC_CONS", out); break;
	    case BC_CAR: hfputs ("BC_CAR", out); break;
	    case BC_CDR: hfputs ("BC_CDR", out); break;
	    case BC_REFVAL: hfputs ("BC_REFVAL", out); break;
	    case BC_KTUPLE2: hfputs ("BC_KTUPLE2", out); break;
	    case BC_MAKETUPLE2: hfputs ("BC_MAKETUPLE2", out); break;
	    case BC_KTUPLEGNRL: hfputs ("BC_KTUPLEGNRL", out); break;
	    case BC_MAKETUPLEGNRL: hfputs ("BC_MAKETUPLEGNRL", out); break;
	    case BC_KRECORDGNRL: hfputs ("BC_KRECORDGNRL", out); break;
	    case BC_MAKERECORDGNRL: hfputs ("BC_MAKERECORDGNRL", out); break;
	    case BC_MAPLET: hfputs ("BC_MAPLET", out); break;
	    case BC_OVERWRITE: hfputs ("BC_OVERWRITE", out); break;
	    case BC_FAKEENV: hfputs ("BC_FAKEENV", out); break;
	    case BC_QENV: hfputs ("BC_QENV", out); break;
	    case BC_MAPGET: hfputs ("BC_MAPGET", out); break;
	    case BC_NCARGET: hfputs ("BC_NCARGET", out); break;
	    case BC_MAPREMOVE: hfputs ("BC_MAPREMOVE", out); break;
	    case BC_MAPADD: hfputs ("BC_MAPADD", out); break;
	    case BC_KMUGNRL: hfputs ("BC_KMUGNRL", out); break;
	    case BC_MUGNRL: hfputs ("BC_MUGNRL", out); break;
	    case BC_QUEUELIST: hfputs ("BC_QUEUELIST", out); break;
	    case BC_MAPDOMRESTRTO: hfputs ("BC_MAPDOMRESTRTO", out); break;
	    case BC_MAPDOMRESTRBY: hfputs ("BC_MAPDOMRESTRBY", out); break;
	    case BC_DIFF: hfputs ("BC_DIFF", out); break;
	    case BC_DELTA: hfputs ("BC_DELTA", out); break;
	    case BC_STARTMAPSWEEP: hfputs ("BC_STARTMAPSWEEP", out); break;
	    case BC_QUEUEUNSHAREDLIST: hfputs ("BC_QUEUEUNSHAREDLIST", out); break;
	    case BC_EXCEPTION: hfputs ("BC_EXCEPTION", out); break;
	    case BC_DELAY: hfputs ("BC_DELAY", out); break;
	    case BC_PACK: hfputs ("BC_PACK", out); break;
	    case BC_IADD: hfputs ("BC_IADD", out); break;
	    case BC_ISUB: hfputs ("BC_ISUB", out); break;
	    case BC_IMUL: hfputs ("BC_IMUL", out); break;
	    case BC_IDIV: hfputs ("BC_IDIV", out); break;
	    case BC_IMOD: hfputs ("BC_IMOD", out); break;
	    case BC_IOR: hfputs ("BC_IOR", out); break;
	    case BC_IXOR: hfputs ("BC_IXOR", out); break;
	    case BC_IAND: hfputs ("BC_IAND", out); break;
	    case BC_ILSL: hfputs ("BC_ILSL", out); break;
	    case BC_ILSR: hfputs ("BC_ILSR", out); break;
	    case BC_IASR: hfputs ("BC_IASR", out); break;
	    case BC_IABS: hfputs ("BC_IABS", out); break;
	    case BC_INEG: hfputs ("BC_INEG", out); break;
	    case BC_ISQR: hfputs ("BC_ISQR", out); break;
	    case BC_INOT: hfputs ("BC_INOT", out); break;
	    case BC_IMIN: hfputs ("BC_IMIN", out); break;
	    case BC_IMAX: hfputs ("BC_IMAX", out); break;
	    case BC_RMIN: hfputs ("BC_RMIN", out); break;
	    case BC_RMAX: hfputs ("BC_RMAX", out); break;
	    case BC_FLOOR: hfputs ("BC_FLOOR", out); break;
	    case BC_CEIL: hfputs ("BC_CEIL", out); break;
	    case BC_AREF: hfputs ("BC_AREF", out); break;
	    case BC_IAREF: hfputs ("BC_IAREF", out); break;
	    case BC_NUMINT: hfputs ("BC_NUMINT", out); break;
	    case BC_INTNUM: hfputs ("BC_INTNUM", out); break;
	    case BC_SIZE: hfputs ("BC_SIZE", out); break;
	    case BC_ORD: hfputs ("BC_ORD", out); break;
	    case BC_ORDOF: hfputs ("BC_ORDOF", out); break;
	    case BC_SUBSTR: hfputs ("BC_SUBSTR", out); break;
	    default: hfprintf (out, "OP[%d]", IVAL(CAR(e))); break;
	    }
	  hfputs ("(", out);
	  {
	    ExprPtr args = CDR(e);

	    switch (args->kind)
	      {
	      case E_TUPLE_2: /* binary op */
		bcPrtExpr (out, CAR(args));
		hfputs(", ", out);
		bcPrtExpr (out, CDR(args));
		break;
	      case E_RECORD_GNRL:
		{
		  ExprPtr shape = ASHAPE (args);

		  if (shape->kind==(CHAR)E_INT)
		    hfprintf (out, "<shape:%d>", IVAL(shape));
		  /*fallthrough*/
		}
	      case E_TUPLE_GNRL: /* usually ternary op, may be of higher arity. */
		{
		  COUNTER i, n;
		  CHAR *delim = (CHAR *)"";

		  for (n=AVAL(args)->n, i=0; i<n; i++)
		    {
		      hfputs ((char *)delim, out);
		      delim = (CHAR *)", ";
		      bcPrtExpr (out, AVAL(args)->expr[i]);
		    }
		  break;
		}
	      default:
		bcPrtExpr (out, args);
		break;
	      }
	  }
	  hfputs (")", out);
	  break;
	case M_KDATA:
	  hfprintf (out, "KDATA<shape:%d>[case:%d](", IVAL(CAR(e)), COUNTEDN(CDR(e)));
	  bcPrtExpr (out, COUNTEDVAL(CDR(e)));
	  hfputs (")", out);
	  break;
	case M_DATA:
	  hfprintf (out, "DATA<shape:%d>[case:%d](", IVAL(CAR(e)), COUNTEDN(CDR(e)));
	  bcPrtExpr (out, COUNTEDVAL(CDR(e)));
	  hfputs (")", out);
	  break;
	case M_KRECORD1:
	  hfprintf (out, "KRECORD1<shape:%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_RECORD1:
	  hfprintf (out, "RECORD1<shape:%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_KMU1:
	  hfprintf (out, "KMU1(");
	  bcPrtExpr (out, CAR(e));
	  hfprintf (out, ", lab:%d => ", COUNTEDN(CDR(e)));
	  bcPrtExpr (out, COUNTEDVAL(CDR(e)));
	  hfputs (")", out);
	  break;
	case M_MU1:
	  hfprintf (out, "MU1(");
	  bcPrtExpr (out, CAR(e));
	  hfprintf (out, ", lab:%d => ", COUNTEDN(CDR(e)));
	  bcPrtExpr (out, COUNTEDVAL(CDR(e)));
	  hfputs (")", out);
	  break;
	case M_CST:
	  hfprintf (out, "CST<%d>", IVAL(CAR(e)));
	  break;
	case M_CGLOB:
	  hfprintf (out, "CGLOB<%d>", IVAL(CAR(e)));
	  break;
	case M_GLOB:
	  hfprintf (out, "GLOBVAR<%d>", IVAL(CAR(e)));
	  break;
	case M_INP:
	  hfprintf (out, "INP<%d>", IVAL(CAR(e)));
	  break;
	case M_VARREF:
	  hfprintf (out, "VARREF[+%d,%d] (", IVAL(CAR(CAR(e))), IVAL(CDR(CAR(e))));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_CLOS:
	  hfprintf (out, "CLOS<code:%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_CLOS0:
	  hfprintf (out, "CLOS0<code:%d>()", IVAL(CAR(e)));
	  break;
	case M_CONSENV:
	  if (IVAL(CAR(e))==-1)
	    hfputs ("CONSMEMO(", out);
	  else
	    hfprintf (out, "CONSENV<#entries:%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_NTH:
	  hfprintf (out, "NTH<%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	case M_SELECT:
	  hfprintf (out, "SELECT<lab:%d>(", IVAL(CAR(e)));
	  bcPrtExpr (out, CDR(e));
	  hfputs (")", out);
	  break;
	default:
	  hfprintf (out, "*weird*[magic=%d]", e->magic);
	  break;
	}
      break;
    case E_CONS:
      bcPrtExpr (out, CAR(e));
      hfputs ("+", out);
      bcPrtExpr (out, CDR(e));
      break;
    default:
      hfprintf (out, "*weird*[kind=%d]", e->kind);
      break;
    }
}

PRIVATE bcPrtInfo (out, info, depth)
     IN CONST hFILE *out;
     IN CONST MapPtr info;
     IN COUNTER depth;
{
  COUNTER i;
  ExprPtr e, ei;
  Boolean found;

  for (i=0; i<depth; i++)
    {
      ei = Kint(i);
      e = MapGet (ei, info, &found);
      hfputs (" ", out);
      if (found)
	bcPrtExpr (out, e);
      else hfputs ("*bug*", out);
    }
}
#undef CASE
#define CASE(op) case op: hfprintf (hstderr, "%d: %s =>", off, #op);
#endif

PRIVATE ExprPtr bcAnalyze (bc)
     IN CONST byteCodeChunk *bc;
{
#define N 11
#define A_LABS _x.p[0]
#define A_OFFSTK _x.p[1]
#define A_DEPTHSTK _x.p[2]
#define A_MAPDEPTHSTK _x.p[3]
#define A_EXCDEPTHSTK _x.p[4]
#define A_OFF _x.p[5]
#define A_SEEN _x.p[6]
#define A_INFO _x.p[7]
#define A_INFOSTK _x.p[8]
#define A_NEXTINFO _x.p[9]
#define A_NEXTINFOSTK _x.p[10]
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[N]; } _x;
  COUNTER depth, mapdepth, excdepth, maxexcdepth;
  OPCODE op, *pc;
  COUNTER off, dpc;
  Boolean found;
  ExprPtr seeninfo;
  COUNTER nCalls;
  ENTER;

  A_LABS = NewMap();
  A_OFFSTK = NIL;
  A_DEPTHSTK = NIL;
  A_MAPDEPTHSTK = NIL;
  A_EXCDEPTHSTK = NIL;
  A_OFF = NIL;
  A_SEEN = NewMap(); /* : int -m> int * (int -m> var expr set)
			map each offset to a mutable pair:
			- the depth l at this offset;
			- an A_INFO map from levels (0..l-1) to their values,
			as in the description of 'mean' above.
		     */
  A_INFO = NewMap(); /* : int -m> var expr set */
  A_INFOSTK = NIL;
  A_NEXTINFO = NewMap(); /* : int -m> var expr set,
			    map each i to {VAR (i, pc)} for current pc,
			    i=0..deph-1. */
  A_NEXTINFOSTK = NIL;
  _x.n = N;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  nCalls = 0;

  {
    COUNTER i;

    for (i=0; i<=bc->nargs; i++)
      {
	ExprPtr e, ei;

	ei = KINT (i);
	e = Kdata (M_INP, ei, NIL);
	(void) KEEP (e);
	A_INFO = MapAdd (ei, e, A_INFO);
	QUIT;
      }
  }
  A_NEXTINFO = bcNextInfo (bc->nargs+1, 0);
  QUIT;
#ifdef DEBUG_ANLZ
  {
    ExprPtr name = COUNTEDVAL(bcXName (bc));
    CHAR *buf = (CHAR *)alloctmp_anchored (SLEN(name)+1, (char *)&_x);

    memcpy (buf, STEXT(name), SLEN(name));
    buf[SLEN(name)] = 0;
    hfprintf (hstderr, "*Start:%s* (nargs=%d).\n", buf, bc->nargs);
    hfflush (hstderr);
  }
#endif

  /* Records:
     - targets of gotos in A_LABS;
     - number of stacked exception handlers in excdepth (max in maxexcdepth);
     - map A_SEEN : int -m> int * (int -m> var expr set)
			map each offset to a mutable pair:
			- the depth l at this offset;
			- an A_INFO map from levels (0..l-1) to their values,
			as in the description of 'mean' above.
     - number of call sites (BC_CALL, BC_NCALL, BC_CALLBIND).

     Should record:
     - Horn clauses stating which entries of _x.p[] are protected by GC;
     - Horn clauses stating which entries of _x.p[] need to remain in _x.p[] (see MakeTuple);
       => compute least model.  If entry need not be protected against GC and need not
       remain in _x.p[], then get it out of _x.p[] and in some local variable in bcXlate0().
     - entries of _x.p[] that denote constant closures, and which function they denote;
       (then, in compiling CALL or N_CALL, replaced (*ACTION(_x.p[...])) by the name of
       the function.)
     - eliminate dead code because of tags.
  */

  depth = bc->nargs+1;
  mapdepth = 0;
  excdepth = 0;
  maxexcdepth = 0;
#define POPTOP_IMM(n) depth -= n
#define PUSHTOP_IMM(n) depth += n
#define SYNCTOP()

  pc = bc->instr;
  for (;;)
    {
      off = pc-bc->instr;
      A_OFF = Kint(off);
      /* Here,
	 - A_INFO maps each i=0..depth-1 to the var expr it denotes.
	 The var expr set uses only variables from previous pc, or M_INP, or M_CST, or M_GLOB, or M_CGLOB.
	 - A_NEXTINFO maps each i=0..depth-1 to M_VAR (i, previous-pc).
      */
      seeninfo = MapGet (A_OFF, A_SEEN, &found);
      if (found)
	{
	  /* remember that current lab will be the target of a goto. */
	  A_LABS = MapAdd (A_OFF, UNIT, A_LABS);
	  /* merge A_INFOs. */
	  {
	    COUNTER i;
	    ExprPtr ei, e;
	    MapPtr m, info;

#ifdef DEBUG_ANLZ
	    if (depth!=IVAL(CAR(seeninfo)))
	      {
		hfprintf (hstderr, "bcAnalyze: depths do not match on merge; remembered: %d, new: %d.\n",
			  IVAL(CAR(seeninfo)), depth);
		hfflush (hstderr);
		abort ();
	      }
	    hfprintf (hstderr, "At off=%d,\n", off);
	    hfprintf (hstderr, "A_INFO = ");
	    bcPrtInfo (hstderr, A_INFO, depth);
	    hfprintf (hstderr, "\n");
	    hfflush (hstderr);
#endif
	    e = CDR (seeninfo);
#ifdef DEBUG_ANLZ
	    hfprintf (hstderr, "seen info = ");
	    bcPrtInfo (hstderr, e, depth);
	    hfprintf (hstderr, "\n");
	    hfflush (hstderr);
#endif
	    info = NewMap ();
	    for (i=0; i<depth; i++)
	      {
		ei = KINT (i);
		m = KONS (MapGetSure (ei, e),
			  MapGetSure (ei, A_INFO));
		info = MapPut (ei, m, info);
	      }
#ifdef DEBUG_ANLZ
	    hfprintf (hstderr, "*Merge* => ");
	    bcPrtInfo (hstderr, info, depth);
	    hfprintf (hstderr, "\n");
	    hfflush (hstderr);
#endif
	    SETCDR (seeninfo, info);
	    QUIT;
	  }
	pop:
	  if (A_OFFSTK==NIL)
	    break;
	  A_OFF = CAR(A_OFFSTK);
	  A_OFFSTK = CDR(A_OFFSTK);
	  off = IVAL(A_OFF);
	  pc = bc->instr + off;
	  depth = IVAL(CAR(A_DEPTHSTK));
	  A_DEPTHSTK = CDR(A_DEPTHSTK);
	  mapdepth = IVAL(CAR(A_MAPDEPTHSTK));
	  A_MAPDEPTHSTK = CDR(A_MAPDEPTHSTK);
	  excdepth = IVAL(CAR(A_EXCDEPTHSTK));
	  A_EXCDEPTHSTK = CDR(A_EXCDEPTHSTK);
	  A_INFO = CAR(A_INFOSTK);
	  A_INFOSTK = CDR(A_INFOSTK);
	  A_NEXTINFO = CAR(A_NEXTINFOSTK);
	  A_NEXTINFOSTK = CDR(A_NEXTINFOSTK);
#ifdef DEBUG_ANLZ
	  hfprintf (hstderr, "\n[POP] NEXTINFO:\t");
	  bcPrtInfo (hstderr, A_NEXTINFO, depth);
	  hfprintf (hstderr, "\n");
#endif
	  continue;
	}
      {
	ExprPtr e;

	e = KINT (depth);
	e = MakePair (E_TUPLE_2, e, A_INFO);
	(void) KEEP (e);
	A_SEEN = MapAdd (A_OFF, e, A_SEEN);
	QUIT;
      }
      A_INFO = A_NEXTINFO; /* By default, the effect of op is to copy what was previously on stack. */
      op = *pc++;
      switch (op)
	{
	CASE(BC_LOOP)
	  break;
	CASE(BC_FUNCTION)
	  break;
	CASE(BC_MERGE)
	  break;
	CASE(BC_BREAK)
	  break;
	CASE(BC_GOTO)
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    POPTOP_IMM(n);
	  }
	  pc += *pc;
	  break;
	CASE(BC_GOTO_0)
	  pc += pc[1];
	  break;
	CASE(BC_IF)
	  POPTOP_IMM(1);
	  dpc = 2;
#define DO_GOTO(xdrop)				\
	  {					\
	    REGISTER COUNTER n;			\
	    REGISTER COUNTER newoff;		\
	    ExprPtr saved;			\
	    					\
	    n = (COUNTER)pc[0];				      \
	    POPTOP_IMM(n+xdrop);			      \
	    newoff = (pc + 1 + pc[1]) - bc->instr;		      \
	    saved = KINT (newoff);			      \
	    A_LABS = MapAdd (saved, UNIT, A_LABS);	      \
	    A_OFFSTK = MakePair (E_CONS, saved, A_OFFSTK);	    \
	    saved = KINT (depth);				    \
	    A_DEPTHSTK = MakePair (E_CONS, saved, A_DEPTHSTK); \
            saved = KINT (mapdepth);			    \
            A_MAPDEPTHSTK = MakePair (E_CONS, saved, A_MAPDEPTHSTK); \
            saved = KINT (excdepth);			    \
            A_EXCDEPTHSTK = MakePair (E_CONS, saved, A_EXCDEPTHSTK); \
            A_INFOSTK = MakePair (E_CONS, A_INFO, A_INFOSTK); \
	    saved = bcNextInfo (depth, off); \
            A_NEXTINFOSTK = MakePair (E_CONS, saved, A_NEXTINFOSTK); \
	    depth += n+xdrop;						\
	    QUIT;						    \
	  }

	_goto:
	  DO_GOTO(0);
	  pc += dpc;
	  break;
	CASE(BC_CGLOBAL)
	  A_INFO = bcSetCGlob (depth, (COUNTER)*pc++, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CONST)
	  A_INFO = bcSetCst (depth, (COUNTER)*pc++, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_GLOBVARREF)
	  A_INFO = bcSetGlob (depth, (COUNTER)*pc++, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_DROP)
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_SWDROP)
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    A_INFO = bcSetStk (depth+n, depth-1, A_INFO);
	    n++;
	    POPTOP_IMM((-n));
	    break;
	  }
	CASE(BC_NSWDROP)
	  {
	    REGISTER COUNTER n, k;

	    n = (COUNTER)*pc++;
	    k = (COUNTER)*pc++;
	    for (; k<0; n++, k++)
	      A_INFO = bcSetStk (depth+n, depth+k, A_INFO);
	    POPTOP_IMM((-n));
	    break;
	  }
	CASE(BC_REF)
	  A_INFO = bcSetOp1 (depth-1, BC_REF, depth-1, A_INFO);
	  break;
	CASE(BC_KTUPLE1)
	  A_INFO = bcSetOp1 (depth-1, BC_KTUPLE1, depth-1, A_INFO);
	  break;
	CASE(BC_MAKETUPLE1)
	  A_INFO = bcSetOp1 (depth-1, BC_MAKETUPLE1, depth-1, A_INFO);
	  break;
	CASE(BC_KONS)
	  A_INFO = bcSetOp2 (depth-2, BC_KONS, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_CONS)
	  A_INFO = bcSetOp2 (depth-2, BC_CONS, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_CONSMEMO)
	  A_INFO = bcSetConsEnv (depth-2, (COUNTER)-1, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_KDATA)
	  {
	    COUNTER magic = (COUNTER)*pc++;
	    COUNTER shape_index = (COUNTER)*pc++;

	    A_INFO = bcSetKData (depth-1, shape_index, magic, depth-1, A_INFO);
	  }
	  SYNCTOP();
	  break;
	CASE(BC_DATA)
	  {
	    COUNTER magic = (COUNTER)*pc++;
	    COUNTER shape_index = (COUNTER)*pc++;

	    A_INFO = bcSetData (depth-1, shape_index, magic, depth-1, A_INFO);
	  }
	  SYNCTOP();
	  break;
	CASE(BC_FETCH) goto fetch;
	CASE(BC_STARTLISTSWEEP) goto fetch;
	CASE(BC_GETMAPSWEEP) goto fetch;
	CASE(BC_GETSUBMAPSWEEP)
	fetch:
	  A_INFO = bcSetStk (depth, depth + (*pc++), A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m8)
	  A_INFO = bcSetStk (depth, depth-8, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m7)
	  A_INFO = bcSetStk (depth, depth-7, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m6)
	  A_INFO = bcSetStk (depth, depth-6, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m5)
	  A_INFO = bcSetStk (depth, depth-5, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m4)
	  A_INFO = bcSetStk (depth, depth-4, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m3)
	  A_INFO = bcSetStk (depth, depth-3, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m2)
	  A_INFO = bcSetStk (depth, depth-2, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m1)
	  A_INFO = bcSetStk (depth, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CAR)
	  A_INFO = bcSetOp1 (depth-1, BC_CAR, depth-1, A_INFO);
	  break;
	CASE(BC_FETCHCAR)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth + (*pc++), A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m8)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-8, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m7)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-7, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m6)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-6, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m5)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-5, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m4)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-4, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m3)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-3, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m2)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-2, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCAR_m1)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CDR)
	  A_INFO = bcSetOp1 (depth-1, BC_CDR, depth-1, A_INFO);
	  break;
	CASE(BC_FETCHCDR)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth + (*pc++), A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m8)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-8, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m7)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-7, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m6)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-6, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m5)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-5, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m4)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-4, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m3)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-3, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m2)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-2, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHCDR_m1)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CARCDR)
	  A_INFO = bcSetOp1 (depth, BC_CDR, depth-1, A_INFO);
	  A_INFO = bcSetOp1 (depth-1, BC_CAR, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CDRCAR)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth-1, A_INFO);
	  A_INFO = bcSetOp1 (depth-1, BC_CDR, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_REFVAL)
	  A_INFO = bcSetOp1 (depth-1, BC_REFVAL, depth-1, A_INFO);
	  break;
	CASE(BC_FETCHREFVAL)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth + (*pc++), A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m8)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-8, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m7)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-7, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m6)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-6, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m5)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-5, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m4)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-4, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m3)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-3, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m2)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-2, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m1)
	  A_INFO = bcSetOp1 (depth, BC_REFVAL, depth-1, A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_KRECORD1)
	  A_INFO = bcSetKrecord1 (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  SYNCTOP();
	  break;
	CASE(BC_MAKERECORD1)
	  A_INFO = bcSetRecord1 (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  SYNCTOP();
	  break;
	CASE(BC_KTUPLE2)
	  A_INFO = bcSetOp2 (depth-2, BC_KTUPLE2, depth-2, depth-1, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MAKETUPLE2)
	  A_INFO = bcSetOp2 (depth-2, BC_MAKETUPLE2, depth-2, depth-1, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_KTUPLEGNRL)
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    A_INFO = bcSetTupleGnrl (depth-n, BC_KTUPLEGNRL, -1, n, depth, A_INFO);
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAKETUPLEGNRL)
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    A_INFO = bcSetTupleGnrl (depth-n, BC_MAKETUPLEGNRL, -1, n, depth, A_INFO);
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_KRECORDGNRL)
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    A_INFO = bcSetTupleGnrl (depth-n, BC_KRECORDGNRL, (COUNTER)*pc++, n, depth, A_INFO);
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAKERECORDGNRL)
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    A_INFO = bcSetTupleGnrl (depth-n, BC_MAKERECORDGNRL, (COUNTER)*pc++, n, depth, A_INFO);
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAPLET)
	  A_INFO = bcSetOp2 (depth-2, BC_MAPLET, depth-2, depth-1, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_OVERWRITE)
	  A_INFO = bcSetOp2 (depth-2, BC_OVERWRITE, depth-2, depth-1, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_UNDERWRITE)
	  A_INFO = bcSetOp2 (depth-2, BC_OVERWRITE, depth-1, depth-2, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_VARREF0)
	  A_INFO = bcSetVarRef (depth, 0, (COUNTER)pc[1], depth + pc[0], A_INFO);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF1)
	  A_INFO = bcSetVarRef (depth, 1, (COUNTER)pc[1], depth + pc[0], A_INFO);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF2)
	  A_INFO = bcSetVarRef (depth, 2, (COUNTER)pc[1], depth + pc[0], A_INFO);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF3)
	  A_INFO = bcSetVarRef (depth, 3, (COUNTER)pc[1], depth + pc[0], A_INFO);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF)
	  A_INFO = bcSetVarRef (depth, (COUNTER)pc[1], (COUNTER)pc[2], depth + pc[0], A_INFO);
	  pc += 3;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CLOSURE)
	  A_INFO = bcSetClos (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  SYNCTOP();
	  break;
	CASE(BC_CLOSURE0)
	  A_INFO = bcSetClos0 (depth, (COUNTER)*pc++, A_INFO);
	  SYNCTOP();
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_NOEXPR)
	  {
	    REGISTER COUNTER i;
	    ExprPtr e;

	    i = (COUNTER)*pc++;
	    while (i--)
	      {
		A_INFO = bcSet (depth, A_NOEXPR, A_INFO);
		PUSHTOP_IMM(1);
	      }
	    break;
	  }
	CASE(BC_TUCK)
	  POPTOP_IMM(1);
	  A_INFO = bcSetStk (depth + (*pc++), depth, A_INFO);
	  break;
	CASE(BC_FAKEENV)
	  A_INFO = bcSetOp1 (depth-1, BC_FAKEENV, depth-1, A_INFO);
	  SYNCTOP();
	  break;
	CASE(BC_CONSENV)
	  A_INFO = bcSetConsEnv (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  SYNCTOP();
	  break;
	CASE(BC_LET)
	  /* Since we do not record any of the values inside allocated environments,
	   we do not need to modify A_INFO here.
	  */
	  {
	    COUNTER k;

	    pc++;
	    while ((k = *pc++))
	      {
		pc++;
	      }
	    break;
	  }
	CASE(BC_LET1)
	  /* Since we do not record any of the values inside allocated environments,
	   we do not need to modify A_INFO here.
	  */
	  {
	    COUNTER k;

	    pc++;
	    k = *pc++;
	    pc++;
	    break;
	  }
	CASE(BC_QENV)
	  A_INFO = bcSetOp1 (depth-1, BC_QENV, depth-1, A_INFO);
	  break;
	CASE(BC_IS)
	  dpc = 3;
	  goto _goto;
	CASE(BC_ISDATA)
	  dpc = 3;
	  goto _goto;
	CASE(BC_SWITCH)
	  {
	    COUNTER m, n, i;

	    m = ((COUNTER)*pc++) - 2;
	    n = (COUNTER)*pc++;
	    for (i=-m; i<=n; i++)
	      {
		REGISTER COUNTER newoff;
		ExprPtr saved;

		newoff = (pc + (COUNTER)*pc++) - bc->instr;
		saved = KINT (newoff);
		A_LABS = MapAdd (saved, UNIT, A_LABS);
		A_OFFSTK = MakePair (E_CONS, saved, A_OFFSTK);
		saved = KINT (depth);
		A_DEPTHSTK = MakePair (E_CONS, saved, A_DEPTHSTK);
		saved = KINT (mapdepth);
		A_MAPDEPTHSTK = MakePair (E_CONS, saved, A_MAPDEPTHSTK);
		saved = KINT (excdepth);
		A_EXCDEPTHSTK = MakePair (E_CONS, saved, A_EXCDEPTHSTK);
		A_INFOSTK = MakePair (E_CONS, A_INFO, A_INFOSTK);
		saved = bcNextInfo (depth, off);
		A_NEXTINFOSTK = MakePair (E_CONS, saved, A_NEXTINFOSTK);
		QUIT;
	      }
	    goto pop;
	  }
	CASE(BC_ISEQUAL)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_NTH)
	  A_INFO = bcSetNth (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_0)
	  A_INFO = bcSetNth (depth-1, 0, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_1)
	  A_INFO = bcSetNth (depth-1, 1, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_2)
	  A_INFO = bcSetNth (depth-1, 2, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_3)
	  A_INFO = bcSetNth (depth-1, 3, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_4)
	  A_INFO = bcSetNth (depth-1, 4, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_5)
	  A_INFO = bcSetNth (depth-1, 5, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_6)
	  A_INFO = bcSetNth (depth-1, 6, depth-1, A_INFO);
	  break;
	CASE(BC_NTH_7)
	  A_INFO = bcSetNth (depth-1, 7, depth-1, A_INFO);
	  break;
	CASE(BC_3ARGS)
	  A_INFO = bcSetNth (depth+1, 2, depth-1, A_INFO);
	  A_INFO = bcSetNth (depth, 1, depth-1, A_INFO);
	  A_INFO = bcSetNth (depth-1, 0, depth-1, A_INFO);
	  PUSHTOP_IMM(2);
	  break;
	CASE(BC_NARGS)
	  {
	    COUNTER i, n = (COUNTER)*pc++;

	    for (i=n; --i>=0; )
	      A_INFO = bcSetNth (depth+i-1, i, depth-1, A_INFO);
	    PUSHTOP_IMM(n-1);
	    break;
	  }
	CASE(BC_TYPEMATCH)
	  dpc = 3;
	  DO_GOTO(1);
	  POPTOP_IMM(1);
	  pc += dpc;
	  break;
	CASE(BC_CARDEQ)
	  dpc = 3;
	  goto _goto;
	CASE(BC_CARDGEQ)
	  dpc = 3;
	  goto _goto;
	CASE(BC_SELECT)
	  A_INFO = bcSetSelect (depth-1, (COUNTER)*pc++, depth-1, A_INFO);
	  break;
	CASE(BC_ISNEMAP)
	  dpc = 2;
	  goto _goto;
	CASE(BC_MAPGET)
	  POPTOP_IMM(1);
	  dpc = 2;
	  DO_GOTO(0);
	  A_INFO = bcSetOp2 (depth-1, BC_MAPGET, depth, depth-1, A_INFO);
	  /* The BC_MAPGET op applied to m and x denotes ?m x when x is in dom m;
	     it has no meaning otherwise (i.e., if we follow the goto). */
	  pc += dpc;
	  break;
	CASE(BC_NCARGET)
	  dpc = 2;
	  DO_GOTO(0);
	  POPTOP_IMM(1);
	  A_INFO = bcSetOp2 (depth-1, BC_NCARGET, depth, depth-1, A_INFO);
	  pc += dpc;
	  break;
	CASE(BC_NCARPUT)
	  A_INFO = bcSetStk (depth-3, depth-1, A_INFO);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_MAPREMOVE)
	  A_INFO = bcSetOp2 (depth-2, BC_MAPREMOVE, depth-2, depth-1, A_INFO);
	  SYNCTOP();
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RETURN)
	  goto pop;
	CASE(BC_MAPADD)
	  A_INFO = bcSetOp3 (depth-3, BC_MAPADD, depth-3, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_MAPADD2)
	  A_INFO = bcSetOp3 (depth-3, BC_MAPADD, depth-2, depth-1, depth-3, A_INFO);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_SETCAR)
	  /* Since we do not record the CAR of the CONSENV used in memo functions,
	     we do not need to modify A_INFO here.
	  */
	  POPTOP_IMM(2);
	  break;
	CASE(BC_N_CALL)
	  {
	    COUNTER n = (COUNTER)*pc++;

	    POPTOP_IMM(n);
	    A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO);
	    nCalls++;
	    break;
	  }
	CASE(BC_CALL)
	  POPTOP_IMM(1);
	  A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO);
	  nCalls++;
	  break;
	CASE(BC_SETREF)
	  /* We do not record the side effect; this means we should not
	     pay too much attention to contents of REFs, unless we do some
	     alias analysis first.
	  */
	  POPTOP_IMM(2);
	  break;
	CASE(BC_KMU1)
	  {
	    COUNTER lab;

	    SYNCTOP();
	    lab = (COUNTER)*pc++;
	    POPTOP_IMM(1);
	    A_INFO = bcSetKmu1 (depth-1, lab, depth-1, depth, A_INFO);
	    break;
	  }
	CASE(BC_MU1)
	  {
	    COUNTER lab;

	    SYNCTOP();
	    lab = (COUNTER)*pc++;
	    POPTOP_IMM(1);
	    A_INFO = bcSetMu1 (depth-1, lab, depth-1, depth, A_INFO);
	    break;
	  }
	CASE(BC_KMUGNRL)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_KMUGNRL, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MUGNRL)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_MUGNRL, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_SWAP)
	  A_INFO = bcSetSwap (depth-2, depth-1, A_INFO);
	  break;
	CASE(BC_NEWQUEUE)
	  A_INFO = bcSet (depth, A_VOLATILE, A_INFO); /* We ignore what's inside a new queue. */
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CONCATQUEUE)
	  pc++;
	  A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO); /* We ignore what's inside a queue. */
	  break;
	CASE(BC_NEW1QUEUE)
	  SYNCTOP();
	  A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO); /* We ignore what's inside a new queue. */
	  break;
	CASE(BC_QUEUELIST)
	  SYNCTOP();
	  A_INFO = bcSetOp1 (depth-1, BC_QUEUELIST, depth-1, A_INFO);
	  break;
	CASE(BC_LISTQUEUE)
	  SYNCTOP();
	  A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO); /* We ignore what's inside a new queue. */
	  break;
	CASE(BC_MAPINTER) goto mapdrt;
	CASE(BC_MAPDOMRESTRTO)
	mapdrt:
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_MAPDOMRESTRTO, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MAPDOMRESTRBY)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_MAPDOMRESTRBY, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_DIFF)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_DIFF, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_DELTA)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_DELTA, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_STARTMAPSWEEP)
	  {
	    COUNTER j = depth + (*pc++);
	    COUNTER i;

	    mapdepth += 2;
	    for (i=0; i<=MAP_CHAIN_LENGTH; i++)
	      A_INFO = bcSetOp1 (depth+i, BC_STARTMAPSWEEP, j, A_INFO);
	    /* We record that all data pushed on the stack is a BC_STARTMAPSWEEP(var #j),
	       to be able to realize that this data does not need to be protected against
	       the GC.  But BC_STARTMAPSWEEP(...) data are volatile: they may change
	       with BC_INCMAPSWEEP.
	    */
	    PUSHTOP_IMM(1+MAP_CHAIN_LENGTH);
	    break;
	  }
	CASE(BC_STARTSUBMAPSWEEP)
	  pc++;
	  A_INFO = bcSet (depth, A_VOLATILE, A_INFO);
	  A_INFO = bcSet (depth+1, A_VOLATILE, A_INFO);
	  /* The two elements pushed on top of stack will change through BC_INCSUBMAPSWEEP.
	     Moreover, they should be protected against GC.
	  */
	  PUSHTOP_IMM(2);
	  break;
	CASE(BC_GETLISTSWEEP)
	  A_INFO = bcSetOp1 (depth, BC_CAR, depth + (*pc++), A_INFO);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_INCLISTSWEEP)
	  /* We don't do anything on A_INFO: we trust that BC_INCLISTSWEEP is correctly
	     used.
	  */
	  dpc = 3;
	  DO_GOTO(0);
	  pc += dpc;
	  break;
	CASE(BC_INCMAPSWEEP)
	  /* We don't do anything on A_INFO: we trust that BC_INCMAPSWEEP is correctly
	     used.
	  */
	  DO_GOTO(0);
	  pc += 4;
	  break;
	CASE(BC_INCSUBMAPSWEEP)
	  /* We don't do anything on A_INFO: we trust that BC_INCSUBMAPSWEEP is correctly
	     used.
	  */
	  SYNCTOP();
	  DO_GOTO(0);
	  pc += 3;
	  break;
	CASE(BC_MAPDROP)
	  /* No effect on A_INFO. */
	  mapdepth -= (COUNTER)*pc++;
	  break;
	CASE(BC_QUEUEUNSHAREDLIST)
	  A_INFO = bcSetOp1 (depth-1, BC_QUEUEUNSHAREDLIST, depth-1, A_INFO);
	  break;
	CASE(BC_HANDLE)
	  excdepth++;
	  if (excdepth>maxexcdepth)
	    maxexcdepth = excdepth;
	  {
	    ExprPtr offset, saved, e;

	    offset = KINT ((pc + 1 + pc[1]) - bc->instr);
	    A_LABS = MapAdd (offset, UNIT, A_LABS);
	    A_OFFSTK = MakePair (E_CONS, offset, A_OFFSTK);
	    offset = KINT (depth+1); /* +1 to account for the exception that will be pushed on the stack
					on entry to the handler. */
	    A_DEPTHSTK = MakePair (E_CONS, offset, A_DEPTHSTK);
	    offset = KINT (mapdepth);
	    A_MAPDEPTHSTK = MakePair (E_CONS, offset, A_MAPDEPTHSTK);
	    offset = KINT (excdepth);
	    A_EXCDEPTHSTK = MakePair (E_CONS, offset, A_EXCDEPTHSTK);
	    e = bcSet (depth, A_VOLATILE, A_INFO); /* var #depth is some unknown exception. */
            A_INFOSTK = MakePair (E_CONS, e, A_INFOSTK);
	    saved = bcNextInfo (depth+1, off);
	    A_NEXTINFOSTK = MakePair (E_CONS, saved, A_NEXTINFOSTK);
	    QUIT;
	  }
	  pc += 2;
	  break;
	CASE(BC_NEXTHANDLE) goto nexthandle;
	CASE(BC_NEXTHANDLE_NODEBUG)
	nexthandle:
	  SYNCTOP();
	  /* No effect on A_INFO. */
	  --excdepth;
	  break;
	CASE(BC_RAISE)
	  SYNCTOP();
	  /* No effect on A_INFO. */
	  goto pop;
	CASE(BC_EXCEPTION)
	CASE(BC_EXCEPTION_BENIGN)
	  SYNCTOP();
	  A_INFO = bcSetOp1 (depth-1, BC_EXCEPTION, depth-1, A_INFO);
	  break;
	CASE(BC_DELAY)
	  SYNCTOP();
	  A_INFO = bcSetOp1 (depth-1, BC_DELAY, depth-1, A_INFO);
	  break;
	CASE(BC_PACK)
	  SYNCTOP();
	  A_INFO = bcSetOp2 (depth-2, BC_PACK, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IADD)
	  A_INFO = bcSetOp2 (depth-2, BC_IADD, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ISUB)
	  A_INFO = bcSetOp2 (depth-2, BC_ISUB, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMUL)
	  A_INFO = bcSetOp2 (depth-2, BC_IMUL, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IDIV)
	  A_INFO = bcSetOp2 (depth-2, BC_IDIV, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMOD)
	  A_INFO = bcSetOp2 (depth-2, BC_IMOD, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ISNZERO)
	  dpc = 2;
	  goto _goto;
	CASE(BC_IOR)
	  A_INFO = bcSetOp2 (depth-2, BC_IOR, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IXOR)
	  A_INFO = bcSetOp2 (depth-2, BC_IXOR, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IAND)
	  A_INFO = bcSetOp2 (depth-2, BC_IAND, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ILSL)
	  A_INFO = bcSetOp2 (depth-2, BC_ILSL, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ILSR)
	  A_INFO = bcSetOp2 (depth-2, BC_ILSR, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IASR)
	  A_INFO = bcSetOp2 (depth-2, BC_IASR, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IABS)
	  A_INFO = bcSetOp1 (depth-1, BC_IABS, depth-1, A_INFO);
	  break;
	CASE(BC_INEG)
	  A_INFO = bcSetOp1 (depth-1, BC_INEG, depth-1, A_INFO);
	  break;
	CASE(BC_ISQR)
	  A_INFO = bcSetOp1 (depth-1, BC_ISQR, depth-1, A_INFO);
	  break;
	CASE(BC_INOT)
	  A_INFO = bcSetOp1 (depth-1, BC_INOT, depth-1, A_INFO);
	  break;
	CASE(BC_ILT)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_ILE)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IGT)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IGE)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IMIN)
	  A_INFO = bcSetOp2 (depth-2, BC_IMIN, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMAX)
	  A_INFO = bcSetOp2 (depth-2, BC_IMAX, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RLT)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RLE)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RGT)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RGE)
	  POPTOP_IMM(2);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RMIN)
	  A_INFO = bcSetOp2 (depth-2, BC_RMIN, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RMAX)
	  A_INFO = bcSetOp2 (depth-2, BC_RMAX, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_FLOOR)
	  A_INFO = bcSetOp1 (depth-1, BC_FLOOR, depth-1, A_INFO);
	  break;
	CASE(BC_CEIL)
	  A_INFO = bcSetOp1 (depth-1, BC_CEIL, depth-1, A_INFO);
	  break;
	CASE(BC_AREF)
	  A_INFO = bcSetOp2 (depth-2, BC_AREF, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ASET)
	  /* We do not record the side effect; this means we should not
	     pay too much attention to contents of AREFs, unless we do some
	     alias analysis first.
	  */
	  POPTOP_IMM(3);
	  break;
	CASE(BC_IAREF)
	  A_INFO = bcSetOp2 (depth-2, BC_IAREF, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IASET)
	  /* We do not record the side effect; this means we should not
	     pay too much attention to contents of IAREFs, unless we do some
	     alias analysis first.
	  */
	  POPTOP_IMM(3);
	  break;
	CASE(BC_NUMINT)
	  A_INFO = bcSetOp1 (depth-1, BC_NUMINT, depth-1, A_INFO);
	  break;
	CASE(BC_INTNUM)
	  A_INFO = bcSetOp1 (depth-1, BC_INTNUM, depth-1, A_INFO);
	  break;
	CASE(BC_SIZE)
	  A_INFO = bcSetOp1 (depth-1, BC_SIZE, depth-1, A_INFO);
	  break;
	CASE(BC_ORD)
	  A_INFO = bcSetOp1 (depth-1, BC_ORD, depth-1, A_INFO);
	  break;
	CASE(BC_ORDOF)
	  A_INFO = bcSetOp2 (depth-2, BC_ORDOF, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_SUBSTR)
	  A_INFO = bcSetOp3 (depth-3, BC_SUBSTR, depth-3, depth-2, depth-1, A_INFO);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_CALLBIND)
	  SYNCTOP();
	  pc++;
	  POPTOP_IMM(1);
	  A_INFO = bcSet (depth-1, A_VOLATILE, A_INFO);
	  nCalls++;
	  /* We consider we do not know what CALLBIND returns;
	     although it must be a Boolean. */
	  break;
	CASE(BC_ILLEGAL) goto illegal;
	default:
	  if (op>=BC_HASTAG && op<BC_HASTAG+E_UNUSED)
	    {
#ifdef DEBUG_ANLZ
	      hfprintf (hstderr, "%d: HASTAG %d =>", off, op-BC_HASTAG);
#endif
	      dpc = 2;
	      goto _goto;
	    }
	illegal:
	  kinderror ("bcAnalyze", op);
	  break;
	}
      A_NEXTINFO = bcNextInfo (depth, pc-bc->instr); /* prepare nextinfo for next turn. */
      A_INFO = MapDRT (A_INFO, A_NEXTINFO); /* to take care of data possibly dropped from stack. */
#ifdef DEBUG_ANLZ
      hfprintf (hstderr, "\t[depth=%d]", depth);
      bcPrtInfo (hstderr, A_INFO, depth);
      hfprintf (hstderr, "\n\tNEXTINFO:\t");
      bcPrtInfo (hstderr, A_NEXTINFO, depth);
      hfprintf (hstderr, "\n");
      hfflush (hstderr);
#endif
      QUIT;
#undef N
    }
#ifdef OBSOLETE
  if (excdepth!=0 || mapdepth!=0)
    {
      hfprintf (hstderr, "bcAnalyze: inconsistency, excdepth=%d, mapdepth=%d, depth=%d.\n",
		excdepth, mapdepth, depth);
      hfflush (hstderr);
      abort ();
    }
#endif

#undef CASE
#undef DO_GOTO

#undef SYNCTOP
#undef PUSHTOP_IMM
#undef POPTOP_IMM
  {
    ExprPtr e, aux;

    e = NIL;
    aux = KINT (nCalls);
    e = MakePair (E_CONS, aux, e);
    KEEPONLY (e);
    e = MakePair (E_CONS, A_SEEN, e);
    KEEPONLY (e);
    aux = KINT (maxexcdepth);
    e = MakePair (E_CONS, aux, e);
    KEEPONLY(e);
    e = MakePair (E_CONS, A_LABS, e);
    KEEPONLY(e);
    e = ListToArray0 (e);
    KEEPONLY (e);
    /* return [| labs, maxexcdepth, seen, ncalls |] */
    theStackFrames = _x.next;
    return e;
  }
}

PRIVATE ExprPtr bcTopVal0 (vals, off, i)
     IN CONST MapPtr vals;
     IN COUNTER off, i;
{
  ExprPtr eoff, ei, e;
  Boolean found;
  ENTER;

  eoff = KINT (off);
  e = MapGet (eoff, vals, &found);
  if (found)
    {
      e = CDR (e);
      ei = KINT (i);
      e = MapGet (ei, e, &found);
      if (!found)
	e = NOEXPR;
    }
  else e = NOEXPR;
  QUIT;
  return e;
}

PRIVATE ExprPtr bcTopVal (vals, off, i)
     IN CONST MapPtr vals;
     IN COUNTER off, i;
{
  ExprPtr e;
  Boolean found;
  ENTER;

 again:
  e = bcTopVal0 (vals, off, i);
  if (e==NOEXPR)
    {
      QUIT;
      return e;
    }
  if (M_ISVAR(e))
    {
      off = M_VAR_OFF(e);
      i = M_VAR_I (e);
      goto again;
    }
  QUIT;
  return e;
}

PRIVATE ExprPtr whnf (vals, val) /* reduce at the top, using only rules of the form x := r
				    with x a variable (i.e., reduces until we reach something
				    that is not a variable). */
     IN CONST MapPtr vals;
     IN CONST ExprPtr val;
{
  if (M_ISVAR (val))
    return bcTopVal (vals, M_VAR_OFF(val), M_VAR_I(val));
  return val;
}

PRIVATE COUNTER findSameVal (vals, off, maxdepth, val) /* Try to find some variable _x.p[i]
							  with 0 <= i < maxdepth
							  that has the same value as val. */
     IN CONST MapPtr vals;
     IN COUNTER off, maxdepth;
     IN CONST ExprPtr val;
{
  COUNTER i;

  val = whnf (vals, val);
  if (m_volatile(val))
    return -1;
  for (i=maxdepth; --i>=0; )
    if (val==bcTopVal (vals, off, i))
      return i;
  return -1;
}

PRIVATE COUNTER gcProtected (vals, off, maxdepth, val)
/* Try to find some variable _x.p[i] with 0<= i < maxdepth
   whose value is val, or contains val as a (direct) component.
   Then return i.
   Or discover that val is itself protected (e.g., a global constant), then return maxdepth.
   Otherwise return -1.
*/
     IN CONST MapPtr vals;
     IN COUNTER off, maxdepth;
     IN CONST ExprPtr val;
{
  COUNTER i;
  ExprPtr var;

  val = whnf (vals, val);
  if (M_ISGLOB (val) || M_ISCST (val) || M_ISINP (val) || M_ISVARREF (val))
    return maxdepth;
  i = findSameVal (vals, off, maxdepth, val);
  if (i>=0)
    return i;
  if (val!=NOEXPR && val->kind==(CHAR)E_DATA)
    switch (val->magic)
      {
      case M_NTH:
      case M_SELECT:
	var = CDR(val);
	break;
      case M_OP:
	switch (IVAL(CAR(val)))
	  {
	  case BC_CAR: case BC_CDR:
	    /*case BC_REFVAL: case BC_AREF: case BC_IAREF:
	      useless since the parent (ref, array) can be modified
	      at any time. */
	    var = CDR(val);
	    break;
	  case BC_MAPGET: case BC_NCARGET:
	    var = CAR(CDR(val));
	    break;
	  default:
	    return -1;
	  }
      default:
	return -1;
      }
  else return -1;
}

PRIVATE COUNTER bcFindCar (vals, off, depth, j)
     IN CONST MapPtr vals;
     IN COUNTER off, depth, j;
{
#ifdef OPT
  ExprPtr consval = bcTopVal (vals, off, j);
  COUNTER i;
  COUNTER op;

  if (!M_ISOP2(consval))
    return -1;
  op = M_OP2_I(consval);
  if (op!=BC_CONS && op!=BC_KONS && op!=BC_KTUPLE2 && op!=BC_MAKETUPLE2
      && op!=BC_KRECORD1 && op!=BC_MAKERECORD1 && op!=BC_PACK)
    return -1;
  i = findSameVal (vals, off, depth, M_OP2_CAR(consval));
  return i;
#else
  return -1;
#endif
}

PRIVATE COUNTER bcFindCdr (vals, off, depth, j)
     IN CONST MapPtr vals;
     IN COUNTER off, depth, j;
{
#ifdef OPT
  ExprPtr consval = bcTopVal (vals, off, j);
  COUNTER i;
  COUNTER op;

  if (!M_ISOP2(consval))
    return -1;
  op = M_OP2_I(consval);
  if (op!=BC_CONS && op!=BC_KONS && op!=BC_KTUPLE2 && op!=BC_MAKETUPLE2
      && op!=BC_KRECORD1 && op!=BC_MAKERECORD1 && op!=BC_PACK)
    return -1;
  i = findSameVal (vals, off, depth, M_OP2_CDR(consval));
  return i;
#else
  return -1;
#endif
}

PRIVATE COUNTER bcFindNth (vals, off, depth, j, idx)
     IN CONST MapPtr vals;
     IN COUNTER off, depth, j, idx;
{
#ifdef OPT
  ExprPtr tupleval = bcTopVal (vals, off, j);
  COUNTER i;
  COUNTER op;

  if (!M_ISOP2(tupleval))
    return -1;
  op = M_OP2_I(tupleval);
  if (op!=BC_KTUPLEGNRL && op!=BC_MAKETUPLEGNRL && op!=BC_KRECORDGNRL && op!=BC_MAKERECORDGNRL)
    return -1;
  i = findSameVal (vals, off, depth, AVAL(M_OP2_CDR(tupleval))->expr[idx]);
  return i;
#else
  return -1;
#endif
}

PRIVATE ExprPtr regAlloc (bc, vals)
     IN CONST byteCodeChunk *bc;
     IN CONST MapPtr vals;
{
#define N 2
#define RA_REG _x.p[0]
  /* maps each off to set of positions j on stack that are protected
     against GC, at off and at all later offsets as long as position
     j exists (has not been popped off).
     RA_REG(off) = RA_PROT(off) &
     inter_{off' / off -> off'} [depth(off'), +infty[ U RA_REG(off')

     Can be compiled as Horn clauses.  Create a variable
     gc[off,j] to say that j is not in RA_REG(off).
     Then:
     gc[off,j] <= "j is not in RA_PROT(off)"   (where "..." is a Boolean constant).
     For every off' such that off -> off', and j < depth(off')
     gc[off,j] <= gc[off',j]
     
  */
#define RA_PROT _x.p[1]
  /* same, except only counting those that are protected right now.
   Computed using gcProtected(). */
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[N]; } _x;
  COUNTER depth, maxdepth;
  OPCODE op, *pc;
  COUNTER off, dpc;
  Boolean found;
  ENTER;

  RA_REG = NewMap();
  RA_PROT = NewMap();
  _x.n = N;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;

  /*!!!*/

  {
    ExprPtr e;

    e = RA_REG;
    KEEPONLY(e);
    theStackFrames = _x.next;
    return e;
  }
#undef N
}

PRIVATE COUNTER xlateGenSym = 0;
EXTERN MapPtr bcXlateMap;
EXTERN COUNTER nbcXlate;

  /* Try and guess a name for current function name.
   Return COUNTED(name, serial number). */
PUBLIC ExprPtr bcXName(bc)
     IN CONST byteCodeChunk *bc;
{
  ExprPtr name, e;
  CHAR *buf, *t;
  Boolean found;
  ENTER;

  name = MapGet ((ExprPtr)bc, bcXlateMap, &found);
  if (found && COUNTEDVAL(name)!=NIL)
    return name;
  buf = NULL;
  if (bc->nConsts>=1 && (e=bc->consts[0], e!=NOEXPR && e->kind==(CHAR)BUILD_HEADER_KIND))
    {
      e = CAR(ARGS(e));
      if (e->kind==(CHAR)E_STRING)
	{
	  CHAR *s, *tend, c;

	  buf = (CHAR *)alloctmp_anchored (2*SLEN(e)+48, (char *)&found);
	  s = buf;
	  *s++ = '_';
	  *s++ = '_';
	  for (t = STEXT(e), tend = t+SLEN(e); t<tend; )
	    {
	      c = *t++;
	      switch ((int) c)
		{
		case '_': *s++ = '_'; *s++ = c; break;
		case '*': *s++ = '_'; *s++ = 's'; break;
		case '!': *s++ = '_'; *s++ = 'b'; break;
		case '%': *s++ = '_'; *s++ = 'p'; break;
		case '&': *s++ = '_'; *s++ = 'a'; break;
		case '$': *s++ = '_'; *s++ = 'd'; break;
		case '#': *s++ = '_'; *s++ = 'h'; break;
		case '+': *s++ = '_'; *s++ = 'P'; break;
		case '-': *s++ = '_'; *s++ = 'M'; break;
		case '/': *s++ = '_'; *s++ = 'D'; break;
		case ':': *s++ = '_'; *s++ = 'c'; break;
		case '<': *s++ = '_'; *s++ = 'L'; break;
		case '=': *s++ = '_'; *s++ = 'e'; break;
		case '>': *s++ = '_'; *s++ = 'G'; break;
		case '?': *s++ = '_'; *s++ = 'q'; break;
		case '@': *s++ = '_'; *s++ = 'A'; break;
		case '\\': *s++ = '_'; *s++ = 'B'; break;
		case '~': *s++ = '_'; *s++ = 'n'; break;
		case '\'': *s++ = '_'; *s++ = 't'; break;
		case '^': *s++ = '_'; *s++ = 'x'; break;
		case '|': *s++ = '_'; *s++ = 'o'; break;
		case '`': *s++ = '_'; *s++ = 'z'; break;
		default: *s++ = c; break;
		}
	    }
	  t = s;
	}
    }
  if (buf==NULL)
    {
      buf = (CHAR *)alloctmp_anchored (48, (char *)&found);
      strcpy ((char *)buf, "__himml_");
      t = buf+strlen((char *)buf);
    }
  sprintf ((char *)t, "_%d", ++xlateGenSym);
  if (!found)
    {
      name = MakeCounted (E_COUNTED, NIL, nbcXlate++);
      (void) KEEP (name);
      setMapPut ((ExprPtr) bc, name, &bcXlateMap);
    }
  e = Kstr (buf);
  SETREFVAL (name, e);
  QUIT;
  return name;
}

PUBLIC void bcXlate0(out, bc)
     IN hFILE *out;
     IN CONST byteCodeChunk *bc;
{
#define NX 11
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NX]; } _x;
#define BC_NAME _x.p[0]
#define BC_SEEN _x.p[1] /* : int -m> unit */
#define BC_OFF _x.p[2] /* : int */
#define BC_OFFSTK _x.p[3] /* : int list */
#define BC_DEPTH _x.p[4] /* : int */
#define BC_DEPTHSTK _x.p[5] /* : int list, same length as BC_OFFSTK */
#define BC_MAPDEPTH _x.p[6] /* : int */
#define BC_MAPDEPTHSTK _x.p[7] /* : int list, same length as BC_OFFSTK */
#define BC_EXCINFO _x.p[8] /* : (int * int) list, map exception number to offset in code. */
#define BC_LABS _x.p[9] /* : int set, set of goto targets. */
#define BC_VALS _x.p[10] /* : int -m> int * (int -m> var expr set)
			map each offset to a mutable pair:
			- the depth l at this offset;
			- an A_INFO map from levels (0..l-1) to their values,
			as in the description of 'mean' above. */
  CHAR *buf, *t;
  OPCODE op, *pc;
  COUNTER off;
  Boolean found;
  COUNTER depth, mapdepth;
  COUNTER dpc;
  COUNTER nMaps, maxexcdepth;
  COUNTER nHandlers = 0, nRaises = 0;
  COUNTER nCalls;
  Boolean heap_x;
  ENTER;

  _x.n = 0;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;

  BC_NAME = bcXName (bc);
  _x.n = 1;
  BC_NAME = COUNTEDVAL(BC_NAME);
  buf = (CHAR *)alloctmp_anchored (SLEN(BC_NAME)+1, (char *)&_x);
  memcpy (buf, STEXT(BC_NAME), SLEN(BC_NAME));
  buf[SLEN(BC_NAME)] = 0;
  BC_SEEN = NewMap();
  BC_OFF = NIL;
  BC_OFFSTK = NIL;
  BC_DEPTH = NIL;
  BC_DEPTHSTK = NIL;
  BC_MAPDEPTH = NIL;
  BC_MAPDEPTHSTK = NIL;
  BC_EXCINFO = NIL;
  BC_LABS = NewMap ();
  BC_VALS = NewMap ();
  _x.n = NX;

  {
    ExprPtr e = bcAnalyze (bc);

    BC_LABS = AVAL(e)->expr[0];
    maxexcdepth = IVAL(AVAL(e)->expr[1]);
    BC_VALS = AVAL(e)->expr[2];
    nCalls = IVAL(AVAL(e)->expr[3]);
  }

  hfprintf (out, "PUBLIC ExprPtr %s_consts[%d];\n", buf, bc->nConsts); /* The function's constant pool. */
  if (nCalls>0)
    {
      hfprintf (out, "DEF_PRIVATE_ACTION_1ARG(%s)\n{\n  ExprPtr res, *data = (ExprPtr *)ENV_UNDERLYING(env);\n  M_VARCALL_CC(res,%s,data[0],ENV_ARITY(_env),data); return res;\n}\nACTION_END\n",
		buf, buf); /* Install CallCC trampoline. */
    }
  hfprintf (out, "PUBLIC FUN_HEADER(%s)\n{\n",
	    buf);
  heap_x = (bc->stackSize*sizeof(ExprPtr) > SECURITY/2);
  if (heap_x)
    hfprintf (out, "  REGISTER gcStackView *_p;\n#define _x (*_p)\n");
  else
    hfprintf (out, "  struct _x { COUNTER n; gcStackView *next; ExprPtr p[%d]; } _x;\n",
	      bc->stackSize);

  nMaps = (bc->excStackOff - bc->mapStackOff) / sizeof (ExprPtr *); /* twice the max number of nestings
								       of map iterations; each one consumes
								       two pointers. */
  if (nMaps > 0)
    hfprintf (out, "  ExprPtr * VOLATILE maptop[%d];\n", nMaps);
  if (maxexcdepth>0)
    {
      hfprintf (out, "  struct xhandlerinfo {\n    VOLATILE struct xhandler *curxhandler, *endxhandler;\n    struct xhandler exc[%d];\n  } xhi;\n\n  xhi.curxhandler = xhi.endxhandler = NULL;\n",
		maxexcdepth);
    }

  /* Add CallCC() trampoline to prevent stack starvation. */
  if (nCalls>0)
    {
      hfprintf (out, "  if (STKFULL(env))\n    {\n      COUNTER i, n = ENV_ARITY(env);\n      ExprPtr *data = alloctmp ((n+1)*sizeof(ExprPtr));\n      ExprPtr res;\n      va_list _args;\n");
      hfprintf (out, "      data[0] = env; VA_START; for (i=0; i<n; ) data[++i] = va_arg(_args,ExprPtr); va_end(_args);\n");
      hfprintf (out, "      res = CallCC(ACTION_NAME(%s),(ExprPtr)data,true);\n      freetmp(data);\n      return res;\n    }\n",
		buf);
    }

  if (heap_x)
    hfprintf (out, "  _p = alloctmp_anchored (OFFSETOF(gcStackView,p[%d]), (char *)&env);\n",
	      bc->stackSize);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
  mapdepth = 0;
  if (bc->nargs==0)
    {
      hfputs ("\n  _x.n = 1; _x.p[0] = ENV_UNDERLYING(env);\n", out);
      depth = 1;
    }
  else if (bc->nargs==1) /* regular unary function */
    {
      hfputs ("\n  _x.n = 2; _x.p[0] = ENV_UNDERLYING(env); ", out);
      if (bc->ntargs==0)
	{
#ifdef EMIT_ASSERT
	  hfputs ("assert(ENV_ARITY(env)<=1); ", out);
#endif
	  hfputs ("_x.p[1] = tuple0;\n", out);
	}
      else if (bc->ntargs==1)
	{
#ifdef EMIT_ASSERT
	  hfputs ("assert(ENV_ARITY(env)==1); ", out);
#endif
	  hfputs ("{ va_list _args; VA_START; _x.p[1] = va_arg(_args,ExprPtr); va_end(_args); }\n", out);
	}
      else hfputs ("{ FUN_COLLECT(FC_IGN); _x.p[1] = arg; }\n", out);
      /* Warning: because of FC_IGN, when 2 or more arguments are passed and are collected
	 as a tuple into _x.p[1], _x.p[1] is not protected against GC until _x is linked
	 into theStackFrames below. */
      depth = 2;
    }
  else
    {
      COUNTER i;

      hfprintf (out, "\n  _x.n = %d; _x.p[0] = ENV_UNDERLYING(env);", bc->nargs+1);
      hfprintf (out, "  {\n    va_list _args;\n    VA_START;\n    if (ENV_ARITY(env)==%d) {\n   ",
		bc->nargs);
      for (i=1; i<=bc->nargs; i++)
	hfprintf (out, " _x.p[%d] = va_arg(_args,ExprPtr);", i);
      hfprintf (out, "\n    }\n    else {\n      ExprPtr _arg = va_arg(_args,ExprPtr);\n");
      if (bc->nargs==2)
	hfprintf (out, "      _x.p[1] = CAR(_arg); _x.p[2] = CDR(_arg);");
      else
	{
	  hfprintf (out, "      VecPtr vp = AVAL(_arg);\n     ");
	  for (i=1; i<=bc->nargs; i++)
	    hfprintf (out, " _x.p[%d] = vp->expr[%d];", i, i-1);
	}
      hfprintf (out, "\n    }\n    va_end(_args);\n  }\n");
      depth = bc->nargs+1;
    }
  if (heap_x)
    hfputs ("  _x.next = theHeapFrames;\n  theHeapFrames = _p;\n", out);
  else
    hfputs ("  _x.next = theStackFrames;\n  theStackFrames = (gcStackView *)&_x;\n", out);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif

#ifdef EMIT_ASSERT
#define ASSERTTOP() { hfprintf (out, " assert(_x.n==%d);\n", depth); }
#else
#define ASSERTTOP() {}
#endif

#ifdef OBSOLETE
#define POPTOP_IMM(n) { depth -= n; if (n!=0) { hfprintf (out, "  _x.n -= %d;", n); ASSERTTOP(); } }
#define PUSHTOP_IMM(n) { depth += n; if (n!=0) { hfprintf (out, "  _x.n += %d;", n); ASSERTTOP(); } }
#define SYNCTOP() { ASSERTTOP(); }
#endif

#define POPTOP_IMM(n) { depth -= n; if (n!=0) { if (depth<8) hfprintf (out, "  _x.n = %d;", depth); else hfprintf (out, "  _x.n -= %d;", n); } }
#define PUSHTOP_IMM(n) { depth += n; if (n!=0) { if (depth<8) hfprintf (out, "  _x.n = %d;", depth); else hfprintf (out, "  _x.n += %d;", n);  } }
#define SYNCTOP() { ASSERTTOP(); }

  pc = bc->instr;
  for (;;)
    {
      off = pc-bc->instr;
      BC_OFF = Kint(off);
      (void) MapGet (BC_OFF, BC_SEEN, &found);
      if (found)
	{
	  hfprintf (out, "  goto lab%d;\n", off);
	pop:
	  if (BC_OFFSTK==NIL)
	    break;
	  BC_OFF = CAR(BC_OFFSTK);
	  BC_OFFSTK = CDR(BC_OFFSTK);
	  off = IVAL(BC_OFF);
	  pc = bc->instr+off;
	  BC_DEPTH = CAR(BC_DEPTHSTK);
	  BC_DEPTHSTK = CDR(BC_DEPTHSTK);
	  depth = IVAL(BC_DEPTH);
	  BC_MAPDEPTH = CAR(BC_MAPDEPTHSTK);
	  BC_MAPDEPTHSTK = CDR(BC_MAPDEPTHSTK);
	  mapdepth = IVAL(BC_MAPDEPTH);
	  continue;
	}
      (void) MapGet (BC_OFF, BC_LABS, &found);
      if (found)
	hfprintf (out, "lab%d:", off);
#define CASE(op) case op: hfputs (" /* " #op " */\n", out); MERGE(lab_,op)
      BC_SEEN = MapAdd (BC_OFF, UNIT, BC_SEEN);
      op = *pc++;
      switch (op)
	{
	CASE(BC_LOOP): break;
	CASE(BC_FUNCTION):
	  hfputs ("#if OS==AMIGA_OS\n  if (++__spin_loop==32) { __spin_loop=0; chkabort(); }\n#endif\n", out);
	  break;
	CASE(BC_MERGE): break;
	CASE(BC_BREAK): break;
	CASE(BC_GOTO):
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    POPTOP_IMM(n);
	  }
	  pc += *pc;
	  break;
#ifdef WRONG_GOTO
	  hfprintf (out, "  goto lab%d;\n", pc-bc->instr);
	  goto pop;
#endif
	CASE(BC_GOTO_0):
	  pc += pc[1];
	  break;
#ifdef WRONG_GOTO
	  hfprintf (out, "  goto lab%d;\n", pc-bc->instr);
	  goto pop;
#endif
	CASE(BC_IF):
	  POPTOP_IMM(1);
	  hfprintf (out, "  if (_x.p[%d]==falseValue) {\n", depth);
	  dpc = 2;
#define DO_GOTO(xdrop)				\
	  {					\
	    REGISTER COUNTER n;			\
	    REGISTER COUNTER newoff;		\
	    ExprPtr saved;			\
	    					\
	    n = (COUNTER)pc[0];				      \
	    POPTOP_IMM(n+xdrop);			      \
	    newoff = (pc + 1 + pc[1]) - bc->instr;		      \
	    hfprintf (out, "    goto lab%d;\n  }\n", newoff); \
	    saved = KINT (newoff);			      \
	    BC_OFFSTK = MakePair (E_CONS, saved, BC_OFFSTK);  \
	    BC_DEPTH = Kint (depth);				    \
	    BC_DEPTHSTK = MakePair (E_CONS, BC_DEPTH, BC_DEPTHSTK); \
            BC_MAPDEPTH = Kint (mapdepth);			    \
            BC_MAPDEPTHSTK = MakePair (E_CONS, BC_MAPDEPTH, BC_MAPDEPTHSTK); \
	    depth += n+xdrop;						\
	    QUIT;						    \
	  }

	_goto:
	  DO_GOTO(0);
	  pc += dpc;
	  break;
	CASE(BC_CGLOBAL):
	  hfprintf (out, "  _x.p[%d] = CGlobals[%d];\n", depth, *pc++);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CONST):
	  hfprintf (out, "  _x.p[%d] = %s_consts[%d];\n", depth, buf, *pc++);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_GLOBVARREF):
	  hfprintf (out, "  _x.p[%d] = EVARS(GLOBENV)->expr[%d];\n", depth, *pc++);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_DROP):
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_SWDROP):
	  {
	    REGISTER COUNTER n;

	    n = (COUNTER)*pc++;
	    hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth+n, depth-1);
	    n++;
	    POPTOP_IMM((-n));
	    break;
	  }
	CASE(BC_NSWDROP):
	  {
	    REGISTER COUNTER n, k;

	    n = (COUNTER)*pc++;
	    k = (COUNTER)*pc++;
	    for (; k<0; n++, k++)
	      hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth+n, depth+k);
	    POPTOP_IMM((-n));
	    break;
	  }
	CASE(BC_REF):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakeSing (E_REF, _x.p[%d]);\n", depth-1, depth-1);
	  break;
	CASE(BC_KTUPLE1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Ktuple1 (_x.p[%d]);\n", depth-1, depth-1);
	  break;
	CASE(BC_MAKETUPLE1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakeSing (E_UNSHARED_TUPLE_1, _x.p[%d]);\n", depth-1, depth-1);
	  break;
	CASE(BC_KONS):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Kons (_x.p[%d], _x.p[%d]);\n", depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_CONS):
	CASE(BC_CONSMEMO):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakePair (E_CONS, _x.p[%d], _x.p[%d]);\n", depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_KDATA):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Kdata ((SHORT)%d, %s_consts[%d], _x.p[%d]);\n",
		    depth-1, pc[0], buf, pc[1], depth-1);
	  pc += 2;
	  break;
	CASE(BC_DATA):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakeData ((SHORT)%d, %s_consts[%d], _x.p[%d]);\n",
		    depth-1, pc[0], buf, pc[1], depth-1);
	  pc += 2;
	  break;
	CASE(BC_FETCH): goto fetch;
	CASE(BC_STARTLISTSWEEP): goto fetch;
	CASE(BC_GETMAPSWEEP): goto fetch;
	CASE(BC_GETSUBMAPSWEEP):
	fetch:
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth + (*pc++));
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m8):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-8);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m7):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-7);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m6):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-6);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m5):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-5);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m4):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-4);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m3):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-3);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m2):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-2);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCH_m1):
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth, depth-1);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CAR):
	  {
	    COUNTER i = bcFindCar (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, depth-1);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth-1, depth-1);
	  }
	  break;
	CASE(BC_FETCHCAR):
	  {
	    COUNTER j = depth + (*pc++);
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m8):
	  {
	    COUNTER j = depth - 8;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m7):
	  {
	    COUNTER j = depth - 7;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m6):
	  {
	    COUNTER j = depth - 6;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m5):
	  {
	    COUNTER j = depth - 5;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m4):
	  {
	    COUNTER j = depth - 4;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m3):
	  {
	    COUNTER j = depth - 3;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m2):
	  {
	    COUNTER j = depth - 2;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCAR_m1):
	  {
	    COUNTER j = depth - 1;
	    COUNTER i = bcFindCar (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CAR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CAR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_CDR):
	  {
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, depth-1);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth-1, depth-1);
	  }
	  break;
	CASE(BC_FETCHCDR):
	  {
	    COUNTER j = depth + (*pc++);
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m8):
	  {
	    COUNTER j = depth - 8;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m7):
	  {
	    COUNTER j = depth - 7;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m6):
	  {
	    COUNTER j = depth - 6;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m5):
	  {
	    COUNTER j = depth - 5;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m4):
	  {
	    COUNTER j = depth - 4;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m3):
	  {
	    COUNTER j = depth - 3;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m2):
	  {
	    COUNTER j = depth - 2;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_FETCHCDR_m1):
	  {
	    COUNTER j = depth - 1;
	    COUNTER i = bcFindCdr (BC_VALS, off, depth, j);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==CDR(_x.p[%d]));\n", i, j);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* CDR optimized away */\n", depth, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = CDR(_x.p[%d]);\n", depth, j);
	    PUSHTOP_IMM(1);
	  }
	  break;
	CASE(BC_CARCDR):
	  hfprintf (out, "  {\n    REGISTER ExprPtr a;\n    a = _x.p[%d]; _x.p[%d] = CAR(a); _x.p[%d] = CDR(a);\n  }\n",
		    depth-1, depth-1, depth);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CDRCAR):
	  hfprintf (out, "  {\n    REGISTER ExprPtr a;\n    a = _x.p[%d]; _x.p[%d] = CDR(a); _x.p[%d] = CAR(a);\n  }\n",
		    depth-1, depth-1, depth);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_REFVAL):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth-1, depth-1);
	  break;
	CASE(BC_FETCHREFVAL):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth + (*pc++));
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m8):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-8);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m7):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-7);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m6):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-6);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m5):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-5);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m4):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-4);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m3):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-3);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m2):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-2);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_FETCHREFVAL_m1):
	  hfprintf (out, "  _x.p[%d] = REFVAL(_x.p[%d]);\n", depth, depth-1);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_KRECORD1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Krecord1(%s_consts[%d], _x.p[%d]);\n",
		    depth-1, buf, *pc++, depth-1);
	  break;
	CASE(BC_MAKERECORD1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakePair(E_RECORD_1, %s_consts[%d], _x.p[%d]);\n",
		    depth-1, buf, *pc++, depth-1);
	  break;
	CASE(BC_KTUPLE2):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Ktuple2 (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MAKETUPLE2):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakePair (E_TUPLE_2, _x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_KTUPLEGNRL):
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    hfprintf (out, "  _x.p[%d] = KtupleGnrl (%d, &_x.p[%d]);\n",
		      depth-n, n, depth-n);
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAKETUPLEGNRL):
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    hfprintf (out, "  _x.p[%d] = MakeRecord (E_TUPLE_GNRL, NIL, %d, &_x.p[%d]);\n",
		      depth-n, n, depth-n);
	    /* Here it is important that the arguments are actually put in _x.p[]! */
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_KRECORDGNRL):
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    hfprintf (out, "  _x.p[%d] = KrecordGnrl (%s_consts[%d], %d, &_x.p[%d]);\n",
		      depth-n, buf, *pc++, n, depth-n);
	    /* Here it is important that the arguments are actually put in _x.p[]!
	     Also, KrecordGnrl() should not CallCC(), otherwise _x might be swapped out! */
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAKERECORDGNRL):
	  SYNCTOP();
	  {
	    REGISTER INT n;

	    n = (INT)*pc++;
	    hfprintf (out, "  _x.p[%d] = MakeRecord (E_RECORD_GNRL, %s_consts[%d], %d, &_x.p[%d]);\n",
		      depth-n, buf, *pc++, n, depth-n);
	    /* Here it is important that the arguments are actually put in _x.p[]!
	     Also, MakeRecord() should not CallCC(), otherwise _x might be swapped out!  */
	    n--;
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_MAPLET):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapletKons (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_OVERWRITE):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapOverwrite (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_UNDERWRITE):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapOverwrite (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-1, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_VARREF0):
	  hfprintf (out, "  _x.p[%d] = EVARS(_x.p[%d])->expr[%d];\n",
		    depth, depth+pc[0], pc[1]);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF1):
	  hfprintf (out, "  _x.p[%d] = EVARS(ENEXT(_x.p[%d]))->expr[%d];\n",
		    depth, depth+pc[0], pc[1]);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF2):
	  hfprintf (out, "  _x.p[%d] = EVARS(ENEXT(ENEXT(_x.p[%d])))->expr[%d];\n",
		    depth, depth+pc[0], pc[1]);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF3):
	  hfprintf (out, "  _x.p[%d] = EVARS(ENEXT(ENEXT(ENEXT(_x.p[%d]))))->expr[%d];\n",
		    depth, depth+pc[0], pc[1]);
	  pc += 2;
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_VARREF):
	  {
	    REGISTER COUNTER n;

	    hfprintf (out, "  {\n    ExprPtr env = _x.p[%d];\n", depth + (*pc++));
	    n = (COUNTER)*pc++;
	    for (; n--!=0; )
	      hfputs ("    env = ENEXT(env);\n", out);
	    hfprintf (out, "    _x.p[%d] = EVARS(env)->expr[%d];\n  }\n",
		      depth, *pc++);
	    PUSHTOP_IMM(1);
	    break;
	  }
	CASE(BC_CLOSURE):
	  SYNCTOP();
	  /* In the interpreted world, bc->consts[*pc++]
	     is an E_TUPLE_2 (bc, NIL).  We shouldn't care:
	     it should be overridden by the use of bcXlateMap.
	  */
	  {
	    ExprPtr fname;
	    byteCodeChunk *bcc;
	    Boolean found;
	    COUNTER i;
	    CHAR *s;

	    i = *pc++;
	    bcc = (byteCodeChunk *)CAR(bc->consts[i]);
	    fname = bcXName (bcc);
	    fname = COUNTEDVAL(fname);
	    (void) KEEP (fname);
	    s = alloctmp_anchored (SLEN(fname)+1, (char *)&_x);
	    memcpy (s, STEXT(fname), SLEN(fname));
	    s[SLEN(fname)] = 0;
	    hfprintf (out, "  _x.p[%d] = MakeCompiled (E_COMPILED_CLOSURE, %s, _x.p[%d]);\n",
		      depth-1, s, depth-1);
	    freetmp (s);
	    QUIT;
	    break;
	  }
	CASE(BC_CLOSURE0):
	  SYNCTOP();
	  {
	    ExprPtr fname;
	    byteCodeChunk *bcc;
	    Boolean found;
	    COUNTER i;
	    CHAR *s;

	    i = *pc++;
	    bcc = (byteCodeChunk *)CAR(bc->consts[i]);
	    fname = bcXName (bcc);
	    fname = COUNTEDVAL(fname);
	    (void) KEEP (fname);
	    s = alloctmp_anchored (SLEN(fname)+1, (char *)&_x);
	    memcpy (s, STEXT(fname), SLEN(fname));
	    s[SLEN(fname)] = 0;
	    hfprintf (out, "  _x.p[%d] = MakeCompiled (E_COMPILED_CLOSURE, %s, NIL);\n",
		      depth, s);
	    PUSHTOP_IMM(1);
	    freetmp (s);
	    QUIT;
	    break;
	  }
	CASE(BC_NOEXPR):
	  {
	    REGISTER COUNTER i;

	    i = (COUNTER)*pc++;
	    while (i--)
	      {
		hfprintf (out, "  _x.p[%d] = NOEXPR; ", depth);
		PUSHTOP_IMM(1);
	      }
	    break;
	  }
	CASE(BC_TUCK):
	  POPTOP_IMM(1);
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n",
		    depth + (*pc++), depth);
	  break;
	CASE(BC_FAKEENV):
	  SYNCTOP();
	  hfprintf (out, "  {\n    ExprPtr env;\n    MAKEENV1(EVARS(GLOBENV), _x.p[%d], env);\n",
		    depth-1);
	  hfprintf (out, "    _x.p[%d] = env;\n  }\n", depth-1);
	  break;
	CASE(BC_CONSENV):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = envCons ((COUNTER)%d, _x.p[%d]);\n",
		    depth-1, *pc++, depth-1);
	  break;
	CASE(BC_LET):
	  {
	    COUNTER k;

	    hfprintf (out, "  {\n    ExprPtr env = _x.p[%d];\n    REGISTER ExprPtr *ep = EVARS(env)->expr;\n",
		      depth + (*pc++));
	    while ((k = *pc++))
	      {
		hfprintf (out, "    ep[%d] = _x.p[%d];\n",
			  *pc++, depth+k);
	      }
	    hfputs ("    HITBARRIER(env);\n  }\n", out);
	    break;
	  }
	CASE(BC_LET1):
	  {
	    COUNTER k;

	    hfprintf (out, "  {\n    ExprPtr env = _x.p[%d];\n",
		      depth + (*pc++));
	    k = *pc++;
	    hfprintf (out, "    EVARS(env)->expr[%d] = _x.p[%d];\n    HITBARRIER(env);\n  }\n",
		      *pc++, depth+k);
	    break;
	  }
	CASE(BC_QENV):
	  hfprintf (out, "  {\n    ExprPtr e = _x.p[%d];\n",
		    depth-1);
	  hfprintf (out, "    _x.p[%d] = (e->kind==(CHAR)E_INTERPRETED_CLOSURE)?(CDR(QENV(e))):QENV(e);\n  }\n",
		    depth-1);
	  /* Useless if we don't have a bytecode interpreter! */
	  break;
	CASE(BC_IS):
	  hfprintf (out, "  if (_x.p[%d]!=%s_consts[%d]) {\n",
		    depth-1, buf, pc[2]);
	  dpc = 3;
	  goto _goto;
	CASE(BC_ISDATA):
	  hfprintf (out, "  if (_x.p[%d]->magic!=(SHORT)%d) {\n",
		    depth-1, pc[2]);
	  dpc = 3;
	  goto _goto;
	CASE(BC_SWITCH):
	  {
	    COUNTER m, n, i;

	    m = ((COUNTER)*pc++) - 2;
	    n = (COUNTER)*pc++;
	    hfprintf (out, "  switch (_x.p[%d]->magic) {\n", depth-1);
	    for (i=-m; i<=n; i++)
	      {
		REGISTER COUNTER newoff;
		ExprPtr saved;

		newoff = (pc + (COUNTER)*pc++) - bc->instr;
		hfprintf (out, "    case %d: goto lab%d;\n",
			  i, newoff);
		saved = KINT (newoff);
		BC_OFFSTK = MakePair (E_CONS, saved, BC_OFFSTK);
		BC_DEPTH = Kint (depth);
		BC_DEPTHSTK = MakePair (E_CONS, BC_DEPTH, BC_DEPTHSTK);
		BC_MAPDEPTH = Kint (mapdepth);
		BC_MAPDEPTHSTK = MakePair (E_CONS, BC_MAPDEPTH, BC_MAPDEPTHSTK);
		QUIT;
	      }
	    hfputs ("    default: assert(0); break;\n    }\n", out);
	    goto pop;
	  }
	CASE(BC_ISEQUAL):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (_x.p[%d]!=_x.p[%d]) {\n", depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_NTH):
	  {
	    COUNTER idx = *pc++;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_0):
	  {
	    COUNTER idx = 0;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_1):
	  {
	    COUNTER idx = 1;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_2):
	  {
	    COUNTER idx = 2;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_3):
	  {
	    COUNTER idx = 3;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_4):
	  {
	    COUNTER idx = 4;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_5):
	  {
	    COUNTER idx = 5;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_6):
	  {
	    COUNTER idx = 6;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_NTH_7):
	  {
	    COUNTER idx = 7;
	    COUNTER i = bcFindNth (BC_VALS, off, depth, depth-1);

	    if (i>=0)
	      {
#ifdef DEBUG_OPT
		hfprintf (out, "  assert (_x.p[%d]==AVAL(_x.p[%d])->expr[%d]);\n", i, depth-1, idx);
#endif
		hfprintf (out, "  _x.p[%d] = _x.p[%d]; /* NTH optimized away */\n", depth-1, i);
	      }
	    else
	      hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[%d];\n",
			depth-1, depth-1, idx);
	  }
	  break;
	CASE(BC_3ARGS):
	  hfprintf (out, "  {\n    ExprPtr a = _x.p[%d];\n", depth-1);
	  hfprintf (out, "    _x.p[%d] = AVAL(a)->expr[2];\n", depth+1);
	  hfprintf (out, "    _x.p[%d] = AVAL(a)->expr[1];\n", depth);
	  hfprintf (out, "    _x.p[%d] = AVAL(a)->expr[0];\n  }\n", depth-1);
	  PUSHTOP_IMM(2);
	  break;
	CASE(BC_NARGS):
	  {
	    COUNTER i, n = *pc++;

	    hfprintf (out, "  {\n    VecPtr vp = AVAL(_x.p[%d]);\n", depth-1);
	    for (i=0; i<n; i++)
	      hfprintf (out, "    _x.p[%d] = vp->expr[%d];\n", depth+i-1, i);
	    PUSHTOP_IMM(n-1);
	    hfprintf (out, "  }\n");
	    break;
	  }
	CASE(BC_TYPEMATCH):
	  hfprintf (out, "  if (TypeMatch (_x.p[%d], %s_consts[%d])) {\n",
		    depth-1, buf, pc[2]);
	  dpc = 3;
	  DO_GOTO(1);
	  POPTOP_IMM(1);
	  pc += dpc;
	  break;
	CASE(BC_CARDEQ):
	  hfprintf (out, "  if (MapCardDec (_x.p[%d], (INT)%d)!=0) {\n",
		    depth-1, pc[2]);
	  dpc = 3;
	  goto _goto;
	CASE(BC_CARDGEQ):
	  hfprintf (out, "  if (MapCardDec (_x.p[%d], (INT)%d)>0) {\n",
		    depth-1, pc[2]);
	  dpc = 3;
	  goto _goto;
	CASE(BC_SELECT):
	  hfprintf (out, "  _x.p[%d] = AVAL(_x.p[%d])->expr[IVAL(MapGetSure(%s_consts[%d], ASHAPE(_x.p[%d])))-1];\n",
		    depth-1, depth-1, buf, *pc++, depth-1);
	  break;
	CASE(BC_ISNEMAP):
	  hfprintf (out, "  if (_x.p[%d]==NewMap()) {\n",
		    depth-1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_MAPGET):
	  POPTOP_IMM(1);
	  hfprintf (out, "  {\n    Boolean found;\n    ExprPtr val = MapGet (_x.p[%d], _x.p[%d], &found);\n",
		    depth, depth-1);
	  hfprintf (out, "    if (!found) {\n");
	  dpc = 2;
	  DO_GOTO(0);
	  hfprintf (out, "    _x.p[%d] = val;\n  }\n", depth-1);
	  pc += dpc;
	  break;
	CASE(BC_NCARGET):
	  hfprintf (out, "  {\n    Boolean found;\n    ExprPtr val = NGet (_x.p[%d], CAR(_x.p[%d]), &found);\n",
		    depth-1, depth-2);
	  hfprintf (out, "    if (!found) {\n");
	  dpc = 2;
	  DO_GOTO(0);
	  POPTOP_IMM(1);
	  hfprintf (out, "    _x.p[%d] = val;\n  }\n", depth-1);
	  pc += dpc;
	  break;
	CASE(BC_NCARPUT):
	  hfprintf (out, "  setNPut (_x.p[%d], _x.p[%d], &CAR(_x.p[%d]), _x.p[%d]);\n",
		    depth-2, depth-1, depth-3, depth-3);
	  POPTOP_IMM(2);
	  hfprintf (out, "  _x.p[%d] = _x.p[%d];\n", depth-1, depth+1);
	  break;
	CASE(BC_MAPREMOVE):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapKill (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RETURN):
	  hfprintf (out, "  _x.p[0] = _x.p[%d];\n", depth-1);
	  if (maxexcdepth>0)
	    hfputs ("  if (xhi.curxhandler!=NULL) { excHandler *exc = excStack->next; freetmp (excStack); excStack = exc; }\n", out);
	  /* if we have installed an exception handler in excStack, pop it. */
	  hfputs ("  goto end;\n", out);
	  goto pop;
	CASE(BC_MAPADD):
	  hfprintf (out, "  _x.p[%d] = MapAdd (_x.p[%d], _x.p[%d], _x.p[%d]);\n",
		    depth-3, depth-3, depth-2, depth-1);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_MAPADD2):
	  hfprintf (out, "  _x.p[%d] = MapAdd (_x.p[%d], _x.p[%d], _x.p[%d]);\n",
		    depth-3, depth-2, depth-1, depth-3);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_SETCAR):
	  hfprintf (out, "  SETCAR(_x.p[%d], _x.p[%d]);\n", depth-1, depth-2);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_N_CALL):
	  {
	    COUNTER n = (COUNTER)*pc++;
	    COUNTER i;
	    ExprPtr funval;

	    funval = bcTopVal (BC_VALS, off, depth-n-1);
	    hfprintf (out, "  _x.p[%d] = ", depth-n-1);
#ifdef DEBUG_ANLZ
	    if (funval==NOEXPR)
	      hfprintf (hstderr, "BC_N_CALL, funval = NOEXPR.\n");
	    else
	      {
		hfprintf (hstderr, "BC_N_CALL, funval = ");
		bcPrtExpr (hstderr, funval);
		hfprintf (hstderr, ".\n");
	      }
	    hfflush (hstderr);
#endif
	    if (M_ISCLOS(funval))
	      {
		ExprPtr fname = bcXName ((byteCodeChunk *)CAR (bc->consts[M_CLOS_I(funval)]));
		CHAR *s;

		fname = COUNTEDVAL(fname);
		(void) KEEP (fname);
		s = alloctmp_anchored (SLEN(fname)+1, (char *)&_x);
		memcpy (s, STEXT(fname), SLEN(fname));
		s[SLEN(fname)] = 0;
		hfputs ((char *)s, out);
	      }
	    /* No need to check whether funval is in GLOBVAR, funNames[]: there
	       are no N_LAMBDAs in GLOBENV. */
	    else
	      {
		hfprintf (out, "(*ACTION(_x.p[%d]))",
			  depth-n-1);
	      }
	    hfprintf (out, " (ENV_CONS(QENV(_x.p[%d]),%d)", depth-n-1, n);
	    for (i=depth-n; i<depth; i++)
	      hfprintf (out, ", _x.p[%d]", i);
	    hfprintf (out, ");\n");
	    POPTOP_IMM(n);
	    break;
	  }
	CASE(BC_CALL):
	  {
	    ExprPtr funval;

	    funval = bcTopVal (BC_VALS, off, depth-2);
	    hfprintf (out, "  _x.p[%d] = ",
		      depth-2);
#ifdef DEBUG_ANLZ
	    if (funval==NOEXPR)
	      hfprintf (hstderr, "BC_CALL, funval = NOEXPR.\n");
	    else
	      {
		hfprintf (hstderr, "BC_CALL, funval = ");
		bcPrtExpr (hstderr, funval);
		hfprintf (hstderr, ".\n");
	      }
	    hfflush (hstderr);
#endif
	    if (M_ISCLOS(funval))
	      {
		ExprPtr fname = bcXName ((byteCodeChunk *)CAR (bc->consts[M_CLOS_I(funval)]));
		CHAR *s;

		fname = COUNTEDVAL(fname);
		(void) KEEP (fname);
		s = alloctmp_anchored (SLEN(fname)+1, (char *)&_x);
		memcpy (s, STEXT(fname), SLEN(fname));
		s[SLEN(fname)] = 0;
		hfputs ((char *)s, out);
	      }
	    else if (M_ISGLOB(funval) && funNames[M_GLOB_I(funval)]!=NULL)
	      {
		hfprintf (out, "ACTION_NAME(%s)", funNames[M_GLOB_I(funval)]);
	      }
	    else
	      hfprintf (out, "(*ACTION(_x.p[%d]))",
			depth-2);
	    hfprintf (out, "(ENV_CONS(QENV(_x.p[%d]),1), _x.p[%d]);\n",
		      depth-2, depth-1);
	    POPTOP_IMM(1);
	    break;
	  }
	CASE(BC_SETREF):
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  hfprintf (out, "  SETREFVAL(_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-1);
	  POPTOP_IMM(2);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  break;
	CASE(BC_KMU1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Kmu1 (_x.p[%d], %s_consts[%d], _x.p[%d]);\n",
		    depth-2, depth-2, buf, *pc++, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MU1):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = Mu1 (_x.p[%d], %s_consts[%d], _x.p[%d]);\n",
		    depth-2, depth-2, buf, *pc++, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_KMUGNRL):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = KmuGnrl (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MUGNRL):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MuGnrl (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_SWAP):
	  hfprintf (out, "  {\n    ExprPtr aux = _x.p[%d];\n    _x.p[%d] = _x.p[%d];\n    _x.p[%d] = aux;\n  }\n",
		    depth-1, depth-1, depth-2, depth-2);
	  break;
	CASE(BC_NEWQUEUE):
	  hfprintf (out, "  _x.p[%d] = NIL;\n", depth);
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_CONCATQUEUE):
	  hfprintf (out, "  {\n    REGISTER ExprPtr q = _x.p[%d];\n", depth + (*pc++));
	  hfprintf (out, "    if (q==NIL)\n      ;\n    else if (_x.p[%d]==NIL)\n      _x.p[%d] = q;\n    else\n      {\n        REGISTER ExprPtr e1 = CDR(q), e2 = CDR(_x.p[%d]);\n        SETCDR(q,e2);\n        SETCDR(_x.p[%d],e1);\n      }\n  }\n",
		    depth-1, depth-1, depth-1, depth-1);
	  break;
	CASE(BC_NEW1QUEUE):
	  SYNCTOP();
	  hfprintf (out, "  {\n    REGISTER ExprPtr e;\n    EALLOC_HOMO(e);\n    e->kind = (CHAR)E_CONS;\n    CAR(e) = _x.p[%d];\n    CDR(e) = e;\n    _x.p[%d] = e;\n  }\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_QUEUELIST):
	  SYNCTOP();
	  hfprintf (out, "  if (_x.p[%d]!=NIL)\n    _x.p[%d] = QueueToList (_x.p[%d]);\n",
		    depth-1, depth-1, depth-1);
	  break;
	CASE(BC_LISTQUEUE):
	  SYNCTOP();
	  hfprintf (out, "  if (_x.p[%d]!=NIL)\n    _x.p[%d] = ListToQueue (_x.p[%d]);\n",
		    depth-1, depth-1, depth-1);
	  break;
	CASE(BC_MAPINTER): goto mapdrt;
	CASE(BC_MAPDOMRESTRTO):
	mapdrt:
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapDRT (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-1, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_MAPDOMRESTRBY):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapDRB (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-1, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_DIFF):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapDRB (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_DELTA):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MapD (_x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-1, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_STARTMAPSWEEP):
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  hfprintf (out, "  {\n    REGISTER ExprPtr e = _x.p[%d], *starttop = &_x.p[%d], *top = starttop+1;\n",
		    depth + (*pc++), depth);
	  hfprintf (out, "    maptop[%d] = top;\n",
		    mapdepth);
	  hfprintf (out, "    while (e->kind!=(CHAR)E_MAPLET)\n      {\n        REGISTER ExprPtr aux = MAPLEFT(e);\n        if (aux==empty_map)\n          e = MAPRIGHT(e);\n        else\n          {\n            *top++ = MAPRIGHT(e);\n            e = aux;\n          }\n      }\n    *starttop = e;\n    maptop[%d] = top;\n",
		    mapdepth+1);
	  mapdepth += 2;
	  hfprintf (out, "    for (starttop += (1+MAP_CHAIN_LENGTH); top<starttop;) *top++ = NOEXPR;\n  }\n");
	  PUSHTOP_IMM(1+MAP_CHAIN_LENGTH);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  break;
	CASE(BC_STARTSUBMAPSWEEP):
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  hfprintf (out, "  {\n    ExprPtr map = _x.p[%d];\n    _x.p[%d] = NewMap();\n    _x.p[%d] = NIL;\n  ",
		    depth + (*pc++), depth, depth+1);
	  PUSHTOP_IMM(2);
	  hfprintf (out, "    (void) MapSweep (map, do_initialsubmap, (char *)&_x.p[%d]);\n  }\n",
		    depth-1);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  break;
	CASE(BC_GETLISTSWEEP):
	  hfprintf (out, "  _x.p[%d] = CAR(_x.p[%d]);\n",
		    depth, depth + (*pc++));
	  PUSHTOP_IMM(1);
	  break;
	CASE(BC_INCLISTSWEEP):
	  hfprintf (out, "  {\n    REGISTER ExprPtr *ep = &_x.p[%d];\n",
		    depth + pc[2]);
	  hfprintf (out, "    if ((*ep = CDR(*ep)) == NIL) {");
	  dpc = 3;
	  DO_GOTO(0);
	  pc += dpc;
	  hfputs ("  }\n", out);
	  break;
	CASE(BC_INCMAPSWEEP):
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  hfprintf (out, "  {\n    REGISTER ExprPtr e, **mtop = (ExprPtr **)&maptop[%d], *tpp = (ExprPtr *)maptop[%d];\n",
		    mapdepth-pc[3], mapdepth-pc[3]+1);
	  hfprintf (out, "    do {\n      if (tpp==mtop[0]) {");
	  DO_GOTO(0);
	  hfprintf (out, "      e = *--tpp;\n      *tpp = NOEXPR;\n    } while (e==empty_map);\n    while (e->kind!=(CHAR)E_MAPLET)\n      {\n        REGISTER ExprPtr aux = MAPLEFT(e);\n        if (aux==empty_map)\n          e = MAPRIGHT(e);\n        else\n          {\n            *tpp++ = MAPRIGHT(e);\n            e = aux;\n          }\n      }\n    mtop[1] = tpp;\n    _x.p[%d] = e;\n  }\n",
		    depth + pc[2]);
	  pc += 4;
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  break;
	CASE(BC_INCSUBMAPSWEEP):
	  SYNCTOP();
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  hfprintf (out, "  {\n    REGISTER ExprPtr *ep = &_x.p[%d], l = ep[1];\n",
		    depth + pc[2]);
	  hfprintf (out, "    while (true)\n      {\n        switch (l->kind)\n          {\n          case E_NIL: {");
	  DO_GOTO(0);
	  hfprintf (out, "          case E_OR:\n            l->kind = (CHAR)E_AND;\n            ep[0] = MapPutMapletTrie (CAR(l), ep[0], (CODE)(1L<<NLOWBITS));\n            goto lab%d;\n",
		    (pc - bc->instr)+3);
	  pc += 3;
	  hfprintf (out, "          case E_AND:\n           l->kind = (CHAR)E_OR;\n            ep[0] = MapRemoveTrie (CAR(CAR(l)), ep[0], (CODE)(1L<<NLOWBITS));\n            break;\n          }\n        l = CDR(l);\n      }\n    /*NOTREACHED*/\n  }\n");
	  break;
	CASE(BC_MAPDROP):
	  mapdepth -= (COUNTER)*pc++;
	  break;
	CASE(BC_QUEUEUNSHAREDLIST):
	  hfprintf (out, "  if (_x.p[%d]!=NIL)\n    {\n      REGISTER ExprPtr aux = _x.p[%d];\n      _x.p[%d] = CDR(aux);\n      SETCDR(aux,NIL);\n    }\n",
		    depth-1, depth-1, depth-1);
	  break;
	CASE(BC_HANDLE):
	  if (maxexcdepth==0)
	    {
	      hfprintf (hstderr, "bcXlate0: maxexcdepth = 0, but encountered BC_HANDLE.\n");
	      hfflush (hstderr);
	      abort ();
	    }
	  hfprintf (out, "  if (xhi.curxhandler==NULL)\n    {\n      {\n        REGISTER excHandler *exc = (excHandler *)alloctmp_anchored (sizeof(excHandler), (char *)&env);\n        exc->next = excStack;\n        exc->tag = X_HANDLER;\n        exc->data = NIL;\n        exc->code = NIL;\n        excStack = exc;\n");
	  /* Note: Since xhi.xhandlers is only allocated, not freed, this code, guarded
	     by (xhi.xhandlers==NULL) will only be executed at most once in the current function.
	     This makes calling SETJUMP() safe here.
	  */
	  hfprintf (out, "        if (SETJUMP(exc->jb)!=0)\n          {\n");
	  /* deal with exception raising: find handler. */
	  hfprintf (out, "            (void) alloctmp (0);\n            goto do_raise;\n          }\n      }\n",
		    depth-1);
	  nRaises++;
	  /* Here SETJUMP() returned 0: install exception handler. */
	  hfprintf (out, "      xhi.curxhandler = &xhi.exc[0];\n      xhi.endxhandler = &xhi.exc[%d];\n    }\n",
		    maxexcdepth);
	  nHandlers++;
	  hfprintf (out, "  {\n    REGISTER struct xhandler *xh = (struct xhandler *)xhi.curxhandler++;\n    xh->handlerNum = %d;\n    xh->n = _x.n;\n  }\n",
		    nHandlers);
	  /* Ignore pc[0] = number of things to drop from stack.
	     Apparently bcEval0() ignores it, too.
	     And bc.c always puts 0 as drop number with BC_HANDLER.
	  */
	  {
	    ExprPtr xnum, offset, e;

	    xnum = KINT (nHandlers);
	    offset = KINT ((pc + 1 + pc[1]) - bc->instr);
	    e = MakePair (E_CONS, xnum, offset);
	    (void) KEEP(e);
	    BC_EXCINFO = MakePair (E_CONS, e, BC_EXCINFO);
	    BC_OFFSTK = MakePair (E_CONS, offset, BC_OFFSTK);
	    BC_DEPTH = Kint (depth+1); /* +1 to account for the exception that will be pushed on the stack
					  on entry to the handler. */
	    BC_DEPTHSTK = MakePair (E_CONS, BC_DEPTH, BC_DEPTHSTK);
	    BC_MAPDEPTH = Kint (mapdepth);
	    BC_MAPDEPTHSTK = MakePair (E_CONS, BC_MAPDEPTH, BC_MAPDEPTHSTK);
	    QUIT;
	  }
	  pc += 2;
	  break;
	CASE(BC_NEXTHANDLE): goto nexthandle;
	CASE(BC_NEXTHANDLE_NODEBUG):
	nexthandle:
	  SYNCTOP();
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
	  if (maxexcdepth==0)
	    {
	      hfprintf (hstderr, "bcXlate0: maxexcdepth = 0, but encountered BC_NEXTHANDLE.\n");
	      hfflush (hstderr);
	      abort ();
	    }
#ifdef OLD_HANDLERS
	  hfprintf (out, "  if (--xhi.curxhandler==xhi.xhandlers->exc)\n    {\n      struct xhandlers *next = xhi.xhandlers->next;\n      if (next==NULL)\n        ;\n      else\n        {\n          freetmp ((char *)xhi.xhandlers);\n          xhi.xhandlers = next;\n          xhi.endxhandler = &xhi.xhandlers->exc[MAX_EXC_NEST];\n          xhi.curxhandler = xhi.endxhandler;\n        }\n    }\n");
#else
	  hfprintf (out, "  --xhi.curxhandler;\n");
#endif
	  break;
	CASE(BC_RAISE):
	  SYNCTOP();
	  hfprintf (out, "  THEEXPR = _x.p[%d]; goto do_raise;\n", depth-1);
	  nRaises++;
	  goto pop;
	CASE(BC_EXCEPTION):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakeException (_x.p[%d], EXC_CHECKED);\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_EXCEPTION_BENIGN):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakeException (_x.p[%d], EXC_BENIGN);\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_DELAY):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakePromise (_x.p[%d]);\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_PACK):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = MakePack (E_PACK, _x.p[%d], _x.p[%d]);\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IADD):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) + IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ISUB):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) - IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMUL):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) * IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IDIV):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) / IVAL(_x.p[%d]));\n",
		    /* instead of idiv() */
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMOD):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) % IVAL(_x.p[%d]));\n",
		    /* instead of imod() */
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ISNZERO):
	  hfprintf (out, "  if (_x.p[%d]==izero) {", depth-1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IOR):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) | IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IXOR):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) ^ IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IAND):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) & IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ILSL):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) << IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ILSR):
	  hfprintf (out, "  _x.p[%d] = Kint (((unsigned long)IVAL(_x.p[%d])) >> IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IASR):
	  hfprintf (out, "  _x.p[%d] = Kint (IVAL(_x.p[%d]) >> IVAL(_x.p[%d]));\n",
		    depth-2, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IABS):
	  hfprintf (out, "  {\n    REGISTER INT n = IVAL(_x.p[%d]);\n", depth-1);
	  hfprintf (out, "    if (n<0)\n      _x.p[%d] = Kint (-n);\n  }\n", depth-1);
	  break;
	CASE(BC_INEG):
	  hfprintf (out, "  _x.p[%d] = Kint (-IVAL(_x.p[%d]));\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_ISQR):
	  hfprintf (out, "  {\n    REGISTER INT n = IVAL(_x.p[%d]);\n", depth-1);
	  hfprintf (out, "    _x.p[%d] = Kint (n*n);\n  }\n", depth-1);
	  break;
	CASE(BC_INOT):
	  hfprintf (out, "  _x.p[%d] = Kint (~IVAL(_x.p[%d]));\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_ILT):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (IVAL(_x.p[%d]) >= IVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_ILE):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (IVAL(_x.p[%d]) > IVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IGT):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (IVAL(_x.p[%d]) <= IVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IGE):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (IVAL(_x.p[%d]) < IVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_IMIN):
	  hfprintf (out, "  if (IVAL(_x.p[%d]) > IVAL(_x.p[%d])) _x.p[%d] = _x.p[%d];\n",
		    depth-2, depth-1, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IMAX):
	  hfprintf (out, "  if (IVAL(_x.p[%d]) < IVAL(_x.p[%d])) _x.p[%d] = _x.p[%d];\n",
		    depth-2, depth-1, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RLT):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth, depth+1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) >= XVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RLE):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth, depth+1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) > XVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RGT):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth, depth+1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) <= XVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RGE):
	  POPTOP_IMM(2);
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth, depth+1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) < XVAL(_x.p[%d])) {",
		    depth, depth+1);
	  dpc = 2;
	  goto _goto;
	CASE(BC_RMIN):
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth-2, depth-1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) > XVAL(_x.p[%d]))\n    _x.p[%d] = _x.p[%d];\n",
		    depth-2, depth-1, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_RMAX):
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL || _x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth-2, depth-1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  if (XVAL(_x.p[%d]) < XVAL(_x.p[%d]))\n    _x.p[%d] = _x.p[%d];\n",
		    depth-2, depth-1, depth-2, depth-1);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_FLOOR):
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth-1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  {\n    REAL x = floor (XVAL(_x.p[%d]));\n", depth-1);
	  hfprintf (out, "    _x.p[%d] = Kreal (&x);\n  }\n",
		    depth-1);
	  break;
	CASE(BC_CEIL):
	  hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)E_REAL)\n    {\n",
		    depth-1);
	  hfprintf (out, "      THEEXPR = INCOMPEXC; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  {\n    REAL x = ceil (XVAL(_x.p[%d]));\n", depth-1);
	  hfprintf (out, "    _x.p[%d] = Kreal (&x);\n  }\n",
		    depth-1);
	  break;
	CASE(BC_AREF):
	  hfprintf (out, "  {\n    REGISTER COUNTER m = IVAL(_x.p[%d]);\n", depth-1);
	  hfprintf (out, "    REGISTER ExprPtr a;\n    REGISTER VecPtr vp;\n    if (m<0) {\nsub%d:\n", off);
	  hfprintf (out, "      THEEXPR = subscriptException; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  else\n    {\n      a = _x.p[%d];\n      vp = AVAL(a);\n", depth-2);
	  hfprintf (out, "      if (m>=vp->n) goto sub%d;\n      else _x.p[%d] = vp->expr[m];\n    }\n  }\n", off, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ASET):
	  hfprintf (out, "  {\n    REGISTER ExprPtr a, y;\n    REGISTER VecPtr vp;\n    REGISTER COUNTER m = IVAL(_x.p[%d]);\n",
		    depth-2);
	  hfprintf (out, "    if (m<0) {\nsub%d:\n", off);
	  hfprintf (out, "      THEEXPR = subscriptException; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  else\n    {\n      a = _x.p[%d];\n      y = _x.p[%d];\n",
		    depth-3, depth-1);
	  hfprintf (out, "      vp = AVAL(a);\n      if (m>=vp->n) goto sub%d;\n      else\n        {\n          vp->expr[m] = y;\n          HITBARRIER(a);\n        }\n    }\n  }\n",
		    off);
	  POPTOP_IMM(3);
	  break;
	CASE(BC_IAREF):
	  hfprintf (out, "  {\n    REGISTER INT m = IVAL(_x.p[%d]), *ip;\n",
		    depth-1);
	  hfprintf (out, "    if (m<0) {\nsub%d:\n", off);
	  hfprintf (out, "      THEEXPR = subscriptException; goto do_raise;\n");
	  nRaises++;
	  hfprintf (out, "      }\n    ip = (INT *)EXTDATA (_x.p[%d]);\n",
		    depth-2);
	  hfprintf (out, "    if (m>=ip[0]) goto sub%d;\n", off);
	  hfprintf (out, "    ip++;\n    _x.p[%d] = Kint (ip [m]);\n  }\n",
		    depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_IASET):
	  hfprintf (out, "  {\n    REGISTER ExprPtr a = _x.p[%d];\n    REGISTER INT m = IVAL(_x.p[%d]), ival = IVAL(_x.p[%d]), *ip;\n",
		    depth-3, depth-2, depth-1);
	  hfprintf (out, "    if (m<0) {\nsub%d:\n", off);
	  hfprintf (out, "      THEEXPR = subscriptException; goto do_raise;\n");
	  nRaises++;
	  hfprintf (out, "      }\n    ip = (INT *)EXTDATA(a);\n    if (m>=ip[0]) goto sub%d;\n", off);
	  hfprintf (out, "    ip++;\n    ip[m] = ival;\n  }\n");
	  POPTOP_IMM(3);
	  break;
	CASE(BC_NUMINT):
	  hfprintf (out, "  _x.p[%d] = Kint ((INT)XVAL(_x.p[%d]));\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_INTNUM):
	  hfprintf (out, "  {\n    REAL x = (REAL)IVAL(_x.p[%d]);\n", depth-1);
	  hfprintf (out, "    _x.p[%d] = Kreal (&x);\n  }\n",
		    depth-1);
	  break;
	CASE(BC_SIZE):
	  hfprintf (out, "  _x.p[%d] = Kint (SLEN(_x.p[%d]));\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_ORD):
	  hfprintf (out, "  if (SLEN(_x.p[%d])==0)\n    {\n",
		    depth-1);
	  hfprintf (out, "      THEEXPR = stringnthException; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "  _x.p[%d] = Kint ((INT)STEXT(_x.p[%d])[0]);\n",
		    depth-1, depth-1);
	  break;
	CASE(BC_ORDOF):
	  hfprintf (out, "  {\n    INT n = IVAL(_x.p[%d]);\n",
		    depth-1);
	  hfprintf (out, "    if (n<0 || n>=SLEN(_x.p[%d]))\n    {\n",
		    depth-2);
	  hfprintf (out, "      THEEXPR = stringnthException; goto do_raise;\n    }\n");
	  nRaises++;
	  hfprintf (out, "    _x.p[%d] = Kint ((INT)STEXT(_x.p[%d])[n]);\n  }\n",
		    depth-2, depth-2);
	  POPTOP_IMM(1);
	  break;
	CASE(BC_SUBSTR):
	  hfprintf (out, "  {\n    INT n, m = IVAL(_x.p[%d]);\n",
		    depth-2);
	  hfprintf (out, "    if (m<0) { THEEXPR = stringnthException; goto do_raise; }\n");
	  nRaises++;
	  hfprintf (out, "    n = IVAL(_x.p[%d]);\n",
		    depth-1);
	  hfprintf (out, "    if (n>SLEN(_x.p[%d])) { THEEXPR = stringnthException; goto do_raise; }\n",
		    depth-3);
	  nRaises++;
	  hfprintf (out, "    if (n<m) n=m;\n");
	  hfprintf (out, "#if STOP_AND_COPY_VECTORS\n    {\n      REGISTER INT size = n-m;\n      CHAR *s = (CHAR *)alloctmp_anchored (size, (char *)&env);\n      safememcpy (s, (char *)&STEXT(_x.p[%d])[m], size);\n      _x.p[%d] = Kstring (size, s);\n      freetmp (s);\n    }\n#else\n    _x.p[%d] = Kstring (n-m, &STEXT(_x.p[%d])[m]);\n#endif\n  }\n",
		    depth-3, depth-3, depth-3, depth-3);
	  POPTOP_IMM(2);
	  break;
	CASE(BC_CALLBIND):
	  SYNCTOP();
	  hfprintf (out, "  _x.p[%d] = (*ACTION(%s_consts[%d]))(ENV_CONS(_x.p[%d],1), _x.p[%d]);\n",
		    depth-2, buf, *pc++, depth-2, depth-1);
	  /*!!! does not work in case %s_consts[%d] is an E_TUPLE_2(E_BYTECODE(bc), env).
	    In that case, we should call bcEval().
	    A similar problem occurs with BC_CALL and BC_N_CALL above, where we estimate
	    that all functions are compiled, not in bytecode form.
	    In case one day I add, say, run-time [bytecode] compilation, I would need
	    to correct all that.
	  */
	  POPTOP_IMM(1);
	  break;
	CASE(BC_ILLEGAL): goto illegal;
	default:
	  if (op>=BC_HASTAG && op<BC_HASTAG+E_UNUSED)
	    {
	      hfprintf (out, " /* BC_HASTAG %d */\n", op-BC_HASTAG);
	      hfprintf (out, "  if (_x.p[%d]->kind!=(CHAR)%d) {",
			depth-1, op-BC_HASTAG);
	      dpc = 2;
	      goto _goto;
	    }
	illegal:
	  kinderror ("bcXlate0", op);
	  break;
	}
    }
  hfputs ("\nend:\n", out);
#ifdef DEBUG
  hfputs ("check_stack_frames ();\n", out);
#endif
  if (heap_x)
    hfputs ("  theHeapFrames = _x.next;\n  freetmp (_p);\n", out);
  else
    hfputs ("  theStackFrames = _x.next;\n", out);
  hfputs ("  return _x.p[0];\n", out);

  if (nRaises>0)
    {
      hfputs ("do_raise:\n", out);
      if (maxexcdepth>0)
	{
	  hfprintf (out, "  if (xhi.curxhandler==NULL || xhi.curxhandler==xhi.exc)\n    {\n      if (xhi.curxhandler!=NULL) { excHandler *exc = excStack->next; freetmp (excStack); excStack = exc; }\n");
	  if (heap_x)
	    hfprintf (out, "      freetmp (_p);\n");
	  hfprintf (out, "      (void) Raise (THEEXPR);\n      /*NOTREACHED*/\n    }\n");
	}
      else
	{
	  if (heap_x)
	    hfputs ("  freetmp (_p);\n", out);
	  hfputs ("  (void) Raise (THEEXPR);\n  /*NOTREACHED*/\n", out);
	}
      if (maxexcdepth>0 && nHandlers==0)
	{
	  hfprintf (hstderr, "bcXlate0: maxexcdepth=%d, but nHandlers=0.\n", maxexcdepth);
	  hfflush (hstderr);
	  abort ();
	}
      if (nHandlers>0)
	{
	  ExprPtr excInfo, e;

	  if (maxexcdepth==0)
	    {
	      hfprintf (hstderr, "bcXlate0: maxexcdepth=0, but nHandlers=%d.\n", nHandlers);
	      hfflush (hstderr);
	      abort ();
	    }
	  hfputs ("\nraise:\n  {\n    REGISTER struct xhandler *xh = &((struct xhandler *)xhi.curxhandler)[-1];\n", out);
	  hfprintf (out, "    _x.n = xh->n;\n    _x.p[_x.n++] = THEEXPR;\n    switch (xh->handlerNum) {\n");
	  for (excInfo = BC_EXCINFO; excInfo!=NIL; excInfo = CDR(excInfo))
	    {
	      e = CAR(excInfo);
	      hfprintf (out, "      case %d: goto lab%d;\n", IVAL(CAR(e)), IVAL(CDR(e)));
	    }
	  hfprintf (out, "      default: /*NOTREACHED*/ assert(0); break;\n    }\n  }\n");
	}
      else
	hfputs ("  /*NOTREACHED*/\n  assert(0);\n", out);
    }
  if (heap_x)
    hfputs("#undef _x\n", out);
  hfputs ("}\n", out);
  QUIT;
  theStackFrames = _x.next;
}

MAPSWEEPFUNC(do_bcxoutprotos)
{
  hFILE *out = MAPDATA(hFILE *);
  byteCodeChunk *bc = (byteCodeChunk *)x;
  ExprPtr name;
  CHAR *buf;

  name = bcXName (bc);
  name = COUNTEDVAL(name);
  buf = alloctmp (SLEN(name)+1);
  memcpy (buf, STEXT(name), SLEN(name));
  buf[SLEN(name)] = 0;
  hfprintf (out, "IMPORT ExprPtr %s P((IN CONST ExprPtr env, ...));\n",
	    buf);
  return false;
}

PUBLIC void bcXoutProtos (out)
     IN hFILE *out;
{
  (void) MapSweep (bcXlateMap, do_bcxoutprotos, out);
}

MAPSWEEPFUNC(do_bcxout)
{
  hFILE *out = MAPDATA(hFILE *);
  byteCodeChunk *bc = (byteCodeChunk *)x;

  bcXlate0 (out, bc);
  hfputs ("\n", out);
  return false;
}

PUBLIC void bcXout (out)
     IN hFILE *out;
{
  (void) MapSweep (bcXlateMap, do_bcxout, out);
}

MAPSWEEPFUNC(do_bcxoutcode)
{
  ExprPtr *table = MAPDATA(ExprPtr *);
  byteCodeChunk *bc = (byteCodeChunk *)x;
  ExprPtr name = bcXName (bc);

  table[COUNTEDN(name)] = COUNTEDVAL(name);
  return false;
}

PUBLIC void bcXoutCode (out)
     IN hFILE *out;
{
  ExprPtr *table, name;
  COUNTER i;

  table = alloctmp (nbcXlate * sizeof (ExprPtr));
  for (i=0; i<nbcXlate; i++)
    table[i] = NOEXPR;
  (void) MapSweep (bcXlateMap, do_bcxoutcode, table);
  for (i=0; i<nbcXlate; i++)
    {
      if (i % 4 == 0)
	hfputs ("\n  ", out);
      name = table[i];
      if (name==NOEXPR)
	hfputs ("NULL, ", out);
      else
	{
	  CHAR *buf = alloctmp (SLEN(name)+1);

	  memcpy (buf, STEXT(name), SLEN(name));
	  buf[SLEN(name)] = 0;
	  hfprintf (out, "(compiledAction)%s, ", buf);
	  freetmp (buf);
	}
    }
  /*!!!*/
}

MAPSWEEPFUNC(do_bcxdefs)
{
  byteCodeChunk *bc = (byteCodeChunk *)x;
  ExprPtr e, name, l, lr = MAPDATA(ExprPtr);
  ENTER;

  if (bc->nConsts!=0)
    {
      name = bcXName (bc);
      name = COUNTEDVAL (name);
      (void) KEEP(name);
      e = MakeRecord (E_TUPLE_GNRL, name, bc->nConsts, bc->consts);
      (void) KEEP (e);
      l = MakePair (E_CONS, e, REFVAL(lr));
      SETREFVAL (lr, l);
    }
  QUIT;
  return false;
}

PUBLIC ExprPtr bcXdefs (out)
     IN hFILE *out;
{
  ExprPtr e, l, lr, name;
  COUNTER i, n;
  CHAR *buf;
  ENTER;

  lr = MakeSing (E_REF, NIL);
  (void) KEEP (lr);
  (void) MapSweep (bcXlateMap, do_bcxdefs, lr);
  hfprintf (out, "PUBLIC void __init_consts (l)\nIN CONST ExprPtr l;\n{\n  ExprPtr e;\n\n");
  for (l=REFVAL(lr); l!=NIL; l=CDR(l))
    {
      e = CAR(l);
      n = AVAL(e)->n;
      name = ASHAPE (e);
      buf = alloctmp_anchored (SLEN (name)+1, (char *)&buf);
      memcpy (buf, STEXT(name), SLEN(name));
      buf[SLEN(name)] = 0;
      hfprintf (out, "  e = CAR(l); l = CDR(l); memcpy (%s_consts, AVAL(e)->expr, %d * sizeof(ExprPtr));\n",
		buf, n);
      freetmp (buf);
    }
  hfprintf (out, "}\n");
  QUIT;
  return REFVAL(lr);
}

  /*!!!
    For each code <name>, bcXlate0() creates a function by DEF_ACTION(<name>_<n>),
    where <n> is a gensym.
    It also relies on having a constant pool
    PUBLIC ExprPtr <name>_<n>_consts[];
    This constant pool is declared by bcXlate0(), but not filled in!
    It should be filled in by the relevant version of inByteCode(),
    and should be protected against GC by some means.
   */
